C+
      INTEGER FUNCTION TAU_CALL (COMMAND)
C
C                           F I G A R O _ C A L L
C
C  Routine name:
C     FIGARO_CALL
C
C  Function:
C     Executes a single Figaro command.
C
C  Description:
C     This is the main routine of the callable Figaro system.  It executes
C     a single Figaro command, allowing a high level language to be used
C     to control a sequence of such commands.  Note that while this
C     routine may be called directly, it is intended that most calls to
C     the callable Figaro system will be made via the routine FIGARO
C     which concatenates a variable number of arguments and then passes
C     the resulting command string on to this routine.
C
C  Language:
C     FORTRAN
C
C  Call:
C     STATUS = FIGARO_CALL (COMMAND)
C
C  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
C
C     (>) COMMAND       (Fixed string,descr) The Figaro command
C                       to be executed.  This should be exactly the
C                       same as a Figaro command line as given to DCL
C                       (although without the '$' at the start).
C  Returns:
C
C     (<) STATUS        (Integer, function value) Return status code.
C                       0 => OK, non-zero => error.
C
C  Prior requirements: 
C     If a previous call to FIGARO_CALL had failed, ie had returned with
C     non-zero status, FIGARO_CALL will simply return immediately, the
C     function value returned being that non-zero status.  This
C     condition may be cleared by a call to FIGARO_CLEAR (which
C     sets the DCL symbol FIGARO_ERROR to 0).  No explicit initialisation
C     of the callable Figaro system is needed.  
C
C  Common variables used:  
C     Internal variables used by the callable Figaro system.
C
C  External subroutines / functions used:
C
C     All the Figaro routines defined by the include files
C
C  Author: K. Shortridge, AAO
C
C  Date: 3rd June 1993.
C-
C  History:
C     27th Jun 1986.  Original version. KS/AAO.
C     21st Dec 1992.  Modified for the portable version of Figaro. KS/AAO.
C      3rd Jun 1993.  Now supports the repeated execution of routines that have
C                     parameter lists specified. KS/AAO.
C
C  Subroutine details:
C
C  Compiling:
C
C     This routine uses two include files, FIGNAMES.INC and FIGCALL.INC,
C     which are created from the COMMANDS.DAT file by the program GFIG.
C+
      IMPLICIT NONE
C
C     Parameters
C
      CHARACTER*(*) COMMAND
C
C     Functions used
C
      INTEGER FIGSYS_CLOSE
      INTEGER FIGSYS_ERROR
      LOGICAL PAR_REPEAT
      INTEGER ICH_FOLD
      INTEGER ICH_LEN
C
C     Local variables
C
      INTEGER ISPACE             ! Position of first space in COMMAND
      INTEGER LENGTH             ! Length of Figaro command
      CHARACTER NAME*16          ! Name of Figaro command
      LOGICAL NOBATCH            ! Command cannot be used in batch mode
      LOGICAL NOCMD              ! Indicates routine parses its own command
      LOGICAL NOPAR              ! Indicates command has no parameters
      INTEGER NUMBER             ! Number of Figaro routine to be executed
      LOGICAL PARS               ! Indicates if command has parameters
      LOGICAL REPEAT             ! Should command be repeated?
      INTEGER STAT               ! Position of %STAT string in command
      INTEGER STATUS             ! Returned status value
      CHARACTER UC_COMMAND*256   ! Upper case version of COMMAND.
C
C     Common variables
C
      LOGICAL INIT_FLAG          ! Indicates if Figaro system initialised
      LOGICAL ERROR_FLAG         ! Indicates an error has occurred
      LOGICAL VERIFY_FLAG        ! Indicates commands are to be logged
      LOGICAL BATCH_MODE         ! Indicates running in batch mode
C
      COMMON /FIGARO_CALL_COM/ INIT_FLAG,ERROR_FLAG,VERIFY_FLAG
     :                                             ,BATCH_MODE
C
C     Make sure system is initialised
C
      IF (.NOT.INIT_FLAG) CALL FIGARO_OPEN
C
C     Determine length of command and get an upper case version of it.
C
      UC_COMMAND=COMMAND
      LENGTH=ICH_FOLD(UC_COMMAND)
C
C     Don't continue on error
C
      IF (ERROR_FLAG) THEN
         IF (VERIFY_FLAG) THEN
            CALL PAR_WRUSER('Not executed: '//COMMAND(:LENGTH),
     :                                                     STATUS)
         END IF
         TAU_CALL=1
         RETURN
      END IF
C
C     Establish condition handler for numeric errors
C
      CALL TRPZ_START
C
C     Log command
C
      IF (VERIFY_FLAG) CALL PAR_WRUSER(COMMAND(:LENGTH),STATUS)
C
C     %STAT in command?  If so remove it and start timer
C
      STAT=INDEX(UC_COMMAND,'%STAT')
      IF (STAT.NE.0) THEN
         COMMAND(STAT:STAT+5)=' '
         CALL FIGSYS_TIMER('START')
      END IF
C
C     Get command name out of command string
C
      ISPACE=INDEX(COMMAND,' ')
      IF (ISPACE.EQ.0) ISPACE=LEN(COMMAND)+1
      NAME=COMMAND(:ISPACE-1)
C
C     See if we know this command.  The include file is a section of
C     code that tests NAME against the known Figaro command names and
C     sets NUMBER to 0 (if no match is found) or to the number of the
C     subroutine to be called to service the command.  It also sets
C     the variables NOPAR, NOBATCH and NOCMD if these were indicated
C     as qualifiers for the command in the original table.
C
      INCLUDE 'fignames.inc'
C
      IF (NUMBER.GT.0) THEN
C
C        Test for an invalid attempt to use batch mode
C
         IF (BATCH_MODE.AND.NOBATCH) THEN
            CALL PAR_WRUSER('The command '//NAME(:ICH_LEN(NAME))
     :                   //' cannot be used in batch mode.',STATUS)
            CALL FIG_SETERR
         ELSE
C
C           Initialise command processor routines.  Note that the
C           NOCMD case (which is unusual) has to be treated specially.
C
            PARS=.NOT.NOPAR
            IF (NOCMD) THEN
               CALL PAR_INIT(NAME,COMMAND(ISPACE:),0,PARS,STATUS)
            ELSE
               CALL PAR_INIT(NAME,COMMAND,ISPACE,PARS,STATUS)
            END IF
C
C           Perform any necessary Figaro system initialisation.
C
            CALL FIGSYS_INIT(NAME)
C
C           Call the application subroutine.  The include file is a
C           section of code that calls the Figaro subroutine indicated
C           by NUMBER. The actual call is repeated if the parameter system
C           indicates that a parameter list is being processed.
C
            REPEAT=.TRUE.
            DO WHILE (REPEAT)

               INCLUDE 'figcall.inc'

               REPEAT=PAR_REPEAT()
               IF (FIGSYS_ERROR().NE.0) REPEAT=.FALSE.
            END DO
C
C           Log any trapped errors
C
            CALL FIGSYS_TRPLST(NAME)
            CALL TRPZ_STOP
C
C           Perform any necessary Figaro system closedown.
C
            STATUS=FIGSYS_CLOSE(NAME)
C
C           Output statistics if required
C
            IF (STAT.NE.0) THEN
               CALL FIGSYS_TIMER('REPORT')
               CALL FIGSYS_TIMER('STOP')
            END IF
         END IF
C
      ELSE
         TAU_CALL=1
         ERROR_FLAG=.TRUE.
         CALL PAR_WRUSER('Undefined Figaro command: '//
     :                                                 NAME,STATUS)
      END IF
C
      END
