      subroutine calbasic
C+-----------------------------------------------------------------------------
C
C   ---------------
C   C A L B A S I C
C   ---------------
C
C   Description
C   -----------
C   Creates a basic TAURUS calibration parameters structure. The basic
C   structure contains objects for constant information about the telescope,
C   TAURUS instrument, collimator, etalon, camera, and detector. One or more
C   .RUN structures, each holding information for a specific TAURUS run
C   (the run type may be CAL, MASK, OBS, or RING), may be added using the
C   program CALRUN.
C
C
C   Scope of program
C   ----------------
C   - Image dimensionality not relevant.
C   - Data array type not relevant.
C   - Subsetting not relevant.
C   - Magic values not relevant.
C   - Quality and variance arrays not relevant.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read or written)
C   ---------------------------
C   CALOUT  Name of the file which will contain the basic structure.
C           (character, prompted for).
C
C
C   Keywords
C   --------
C   None.
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - File name and structure definition file (TAU_CAL_BASIC.DEF) are
C     read and the structure is created.
C   - The user is prompted for the setup which has been used for the
C     data.  These are from three files: setups.std, etalons.std, and
C     detectors.std.
C   - The values from the files corresponding to the user's setup are
C     added into the structure.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_CREATE_NAMED_STRUCTURE
C     DSA_FREE_LU
C     DSA_GET_LU
C     DSA_OPEN
C     DSA_READ_STRUCT_DEF
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library FIG:
C     FIG_OPFIL2
C
C   Library ICH:
C     ICH_LEN
C
C   Library PAR:
C     PAR_RDCHAR
C
C   Library TAU:
C     TAU_CAL_SETCHR
C     TAU_CAL_SETNUM
C
C
C   Internal subroutines called
C   ---------------------------
C   CLEAN_STR
C   FILL
C   FILLDET
C   FILLETA
C   FILLSET
C   GETCHR
C   GETFLO
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'TAU_COMMON'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / Names > 6 characters / INCLUDE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C                (Loosely based on Nick Fuller's program of
C                the same name)
C
C
C   History
C   -------
C   29-JUN-1990   - Original program
C   14-NOV-1990   - Adds in mask pinhole separation now. (JRL)
C   31-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C+-----------------------------------------------------------------------------
      implicit none
C
C     Local variables
C
      character file_name*80   ! File name
      integer   status         ! DSA status flag
C
C     TAUCAL common block
C
      include 'TAU_COMMON'
C
C     Initialise and open DSA
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Get file name
C
      call par_rdchar('calout'//'/nocheck',' ',file_name)
C
C     Read structure definition file
C
      call dsa_read_struct_def('TAU_CAL_BASIC.DEF',status)
      if (status .ne. 0) go to 500
C
C     Create the structure
C
      call dsa_create_named_structure('CAL',file_name,'cal_struct',
     : status)
      if (status .ne. 0) go to 500
C
C     Get DTA info on TAUCAL file
C
      call tau_cal_dta_name('cal',file_name,dta_name,lendta,status)
      if (status .ne. 0) go to 500
C
C     Add in the necessary info...
C
      call fill('cal')
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine fill(ref_name)
C+--------------------------------------------------------------------------------------
C     FILL       --- Subroutine to fill the data into the CAL structure.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) REF_NAME:  (character) Reference name of the structure
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      character ref_name*(*)
C
C     Functions
C
      integer ich_len
      logical par_qnum
C
C     Standard information files
C
      character files(3)*10
      data files /'SETUPS','ETALONS','DETECTORS'/
C
C     Local variables
C
      integer status,unit,nset,iset,nskip,iskip,i
      real dummy
      character data*80
      logical answer
C
C     Do for the three types of information -- Setups (standard),
C     Etalons and Detectors
C
      do i = 1,3
C
C        Get standard file
C
         status = 0
         call dsa_get_lu(unit,status)
         call fig_opfil2(files(i),'STD',unit,'old',status)
         if (status .ne. 0) then
            call dsa_wruser('Couldn''t find standard '//files(i)//
     :      ' file')
            call dsa_wrflush
            go to 500
         end if
C
C        Now read and write out the first few lines to give the
C        user a chose of objects in the file
C
         read(unit,100,err=500,iostat=status) data
  100    format(a80)
         iset = 0
         do while (data(1:1) .ne. '*')
            call dsa_wruser(data(:ich_len(data)))
            call dsa_wrflush
            iset = iset + 1
            read(unit,100,iostat=status) data
            if (status .ne. 0) then
               call dsa_wruser('Error in reading '//files(i)//' file')
               call dsa_wrflush
               call dsa_wruser('Probable premature end of file')
               call dsa_wrflush
               go to 500
            end if
         end do
         if (iset .eq. 0) then
            call dsa_wruser('Nothing in '//files(i)//' file...')
            call dsa_wrflush
            go to 500
         end if
         answer = par_qnum('Which do you want?',1.0,float(iset),1.0,
     :   .false.,' ',dummy)
         nset = nint(dummy)
C
C        Now read through the file until you find the desired setup
C
         nskip = nset - 1
         iskip = 0
         do while (iskip .ne. nskip)
            read(unit,100,iostat=status) data
            if (status .ne. 0) then
               call dsa_wruser('Error in reading '//files(i)//' file')
               call dsa_wrflush
               call dsa_wruser('Probable premature end of file')
               call dsa_wrflush
            end if
            do while (data(1:1) .ne. '*')
               read(unit,100,iostat=status) data
               if (status .ne. 0) then
                  call dsa_wruser('Error in reading '//files(i)//
     :            ' file')
                  call dsa_wrflush
                  call dsa_wruser('Probable premature end of file')
                  call dsa_wrflush
               end if
            end do
            iskip = iskip + 1
         end do
         if (i .eq. 1) then
            call fillset(ref_name,unit,status)
         else if (i .eq. 2) then
            call filleta(ref_name,unit,status)
         else if (i .eq. 3) then
            call filldet(ref_name,unit,status)
         end if
C
C        Look for any errors...
C
         if (status .ne. 0) then
            call dsa_wruser('***An error has occurred either during')
            call dsa_wrflush
            call dsa_wruser('the reading of the '//files(i)//' file')
            call dsa_wrflush
            call dsa_wruser('or the writing of the CAL file.')
            call dsa_wrflush
         end if
         close(unit)
         status = 0
         call dsa_free_lu(unit,status)
      end do
  500 continue
      end



      subroutine fillset(ref_name,unit,status)
C+--------------------------------------------------------------------------------------
C     FILLSET    --- Subroutine to fill in the general setup info.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) REF_NAME:  (character) Reference name of the structure
C     (>) UNIT:      (integer) Logical unit of the setup file
C     (!) STATUS:    (integer) Status flag
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:      14-NOV-1990 -- Now adds in mask pinhole separation (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer unit,status
      character ref_name*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      real dummy
      character data*80,string*80
C
C     Now read the desired setup and set the variables in the file
C     First the telescope parameters...
C
      read(unit,100,err=500,iostat=status) data
  100 format(a80)
      call getchr(data,string)
      call tau_cal_setchr(ref_name,0,'telescope.name',.true.,
     : ich_len(string),string,status)
      read(unit,100,err=500,iostat=status) data
      call getchr(data,string)
      call tau_cal_setchr(ref_name,0,'telescope.station',.true.,
     : ich_len(string),string,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'telescope.scale',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'telescope.focnum',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'telescope.foclen',.true.,
     : 1,1,1,dummy,status)
C
C     Now the parameters for TAURUS, the collimator and the camera
C
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'taurus.version',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'collimator.foclen',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'camera.focnum',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'camera.foclen',.true.,
     : 1,1,1,dummy,status)
C
C     Finally get the mask separation
C
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'mask.separation',.true.,
     : 1,1,1,dummy,status)
  500 end



      subroutine filleta(ref_name,unit,status)
C+--------------------------------------------------------------------------------------
C     FILLETA    --- Subroutine to fill in the etalon setup info.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) REF_NAME:  (character) Reference name of the structure
C     (>) UNIT:      (integer) Logical unit of the setup file
C     (!) STATUS:    (integer) Status flag
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer unit,status
      character ref_name*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      real dummy,arr(2)
      character data*80,string*80
C
C     Now read the desired etalon variables and write them to the
C     CAL file
C
      read(unit,100,err=500,iostat=status) data
  100 format(a80)
      call getchr(data,string)
      call tau_cal_setchr(ref_name,0,'etalon.name',.true.,
     : ich_len(string),string,status)
      read(unit,100,err=500,iostat=status) data
      call getarr(data,2,arr)
      call tau_cal_setnum(ref_name,0,'etalon.range',.true.,
     : 1,2,2,arr,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'etalon.reflect',.true.,
     : 1,1,1,dummy,status)
      read(unit,100,err=500,iostat=status) data
      call getflo(data,dummy)
      call tau_cal_setnum(ref_name,0,'etalon.nomgap',.true.,
     : 1,1,1,dummy,status)
  500 end


      subroutine filldet(ref_name,unit,status)
C+--------------------------------------------------------------------------------------
C     FILLDET    --- Subroutine to fill in the detector setup info.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) REF_NAME:  (character) Reference name of the structure
C     (>) UNIT:      (integer) Logical unit of the setup file
C     (!) STATUS:    (integer) Status flag
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer unit,status
      character ref_name*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      character data*80,string*80
C
C     Now read the desired detector variables and write them to the
C     CAL file
C
      read(unit,100,err=500,iostat=status) data
  100 format(a80)
      call getchr(data,string)
      call tau_cal_setchr(ref_name,0,'detector.name',.true.,
     : ich_len(string),string,status)
  500 end


      subroutine getchr(char1,char2)
C+--------------------------------------------------------------------------------------
C     GETCHR     --- Subroutine to get character information from the
C                    setup file data line.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) CHAR1:     (character) Data line from the setup file
C     (<) CHAR2:     (character) Returned character variable
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      character char1*(*),char2*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer ind1,ind2,length
C
C     Make sure 'CHAR2' is 'clean' (all blank...)
C
      call clean_str(char2)
C
C     Find the location of the ! and extract the data
C
      ind1 = index(char1,'!') + 2
      ind2 = ich_len(char1)
      length = ind2 - ind1 + 1
      char2(1:length) = char1(ind1:ind2)
      end



      subroutine getflo(char,float)
C+--------------------------------------------------------------------------------------
C     GETFLO     --- Subroutine to get a floating point value from the
C                    setup file data line.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) CHAR:      (character) Data line from the setup file
C     (<) FLOAT:     (real) Returned real variable
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real float
      character char*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer ind1,ind2
C
C     Find the location of the ! and extract the data
C
      ind1 = index(char,'!') + 2
      ind2 = ich_len(char)
      read(char(ind1:ind2),*) float
      end


      subroutine getarr(char,n,arr)
C+--------------------------------------------------------------------------------------
C     GETARR     --- Subroutine to get an array of floating point values from the
C                    setup file data line.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) CHAR:      (character) Data line from the setup file
C     (>) N:         (integer) Number of values in array
C     (<) ARR:       (real array ARR(N)) Returned array
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer n
      real arr(n)
      character char*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer ind1,ind2,j
C
C     Find the location of the ! and extract the data
C
      ind1 = index(char,'!') + 2
      ind2 = ich_len(char)
      read(char(ind1:ind2),*) (arr(j),j=1,n)
      end



      subroutine clean_str(string)
C+--------------------------------------------------------------------------------------
C     CLEAN_STR  --- Subroutine to make a string all blanks
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (!) STRING:    (character) String to be cleaned
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-JUN-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      character string*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer ind1,ind2,j
C
C     Fill the string with blanks
C
      ind1 = 1
      ind2 = ich_len(string)
      do j = ind1,ind2
         string(j:j) = ' '
      end do
      end
