      subroutine calrun
C+------------------------------------------------------------------------------
C
C   -----------
C   C A L R U N
C   -----------
C
C   Description
C   -----------
C   Adds one or more .RUN structures, to hold information for a specific
C   TAURUS run, to a basic TAURUS calibration parameters structure. The run
C   type may be CAL, MASK, OBS, or RING. The basic structure, which must
C   already exist, is created using the program CALBASIC.
C
C
C   Scope of program
C   ----------------
C   - Image dimensionality not relevant.
C   - Data array type not relevant.
C   - Subsetting not relevant.
C   - Magic values not relevant.
C   - Quality and variance arrays not relevant.
C   - Batch execution not supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read or written from VARS.DST)
C   ------------------------------------------
C   CALOUT     Name of the file containing the output structure. (character)
C              (prompted for).
C
C   RUNTYPE    Type of run for which run structures are to be added, one of
C              'CAL', 'MASK', 'OBS', or 'RING'. (character)(prompted for).
C
C   NSTRUCT    Number of run structures to be added for the given run type.
C              (integer)(prompted for).
C
C   RUN_NOS    Run numbers to be assigned to the run structures. (integer,
C              array)(prompted for).
C
C   IMAGE      Name of the file corresponding to a certain run number.
C              NB: This file MUST exist and be XYZ sorted! (character)
C              (prompted for)
C
C   MPU0       Z axis zero point in MPUs (real)(prompted for).
C
C   DMPU       Z axis increment in MPUs/frame (real)(prompted for).
C
C   WAVE       Wavelength of a particular run in microns. (real)(prompted for)
C
C
C   Keywords
C   --------
C   AGAIN      Instruction to add more run structures. (prompted for)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The type of structure to be added is given by the user,
C     the appropriate definition file is read and the substructure
C     is created.
C   - The user is prompted for the values to be held for each data
C     item.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C      DSA_ADD_STRUCTURE
C      DSA_CLOSE
C      DSA_CLOSE_STRUCTURE
C      DSA_DATA_SIZE
C      DSA_NAMED_INPUT
C      DSA_OPEN
C      DSA_OUTPUT
C      DSA_READ_STRUCT_DEF
C      DSA_SET_STRUCT_VAR
C      DSA_WRFLUSH
C      DSA_WRUSER
C
C   Library DTA:
C      DTA_ERROR
C      DTA_RDVARC
C      DTA_RDVARF
C      DTA_SZVAR
C      DTA_TYVAR
C      DTA_WRVARC
C
C   Library ICH:
C      ICH_CI
C      ICH_FOLD
C      ICH_LEN
C
C   Library PAR:
C      PAR_CNPAR
C      PAR_RDARY
C      PAR_RDCHAR
C      PAR_RDKEY
C      PAR_RDVAL
C      PAR_SDCHAR
C
C   Library TAU:
C      TAU_CAL_DTA_NAME
C      TAU_CAL_OPEN
C      TAU_CAL_RDNUM
C      TAU_CAL_SETCHR
C      TAU_CAL_SETNUM
C
C
C   Internal subroutines called
C   ---------------------------
C   None.
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'NDP_NUMERIC_RANGES'
C   INCLUDE 'TAU_COMMON'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / NAMES > 6 CHARACTERS / DO WHILE / END DO
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   29-JUN-1990   - Original version based on a program by NJMF but substantially 
C                   altered so that the values get written in as the 
C                   items are created.  (JRL)
C   05-NOV-1990   - Altered to add data into .RVT substructure (JRL)
C   14-NOV-1990   - Took out pinhole separation (JRL)
C   19-NOV-1990   - New version of TAUCAL routines pasted in. (JRL)
C   30-NOV-1990   - Now writes out the data sizes. (JRL)
C   12-FEB-1991   - Minor bug fixed.  (JRL)
C   15-JUL-1991   - Modified to incorporate changes in PAR routines for
C                   FIGARO version 3.
C   03-JAN-1992   - Minor bug fixed.  Also, procedure for adding substructures
C                   changed to make it less fiddly and avoid a nasty bug in
C                   HDS.  Included TAUCAL common block. (JRL)
C   07-APR-1992   - Slight modification to give a default file name in the
C                   form R#### where #### is the run number. (JRL)
C   31-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions.
C
      character ich_ci*8
      integer   ich_len
      integer   ich_fold
C
C     Maximum number of run numbers
C
      integer maxnum
      parameter (maxnum=40)
C
C     Control for DSA_OUTPUT. Note NEW_FILE=0, i.e. a new data structure is NOT
C     created.
C
      integer   new_file,no_data
      parameter (new_file=0,no_data=0)
C
C     Local variables.
C
      logical   again           ! Instruction to repeat loop
      character calout*80       ! Name of TAUCAL file
      integer   dim(3)          ! Dims of input image
      integer   dims            ! Dims of data arrays in .CAL run structure
      integer   dta_status      ! DTA status flag
      real      dumarr(maxnum)  ! REAL array dummy variable
      integer   dumint          ! INTEGER dummy variable
      real      dumreal         ! REAL dummy variable
      character error*80        ! DTA_ error message
      character extype*8        ! Existing run type
      logical   exist           ! Flag for existence of TAUCAL file
      integer   i               ! Loop counter
      character image*80        ! File names
      integer   n               ! Integer counter
      integer   ndim            ! Number of dimensions
      integer   nelm            ! Number of elements in input structure
      integer   numdim          ! Number of dimensions in input structure
      integer   nstruct         ! Number of structures to be added
      character number*8        ! Run number as character string
      integer   numbers(maxnum) ! Run numbers to be assigned to the structures
      real      realnum(maxnum) ! Run numbers as real
      character runtype*8       ! Run type
      integer   status          ! Status code
      character type*8          ! Type returned by DTA_TYVAR
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     TAUCAL common block
C
      include 'TAU_COMMON'
C
C     Initialize.
C
      status = 0
      again = .true.
C
C     Open DSA system.
C
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Look for TAUCAL file and open it
C
      exist = .true.
      call par_rdchar('calout'//'/nocheck',' ',calout)
      call dsa_named_input('calin',calout,status)
      if (status .ne. 0) go to 500
C
C     Open CALOUT file.
C
      call dsa_wruser
     :  ('Note: the current version of CALOUT will be overwritten.')
      call dsa_wrflush
      call dsa_named_output('calout',calout,'calin',no_data,new_file,
     : status)
      if (status .ne. 0) go to 500
C
C     Get DTA name of TAUCAL file
C
      call tau_cal_dta_name('calin',calout,dta_name,lendta,status)
      if (status .ne. 0) go to 500
C
C     Add run structures as required -
C
      do while (again)
C
C        Get the run type and convert it to upper case.
C
         call par_cnpar('runtype')
         call par_rdchar('runtype',' ',runtype)
         dumint = ich_fold(runtype)
C
C        Get the number of runs of this type to be added.
C
         call par_cnpar('nstruct')
         call par_rdval('nstruct',1.0,real(max_int),1.0,' ',dumreal)
         nstruct = int(dumreal)
C
C        Get the run numbers to be assigned to the structures.
C
         call par_cnpar('run_nos')
         call par_rdary
     :   ('run_nos',1.0,real(max_int),' ',' ',nstruct,maxnum,realnum)
         do i = 1,nstruct
            numbers(i) = int(realnum(i))
         end do
C
C        If run type is 'RING' then fix element TAUCAL.RVT.RUNNOS.  This
C        tells how many rings there are and their run numbers...
C
         if (runtype .eq. 'RING') then
C
C           First look to see if it exists in the first place.
C
            call dta_szvar
     :      (dta_name(:lendta)//'.rvt.runnos',1,ndim,dims,dta_status)
C
C           If it doesn't exist, then create it and fill it in...
C
            if (dta_status .ne. 0) then
               dims = nstruct
               call tau_cal_setnum('calout',0,'rvt.runnos',.true.,1,
     :         dims,dims,realnum,status)
               if (status .ne. 0) go to 500
C
C           If it does exist then tack the new numbers onto the end...
C
            else
               call tau_cal_rdnum('calout',0,'rvt.runnos',' ',.true.,
     :         .false.,float(min_int),float(max_int),' ',dims,dumarr,
     :         status)
               if (status .ne. 0) go to 500
               n = 0
               do i = dims+1,dims+nstruct
                  n = n + 1
                  dumarr(i) = realnum(n)
               end do
               dims = dims + nstruct
               call tau_cal_setnum('calout',0,'rvt.runnos',.true.,1,
     :         dims,dims,dumarr,status)
               if (status .ne. 0) go to 500
            end if
C
C           Set up a counter for the number of RVTs that exist...
C
            call dta_rdvarf(dta_name(:lendta)//'.rvt.nrvt',1,dumreal,
     :      dta_status)
            if (dta_status .ne. 0) then
               call tau_cal_setnum('calout',0,'rvt.nrvt',.true.,1,1,1,
     :         0.0,status)
            end if
C
C           Now, how many runs are there?
C
            call tau_cal_setnum('calout',0,'rvt.nruns',.true.,1,1,1,
     :      float(dims),status)
            if (status .ne. 0) go to 500
         end if
C
C        Read the structure definition file appropriate to the run type.
C
         call dsa_read_struct_def
     :   ('TAU_'//runtype(:ich_len(runtype))//'_RUN.DEF',status)
         if (status .ne. 0) go to 500
C
C        Loop through the requested run numbers -
C
         do i = 1,nstruct
            number = ich_ci(numbers(i))
C
C           Check whether the intended .RUNn structure already exists.
C
            call dta_tyvar
     :      (dta_name(:lendta)//'.run'//number(:ich_len(number)),type,
     :      dta_status)
C
C           If it does, read the run type, issue a warning message, and go to the
C           next run number.
C
            if (dta_status .eq. 0) then
               call dta_rdvarc
     :         (dta_name(:lendta)//'.run'//number(:ich_len(number))
     :         //'.'//'type',8,extype,dta_status)
               call dsa_wruser('.run'//number(:ich_len(number)))
               call dsa_wruser(' already exists with run type ')
               call dsa_wruser(extype(:ich_len(extype)))
               call dsa_wruser(', not changed.')
               call dsa_wrflush
C
C              If the intended .RUNn structure does not exist, add a .RUNn structure
C              according to the definition file.
C
            else
               call dsa_set_struct_var('nnnn','.run'//
     :         number(:ich_len(number)),status)
               call dsa_add_structure
     :         ('calout','tau_'//runtype(:ich_len(runtype))//'_run',
     :         runtype(:ich_len(runtype))//'_run_struct',status)
               if (status .ne. 0) go to 500
C
C              Write the run type into the newly added structure.
C
               call dta_wrvarc
     :         (dta_name(:lendta)//'.run'//number(:ich_len(number))//
     :         '.'//'type',8,runtype,dta_status)
               if (dta_status .ne. 0) go to 500
               call dsa_wruser(' ')
               call dsa_wrflush
               call dsa_wruser('.RUN'//number(:ich_len(number)))
               call dsa_wruser(' added with run type ')
               call dsa_wruser(runtype(:ich_len(runtype))//'.')
               call dsa_wrflush
C
C              Set some variables for each image
C
               call par_cnpar('image')
               image = 'R'//number(:ich_len(number))
               call par_sdchar('image',image,dumint)
               call par_rdchar('image',' ',image)
               call tau_cal_setchr('calout',numbers(i),'image',.true.,
     :         80,image,status)
               call dsa_named_input('image',image,status)
               call dsa_data_size('image',3,numdim,dim,nelm,status)
               call dsa_close_structure('image',status)
               call tau_cal_setnum('calout',numbers(i),'nx',.true.,
     :         1,1,1,float(dim(1)),status)
               call tau_cal_setnum('calout',numbers(i),'ny',.true.,
     :         1,1,1,float(dim(2)),status)
               if (numdim .eq. 3) then
                  call tau_cal_setnum('calout',numbers(i),'nz',.true.,
     :            1,1,1,float(dim(3)),status)
               end if
               if (status .ne. 0) go to 500
C
C              Now set some variables depending on the type of the structure
C
               if ((runtype .eq. 'CAL') .or. (runtype .eq. 'OBS')) then
                  call par_cnpar('mpu0')
                  call par_rdval('mpu0',0.0,max_float,1.0,
     :            'microprocessor units',dumreal)
                  call tau_cal_setnum('calout',numbers(i),'mpu0',
     :            .true.,1,1,1,dumreal,status)
                  call par_cnpar('dmpu')
                  call par_rdval('dmpu',0.0,max_float,1.0,
     :            'microprocessor units',dumreal)
                  call tau_cal_setnum('calout',numbers(i),'dmpu',
     :            .true.,1,1,1,dumreal,status)
                  call par_cnpar('wave')
                  call par_rdval('wave',min_float,max_float,1.0,
     :            'microns',dumreal)
                  call tau_cal_setnum('calout',numbers(i),'wave',
     :            .true.,1,1,1,dumreal,status)
               else if (runtype .eq. 'RING') then
                  call par_cnpar('wave')
                  call par_rdval('wave',min_float,max_float,1.0,
     :            'microns',dumreal)
                  call tau_cal_setnum('calout',numbers(i),'wave',
     :            .true.,1,1,1,dumreal,status)
               end if
               if (status .ne. 0) go to 500
            end if
         end do
C
C        Ask whether any more additions are required.
C
         call par_cnpar('again')
         call par_rdkey('again',.true.,again)
         if (again) then
            call dsa_wruser(' ')
            call dsa_wrflush
         end if
      end do
C
C     Look for DTA_ error.
C
  500 if (dta_status .ne. 0) then
         call dta_error(dta_status,error)
         call dsa_wruser(error(:ich_len(error)))
         call dsa_wrflush
      end if
C
C   Tidy and exit.
C
      call dsa_close(status)
      end
               
