      subroutine correct
C+-----------------------------------------------------------------------------
C
C   -------------
C   C O R R E C T
C   -------------
C
C   Description
C   -----------
C   Does the actual phase correction of a TAURUS data cube.
C
C
C   Scope of program
C   ----------------
C   - Handles 3-D images only.
C   - Images MUST be ZXY sorted
C   - SHORT and FLOAT data only supported all others
C     converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values not supported.
C   - Batch execution supported.
C   - Output will also be ZXY sorted
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   OBSRUN      Run number associated with OBS cube. (integer)
C               (prompted for).
C
C   NAME_GRID   Name of input file with grid of X,Y pixel positions.
C               (character)(prompted for)
C
C   NAME_COR    Name of ouput file for corrected OBS cube.  Will be
C               ZXY sorted upon output (character)(prompted for)
C
C   NAME_OBS    Name of input uncorrected OBS cube. Must be ZXY sorted
C               (character)(prompted for)
C
C   INTERP      Flag for the type of interpolation to be done on the
C               data.  At present there are two methods available.
C               (1) linear and (2) sinc.  Sinc sometimes causes
C               rounding problems in SHORT data, and for such data
C               linear is recommended.
C
C   WHICH_MASK  Number of the MASK already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU     Flag for use of TAUCAL file. (prompted for)
C
C   USE_GRID    Instruction to either use the grid of results from
C               MASK and GRID to define pixel-space or not.
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .MASK              NMASK       Number of MASKs already defined in TAUCAL.
C                                  (integer)(read)
C
C   .MASK.MASKn        NGRID       Number of GRIDs already defined in MASK
C                                  sub-structure of TAUCAL. (integer)(read)
C
C   .MASK.MASKn.GRIDn  NAME_GRID   Name of input file with grid of X,Y pixel
C                                  positions. (character)(prompted for)
C
C   (OBS)              DX          The correction to XC from RVT.  (real)(read)
C
C   (OBS)              DY          The correction to YC from RVT.  (real)(read)
C
C   (OBS)              XC          The X position of the optical axis. (real)
C                                  (read)
C
C   (OBS)              YC          The Y position of the optical axis. (real)
C                                  (read)
C
C   (OBS)              DMPU        Z axis increment in MPUs/frame (real)(read)
C
C   (OBS)              NAME_SHIFTS Name of the file containing phase_map.
C                                  (character)(read)
C
C   (OBS)              FSR         Free spectral range in MPUs. (real)(read)
C
C   .MASK.MASKn        PIXSIZE_X   Pixel size as defined by MASK. (real)(read)
C
C   .CAMERA            FOCLEN      Focal length of the camera used. (real)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   - All standard objects except the data array are copied from
C     IMAGE.
C
C
C   Method
C   ------
C   - Each spectrum is copied over to a real array, if necessary.
C   - Spectrum is shifted by subtracting the fractional part of
C     the shift using a the interpolation interpolation method of
C     the user's choice. 
C   - Spectrum is then shifted by subtracting the integer number
C     of pixels
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_GET_WORKSPACE
C     DSA_GET_WORK_ARRAY
C     DSA_INPUT
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_OUTPUT
C     DSA_RESHAPE_AXIS
C     DSA_SIMPLE_OUTPUT
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_BATCH
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_DUMMY_GRID
C
C   Internal subroutines called
C   ---------------------------
C     SAMPLE
C     SINC_INTER
C     SUBTRACT_I
C     SUBTRACT_R
C     SUB_CORRECT
C     TRANSFER_I
C     TRANSFER_R
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   INTEGER*2
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL  or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   23-MAR-90   - Original program
C   25-MAY-90   - A few changes, mainly cosmetic.  (JRL)
C   03-JUL-90   - TAUCAL added with minor alterations. (JRL)
C   26-JUL-90   - Now caters for all data types.  (JRL)
C   03-SEP-90   - Bugs which confused the FSR of the CAL and OBS cubes
C                 repaired. Also now converts shifts from MPUs to pixels (JRL)
C   18-OCT-90   - Changed to deal with wrap around effect more sensibly. (JRL)
C   19-OCT-90   - Linear interpolation scheme added.  Also fixed sinc routines
C                 so that they deal with negative shifts properly.  (JRL)
C   20-NOV-90   - Info now read from MASK sub-structure.  Also pasted in
C                 new versions of TAUCAL routines. (JRL)
C   03-DEC-90   - Fixed bug in interpolation routines. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+-----------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer        dsa_typesize
      integer        dyn_element
C
C     Local variables
C
      integer        address       ! Dynamic memory address
      logical        badpix        ! Flag for magic values in image file
      logical        badpixg       ! Flag for magic values in grid file
      logical        badpixs       ! Flag for magic values in shifts file
      integer        bytes_float   ! Number of bytes in data type FLOAT
      integer        dims(3)       ! Dimensions of obs cube
      real           dmpu          ! Z axis increment in MPUs
      real           dummy         ! Dummy real variable
      real           dx            ! Offset in Xc
      real           dy            ! Offset in Yc
      real           focus         ! Camera focal length
      real           fscale        ! Pixel size/Focal length
      real           fsr_obs       ! FSR of OBS cube
      integer        gdims(3)      ! Grid file dimensions
      integer        gptr          ! Dynamic memory pointer for grid
      character      grid_lab*80   ! Label for GRID sub-structure
      integer        gslot         ! Dynamic memory slot for grid
      integer        i             ! Integer dummy counter
      integer        ignore        ! Dummy status variable
      integer        iptr          ! Dynamic memory pointer for obs cube
      integer        ishpt         ! Dynamic memory pointer for shifts file
      integer        iwpt1         ! Dynamic memory pointer for workspace
      integer        iwpt2         ! Dynamic memory pointer for workspace
      integer        iwpt3         ! Dynamic memory pointer for workspace
      integer        len_masklab   ! Length of MASK label
      integer        len_gridlab   ! Length of GRID label
      character      mask_lab*80   ! Label for MASK sub-structure
      character      name_grid*80  ! Name of GRID file
      character      name_shifts*80! Name of file with phase map.
      integer        nelm          ! Number of elements in a mapped array
      integer        ndim          ! Number of dimensions in obs cube
      integer        ndimsh        ! Number of dimensions in shifts cube
      integer        ngdim         ! Number of grid file dimensions
      integer        obsrun        ! Run number of observation cube
      integer        optr          ! Dynamic memory pointer for output file
      integer        oslot         ! Dynamic memory slot for output file
      real           pixsize       ! Pixel size defined in MASK
      integer        shdims(2)     ! Shift file dimensions
      integer        shslot        ! Dynamic memory slot for shifts file
      logical        sinc          ! Flag for whether to use sinc interpolation or not
      integer        slot          ! Dynamic memory slot for obs cube
      integer        status        ! DSA and TAUCAL status variable
      logical        tauexist      ! Flag for existence of TAUCAL file
      character      type*8        ! Data type for image file
      character      typeg*8       ! Data type for grid file
      character      types*8       ! Data type for shifts file
      logical        use_grid      ! Flag for existence of grid defined by MASK
      integer        wslot         ! Dynamic memory slot for workspace
      real           xcen          ! X coord of ring centre
      real           ycen          ! Y coord of ring centre
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get run numbers of OBS image
C
      if (tauexist) then
         call par_rdval('obsrun',1.0,float(max_int),1.0,' ',dummy)
         obsrun = nint(dummy)
      end if
C
C     Find out which mask you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,mask_lab,
     : len_masklab,status)
      if (status .ne. 0) go to 500
C
C     Get info from TAUCAL obs run
C
      call tau_cal_rdchr('cal',obsrun,'name_shifts','name_shifts',
     : tauexist,.true.,80,name_shifts,status)
      call tau_cal_rdnum('cal',obsrun,'dx','dx',tauexist,.true.,
     : min_float,max_float,'pixels',1,dx,status)
      call tau_cal_rdnum('cal',obsrun,'dy','dy',tauexist,.true.,
     : min_float,max_float,'pixels',1,dy,status)
      call tau_cal_rdnum('cal',obsrun,'xc','xcen',tauexist,.true.,
     : min_float,max_float,'pixels',1,xcen,status)
      call tau_cal_rdnum('cal',obsrun,'yc','ycen',tauexist,.true.,
     : min_float,max_float,'pixels',1,ycen,status)
      call tau_cal_rdnum('cal',obsrun,'dmpu','dmpu',tauexist,.true.,
     : min_float,max_float,'MPUs/frame',1,dmpu,status)
      call tau_cal_rdnum('cal',obsrun,'fsr','fsr_obs',tauexist,.true.,
     : min_float,max_float,'MPUs',1,fsr_obs,status)
      if (status .ne. 0) go to 500
C
C     Get necessary info from TAUCAL mask structure
C
      call tau_cal_rdnum('cal',0,mask_lab(:len_masklab)//'.pixsize_x',
     : 'pixsize_x',tauexist,.true.,min_float,max_float,'microns',1,
     : pixsize,status)
      if (status .ne. 0) go to 500
C
C     Get camera focus and convert it to microns
C
      call tau_cal_rdnum('cal',0,'camera.foclen','camera_focus',
     : tauexist,.true.,min_float,max_float,'mm',1,focus,status)
      if (status .ne. 0) go to 500
      focus = focus*1000.0
      fscale = pixsize/focus
C
C     Open IMAGE file for observation cube. Not done with TAUCAL as
C     the file must be ZXY sorted...
C
      call dsa_input('image','name_obs',status)
      if (status .ne. 0) go to 500
C
C     Check image dimensions
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',3,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
      if (ndim .ne. 3) then
         call dsa_wruser ('*** Image must be 3-d ***')
         call dsa_wrflush
         go to 500
      end if
C
C     Map IMAGE data array as short if it is so, otherwise map it
C     as float.
C
      if (type .eq. 'SHORT') then
         call dsa_map_data('image','read','short',address,slot,status)
      else
         call dsa_map_data('image','read','float',address,slot,status)
      end if
      if (status .ne. 0) go to 500
      iptr = dyn_element(address)
C
C     Get shifts map...
C
      call dsa_named_input('shifts',name_shifts,status)
      call ndp_get_image_info('shifts',.true.,.false.,types,badpixs,
     : status)
      call dsa_data_size('shifts',2,ndimsh,shdims,nelm,status)
      if (status .ne. 0) go to 500
      if ((shdims(1) .ne. dims(2)) .or. (shdims(2) .ne. dims(3))) then
         call dsa_wruser
     :   ('Input and Shifts data arrays have incompatible sizes')
         call dsa_wrflush
         go to 500
      end if
      call dsa_map_data('shifts','read','float',address,shslot,status)
      ishpt = dyn_element(address)
      if (status .ne. 0) go to 500
C
C     If using grid open it. If not the create a dummy one...
C
      call par_rdkey('use_grid',.true.,use_grid)
      if (use_grid) then
C
C        Find which grid you want...
C
         call tau_cal_find_struct('cal',mask_lab(:len_masklab),'grid',
     :   tauexist,grid_lab,len_gridlab,status)
         if (status .ne. 0) go to 500
C
C        Get the grid file name, open it and check its size
C
         call tau_cal_rdchr('cal',0,
     :   grid_lab(:len_gridlab)//'.name_grid',
     :   'name_grid',tauexist,.true.,80,name_grid,status)
         call dsa_named_input('grid',name_grid,status)
         call ndp_get_image_info('grid',.true.,.false.,typeg,badpixg,
     :   status)
         call dsa_data_size('grid',3,ngdim,gdims,nelm,status)
         if (status .ne. 0) go to 500
         if ((gdims(2) .ne. dims(2)) .or. (gdims(3) .ne. dims(3))) then
            call dsa_wruser
     :      ('Input and Grid data arrays have incompatible sizes')
            call dsa_wrflush
            go to 500
         end if
C
C        Map grid file
C
         call dsa_map_data('grid','read','float',address,gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
C
C     Otherwise create dummy grid
C
      else
         call dsa_get_work_array(2*dims(2)*dims(3),'float',address,
     :   gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
         call tau_dummy_grid(dims(2),dims(3),dynamic_mem(gptr))
      end if
C
C     Decide how to do interpolation.  If the data are SHORT then a sinc
C     may cause rounding problems which leads to a ring pattern in the
C     background.
C
      if (type .eq. 'SHORT') then
         call var_setnum('interp',0,0,1.0,ignore)
      else
         call var_setnum('interp',0,0,2.0,ignore)
      end if
      call dsa_wruser('Which interpolation scheme to use')
      call dsa_wrflush
      call dsa_wruser('(1)  Linear')
      call dsa_wrflush
      call dsa_wruser('(2)  Sinc')
      call dsa_wrflush
      call par_rdval('interp',1.0,2.0,1.0,' ',dummy)
      sinc = nint(dummy) .eq. 2
C
C     Get some workspace
C
      bytes_float = dsa_typesize('float',status)
      call dsa_get_workspace(3*dims(1)*bytes_float,address,wslot,status)
      if (status .ne. 0) go to 500
      iwpt1 = dyn_element(address)
      iwpt2 = iwpt1 + dims(1)*bytes_float
      iwpt3 = iwpt2 + dims(1)*bytes_float
C
C     Open output file
C
      call dsa_output('output','name_cor','image',1,1,status)
      if (type .eq. 'SHORT') then
         call dsa_simple_output('output','data','short',ndim,dims,
     :   status)
      else
         call dsa_simple_output('output','data','float',ndim,dims,
     :   status)
      end if
      do i = 1,3
         call dsa_reshape_axis('output',i,'image',i,1,dims(i),status)
      end do
      if (status .ne. 0) go to 500
C
C     Map output array in the same way as the input array
C
      if (type .eq. 'SHORT') then
         call dsa_map_data('output','write','short',address,oslot,
     :   status)
      else
         call dsa_map_data('output','write','float',address,oslot,
     :   status)
      end if
      if (status .ne. 0) go to 500
      optr = dyn_element(address)
C
C     Now do the work...
C
      call sub_correct(iptr,type,dims(2),dims(3),dims(1),
     : fscale,dmpu,fsr_obs,dynamic_mem(gptr),xcen,ycen,
     : dynamic_mem(ishpt),dx,dy,sinc,dynamic_mem(iwpt1),
     : dynamic_mem(iwpt2),dynamic_mem(iwpt3),optr)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_correct(iptr,type,nx,ny,nz,fscale,dmpu,fsr_obs,
     : coords,xcen,ycen,shifts,dx,dy,sinc,temp1,temp2,temp3,optr)
C+-------------------------------------------------------------------------------
C     SUB_CORRECT    --- Subroutine to do the shifting of the data
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IPTR:       (integer) Input data dynamic memory pointer
C     (>) TYPE:       (character) Input (and output) data type.
C     (>) NX:         (integer) X dimension of data cube.
C     (>) NY:         (integer) Y dimension of data cube.
C     (>) NZ:         (integer) Z dimension of data cube.
C     (>) FSCALE:     (real) Ratio of pixel size to camera focal length.
C     (>) DMPU:       (real) Z axis increment in MPUs
C     (>) FSR_OBS:    (real) On axis FSR of OBS cube.
C     (>) COORDS:     (real array COORDS(2,NX,NY)) Grid of coordinates defined in GRID
C     (>) XCEN:       (real) X centre of the calibration rings.
C     (>) YCEN:       (real) Y centre of the calibration rings.
C     (>) SHIFTS:     (real array SHIFTS(NX,NY)) Amount in Z to shift each spectrum
C                      as defined in PHASE_MAP
C     (>) DX:         (real) Offset in Xc
C     (>) DY:         (real) Offset in Yc
C     (>) SINC:       (logical) Flag for whether to use sinc interpolation or not
C     (W) TEMP1:      (real array TEMP1(NZ)) Workspace
C     (W) TEMP2:      (real array TEMP2(NZ)) Workspace
C     (W) TEMP3:      (real array TEMP3(NZ)) Workspace
C     (>) OPTR:       (integer) Output data dynamic memory pointer
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           23-MAR-1990
C     UPDATE:         26-JUL-1990 -- Changed to cater for all data types (JRL)
C                     03-SEP-1990 -- Bug where FSR of CAL cube was confused
C                                    with FSR of OBS cube was fixed. Also now
C                                    converts shifts from MPUs to pixels before
C                                    shifting.  (JRL)
C                     18-OCT-1990 -- Now deals with wrap around more sensibly. (JRL)
C                     19-OCT-1990 -- Linear interpolation added (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iptr,optr
      real temp1(nz),temp2(nz),fscale,fsr_obs,xcen,ycen,dmpu,
     : shifts(nx,ny),coords(2,nx,ny),dx,dy,temp3(nz)
      logical sinc
      character type*(*)
C
C     Functions
C
      integer ich_len
      logical par_batch
C
C     Local variables
C
      integer i,j,k,ifsr,iphase
      real fscale2,xnew,ynew,r2,sec_theta,fsr_obs_pix,fsr,phase,
     : dphase,pix_mpu,shift_fsr
      character message*80
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Set up a few things
C
      pix_mpu = 1.0/dmpu
      fscale2 = fscale*fscale
      fsr_obs_pix = pix_mpu*fsr_obs
C
C     Loops for each spectrum
C
      do j = 1,ny
         do i = 1,nx
C
C           Copy each spectrum to a real array
C
            if (type .eq. 'SHORT') then
               call transfer_i(dynamic_mem(iptr),nx,ny,nz,i,j,temp1)
            else
               call transfer_r(dynamic_mem(iptr),nx,ny,nz,i,j,temp1)
            end if
C
C           Find FSR at each pixel...
C
            xnew = coords(1,i,j) - (xcen - dx)
            ynew = coords(2,i,j) - (ycen - dy)
            r2 = xnew*xnew + ynew*ynew
            sec_theta = sqrt(r2*fscale2 + 1)
            fsr = abs(fsr_obs_pix*sec_theta)
C
C           Interpolate to take out fractional shift (subtract)
C
            phase = shifts(i,j)*pix_mpu
            iphase = int(phase)
            dphase = phase - float(iphase)
            if (sinc) then
               call sinc_inter(temp1,nz,fsr,dphase,temp2)
            else
               call line_inter(temp1,nz,fsr,dphase,temp2)
            end if
C
C           Make an extra copy of the shifted spectrum
C
            do k = 1,nz
               temp3(k) = temp2(k)
            end do
C
C           Now interpolate the copy to take out fractional FSR shift...
C
            ifsr = int(fsr)
            shift_fsr = -(fsr - float(ifsr))
            if (sinc) then
               call sinc_inter(temp3,nz,fsr,shift_fsr,temp1)
            else
               call line_inter(temp3,nz,fsr,shift_fsr,temp1)
            end if
C
C           Now, temp2 has the spectrum shifted by the fractional bit of
C           the phase and temp1 has the spectrum shifted by the fractional bit
C           of the FSR. Time to subtract out the shift in whole pixels.
C
            if (type .eq. 'SHORT') then
               call subtract_i(temp2,temp1,nx,ny,nz,i,j,iphase,ifsr,
     :         dynamic_mem(optr))
            else
               call subtract_r(temp2,temp1,nx,ny,nz,i,j,iphase,ifsr,
     :         dynamic_mem(optr))
            end if
         end do
C
C        A little progress report to reassure the user...
C
         if ((mod(j,10) .eq. 0) .and. (.not. par_batch())) then
            write(message,100) j
  100       format(1x,i8,' columns done...')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
         end if
      end do
      end


      subroutine line_inter(datin,nz,fsr,phi,datout)
C+------------------------------------------------------------------------------------
C     LINE_INTER     --- Subroutine to subtract out shifts of a fraction of a
C                        pixel with a linear interpolation.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATIN:      (real array DATIN(NZ)) Array containing the unshifted spectrum
C     (>) NZ:         (integer) Dimension of spectrum
C     (>) FSR:        (real) Free spectral range of the spectrum
C     (>) PHI:        (real) Amount by which to shift the spectrum
C     (<) DATOUT:     (real array DATOUT(NZ)) Array containing the shifted spectrum
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           19-OCT-1990
C     UPDATE:         03-DEC-1990 -- Bug fixed where wrong weights were assigned to
C                                    wrap around pixels. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nz
      real datin(nz),datout(nz),fsr,phi
C
C     Functions
C
      real line_sample
C
C     Local variables
C
      integer ifsr,i
      real dat0,datp1,w(3),dfsr
C
C     Get fraction of current FSR
C
      ifsr = int(fsr)
      dfsr = fsr - float(ifsr)
C
C     Create array of weights
C
      w(1) = line_sample(1+dfsr)
      w(2) = line_sample(dfsr)
      w(3) = line_sample(1.0-dfsr)
C
C     Create a few extra points at the beginning and end of the array
C     to take care of wrap-around effects. (Weights go backwards for
C     the positive extension as this is essentially a negative shift).
C
      dat0 = datin(ifsr-1)*w(1) + datin(ifsr)*w(2) + datin(ifsr+1)*w(3)
      datp1 = datin(nz-ifsr)*w(3) + datin(nz-ifsr+1)*w(2) +
     : datin(nz-ifsr+2)*w(1)
C
C     Create array of weights
C
      w(1) = line_sample(1.0+phi)
      w(2) = line_sample(phi)
      w(3) = line_sample(1.0-phi)
C
C     Do the interpolation for the first nz-1 points
C
      datout(1) = dat0*w(1) + datin(1)*w(2) + datin(2)*w(3)
      do i = 2,nz-1
         datout(i) = datin(i-1)*w(1) + datin(i)*w(2) + datin(i+1)*w(3)
      end do
C
C     Now for the last one...
C
      datout(nz) = datin(nz-1)*w(1) + datin(nz)*w(2) + datp1*w(3)
      end


      real function line_sample(x)
C+------------------------------------------------------------------------------------
C     LINE_SAMPLE    --- Function to calculate the weights for the linear interpolation.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:          (real) Argument of the function
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           19-OCT-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real x
C
C     Local variables
C
      real y
      y = abs(x)
      if (y .ge. 1.0) then
         line_sample = 0.0
      else
         line_sample = 1.0 - y
      end if
      end



      subroutine sinc_inter(datin,nz,fsr,phi,datout)
C+------------------------------------------------------------------------------------
C     SINC_INTER     --- Subroutine to subtract out shifts of a fraction of a
C                        pixel with a sinc interpolation.  The basic algorithm was
C                        stolen from Dave Pike's routine SLIDE.  It has been
C                        modified to make it more efficient and so that it
C                        subtracts rather than adds.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATIN:      (real array DATIN(NZ)) Array containing the unshifted spectrum
C     (>) NZ:         (integer) Dimension of spectrum
C     (>) FSR:        (real) Free spectral range of the spectrum
C     (>) PHI:        (real) Amount by which to shift the spectrum
C     (<) DATOUT:     (real array DATOUT(NZ)) Array containing the shifted spectrum
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           23-MAR-1990
C     UPDATE:         19-OCT-1990 -- Modified so that it handles negative shifts
C                                    properly.  (JRL)
C                     03-DEC-1990 -- Bug fixed where wrong weights were assigned to
C                                    wrap around pixels. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nz
      real datin(nz),datout(nz),fsr,phi
C
C     Functions
C
      real sample
C
C     Local variables
C
      integer ifsr,i,n
      real datm1,dat0,datp1,datp2,w(5),dfsr
C
C     Get fraction of current FSR
C
      ifsr = int(fsr)
      dfsr = fsr - float(ifsr)
C
C     Create array of weights
C
      n = 0
      do i = 2,-2,-1
         n = n + 1
         w(n) = sample(float(i)+dfsr)
      end do
C
C     Create a few extra points at the beginning and end of the array
C     to take care of wrap-around effects. Weights go backwards for positive
C     extensions as this is essentially a negative shift.
C
      datm1 = datin(ifsr-3)*w(1) + datin(ifsr-2)*w(2)
     : + datin(ifsr-1)*w(3) + datin(ifsr)*w(4) + datin(ifsr+1)*w(5)
      dat0 = datin(ifsr-2)*w(1) + datin(ifsr-1)*w(2)
     : + datin(ifsr)*w(3) + datin(ifsr+1)*w(4) + datin(ifsr+2)*w(5)
      datp1 = datin(nz-ifsr-1)*w(5) + datin(nz-ifsr)*w(4)
     : + datin(nz-ifsr+1)*w(3) + datin(nz-ifsr+2)*w(2)
     : + datin(nz-ifsr+3)*w(1)
      datp2 = datin(nz-ifsr)*w(5) + datin(nz-ifsr+1)*w(4)
     : + datin(nz-ifsr+2)*w(3) + datin(nz-ifsr+3)*w(2)
     : + datin(nz-ifsr+4)*w(1)
C
C     Create array of weights
C
      n = 0
      do i = 2,-2,-1
         n = n + 1
         w(n) = sample(float(i)+phi)
      end do
C
C     Do the interpolation for the first nz-2 points
C
      datout(1) = datm1*w(1) + dat0*w(2) + datin(1)*w(3) + datin(2)*w(4)
     : + datin(3)*w(5)
      datout(2) = dat0*w(1) + datin(1)*w(2) + datin(2)*w(3) +
     : datin(3)*w(4) + datin(4)*w(5)
      do i = 3,nz-2
         datout(i) = datin(i-2)*w(1) + datin(i-1)*w(2) + datin(i)*w(3)
     :   + datin(i+1)*w(4) + datin(i+2)*w(5)
      end do
C
C     Now for the last few...
C
      datout(nz-1) = datin(nz-3)*w(1) + datin(nz-2)*w(2)
     : + datin(nz-1)*w(3) + datin(nz)*w(4) + datp1*w(5)
      datout(nz) = datin(nz-2)*w(1) + datin(nz-1)*w(2) + datin(nz)*w(3)
     : + datp1*w(4)+ datp2*w(5)
      end



      real function sample(x)
C+------------------------------------------------------------------------------------
C     SAMPLE         --- Function to calculate the weights for the sinc interpolation.
C                        Basic algorithm lifted from Dave Pike's function SAMPLE, but
C                        again modified for efficiency.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:          (real) Argument of the function
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           23-MAR-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real x
C
C     Local variables
C
      real y
      y = abs(x)
      if (y .lt. 1.0) then
         sample = 1.0 - 2.0*y*y + y*y*y
      else if ((y .ge. 1.0) .and. (y .lt. 2.0)) then
         sample = 4.0 - 8.0*y + 5.0*y*y - y*y*y
      else
         sample = 0.0
      end if
      end



      subroutine transfer_i(data,nx,ny,nz,i,j,temp)
C+-------------------------------------------------------------------------------
C     TRANSFER_I     --- Subroutine to copy and convert individual spectra from
C                        the input cube to the output vector.
C
C     Parameters (">" Input, "<" Output)
C
C     (>) DATA:       (short array DATA(NZ,NX,NY)) Input data cube
C     (>) NX:         (integer) X dimension of obs cube.
C     (>) NY:         (integer) Y dimension of obs cube.
C     (>) NZ:         (integer) Z dimension of obs cube.
C     (>) I:          (integer) X position of spectrum to be copied
C     (>) J:          (integer) Y position of spectrum to be copied
C     (<) TEMP:       (real array TEMP(NZ)) Output spectrum
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,i,j
      integer*2 data(nz,nx,ny)
      real temp(nz)
C
C     Local variables
C
      integer k
C
C     Copy it over...
C
      do k = 1,nz
         temp(k) = float(data(k,i,j))
      end do
      end


      subroutine transfer_r(data,nx,ny,nz,i,j,temp)
C+-------------------------------------------------------------------------------
C     TRANSFER_R     --- Subroutine to copy individual spectra from
C                        the input cube to the output vector.
C
C     Parameters (">" Input, "<" Output)
C
C     (>) DATA:       (real array DATA(NZ,NX,NY)) Input data cube
C     (>) NX:         (integer) X dimension of obs cube.
C     (>) NY:         (integer) Y dimension of obs cube.
C     (>) NZ:         (integer) Z dimension of obs cube.
C     (>) I:          (integer) X position of spectrum to be copied
C     (>) J:          (integer) Y position of spectrum to be copied
C     (<) TEMP:       (real array TEMP(NZ)) Output spectrum
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,i,j
      real temp(nz),data(nz,nx,ny)
C
C     Local variables
C
      integer k
C
C     Copy it over...
C
      do k = 1,nz
         temp(k) = data(k,i,j)
      end do
      end



      subroutine subtract_i(temp2,temp1,nx,ny,nz,i,j,iphase,ifsr,output)
C+-------------------------------------------------------------------------------
C     SUBTRACT_I     --- Subroutine to subtract the integral part of the shifts,
C                        transfer the data to the output cube and convert it.
C
C     Parameters (">" Input, "<" Output)
C
C     (>) TEMP2:      (real array TEMP2(NZ)) Input spectrum shifted by fractional phase
C     (>) TEMP1:      (real array TEMP1(NZ)) Input spectrum shifted by fractional FSR
C     (>) NX:         (integer) X dimension of obs cube.
C     (>) NY:         (integer) Y dimension of obs cube.
C     (>) NZ:         (integer) Z dimension of obs cube.
C     (>) I:          (integer) X position of spectrum to be copied
C     (>) J:          (integer) Y position of spectrum to be copied
C     (>) IPHASE:     (integer) Number of Z pixels by which to be shifted
C     (>) IFSR:       (integer) Free spectral range as an integer
C     (>) OUTPUT:     (short array OUTPUT(NZ,NX,NY)) Output data cube
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:         18-OCT-1990 -- Modified to cope with wrap around better. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,i,j,iphase,ifsr
      integer*2 output(nz,nx,ny)
      real temp2(nz),temp1(nz)
C
C     Local variables
C
      integer itarg,k
C
C     Loop through the output spectrum, find the corresponding
C     pixel in the input spectrum and copy it
C
      do itarg = 1,nz
         k = itarg + iphase
         if ((k .ge. 1) .and. (k .le. nz)) then
            output(itarg,i,j) = nint(temp2(k))
         else if (k .lt. 1) then
            k = k + ifsr
            output(itarg,i,j) = nint(temp1(k))
         else if (k .gt. nz) then
            k = k - ifsr
            output(itarg,i,j) = nint(temp1(k))
         end if
      end do
      end



      subroutine subtract_r(temp2,temp1,nx,ny,nz,i,j,iphase,ifsr,output)
C+-------------------------------------------------------------------------------
C     SUBTRACT_R     --- Subroutine to subtract the integral part of the shifts
C                        and transfer the data to the output cube.
C
C     Parameters (">" Input, "<" Output)
C
C     (>) TEMP2:      (real array TEMP2(NZ)) Input spectrum shifted by fractional phase
C     (>) TEMP1:      (real array TEMP1(NZ)) Input spectrum shifted by fractional FSR
C     (>) NX:         (integer) X dimension of obs cube.
C     (>) NY:         (integer) Y dimension of obs cube.
C     (>) NZ:         (integer) Z dimension of obs cube.
C     (>) I:          (integer) X position of spectrum to be copied
C     (>) J:          (integer) Y position of spectrum to be copied
C     (>) IPHASE:     (integer) Number of Z pixels by which to be shifted
C     (>) IFSR:       (integer) Free spectral range as an integer
C     (>) OUTPUT:     (real array OUTPUT(NZ,NX,NY)) Output data cube
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:         18-OCT-1990 -- Modified to cope with wrap around better. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,i,j,iphase,ifsr
      real temp2(nz),temp1(nz),output(nz,nx,ny)
C
C     Local variables
C
      integer itarg,k
C
C     Loop through the output spectrum, find the corresponding
C     pixel in the input spectrum and copy it
C
      do itarg = 1,nz
         k = itarg + iphase
         if ((k .ge. 1) .and. (k .le. nz)) then
            output(itarg,i,j) = temp2(k)
         else if (k .lt. 1) then
            k = k + ifsr
            output(itarg,i,j) = temp1(k)
         else if (k .gt. nz) then
            k = k - ifsr
            output(itarg,i,j) = temp1(k)
         end if
      end do
      end
