      subroutine cre_rv
C+------------------------------------------------------------------------------
C
C   -----------
C   C R E _ R V
C   -----------
C
C   Description
C   -----------
C   Utility program to create estimate RVT and RVFs from values
C   in TAUCAL file.
C
C
C   Scope of program
C   ----------------
C   - Dimensionality irrelevant
C   - Data type irrelevant
C   - Quality and error arrays irrelevant
C   - Magic values irrelevant
C   - Batch execution not supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE      Name of TAUCAL file (character)(prompted for)
C
C   WHICH_RVT    Which of the available RVTs to use. (integer)(prompted for)
C
C   WHICH_RVF    Which of the available RVFs to use. (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C   RVF          True if the output structure is an RVF. False if it's an RVT.
C
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (CAL)        NZ            Number of frames in the RVF cube
C
C   (CAL)        MPU0          Z axis zeropoint in microprocessor units. (real)
C                              (read)
C
C   (CAL)        DMPU          Z axis increment in microprocessor units/frame
C                              (real)(read)
C
C   .RVF         NRVF          Number of RVFs which exist in TAUCAL. (integer)(read)
C
C   .RVF.RVFn    NSER          Number of series nominated in the cube. (integer)
C                              (read)
C
C   .RVF.RVFn    SER_NOS       The delimiting frame numbers for each series
C                              (integer[2,nser])(read)
C
C   .RVF.RVFn    NAME_RVF      Name of the output RVF file. (character)(read)
C
C   .RVF.RVFn    RING_PARS     Estimated parameters for the rings.
C                              (real[10,2*NSER])(read)
C
C   .RVF.RVFn    RINGRUN       Run number associated with ring data cube.
C                              (integer)(prompted for).
C
C   .RVT         NRVT          Number of RVTs in TAUCAL (integer)(read)
C
C   .RVT.RVTn    NRUNS         Number of runs included in this RVT. (integer)
C                              (read)
C
C   .RVT.RVTn    RUNS          Run numbers included in this RVT.
C                             (integer[NRUNS])(read)
C
C   .RVT.RVTn    NAME_RVT      Name of the output RVT file. (character)(read)
C
C   .RVT.RVTn    RING_PARS     Estimated parameters for the rings. (real[10])
C                              (read)
C
C
C   Propagation of data structure
C   -----------------------------
C   - RVT or RVF of standard format is created.
C
C
C   Method
C   ------
C   - The necessary information is read from the TAUCAL file
C   - For RVFs a pseudo-RVF is set up and the new file is constructed
C     in a similar way to PRERING.
C   - For RVTs the information is copied directly from TAUCAL file
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_CLOSE_STRUCTURE
C     DSA_DATA_SIZE
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_NAMED_OUTPUT
C     DSA_OPEN
C     DSA_SIMPLE_OUTPUT
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library PAR:
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C
C
C   Internal subroutines called
C   ---------------------------
C   SUB_CRE_RVF
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   07-NOV-90   - Original program
C   16-NOV-90   - Now reads RVF data from RVF substructure (JRL)
C   20-NOV-90   - New TAUCAL routines pasted in.  (JRL)
C   11-DEC-90   - Changed to deal with new version of storing info
C                 for RVT. (JRL)
C   21-FEB-91   - Slight change so that parameter NZ is read from TAUCAL. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
C
C     Local variables
C
      integer      address      ! Dynamic memory address
      integer      bytes_float  ! Number of bytes in FLOAT type
      integer      dims(2)      ! Dimension of output file
      real         dmpu         ! Z axis increment of cal cube
      real         dummy        ! Real dummy variable
      integer      len_rvflab   ! RVF sub-structure label length
      integer      len_rvtlab   ! RVT sub-structure label length
      real         mpu0         ! Z axis zeropoint of cal cube
      character    name_rvf*80  ! Name of RVF file
      character    name_rvt*80  ! Name of RVT file
      integer      nruns        ! Number of runs in RVT
      integer      nser         ! Number of series in the RVF
      integer      nz           ! Dimension of Z axis of CAL cube
      integer      optr         ! Dynamic memory pointer for output file
      integer      oslot        ! Dynamic memory slot for output file
      integer      ringrun      ! Run number for cal cube
      logical      rvf          ! True if the run is an RVF
      character    rvf_lab*80   ! RVF substructure label
      character    rvt_lab*80   ! RVT substructure label
      integer      status       ! DSA and TAU status variable
      logical      tauexist     ! Flag for existence of TAUCAL file
      integer      totspace     ! Total amount of space for workspace allocation
      integer      wpt1         ! Dynamic memory pointer for workspace
      integer      wpt2         ! Dynamic memory pointer for workspace
      integer      wslot1       ! Dynamic memory slot for workspace
      integer      wslot2       ! Dynamic memory slot for workspace
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
      if (.not. tauexist) then
         call dsa_wruser('You can''t run this without a TAUCAL')
         call dsa_wrflush
         go to 500
      end if
C
C     Is it an RVF or an RVT you want to create?
C
      call par_rdkey('rvf',.true.,rvf)
C
C     Section for creating an RVF
C
      if (rvf) then
C
C        Which RVF do you want?
C
         call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     :   len_rvflab,status)
         if (status .ne. 0) go to 500
C
C        Find the run number associated with this RVF
C
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.run',
     :   'ringrun',tauexist,.false.,0.0,float(max_int),' ',1,dummy,
     :   status)
         ringrun = nint(dummy)
         if (status .ne. 0) go to 500
C
C        Get the z axis zeropoint and increment in MPUs
C
         call tau_cal_rdnum('cal',ringrun,'mpu0','mpu0',tauexist,
     :   .true.,min_float,max_float,'MPUs',1,mpu0,status)
         call tau_cal_rdnum('cal',ringrun,'dmpu','dmpu',tauexist,
     :   .true.,min_float,max_float,'MPUs/frame',1,dmpu,status)
C
C        Read the number of series...
C
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.nser',
     :   'nser',tauexist,.true.,0.0,float(max_int),' ',1,dummy,status)
         nser = nint(dummy)
         if (nser .eq. 0) then
            call dsa_wruser('No series defined for this run number')
            call dsa_wrflush
            go to 500
         end if
C
C        Get some workspace and put the series numbers in it
C
         bytes_float = dsa_typesize('float',status)
         totspace = nser*2*bytes_float
         call dsa_get_workspace(totspace,address,wslot1,status)
         if (status .ne. 0) then
            call dsa_wruser('Couldn''t get workspace...')
            call dsa_wrflush
            go to 500
         end if
         wpt1 = dyn_element(address)
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.ser_nos',
     :   'ser_nos',tauexist,.true.,0.0,float(max_int),' ',nser*2,
     :   dynamic_mem(wpt1),status)
         if (status .ne. 0) go to 500
C
C        Get the number of frames in original CAL cube
C
         call tau_cal_rdnum('cal',ringrun,'nz','nz',tauexist,
     :   .true.,0.0,float(max_int),' ',1,dummy,status)
         nz = nint(dummy)
         if (status .ne. 0) go to 500
C
C        Get name of RVF and open the file
C
         call tau_cal_rdchr('cal',0,rvf_lab(:len_rvflab)//'.name_rvf',
     :   'name_rvf',tauexist,.true.,80,name_rvf,status)
         call dsa_named_output('output',name_rvf,' ',1,1,status)
         dims(1) = 10
         dims(2) = nz
         call dsa_simple_output('output','data','float',2,dims,status)
C
C        Map the output file
C
         call dsa_map_data('output','write','float',address,oslot,
     :   status)
         if (status .ne. 0) go to 500
         optr = dyn_element(address)
C
C        Now get some workspace and read the ring parameters into it...
C
         totspace = nser*20*bytes_float
         call dsa_get_workspace(totspace,address,wslot2,status)
         if (status .ne. 0) then
             call dsa_wruser('Couldn''t get workspace...')
             call dsa_wrflush
             go to 500
         end if
         wpt2 = dyn_element(address)
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.ring_pars',
     :   'ring_pars',tauexist,.true.,min_float,max_float,' ',nser*20,
     :   dynamic_mem(wpt2),status)
         if (status .ne. 0) go to 500
C
C        Right...Now we're ready to create the file put the data into the file...
C
         call sub_cre_rvf(10,nz,nser,dynamic_mem(wpt1),
     :   dynamic_mem(wpt2),mpu0,dmpu,dynamic_mem(optr))
C
C     Section for creating an RVT
C
      else
C
C        Find out which RVTs to use...
C
         call tau_cal_find_struct('cal','rvt','rvt',tauexist,rvt_lab,
     :   len_rvtlab,status)
         if (status .ne. 0) go to 500
C
C        Find out how many runs were included in this RVT
C
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.nruns',
     :   'nruns',tauexist,.true.,0.0,float(max_int),' ',1,dummy,status)
         if (status .ne. 0) go to 500
         nruns = nint(dummy)
         if (nruns .eq. 0) then
            call dsa_wruser('No runs defined for this RVT')
            call dsa_wrflush
            go to 500
         end if
C
C        Get name of output file and open it...
C
         call tau_cal_rdchr('cal',0,rvt_lab(:len_rvtlab)//'.name_rvt',
     :   name_rvt,tauexist,.true.,80,name_rvt,status)
         call dsa_named_output('output',name_rvt,' ',1,1,status)
         dims(1) = 10
         dims(2) = nruns
         call dsa_simple_output('output','data','float',2,dims,
     :   status)
         if (status .ne. 0) go to 500
C
C        Map the output file
C
         call dsa_map_data('output','write','float',address,
     :   oslot,status)
         if (status .ne. 0) go to 500
         optr = dyn_element(address)
C
C        Read the ring parameters straight into the output file...
C
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.ring_pars',
     :   'ring_pars',tauexist,.true.,min_float,max_float,' ',nruns*10,
     :   dynamic_mem(optr),status)
         if (status .ne. 0) go to 500
      end if
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_cre_rvf(nx,nz,nser,ser_nos,ring_pars,mpu0,dmpu,rvf)
C+--------------------------------------------------------------------------------------
C     SUB_CRE_RVF --- Routine to create an RVF input file for RINGS from
C                     TAUCAL information.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NX:        (integer) First dimension of RVF
C     (>) NZ:        (integer) Second dimension of RVF
C     (>) NSER:      (integer) Number of series considered in PRERING
C     (>) SER_NOS:   (real array SER_NOS(2,NSER)) Delimiters for each series
C     (>) RING_PARS: (real array RING_PARS(NX,2,NSER)) Ring parameters for SER_NOS.
C     (>) MPU0:      (real) Z axis zeropoint in MPUs
C     (>) DMPU:      (real) Z axis increment in MPUs/Frame
C     (<) RVF:       (real array RVF(NX,NY)) Output RVF array
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        07-NOV-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,nz,nser
      real ser_nos(2,nser),ring_pars(nx,2,nser),rvf(nx,nz),mpu0,dmpu
C
C     Local variables
C
      integer i,j,iser1,iser2
      real r1,r2,s1,s2,p1,p2,r_sig_const,peak_const,slope,aint,
     : rr,rad,sig,aproc1,aproc2
C
C     Initialise the whole RVF array
C
      do i = 1,nz
         rvf(1,i) = mpu0 + float(i-1)*dmpu
         do j = 2,nx-1
            rvf(j,i) = 0.0
         end do
         rvf(nx,i) = 2.0
      end do
C
C     Loop for the total number of series
C
      do i = 1,nser
C
C        First get the series delimiters
C
         iser1 = nint(ser_nos(1,i))
         iser2 = nint(ser_nos(2,i))
C
C        Set up the constants...
C
         aproc1 = ring_pars(1,1,i)
         aproc2 = ring_pars(1,2,i)
         r1 = ring_pars(4,1,i)
         r2 = ring_pars(4,2,i)
         s1 = ring_pars(5,1,i)
         s2 = ring_pars(5,2,i)
         p1 = ring_pars(6,1,i)
         p2 = ring_pars(6,2,i)
         r_sig_const = 0.5*(r1*s1 + r2*s2)
         peak_const = 0.5*(p1 + p2)
C
C        Calculate pseudo RVF
C
         slope = (r2*r2 - r1*r1)/(aproc2 - aproc1)
         aint = r1*r1 - slope*aproc1
C
C        Now do the estimates for each ring in the series
C
         do j = iser1,iser2
C
C           First set up flag
C
            rvf(10,j) = -1.0
C
C           Xc and Yc.  These are averages already, so don't bother
C           averaging again...
C
            rvf(2,j) = ring_pars(2,1,i)
            rvf(3,j) = ring_pars(3,1,i)
C
C           Radius vs. frame
C
            rr = slope*float(j) + aint
            if (rr .le. 0.0) then
               rvf(4,j) = 0.0
               rvf(10,j) = 2.0
            else
               rvf(4,j) = sqrt(rr)
            end if
            rad = rvf(4,j)
C
C           Sigma and peak
C
            if (rad .ne. 0.0) then
               sig = r_sig_const/rad
            else
               sig = 0.0
            end if
            rvf(5,j) = sig
            rvf(6,j) = peak_const
C
C           Background, ellipticity and error
C
            rvf(7,j) = ring_pars(7,1,i)
            rvf(8,j) = 1.0
            rvf(9,j) = 0.0
         end do
      end do
      end
