      subroutine exam_rvf
C+------------------------------------------------------------------------------
C
C   ---------------
C   E X A M _ R V F
C   ---------------
C
C   Description
C   -----------
C   Examines the RVF array and calculates the free spectral range,
C   the etalon gap and the mean centre of the rings.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - FLOAT type only supported.
C   - Error arrays supported.
C   - Quality arrays not supported
C   - Magic values not supported.
C   - Batch execution not supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   WHICH_RVF   User's choice of RVF from TAUCAL (integer)(prompted for)
C
C   WHICH_MASK  User's choice of MASK from TAUCAL (integer)(prompted for)
C
C
C   Keywords (read from VARS.DST)
C   -----------------------------
C   USE_TAU     Flag for use of TAUCAL file. (prompted for)
C
C   SAME_SLOPE  Instruction to force all fits to have the same slope.
C               (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .RVF         NRVF          Number of RVFs already defined in TAUCAL.
C                             (integer)(read)
C
C   .RVF.RVFn    NAME_RVF      Name of output RVF file (character)(read)
C
C   .RVF.RVFn    WAVE_CAL      Wavelength of the calibration source. (real)
C                              (read)
C
C   .RVF.RVFn    NSER          Number of series in the cube. (integer)(read)
C
C   .CAMERA      CAMERA_FOCUS  Focal length of the camera used in mm. (real)
C                              (read)
C
C   .MASK        NMASK         Number of MASKs defined in TAUCAL. (integer)
C                              (read)
C
C   .MASK.MASKn  PIXSIZE_X     Pixel size as defined by MASK. (real)(read)
C
C   .MASK.MASKn  FSR           Free spectral range in MPUs. (real)(written)
C
C   .MASK.MASKn  ERRFSR        Error in FSR in MPUs. (real)(written)
C
C   .MASK.MASKn  GAP           Etalon gap in microns. (real)(written)
C
C   .MASK.MASKn  ERRGAP        Error in etalon gap in microns. (real)(written)
C
C   .MASK.MASKn  XC            Mean X centre of the rings used. (real)(written)
C
C   .MASK.MASKn  ERRXC         Error in mean X centre. (real)(written)
C
C   .MASK.MASKn  YC            Mean Y centre of the rings used. (real)(written)
C
C   .MASK.MASKn  ERRYC         Error in mean Y centre. (real)(written)
C
C   .MASK.MASKn  CENPHASE      Central phase of CAL cube. (real)(written)
C
C   .MASK.MASKn  ERRCENPH      Error in central phase of CAL cube. (real)
C                              (written)
C
C   .MASK.MASKn  NSERFIT       Number of series fit. (integer)(written)
C
C   .MASK.MASKn  SERIESn       Series number, slope, error_slope, intercept,
C                              err_int for series #n. (real[5])(written)
C
C   .MASK.MASKn  MODE          Mode of operation flag. =-1 if slopes are < 0.
C                             (integer)(written)
C
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested to see if it is 2-dimensional.
C   - The current display device name is read from the parameter file.
C   - If the user is constraining the slopes to all be the same, then
C     he/she/it is asked which series will be the fiducial one and is
C     asked to mark out the limits of that series on a plot of R**2 vs Z.
C   - User has a choice of fitting both the good points and the
C     dodgy ones (the latter are given half weight) or just the
C     good ones.
C   - A line is fitted.
C   - The user can then change the weighting of individual points if
C     he/she/it so wishes.  The points can be zeroed out, half weighted
C     or full weighted.
C   - A new line is fitted. This changing and refitting is repeated until
C     the user is happy with the fit
C   - The user then repeats the same procedure for every other series.  If
C     he/she/it hasn't constrained the slopes, then it just skips to this
C     part.
C   - The user is given one more chance to throw away the fits just done
C     and start all over again.
C   - Once the fits have been defined, then the mean free spectral range and
C     the mean gap are calculated and written out.
C   - The mean centre of all the used rings is also calculated and written out
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_LU
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_MAP_ERRORS
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_CI
C     ICH_FOLD
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_QUEST
C     PAR_QNUM
C     PAR_RDKEY
C     PAR_RDVAL
C
C   PGPLOT:
C     PGBEGIN
C     PGCURSE
C     PGEND
C     PGENV
C     PGLABEL
C     PGLINE
C     PGPOINT
C     PGSCI
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETNUM
C
C   Library VAR:
C     VAR_GETCHR
C
C   Internal subroutines called
C   ---------------------------
C     COLLECT
C     FITLINE
C     FITSAME
C     MARK_LIMITS
C     PLOTLINE
C     PLOT_POINTS
C     PROCESS_TAUCAL
C     SEARCH
C     SUB_EXAM_RVF
C     WIPE_POINT
C     ZAP
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C   A version which doesn't rely on plotting would be nice.
C   Also a version which would allow several different wavelength
C   lines would be nice.
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL  or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   02-MAR-1990   - Original program
C   03-APR-1990   - Option added to constrain both slopes to be the same. (JRL)
C   23-MAY-1990   - Overhauled so that more than 2 series can be done.  Added
C                   in calculation of mean centre coordinates.  Also new and
C                   simpler fitting routine is used. (JRL)
C   03-JUL-1990   - TAUCAL system pasted in.  A few modifications were necessary
C                   for this. This mostly included passing results back to the
C                   main routine.  Nothing difficult...(JRL)
C   25-JUL-1990   - Mode of operation variable MODE added in to accommodate the
C                   idea that the FSR and GAP should both always be positive
C                   and the somewhat inverse nature of TAURUS 1 data.  (JRL)
C   13-NOV-1990   - Now takes RVF error array into account.  (JRL)
C   16-NOV-1990   - Changed to read from and write to RVF and MASK
C                   sub-structures. (JRL)
C   20-NOV-1990   - New versions of TAUCAL routines pasted in. (JRL)
C   21-FEB-1991   - Now erases old points and old fits before plotting new
C                   ones. (JRL)
C   03-SEP-1991   - Now checks to make sure there are at least 2 good points
C                   between the limits chosen with the cursor. (JRL)
C   31-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions.
C
      integer          dsa_typesize
      integer          dyn_element
C
C     Local variables.
C
      integer          address    ! Dynamic memory address
      logical          badpix     ! Flag for magic values
      integer          bytes_float! Number of bytes in float data type
      integer          bytes_int  ! Number of bytes in integer data type
      integer          cenphase(2)! Central phase and error
      integer          dims(2)    ! Dimensions of RVF file
      real             dummy      ! Dummy real variable
      integer          eptr       ! Dynamic memory pointer
      integer          eslot      ! Dynamic memory slot
      real             errfsr     ! Error in calculated free spectral range
      real             errgap     ! Error in calculated etalon spacing
      real             errxc      ! Error in X centre
      real             erryc      ! Error in Y centre
      real             focus      ! Camera focal length
      real             fscale     ! Ratio of pixel size to focal length
      real             fsrcen     ! Calculated on-axis free spectral range
      real             gap        ! Calculated etalon spacing
      integer          iptr       ! Dynamic memory pointer
      integer          islot      ! Dynamic memory slot
      integer          iwpt1      ! Dynamic memory pointer (workspace)
      integer          iwpt2      ! Dynamic memory pointer (workspace)
      integer          iwpt3      ! Dynamic memory pointer (workspace)
      integer          iwpt4      ! Dynamic memory pointer (workspace)
      integer          iwpt5      ! Dynamic memory pointer (workspace)
      integer          iwpt6      ! Dynamic memory pointer (workspace)
      integer          iwpt7      ! Dynamic memory pointer (workspace)
      integer          iwpt8      ! Dynamic memory pointer (workspace)
      integer          iwpt9      ! Dynamic memory pointer (workspace)
      integer          iwpt10     ! Dynamic memory pointer (workspace)
      integer          iwpt11     ! Dynamic memory pointer (workspace)
      integer          iwpt12     ! Dynamic memory pointer (workspace)
      integer          iwpt13     ! Dynamic memory pointer (workspace)
      integer          iwslt      ! Dynamic memory slot (workspace)
      integer          iwslt2     ! Dynamic memory slot (workspace)
      integer          len_masklab! Length of MASK label
      integer          len_rvflab ! Length of RVF label
      character        mask_lab*80! Label for MASK sub-structure
      integer          mode       ! Mode of operation variable.
      character        name_rvf*80! Name of RVF file
      integer          ndim       ! Number of dimensions in RVF file
      integer          nelm       ! Number of elements in RVF file
      integer          nsermax    ! Maximum number of series to be fit
      integer          nsfit      ! Number of series actually fit
      real             pixsize    ! Pixel size defined by MASK
      character        rvf_lab*80 ! Label for RVF sub-structure
      logical          same_slope ! Flag to constrain both slopes to be the same
      integer          status     ! DSA and TAUCAL status flag
      logical          tauexist   ! Flag for existence of TAUCAL file
      integer          totsize    ! Total amount of memory to be allocated
      character        type*8     ! RVF data type
      real             wave       ! Wavelength of calibration source
      real             xc         ! Mean X centre
      real             yc         ! Mean Y centre
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     If using TAUCAL, then find out how many RVFs are on offer...
C
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      if (status .ne. 0) go to 500
C
C     Open IMAGE file for radius-vs-frame data.
C
      call tau_cal_rdchr('cal',0,rvf_lab(:len_rvflab)//'.name_rvf',
     : 'name_rvf',tauexist,.true.,80,name_rvf,status)
      if (status .ne. 0) then
         call dsa_wruser('No RVF file found for this run')
         call dsa_wrflush
         go to 500
      end if
      call dsa_named_input('image',name_rvf,status)
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',2,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
      if (ndim .lt. 2) then
         call dsa_wruser('This is not a 2-D image.')
         call dsa_wrflush
         go to 500
      end if
C
C     Get parameters.
C
      call tau_cal_rdnum('cal',0,'camera.foclen','camera_focus',
     : tauexist,.true.,min_float,max_float,'mm',1,focus,status)
      focus = focus*1000.0
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.wave',
     : 'wave_cal',tauexist,.true.,min_float,max_float,'microns',1,wave,
     : status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.nser','nser',
     : tauexist,.true.,float(min_int),float(max_int),' ',1,dummy,status)
      nsermax = nint(dummy)
C
C     If using TAUCAL, then find out how many MASKs are on offer...
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,mask_lab,
     : len_masklab,status)
      if (status .ne. 0) go to 500
C
C     Get a few more parameters
C
      call tau_cal_rdnum('cal',0,mask_lab(:len_masklab)//'.pixsize_x',
     : 'pixsize_x',tauexist,.true.,min_float,max_float,'microns',1,
     : pixsize,status)
      if (status .ne. 0) go to 500
      fscale = pixsize/focus
      call par_rdkey('same_slope',.false.,same_slope)
C
C     Map RVF data array.
C
      call dsa_map_data('image','read','float',address,islot,status)
      if (status .ne. 0) go to 500
      iptr = dyn_element(address)
C
C     Map RVF error array
C
      call dsa_map_errors('image','read','float',address,eslot,status)
      if (status .ne. 0) go to 500
      eptr = dyn_element(address)
C
C     Get some work space
C
      bytes_float = dsa_typesize('float',status)
      bytes_int = dsa_typesize('int',status)
      totsize = 6*dims(2)*bytes_float + dims(2)*bytes_int +
     : 2*nsermax*bytes_float
      call dsa_get_workspace(totsize,address,iwslt,status)
      if (status .ne. 0) go to 500
      iwpt1 = dyn_element(address)
      iwpt2 = iwpt1 + dims(2)*bytes_float
      iwpt3 = iwpt2 + dims(2)*bytes_float
      iwpt4 = iwpt3 + dims(2)*bytes_float
      iwpt5 = iwpt4 + dims(2)*bytes_int
      iwpt6 = iwpt5 + dims(2)*bytes_float
      iwpt7 = iwpt6 + dims(2)*bytes_float
      iwpt8 = iwpt7 + dims(2)*bytes_int
      iwpt9 = iwpt8 + nsermax*bytes_float
C
C     Now get some workspace for the results
C
      totsize = 6*nsermax*bytes_float + 2*nsermax*bytes_int
      call dsa_get_workspace(totsize,address,iwslt2,status)
      if (status .ne. 0) go to 500
      iwpt10 = dyn_element(address)
      iwpt11 = iwpt10 + 2*nsermax*bytes_float
      iwpt12 = iwpt11 + 2*nsermax*bytes_float
      iwpt13 = iwpt12 + 2*nsermax*bytes_int
C
C     Process the array.
C
      call sub_exam_rvf
     : (dynamic_mem(iptr),dynamic_mem(eptr),dims(1),dims(2),fscale,wave,
     : same_slope,nsermax,dynamic_mem(iwpt1),dynamic_mem(iwpt2),
     : dynamic_mem(iwpt3),dynamic_mem(iwpt4),dynamic_mem(iwpt5),
     : dynamic_mem(iwpt6),dynamic_mem(iwpt7),dynamic_mem(iwpt8),
     : dynamic_mem(iwpt9),fsrcen,errfsr,gap,errgap,xc,errxc,
     : yc,erryc,dynamic_mem(iwpt10),dynamic_mem(iwpt11),
     : dynamic_mem(iwpt12),dynamic_mem(iwpt13),cenphase,nsfit,mode)
C
C     Update TAUCAL file
C
      call process_taucal('cal',tauexist,rvf_lab,len_rvflab,fsrcen,
     : errfsr,gap,errgap,xc,errxc,yc,erryc,dynamic_mem(iwpt10),
     : dynamic_mem(iwpt11),dynamic_mem(iwpt12),nsermax,nsfit,cenphase,
     : mode,status)
C
C     Tidy and exit.
C
  500 call dsa_close(status)
      end





      subroutine sub_exam_rvf(rvfar,errs,nx,nz,fscale,wave,same_slope,
     : nsermax,zp,r2p,sig,nerr,zfit,r2fit,sigfit,x1,x2,fsrcen,errfsr,
     : gap,errgap,xc,errxc,yc,erryc,slope,aint,iflag,cp,cenphase,nsfit,
     : mode)
C+--------------------------------------------------------------------------------------
C     SUB_EXAM_RVF--- does the actual work in finding the free spectral
C                     range and the etalon gap.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) RVFAR:      (real array RVFAR(NX,NZ)) Input RVF data
C     (>) ERRS:       (real array ERRS(NX,NZ)) RVF error array
C     (>) NX:         (integer) Number of items in RVF
C     (>) NZ:         (integer) Number of frames in calibration cube
C     (>) FSCALE:     (real) Ratio of pixel size to focal length
C     (>) WAVE:       (real) Wavelength of calibration source
C     (>) SAME_SLOPE: (logical) Flag to constrain both slopes to be the same
C     (>) NSERMAX:    (integer) Maximum number of series to be fit
C     (W) ZP:         (real array ZP(NZ)) Workspace
C     (W) R2P:        (real array R2P(NZ)) Workspace
C     (W) SIG:        (real array SIG(NZ)) Workspace
C     (W) NERR:       (integer array NERR(NZ)) Workspace
C     (W) ZFIT:       (real array ZFIT(NZ)) Workspace
C     (W) R2FIT:      (real array R2FIT(NZ)) Workspace
C     (W) SIGFIT:     (real array SIGFIT(NZ)) Workspace
C     (W) X1:         (real array X1(NSERMAX)) Workspace
C     (W) X2:         (real array X2(NSERMAX)) Workspace
C     (<) FSRCEN:     (real) On axis free spectral range
C     (<) ERRFSR:     (real) Error in FSRCEN
C     (<) GAP:        (real) Etalon spacing
C     (<) ERRGAP:     (real) Error in GAP
C     (<) XC:         (real) Mean X centre
C     (<) ERRXC:      (real) Error in X centre
C     (<) YC:         (real) Mean Y centre
C     (<) ERRYC:      (real) Error in Y centre
C     (<) SLOPE:      (real array SLOPE(2,NSERMAX)) Slope and error for each series
C     (<) AINT:       (real array AINT(2,NSERMAX)) Intercept and error for each series
C     (<) IFLAG:      (integer array IFLAG(NSERMAX)) Flag for series having been fit
C     (<) CP:         (real array CP(2,NSERMAX)) Central phase and error for each series
C     (<) CENPHASE:   (real array CENPHASE(2)) Central phase plus error
C     (<) NSFIT:      (integer) Number of series fit.
C     (<) MODE:       (integer) Mode of operation variable.  =-1 if slopes are < 0
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        02-MAR-1990
C     UPDATE:      03-APR-1990 -- Option added to constrain both slopes to be the same.
C                                 (JRL)
C                  23-MAY-1990 -- Greatly changed to allow more than two series.  Also
C                                 added in calculation of mean centre coordinates
C                  25-JUL-1990 -- Changed to allow user to input an estimate of FSR if
C                                 the data have only one series.  Also created mode of
C                                 operation variable to ensure both FSR and gap are
C                                 > 0 and TAURUS 1 data can be accommodated.  (JRL)
C                  13-NOV-1990 -- Added in RVF error array. (JRL)
C                  21-FEB-1991 -- Now erases old fit before refitting. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,nz,nerr(nz),nsermax,nsfit,mode
      integer iflag(nsermax)
      real rvfar(nx,nz),fscale,wave,fsrcen,errfsr,gap,errgap,
     : zp(nz),r2p(nz),sig(nz),zfit(nz),r2fit(nz),sigfit(nz),
     : xc,errxc,yc,erryc,slope(2,nsermax),aint(2,nsermax),
     : cenphase(2),x1(nsermax),x2(nsermax),cp(2,nsermax),
     : errs(nx,nz)
      logical same_slope
C
C     Functions
C
      integer ich_len
      logical par_qnum
      logical par_quest
C
C     Local variables
C
      integer ifid,ifit,nfit,ifirst,i,iser,status,unit,j
      real rmin,rmax,rpad,zmin,zmax,dz,diff,zmid,
     : cenfirst,sum,sumerr,fsr,sumx,sumxx,sumy,sumyy,sumwt,wt,dummy
      logical ok,answer,happy,quit,exist
      character softdev*16,ch1*1,ch2*1,message*80,testfile*15
C
C     Numerical ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Open plot device
C
      call var_getchr('softdev',0,0,softdev,status)
      if (status .ne. 0) then
         call dsa_wruser
     :   ('No screen device selected--use SOFT command')
         call dsa_wrflush
         go to 500
      end if
      call pgbegin(0,softdev,1,1)
C
C     Get range in values of R**2
C
      rmin = 0.0
      rmax = min_float
      do i = 1,nz
         rmax = max(rmax,rvfar(4,i))
      end do
      if (rmin .ge. rmax) then
         call dsa_wruser('rmin >= rmax -- can''t do plot...')
         call dsa_wrflush
         go to 500
      end if
C
C     Pad out the range for plots
C
      rpad = rmax*rmax/10.0
      rmin = rmin*rmin - rpad
      rmax = rmax*rmax + rpad
C
C     Loop until you are satisfied with both fits
C
      ok = .false.
      zmin = max_float
      zmax = min_float
      do while (.not. ok)
C
C        Initialise flags...
C
         do i = 1,nsermax
            iflag(i) = 0
         end do
C
C        Set up input arrays from RVF
C
         do i = 1,nz
            zp(i) = rvfar(1,i)
            zmin = min(zmin,zp(i))
            zmax = max(zmax,zp(i))
            nerr(i) = nint(rvfar(10,i))
            sig(i) = 2.0*rvfar(4,i)*errs(4,i)
            r2p(i) = rvfar(4,i)**2
         end do
         dz = (zmax - zmin)/float(nz)
         zmid = (zmax + zmin)/2.0
         zmin = zmin - dz
         zmax = zmax + dz
C
C        If you want all the lines to have the same slope then fit
C        the fiducial series first...
C
         if (same_slope) then
            call plot_points(nz,zmin,zmax,rmin,rmax,zp,r2p,nerr)
            call dsa_wruser(' ')
            call dsa_wrflush
            answer = par_qnum('Which series is the fiducial one?',
     :      1.0,float(nsermax),1.0,.false.,' ',dummy)
            ifid = nint(dummy)
            call mark_limits(ifid,nz,zp,nerr,ch1,ch2,x1(ifid),
     :      x2(ifid),ifit)
C
C           You can't skip the fiducial series and there isn't much point
C           in quitting either...
c
            if ((ch1 .eq. 'Q') .or. (ch2 .eq. 'Q') .or. (ch1 .eq. 'S')
     :      .or. (ch2 .eq. 'S')) then
               call dsa_wruser('***You can''t SKIP or QUIT on the')
               call dsa_wrflush
               call dsa_wruser('   fiducial series!!***')
               call dsa_wrflush
               go to 500
            end if
C
C           Collect all the points that fall between the two extremes selected
C           and fit a weighted least-squares line to them. Plot the result
C
            call collect(nz,x1(ifid),x2(ifid),zp,r2p,sig,nerr,zfit,
     :      r2fit,sigfit,nfit)
            call fitline(zfit,r2fit,sigfit,nfit,aint(1,ifid),
     :      slope(1,ifid),aint(2,ifid),slope(2,ifid))
            call plotline(slope(1,ifid),aint(1,ifid),x1(ifid),x2(ifid))
C
C           The user can now change the weighting on individual points if
C           he/she/it isn't happy with the fit...
C
            happy = par_quest('Are you happy with this fit?',.true.)
            do while (.not. happy)
               call dsa_wruser
     :         ('Mark the points for changing with the cursor')
               call dsa_wrflush
               call zap(nz,zp,r2p,zmin,zmax,rmin,rmax,nerr,
     :         slope(1,ifid),aint(1,ifid),x1(ifid),x2(ifid))
               call pgsci(0)
               call plotline(slope(1,ifid),aint(1,ifid),x1(ifid),
     :         x2(ifid))
               call pgsci(1)
               call collect(nz,x1(ifid),x2(ifid),zp,r2p,sig,nerr,zfit,
     :         r2fit,sigfit,nfit)
               call fitline(zfit,r2fit,sigfit,nfit,aint(1,ifid),
     :         slope(1,ifid),aint(2,ifid),slope(2,ifid))
               call plotline(slope(1,ifid),aint(1,ifid),x1(ifid),
     :         x2(ifid))
               happy = par_quest('Are you happy with this fit?',.true.)
            end do
            nsfit = 1
            iflag(ifid) = 1
         else
            nsfit = 0
         end if
C
C        Now loop through the maximum number of allowed series, or until
C        the user decides to quit...
C
         iser = 1
         if (ifid .eq. 1) iser = 2
         do while ((iser .le. nsermax) .and. (.not. quit))
C
C           Plot the original set of points plus any fits that have been done
C
            call plot_points(nz,zmin,zmax,rmin,rmax,zp,r2p,nerr)
            do i = 1,nsermax
               if (iflag(i) .eq. 1) then
                  call plotline(slope(1,i),aint(1,i),x1(i),x2(i))
               end if
            end do
C
C           Mark the limits of the new series
C
            call mark_limits(iser,nz,zp,nerr,ch1,ch2,x1(iser),
     :      x2(iser),ifit)
C
C           If there are no more series to be fit then get out
C
            if ((ch1 .eq. 'Q') .or. (ch2 .eq. 'Q')) then
               quit = .true.
C
C           If you don't want to skip the current series then do the next fit...
C
            else if ((ch1 .ne. 'S') .and. (ch2 .ne. 'S')) then
C
C              Collect all the points between the two extrema and fit them
C              according to the type of fit specified by the user. Plot the result
C
               call collect(nz,x1(iser),x2(iser),zp,r2p,sig,nerr,zfit,
     :         r2fit,sigfit,nfit)
               if (.not. same_slope) then
                  call fitline(zfit,r2fit,sigfit,nfit,aint(1,iser),
     :            slope(1,iser),aint(2,iser),slope(2,iser))
               else
                  call fitsame(zfit,r2fit,sigfit,nfit,slope(1,ifid),
     :            slope(2,ifid),aint(1,iser),slope(1,iser),
     :            aint(2,iser),slope(2,iser))
               end if
               call plotline(slope(1,iser),aint(1,iser),x1(iser),
     :         x2(iser))
C
C              Blot out any points you don't like and refit.
C
               happy = par_quest('Are you happy with this fit?',.true.)
               do while (.not. happy)
                  call dsa_wruser
     :            ('Mark the points for changing with the cursor')
                  call dsa_wrflush
                  call zap(nz,zp,r2p,zmin,zmax,rmin,rmax,nerr,
     :            slope(1,iser),aint(1,iser),x1(iser),x2(iser))
                  call pgsci(0)
                  call plotline(slope(1,iser),aint(1,iser),x1(iser),
     :            x2(iser))
                  call pgsci(1)
                  call collect(nz,x1(iser),x2(iser),zp,r2p,sig,nerr,
     :            zfit,r2fit,sigfit,nfit)
                  if (.not. same_slope) then
                     call fitline(zfit,r2fit,sigfit,nfit,aint(1,iser),
     :               slope(1,iser),aint(2,iser),slope(2,iser))
                  else
                     call fitsame(zfit,r2fit,sigfit,nfit,slope(1,ifid),
     :               slope(2,ifid),aint(1,iser),slope(1,iser),
     :               aint(2,iser),slope(2,iser))
                  end if
                  call plotline(slope(1,iser),aint(1,iser),x1(iser),
     :            x2(iser))
                  happy = par_quest('Are you happy with this fit?',
     :            .true.)
               end do
C
C              Flag that this series has been fit successfully and increment
C              the counter for the number of series which have been fit...
C
               iflag(iser) = 1
               nsfit = nsfit + 1
            end if
            iser = iser + 1
            if (iser .eq. ifid) iser = iser + 1
         end do
C
C        If you've made a disgusting mistake somewhere then you can throw away
C        the fits you've just done and start over...
C
         ok = par_quest
     :  ('Keep these fits? (If ''NO'' then start over)',.true.)
      end do
C
C     First write out all of the fits to the user
C
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser('Fits to line:  R**2 = A + B*Z')
      call dsa_wrflush
      call dsa_wruser(' ')
      call dsa_wrflush
      do i = 1,nsermax
         if (iflag(i) .ne. 0) then
            write(message,100) i,aint(1,i),aint(2,i),slope(1,i),
     :      slope(2,i)
  100       format('Series #',i3,'    A =',f12.3,' +/-',f8.3,
     :      '  B =',f12.3,' +/-',f8.3,'')
            call dsa_wrflush
            call dsa_wruser(message(:ich_len(message)))
         end if
      end do
C
C     Find an output file name...
C
      i = 99
      exist = .false.
      do while ((i .gt. 0) .and. (.not. exist))
         write(testfile,101) i
  101    format('exam_rvf.out:',i2.2)
         inquire(file=testfile,exist=exist)
         if (.not. exist) i = i - 1
      end do
      i = i + 1
      write(testfile,101) i
C
C     Open output file...
C
      call dsa_get_lu(unit,status)
      open(unit=unit,file=testfile,status='new')
C
C     If less no series have been fit, then put out a snotty message
C
      if (nsfit .eq. 0) then
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('****No series have been fit, ergo I can''t')
         call dsa_wrflush
         call dsa_wruser('    calculate anything!!!*****')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         fsr = 0.0
         gap = 0.0
         errfsr = 0.0
         errgap = 0.0
         go to 500
C
C     Otherwise, calculate the central phase of each series and its error...
C
      else
         diff = max_float
         do i = 1,nsermax
            if (iflag(i) .ne. 0) then
               cp(1,i) = -aint(1,i)/slope(1,i)
               cp(2,i) = cp(1,i)*sqrt((slope(2,i)/slope(1,i))**2
     :         + (aint(2,i)/aint(1,i))**2)
               if (abs(cp(1,i)-zmid) .lt. diff) then
                  cenphase(1) = cp(1,i)
                  cenphase(2) = cp(2,i)
                  diff = abs(cp(1,i) - zmid)
               end if
            end if
         end do
C
C        Which is the first series to be fit?
C
         i = 1
         do while ((iflag(i) .ne. 1) .and. (i .le. nsermax))
            i = i + 1
         end do
         ifirst = i
         cenfirst = cp(1,ifirst)
C
C        If only one series was fit, then ask for the FSR
C
         if (nsfit .eq. 1) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser('****Only one series has been fit, ergo')
            call dsa_wrflush
            call dsa_wruser('    I can''t calculate both the FSR and')
            call dsa_wrflush
            call dsa_wruser('    the gap.  Give me an estimate of ')
            call dsa_wrflush
            answer = par_qnum('    FSR in MPUs',0.0,max_float,1.0,
     :      .false.,'MPUs',fsrcen)
            call dsa_wruser(' ')
            call dsa_wrflush
            errfsr = 0.0
C
C        Otherwise find mean FSR
C
         else
            sum = 0.0
            sumerr = cp(2,ifirst)**2
            do i = ifirst+1,nsermax
               if (iflag(i) .ne. 0) then
                  fsr = (cp(1,i) - cenfirst)/float(i-ifirst)
                  sum = sum + fsr
                  sumerr = sumerr + cp(2,i)**2
               end if
            end do
            fsrcen = sum/float(nsfit-1)
            errfsr = sqrt(sumerr/float(nsfit-1))
         end if
         write(message,102) fsrcen,errfsr
  102    format('On axis FSR is:',f10.2,' +/-',f8.2)
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        If FSR is negative, then something has gone wrong.  Perhaps
C        the user marked out the series in a silly order?  They must
C        be numbered from left to right!
C
         if (fsrcen .lt. 0.0) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser('****FSR is negative!!  It is probable')
            call dsa_wrflush
            call dsa_wruser('    that you have numbered the series')
            call dsa_wrflush
            call dsa_wruser('    wrongly.  They should be numbered')
            call dsa_wrflush
            call dsa_wruser('    from left to right****')
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            go to 500
         end if
C
C        Now find the mean gap and mode of operation
C
         mode = nint(slope(1,ifirst)/abs(slope(1,ifirst)))
         sum = 0.0
         sumerr = 0.0
         do i = ifirst,nsermax
            if (iflag(i) .ne. 0) then
               gap = wave/(fscale*fscale*abs(slope(1,i))*fsrcen)
               sum = sum + gap
               sumerr = sumerr + gap*gap*((slope(2,i)/slope(1,i))**2 +
     :         (errfsr/fsrcen)**2)
            end if
         end do
         gap = sum/float(nsfit)
         errgap = sqrt(sumerr/float(nsfit))
         write(message,103) gap,errgap
  103    format('Etalon gap is:',f8.2,' +/-',f6.2,' microns')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,104) mode
  104    format('Mode of operation variable: MODE =',i3)
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        Now write FSR and gap into file...
C
         write(unit,*) fsrcen,errfsr,gap,errgap,mode
      end if
C
C     Calculate mean centre
C
      sumx = 0.0
      sumxx = 0.0
      sumy = 0.0
      sumyy = 0.0
      sumwt = 0.0
      do i = 1,nz
         if (nerr(i) .ne. 2) then
            xc = rvfar(2,i)
            yc = rvfar(3,i)
            wt = 0.0
            do j = 1,nsermax
               if ((zp(i) .ge. x1(j)) .and.
     :         (zp(i) .le. x2(j))) then
                  wt = 1.0/(float(1+nerr(i))*sig(i)**2)
               end if
            end do
            sumx = sumx + xc*wt
            sumxx = sumxx + xc*xc*wt
            sumy = sumy + yc*wt
            sumyy = sumyy + yc*yc*wt
            sumwt = sumwt + wt
         end if
      end do
      xc = sumx/sumwt
      errxc = sqrt(sumxx/sumwt - xc*xc)
      yc = sumy/sumwt
      erryc = sqrt(sumyy/sumwt - yc*yc)
C
C     Write results of centre calculation to user and file...
C
      write(message,105) xc,errxc,yc,erryc
  105 format('Mean centre is:',f8.2,' +/-',f6.2,f10.2,' +/-',f6.2)
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
      write(unit,*) xc,errxc,yc,erryc
C
C     Now write out individual fit parameters
C
      do i = 1,nsermax
         if (iflag(i) .ne. 0) then
            write(unit,*) i,aint(1,i),aint(2,i),slope(1,i),slope(2,i)
         end if
      end do
      close(unit)
      call dsa_free_lu(unit,status)
C
C     Tidy and exit
C
      call pgend
  500 end




      subroutine fitsame(x,y,sig,n,slopef,erslf,aint,slope,errain,errsl)
C+--------------------------------------------------------------------------------------
C     FITSAME     --- Routine to fit a line to data given a constrained slope
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:         (real array X(N)) Independent variable
C     (>) Y:         (real array Y(N)) Dependent variable
C     (>) SIG:       (real array SIG(N)) Error array
C     (>) N:         (integer) Number of points being fit
C     (>) SLOPEF:    (real) Fiducial slope
C     (>) ERSLF:     (real) Error in fiducial slope
C     (<) AINT:      (real) Fitted intercept
C     (<) SLOPE:     (real) Output slope
C     (<) ERRAIN:    (real) Error in intercept
C     (<) ERRSL:     (real) Error in slope
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer n
      real x(n),y(n),sig(n),slopef,erslf,aint,slope,errain,errsl
C
C     Local variables
C
      integer i
      real sum,sumsq,sumwt,wt
C
C     Zero some summations
C
      sum = 0.0
      sumsq = 0.0
      sumwt = 0.0
C
C     Set the slopes and their errors equal
C
      slope = slopef
      errsl = erslf
C
C     Now do the necessary summations for the intercept
C
      do i = 1,n
         aint = y(i) - slope*x(i)
         wt = 1.0/sig(i)**2
         sum = sum + aint*wt
         sumsq = sumsq + aint*aint*wt
         sumwt = sumwt + wt
      end do
C
C     Solve for intercept and its error
C
      aint = sum/sumwt
      errain = sqrt(sumsq/sumwt - aint*aint)
      end



      subroutine fitline(x,y,sig,n,aint,slope,errain,errsl)
C+--------------------------------------------------------------------------------------
C     FITLINE     --- Routine to fit a line to data.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:         (real array X(N)) Independent variable
C     (>) Y:         (real array Y(N)) Dependent variable
C     (>) SIG:       (real array SIG(N)) Error array
C     (>) N:         (integer) Number of points being fit
C     (<) AINT:      (real) Fitted intercept
C     (<) SLOPE:     (real) Fitted slope
C     (<) ERRAIN:    (real) Error in intercept
C     (<) ERRSL:     (real) Error in slope
C
C
C     AUTHOR:      Jim Lewis -- RGO, parts stolen from Press et. al. "Numerical
C                               Recipes" pp. 508-9.
C     DATE:        23-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer n
      real x(n),y(n),sig(n),aint,slope,errain,errsl
C
C     Local variables
C
      integer i
      real sx,sy,st2,ss,sxoss,t,wt
C
C     Zero a few summation counters
C
      sx = 0.0
      sy = 0.0
      st2 = 0.0
      slope = 0.0
      ss = 0.0
C
C     Do the necessary sums...
C
      do i = 1,n
         wt = 1.0/sig(i)**2
         ss = ss + wt
         sx = sx + x(i)*wt
         sy = sy + y(i)*wt
      end do
      sxoss = sx/ss
      do i = 1,n
         t = (x(i) - sxoss)/sig(i)
         st2 = st2 + t*t
         slope = slope + t*y(i)/sig(i)
      end do
C
C     Now solve for slope,intercept and their variances
C
      slope = slope/st2
      aint = (sy - sx*slope)/ss
      errain = sqrt((1.0 + (sx*sx/(ss*st2)))/ss)
      errsl = sqrt(1.0/st2)
      end



      subroutine collect(nz,x1,x2,zp,r2p,sig,nerr,zfit,r2fit,sigfit,
     : nfit)
C+--------------------------------------------------------------------------------------
C     COLLECT     --- Routine to collect all the data between two predefined
C                     extremes in Z.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NZ:        (integer) Number of input points
C     (>) X1:        (real) Lower bound in Z
C     (>) X2:        (real) Upper bound in Z
C     (>) ZP:        (real array ZP(NZ)) Input array of Z
C     (>) R2P:       (real array R2P(NZ)) Input array of R**2
C     (>) SIG:       (real array SIG(NZ)) Input array of errors
C     (>) NERR:      (integer array NERR(NZ)) Array of flags for bad fits
C     (<) ZFIT:      (real array ZFIT(NZ)) Output array of Z
C     (<) R2FIT:     (real array R2FIT(NZ)) Output array of R**2
C     (<) SIGFIT:    (real array SIGFIT(NZ)) Output array of errors
C     (<) NFIT:      (integer) Number of points to be fit.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nfit,nz
      integer nerr(nz)
      real x1,x2,zp(nz),r2p(nz),sig(nz),zfit(nz),r2fit(nz),sigfit(nz)
C
C     Local variables
C
      integer i
C
C     Collect all the points between X1 and X2 for fitting
C
      nfit = 0
      do i = 1,nz
         if ((zp(i) .ge. x1) .and. (zp(i) .le. x2) .and.
     :   (nerr(i) .ne. 2)) then
            nfit = nfit + 1
            zfit(nfit) = zp(i)
            r2fit(nfit) = r2p(i)
            sigfit(nfit) = float(nerr(i)+1)*sig(i)
         end if
      end do
      end





      subroutine mark_limits(iser,nz,zp,nerr,ch1,ch2,x1,x2,ifit)
C+--------------------------------------------------------------------------------------
C     MARK_LIMITS --- Routine to allow user mark the Z limits of each series
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) ISER:      (integer) Series number
C     (>) NZ:        (integer) Number of rings
C     (>) ZP:        (real array ZP(NZ)) Z position of rings
C     (>) NERR:      (integer array NERR(NZ)) Error flags for rings
C     (<) CH1:       (character) Character used to mark first end point
C     (<) CH2:       (character) Character used to mark second end point
C     (<) X1:        (real) Lower end point in Z
C     (<) X2:        (real) Upper end point in Z
C     (<) IFIT:      (integer) Flag for fit. = 1 fit just good points;
C                    = 2 fit both good and dodgy points.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-MAY-1990
C     UPDATE:      03-SEP-1991 -- Now checks to make sure at least two good
C                                 points lie between the limits chosen. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer iser,ifit,nz
      integer nerr(nz)
      real x1,x2,zp(nz)
      character ch1*1,ch2*1
C
C     Functions
C
      integer ich_fold
      integer ich_len
      integer par_qnum
C
C     Local variables
C
      integer status,i,itest
      real y1,y2,temp,dummy
      character message*80
      logical ok
C
C     Choose the beginning and end of each series with cursor.
C     Make sure they are in ascending order in X
C
      ok = .false.
      do while (.not. ok)
         write(message,100) iser
         call dsa_wruser(' ')
         call dsa_wrflush
  100    format('Mark x positions where series #',i2,' begins')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser
     :   ('and ends with the cursor (''S'' to skip,''Q'' to quit)')
         call dsa_wrflush
         call pgcurse(x1,y1,ch1)
         status = ich_fold(ch1)
         call pgcurse(x2,y2,ch2)
         status = ich_fold(ch2)
         if (x2 .lt. x1) then
            temp = x2
            x2 = x1
            x1 = temp
         end if
C
C        Fit either just the good points (open circles) or the good and
C        the dodgy ones (Xs).  The bad points (solid circles)
C        are always rejected
C
         if ((ch1 .ne. 'Q') .and. (ch2 .ne. 'Q') .and.
     :   (ch1 .ne. 'S') .and. (ch2 .ne. 'S')) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser('Do you want to:')
            call dsa_wrflush
            call dsa_wruser('(1) Fit just the good ones or')
            call dsa_wrflush
            status = par_qnum
     :      ('(2) Fit both the good and dodgy ones',1.0,2.0,2.0,.true.,
     :      ' ',dummy)
            ifit = nint(dummy)
C
C           Check that there are at least 2 points in this range...
C
            itest = 0
            do i = 1,nz
               if ((zp(i) .ge. x1) .and. (zp(i) .le. x2) .and. 
     :         (nerr(i) .lt. ifit)) itest = itest + 1
            end do
            if (itest .ge. 2) then
               ok = .true.
            else
               call dsa_wruser(' ')
               call dsa_wrflush
               call dsa_wruser('*** Hey, stupid...there aren''t enough')
               call dsa_wruser(' good points between those limits to')
               call dsa_wruser(' fit a line.  Do it again!! ***')
               call dsa_wrflush
               call dsa_wruser(' ')
               call dsa_wrflush
            end if
         end if
      end do
      end




      subroutine plot_points(nz,zmin,zmax,rmin,rmax,xp,yp,nerr)
C+-----------------------------------------------------------------------
C     PLOT_POINTS  -- Plot the RVF points with symbols which
C                     reflect the quality of the fit for the plane.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NZ:         (integer) Number of points to be plotted
C     (>) ZMIN        (real) Minimum value for x axis
C     (>) ZMAX        (real) Maximum value for x axis
C     (>) RMIN:       (real) Minimum value for y axis
C     (>) RMAX:       (real) Maximum value for y axis
C     (>) XP:         (real array XP(NZ)) values of Z for plot
C     (>) YP:         (real array YP(NZ)) values of R**2 for plot
C     (>) NERR:       (integer array NERR(NZ)) quality array.
C                     0 = full weight, 1 = half weight, 2 = zero weight
C                     0 = open circle, 1 = mult. sign, 2 = filled circle
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           02-MAR-1990
C     UPDATE:
C+----------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nz,nerr(nz)
      real xp(nz),yp(nz),rmin,rmax,zmin,zmax
C
C     Local Variables
C
      integer isym(3),jsym,i
C
C     Set up symbols...
C
      isym(1) = 22 !open circle
      isym(2) = 5  !multiplication sign
      isym(3) = 17 !filled circle
C
C     Do plot box
C
      call pgenv(zmin,zmax,rmin,rmax,0,0)
      call pglabel('z (mpu)','R\u2\d (pixels)','RVF results')
C
C     Plot individual points
C
      do i = 1,nz
         jsym = isym(nerr(i)+1)
         call pgpoint(1,xp(i),yp(i),jsym)
      end do
      end

      subroutine zap(nz,zp,r2p,zmin,zmax,rmin,rmax,nerr,slope,
     : aint,x1,x2)
C+----------------------------------------------------------------------
C     ZAP   --- Allows the user to change the weight of points using
C               the cursor.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NZ:         (integer) Number of frames in calibration cube
C     (>) ZP:         (real array ZP(NZ)) Values of Z for the points
C     (>) R2P:        (real array R2P(NZ)) Values of R**2 for the points
C     (>) ZMIN:       (real) Minimum value of Z
C     (>) ZMAX:       (real) Maximum value of Z
C     (>) RMIN:       (real) Minimum value of R**2
C     (>) RMAX:       (real) Maximum value of R**2
C     (>) NERR:       (integer array NERR(NZ)) quality array
C     (>) SLOPE:      (real) Slope if current fit
C     (>) AINT:       (real) Intercept of current fit
C     (>) X1:         (real) First limit of current series
C     (>) X2:         (real) Second limit of current series
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           02-MAR-1990
C     UPDATE:         21-FEB-1991 -- Now erases old marker before displaying
C                                    the new one. (JRL)
C
C+--------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nz,nerr(nz)
      real zp(nz),r2p(nz),rmin,rmax,zmin,zmax,slope,aint,x1,x2
C
C     Functions
C
      integer ich_fold
C
C     Local variables
C
      integer irem,search_stat,dummy
      real xp,yp
      character chr*1
      logical reply,par_quest
C
C     Loop so long as chr .ne. 'Q' (quit)
C
      chr = ' '
      do while (chr .ne. 'Q')
C
C        Read the cursor position and the option implied by chr
C        Change them to upper case.
C
         call pgcurse(xp,yp,chr)
         dummy = ich_fold(chr)
C
C        Zero weight a point.
C
         if (chr .eq. 'Z') then
            call search(nz,zp,r2p,xp,yp,irem,search_stat)
            if (search_stat .eq. 0) then
               call wipe_point(zp(irem),r2p(irem),nerr(irem))
               nerr(irem) = 2
               call pgpoint(1,zp(irem),r2p(irem),17)
            end if
C
C        Half weight a point
C
         else if (chr .eq. 'H') then
            call search(nz,zp,r2p,xp,yp,irem,search_stat)
            if (search_stat .eq. 0) then
               call wipe_point(zp(irem),r2p(irem),nerr(irem))
               nerr(irem) = 1
               call pgpoint(1,zp(irem),r2p(irem),5)
            end if
C
C        Full weight a point
C
         else if (chr .eq. 'F') then
            call search(nz,zp,r2p,xp,yp,irem,search_stat)
            if (search_stat .eq. 0) then
               call wipe_point(zp(irem),r2p(irem),nerr(irem))
               nerr(irem) = 0
               call pgpoint(1,zp(irem),r2p(irem),22)
            end if
C
C        Redraw the plot
C
         else if (chr .eq. 'R') then
            call plot_points(nz,zmin,zmax,rmin,rmax,zp,r2p,nerr)
            call plotline(slope,aint,x1,x2)
C
C        Quit choosing points
C
         else if (chr .eq. 'Q') then
            reply = par_quest('Do you REALLY want to quit?',.true.)
            if (.not. reply) chr = ' '
C
C        HELP!!!!!
C
         else if (chr .eq. '?') then
            call dsa_wruser('')
            call dsa_wrflush
            call dsa_wruser('You get the following options:')
            call dsa_wrflush
            call dsa_wruser('Z:  Zero weight a point')
            call dsa_wrflush
            call dsa_wruser('H:  Half weight a point')
            call dsa_wrflush
            call dsa_wruser('F:  Full weight a point')
            call dsa_wrflush
            call dsa_wruser('R:  Redraw the plot')
            call dsa_wrflush
            call dsa_wruser('Q:  Quit selecting points')
            call dsa_wrflush
            call dsa_wruser('?:  Print up this help message')
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
C
C        Command doesn't exist...
C
         else
            call dsa_wruser('No such command--type ''?'' for help')
            call dsa_wrflush
         end if
      end do
      end



      subroutine search(nz,zp,r2p,xp,yp,irem,status)
C+-------------------------------------------------------------------
C     SEARCH     -- Once a point has been chosen, see if it exists in
C                   the RVF results...
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NZ:        (integer) Number of frames in calibration file
C     (>) ZP:        (real array ZP(NZ)) Values of Z for each frame
C     (>) R2P:       (real array R2P(NZ)) Values of R**2 for each frame
C     (>) XP:        (real) X value marked by cursor
C     (>) YP:        (real) Y value marked by cursor
C     (<) IREM:      (integer) location in the RVF array of the marked point
C     (<) STATUS:    (integer) flag for whether a point was found or not
C
C     AUTHOR:        Jim Lewis -- RGO
C     DATE:          02-MAR-1990
C     UPDATE:
C+--------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nz,irem,status
      real zp(nz),r2p(nz),xp,yp
C
C     Local variables
C
      integer i
      real radmin,dx,dy,rad
C
C     Numeric ranges...
C
      include 'NDP_NUMERIC_RANGES'
C
C     First search to find out which is the closest point in the
C     RVF array to the one marked out by the cursor.
C
      radmin = max_float
      do i = 1,nz
         dx = xp - zp(i)
         dy = sqrt(yp) - sqrt(r2p(i))
         rad = sqrt(dx*dx + dy*dy)
         if (rad .lt. radmin) then
            radmin = rad
            irem = i
         end if
      end do
C
C     If the closest one is too far away, then flag an error
C
      if (radmin .gt. 10.0) then
         call dsa_wruser('No point exists near this position')
         call dsa_wrflush
         status = 1
      else
         status = 0
      end if
      end


      subroutine plotline(sl,aint,x1,x2)
C+------------------------------------------------------------------------
C     PLOTLINE    -- Plot a fitted line
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) SL:         (real) Slope of fitted line
C     (>) AINT:       (real) Intercept of fitted line
C     (>) X1:         (real) Lower limit in X of line
C     (>) X2:         (real) Upper limit in X of line
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           02-MAR-1990
C     UPDATE:
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real sl,aint,x1,x2
C
C     Local variables
C
      real xline(2),yline(2)
C
C     Set up the points defining the line and plot it...
C
      xline(1) = x1
      xline(2) = x2
      yline(1) = sl*x1 + aint
      yline(2) = sl*x2 + aint
      call pgline(2,xline,yline)
      end


      subroutine process_taucal(ref_name,tauexist,label,len,fsr,errfsr,
     : gap,errgap,xc,errxc,yc,erryc,slope,aint,iflag,nsermax,nsfit,
     : cenphase,mode,status)
C+------------------------------------------------------------------------
C     PROCESS_TAUCAL -- Put the results into the TAUCAL file
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) REF_NAME:   (character) Reference name for the TAUCAL file
C     (>) TAUEXIST:   (logical) Flag for existence of TAUCAL file
C     (>) LABEL:      (character) First part of TAUCAL label
C     (>) LEN:        (integer) Length of label
C     (>) FSR:        (real) Free spectral range in MPUs
C     (>) ERRFSR:     (real) Error in free spectral range
C     (>) GAP:        (real) Etalon gap in microns
C     (>) ERRGAP:     (real) Error in gap
C     (>) XC:         (real) X centre of rings used
C     (>) ERRXC:      (real) Error in Xc
C     (>) YC:         (real) Y centre of rings used
C     (>) ERRYC:      (real) Error in Yc
C     (>) SLOPE:      (real array SLOPE(2,NSERMAX)) Slope and error for each series fit
C     (>) AINT:       (real array AINT(2,NSERMAX)) Intercept and error for each series fit
C     (>) IFLAG:      (integer array IFLAG(NSERMAX)) Set = 1 for each series fit.
C     (>) NSERMAX:    (integer) Maximum number of series allowed
C     (>) NSFIT:      (integer) Number of series actually fit
C     (>) CENPHASE:   (real array CENPHASE(2)) Central phase and error of the CAL cube
C     (>) MODE:       (integer) Mode of operation variable.  =-1 if slopes are < 0
C     (!) STATUS:     (integer) Status variable
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           03-JUL-1990
C     UPDATE:         25-JUL-1990 -- Added variable MODE.  (JRL)
C                     16-NOV-1990 -- Now writes to RVF substructure. (JRL)
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nsermax,nsfit,status,mode,len
      integer iflag(nsermax)
      real fsr,errfsr,gap,errgap,xc,errxc,yc,erryc,slope(2,nsermax),
     : aint(2,nsermax),cenphase(2)
      character ref_name*(*),label*(*)
      logical tauexist
C
C     Functions
C
      character ich_ci*8
C
C     Local variables
C
      integer i,j
      real data(5),dummy
      character ns*14
C
C     Store FSR and its error
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.fsr',tauexist,
     : 1,1,1,fsr,status)
      call tau_cal_setnum(ref_name,0,label(:len)//'.errfsr',tauexist,
     : 1,1,1,errfsr,status)
C
C     Store gap and its error
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.gap',tauexist,
     : 1,1,1,gap,status)
      call tau_cal_setnum(ref_name,0,label(:len)//'.errgap',tauexist,
     : 1,1,1,errgap,status)
C
C     Store mode of operation variable
C
      dummy = float(mode)
      call tau_cal_setnum(ref_name,0,label(:len)//'.mode',tauexist,
     : 1,1,1,dummy,status)
C
C     Store Xc and its error
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.xc',tauexist,
     : 1,1,1,xc,status)
      call tau_cal_setnum(ref_name,0,label(:len)//'.errxc',tauexist,
     : 1,1,1,errxc,status)
C
C     Store Yc and its error
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.yc',tauexist,
     : 1,1,1,yc,status)
      call tau_cal_setnum(ref_name,0,label(:len)//'.erryc',tauexist,
     : 1,1,1,erryc,status)
C
C     Store central phase and its error
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.cenphase',tauexist,
     : 1,1,1,cenphase(1),status)
      call tau_cal_setnum(ref_name,0,label(:len)//'.errcenph',tauexist,
     : 1,1,1,cenphase(2),status)
C
C     Store number of series fit
C
      call tau_cal_setnum(ref_name,0,label(:len)//'.nserfit',tauexist,
     : 1,1,1,float(nsfit),status)
C
C     Store series fit parameters
C
      j = 0
      do i = 1,nsermax
         if (iflag(i) .ne. 0) then
            j = j + 1
            data(1) = float(i)
            data(2) = slope(1,i)
            data(3) = slope(2,i)
            data(4) = aint(1,i)
            data(5) = aint(2,i)
            ns = '.series'//ich_ci(j)
            call tau_cal_setnum(ref_name,0,label(:len)//ns,tauexist,
     :      1,5,5,data,status)
         end if
      end do
      end


      subroutine wipe_point(x,y,nerr)
C+-------------------------------------------------------------------
C     WIPE_POINT -- "Erases" a point
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:         (real) X position of point
C     (>) Y:         (real) Y position of point
C     (>) NERR:      (integer) Error flag
C
C     AUTHOR:        Jim Lewis -- RGO
C     DATE:          21-FEB-1991
C     UPDATE:
C
C+--------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nerr
      real x,y
C
C     Change the colour index so that you write in the 
C     background colour
C
      call pgsci(0)
C
C     Now "erase" the old point
C
      if (nerr .eq. 0) then
         call pgpoint(1,x,y,22)
      else if (nerr .eq. 1) then
         call pgpoint(1,x,y,5)
      else if (nerr .eq. 2) then
         call pgpoint(1,x,y,17)
      end if
C
C     Reset the colour index and leave
C
      call pgsci(1)
      end
