      subroutine exam_rvt
C+------------------------------------------------------------------------------
C
C   ---------------
C   E X A M _ R V T
C   ---------------
C
C   Description
C   -----------
C   Finds the offsets in Z, Xc and Yc by an analysis of the RVT
C
C
C   Scope of program
C   ----------------
C   - Data dimension irrelevant
C   - Data type irrelevant
C   - Quality and error arrays not supported
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE    Name of TAUCAL file (character)(prompted for)
C
C   OBSRUN     Run number associated with OBS cube. (integer)
C              (prompted for).
C
C   WHICH_RVT  The number of the RVT to be used if more than
C              one already exist in the TAUCAL file. (integer)(prompted for)
C
C   WHICH_RVF  The number of the RVT to be used if more than
C              one already exist in the TAUCAL file. (integer)(prompted for)
C
C   EST        The estimation method to be employed. There are currently four
C              available. (1) Use the estimates from the previous run number,
C              (2) Use the estimates from the following run number,
C              (3) Use the average of these two, (4) Interpolate between these
C              two. (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU   Flag for use of TAUCAL file. (prompted for)
C
C   DISP      Instruction to plot R vs T, Xc vs T and Yc vs T.
C             (prompted for)
C
C   IGNORE    Instruction to ignore the change in R (and Z) and
C             only do this calculation for Xc and Yc. 
C             (not prompted for)
C
C   CAL_CUBE  Flag that the OBS cube really is the CAL cube
C             used to define the RVT.
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .RVT        NRVT         The number of RVTs available for use (integer)
C                            (read)
C
C   .RVT.RVTn   NAME_RVT     The name of the structure containing the RVT
C                            (character)(read)
C
C   .RVT.RVTn   RUN_USED     Run number of the CAL cube used to define RVT.
C                            (integer)(read)
C
C   .RVF        NRVF         The number of RVFs available for use (integer)
C                            (read)
C
C   .RVF.RVFn   XCEN         The X position of the optical axis. (real)(read)
C
C   .RVF.RVFn   YCEN         The Y position of the optical axis. (real)(read)
C
C   .RVF.RVFn   RUN          The run number used to define that RVF. (integer)
C                            (read)
C
C   (CAL)       NX           X size of CAL cube used to define RVFn 
C                            (integer)(read)
C
C   (CAL)       NY           Y size of CAL cube used to define RVFn
C                            (integer)(read)
C
C   (OBS)       NX           X size of OBS cube (integer)(read)
C
C   (OBS)       NY           Y size of OBS cube (integer)(read)
C
C   .RVF.RVFm   NSERFIT      The number of series fit in EXAM_RVF. (integer)
C                            (read)
C
C   .RVF.RVFm   FITDATA      The series number, slope, error, intercept and
C                            error for each of the fitted series.
C                            (real[5])(read)
C
C   (OBS)       XCEN         The X position of the optical axis. (real)(written)
C
C   (OBS)       YCEN         The Y position of the optical axis. (real)(written)
C
C   (OBS)       DX           The correction to XC from RVT.  (real)(written)
C
C   (OBS)       DY           The correction to YC from RVT.  (real)(written)
C
C   (OBS)       DZ           The correction to central phase from RVT (real)
C                            (written)
C
C   (OBS)       DR           The difference in radii from RVT (real)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested to see if it is 3-dimensional and
C     its z axis is mapped as FLOAT.
C   - RVT is sorted by run number
C   - From the cal run number, the one calibration cube ring is found.
C   - The two ring observations on either side of the observing cube run number
C     are found.  If the number of the observation is either greater than the
C     number for the last ring or less than that for the first ring then the
C     fits are extrapolated.
C   - R at observing time is estimated and hence dr is found.  This
C     leads to dz from the fits from EXAM_RVF
C   - Similarly for Xc and Yc leading to estimates of dx and dy.
C   - These three relationships are plotted if you so desire.
C   - The resulting offsets (dx,dy,dz,dr) are written out.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_LU
C     DSA_FREE_WORKSPACE
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_CI
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_BATCH
C     PAR_RDKEY
C     PAR_RDVAL
C
C   PGPLOT:
C     PGBEGIN
C     PGEND
C     PGENV
C     PGLABEL
C     PGPOINT
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETNUM
C
C   Library VAR:
C     VAR_GETCHR
C
C
C   Internal subroutines called
C   ---------------------------
C   SORT_RUNS
C   SUB_EXAM_RVT
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   05-APR-90   - Original program
C   24-MAY-90   - Now averages all slopes found in EXAM_RVF.  A few
C                 cosmetic changes as well. (JRL)
C   03-JUL-90   - TAUCAL system added with some minor changes.  (JRL)
C   03-SEP-90   - dz now stored in MPUs.  (JRL)
C   07-NOV-90   - Changed to take account of new way of storing RVT info
C                 into TAUCAL.  Also added choice of methods of estimation.
C                 Original technique of interpolating between U.T.s of cube
C                 was replaced in favour of run numbers. (JRL)
C   16-NOV-90   - Changed to get info from RVF sub-structure as well. (JRL)
C   20-NOV-90   - New versions of TAUCAL routines pasted in. (JRL)
C   20-MAR-91   - Included IGNORE keyword. Also checks to see that CAL 
C                 cube being used to define Xc and Yc has the same x,y 
C                 dimensions as the OBS cube. (JRL)
C   12-JUL-91   - Now deals with the CAL cube automatically instead of
C                 leaving it to chance that the user may ask for the wrong
C                 interpolation method.  (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer        dsa_typesize
      integer        dyn_element
      character      ich_ci*8
      logical        par_batch
C
C     Local variables
C
      integer        address      ! Dynamic memory address
      logical        badpixt      ! Flag for magic values in RVT
      integer        bytes_float  ! Number of bytes in float type
      integer        bytes_int    ! Number of bytes in integer type
      integer        calrun       ! Run number of CAL cube
      logical        cal_cube     ! Flag for OBS cube being really the CAL cube
      real           data(5)      ! Dummy real data array
      real           deltax       ! Size correction in X
      real           deltay       ! Size correction in Y
      integer        dimsrvt(2)   ! Dimensions of RVT
      logical        disp         ! Instruction to display RVT, X vs T and Y vs T
      real           dr           ! Resulting change in radius
      real           dummy        ! Dummy real variable
      real           dx           ! Resulting change in Xc
      real           dy           ! Resulting change in Yc
      real           dz           ! Resulting change in Z
      integer        est          ! Estimation method
      logical        exist        ! TRUE if a file exists
      integer        i            ! Integer counter
      logical        ignore       ! Instruction to skip dr and dz
      integer        irvtpt       ! Dynamic memory pointer for RVT
      integer        len_rvflab   ! Length of RVF label
      integer        len_rvtlab   ! Length of RVT label
      character      name_rvt*80  ! Name of RVT file
      integer        ndimrvt      ! Number of dimensions in RVT
      integer        nelm         ! Number of elements
      integer        nserfit      ! Number of series fit from RVF
      integer        nxcal        ! X size of CAL cube
      integer        nycal        ! Y size of CAL cube
      integer        nxobs        ! X size of OBS cube
      integer        nyobs        ! Y size of OBS cube
      integer        obsrun       ! Run number of observation cube
      integer        off_stat     ! Flag for error in sub_exam_rvt
      integer        run_used     ! Run number of CAL cube used in RVT
      character      rvf_lab*80   ! Label for RVT sub-structure
      character      rvt_lab*80   ! Label for RVT sub-structure
      integer        rvtslot      ! Dynamic memory slot for RVT
      integer        status       ! DSA status variable
      logical        tauexist     ! Flag for the existence of TAUCAL file
      real           sl           ! Mean slope from EXAM_RVF
      real           sum          ! Variable for averaging RVF slopes
      real           sumwt        ! Variable for averaging RVF slopes
      character      testfile*15  ! Name for output file
      integer        totspace     ! Total number of bytes for workspace allocation
      character      typet*8      ! Data type for RVT
      integer        unit         ! Logical unit number
      integer        wpt1         ! Dynamic memory pointer for workspace
      integer        wpt2         ! Dynamic memory pointer for workspace
      integer        wslot        ! Dynamic memory slot for workspace
      real           xcen         ! Mean X centre for calibration rings
      real           ycen         ! Mean Y centre for calibration rings
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get run numbers of OBS image.
C
      if (tauexist) then
         call par_rdval('obsrun',1.0,float(max_int),1.0,' ',dummy)
         obsrun = nint(dummy)
      end if
C
C     Get RVT info
C
      call tau_cal_find_struct('cal','rvt','rvt',tauexist,rvt_lab,
     : len_rvtlab,status)
      if (status .ne. 0) go to 500
      call tau_cal_rdchr('cal',0,rvt_lab(:len_rvtlab)//'.name_rvt',
     : 'name_rvt',tauexist,.true.,80,name_rvt,status)
      call dsa_named_input('rvt',name_rvt,status)
      call ndp_get_image_info('rvt',.true.,.false.,typet,badpixt,
     : status)
      call dsa_data_size('rvt',2,ndimrvt,dimsrvt,nelm,status)
      call dsa_map_data('rvt','read','float',address,rvtslot,status)
      if (status .ne. 0) go to 500
      irvtpt = dyn_element(address)
C
C     Check and see if this OBS cube is the same as the CAL cube which
C     helped to define this RVT
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.run_used',
     :   'run_used',tauexist,.false.,min_float,max_float,' ',1,
     :   dummy,status)
         if (status .ne. 0) go to 500
         run_used = nint(dummy)
         cal_cube = .false.
         if (run_used .eq. obsrun) cal_cube = .true.
      else
         call par_rdkey('cal_cube',.false.,cal_cube)
      end if
C
C     Get workspace for run numbers
C
      bytes_float = dsa_typesize('float',status)
      bytes_int = dsa_typesize('int',status)
      totspace = bytes_float*dimsrvt(2) + bytes_int*dimsrvt(2)
      call dsa_get_workspace(totspace,address,wslot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace...')
         call dsa_wrflush
         go to 500
      end if
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + bytes_float*dimsrvt(2)
C
C     Get a few parameters
C
      if (.not. par_batch()) then
         call par_rdkey('disp',.true.,disp)
      else
         disp = .false.
      end if
      if (cal_cube) then
         est = 5
      else
         call par_rdval('est',1.0,4.0,1.0,' ',dummy)
         est = nint(dummy)
      end if
      ignore = .false.
      call par_rdkey('ignore',.false.,ignore)
C
C     Find the number of RVFs in TAUCAL.  If there are more than one,
C     there is a chance that one corresponds to the CAL cube and
C     one corresponds to the OBS cube.
C
      call dsa_wruser('Reading info for RVF corresponding to')
      call dsa_wruser(' OBS cube')
      call dsa_wrflush
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      if (status .ne. 0) go to 500
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.xc','xcen',
     : tauexist,.true.,min_float,max_float,'pixels',1,xcen,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.yc','ycen',
     : tauexist,.true.,min_float,max_float,'pixels',1,ycen,status)
      call dsa_wruser(' ')
      call dsa_wrflush
C
C     Now read dimensional information from TAUCAL file in order to
C     check the size of the CAL cube versus the OBS cube.
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.run',
     :   'calrun',tauexist,.false.,float(min_int),float(max_int),' ',
     :   1,dummy,status)
         calrun = nint(dummy)
         call tau_cal_rdnum('cal',calrun,'nx','nxcal',tauexist,.false.,
     :   float(min_int),float(max_int),'pixels',1,dummy,status)
         nxcal = nint(dummy)
         call tau_cal_rdnum('cal',calrun,'ny','nycal',tauexist,.false.,
     :   float(min_int),float(max_int),'pixels',1,dummy,status)
         nycal = nint(dummy)
         call tau_cal_rdnum('cal',obsrun,'nx','nxobs',tauexist,.false.,
     :   float(min_int),float(max_int),'pixels',1,dummy,status)
         nxobs = nint(dummy)
         call tau_cal_rdnum('cal',obsrun,'ny','nyobs',tauexist,.false.,
     :   float(min_int),float(max_int),'pixels',1,dummy,status)
         nyobs = nint(dummy)
         if (status .ne. 0) go to 500
C
C        Calculate a shift based on the idea that if the frames are a 
C        different size, then they were both centred on the same physical
C        location of the detector
C
         deltax = 0.5*float(nxcal - nxobs)
         deltay = 0.5*float(nycal - nyobs)
      end if

C
C     Get RVF corresponding to CAL cube if there are more than one
C     Get series fits from EXAM_RVF.  Average the slopes.
C
      call dsa_wruser('Reading info for RVF corresponding to')
      call dsa_wruser(' CAL cube')
      call dsa_wrflush
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.nserfit',
     : 'nserfit',tauexist,.true.,min_float,max_float,' ',1,dummy,status)
      if (status .ne. 0) go to 500
      nserfit = nint(dummy)
      sum = 0.0
      sumwt = 0.0
      do i = 1,nserfit
         call tau_cal_rdnum('cal',0,
     :   rvf_lab(:len_rvflab)//'.series'//ich_ci(i),'fitdata',
     :   tauexist,.true.,min_float,max_float,' ',5,data,status)
         if (status .ne. 0) go to 500
         sum = sum + data(2)/(data(3)**2)
         sumwt = sumwt + 1.0/(data(3)**2)
      end do
      sl = sum/sumwt
C
C     Now find the offsets...
C
      call sub_exam_rvt(dynamic_mem(irvtpt),dimsrvt(1),dimsrvt(2),
     : dynamic_mem(wpt1),dynamic_mem(wpt2),obsrun,sl,xcen,
     : ycen,disp,est,ignore,dx,dy,dz,dr,off_stat)
C
C     If all went well, the write the results to a file...
C
      if (off_stat .eq. 0) then
C
C        Find an output file name...
C
         i = 99
         exist = .false.
         do while ((i .gt. 0) .and. (.not. exist))
            write(testfile,100) i
  100       format('exam_rvt.out:',i2.2)
            inquire(file=testfile,exist=exist)
            if (.not. exist) i = i - 1
         end do
         i = i + 1
         write(testfile,100) i
C
C        Now write to it...
C
         call dsa_get_lu(unit,status)
         open(unit=unit,file=testfile,status='new')
         write(unit,*) dx,dy,dz,dr
         close(unit)
         call dsa_free_lu(unit,status)
      end if
C
C     Process TAUCAL
C
      call tau_cal_setnum('cal',obsrun,'dx',tauexist,1,1,1,dx,status)
      call tau_cal_setnum('cal',obsrun,'dy',tauexist,1,1,1,dy,status)
      call tau_cal_setnum('cal',obsrun,'dz',tauexist,1,1,1,dz,status)
      call tau_cal_setnum('cal',obsrun,'dr',tauexist,1,1,1,dr,status)
      call tau_cal_setnum('cal',obsrun,'xc',tauexist,1,1,1,xcen-deltax,
     : status)
      call tau_cal_setnum('cal',obsrun,'yc',tauexist,1,1,1,ycen-deltay,
     : status)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end



      subroutine sub_exam_rvt(rvt,nxt,nyt,runs,nord,obsrun,sl,
     : xcen,ycen,disp,est,ignore,dx,dy,dz,dr,off_stat)
C+--------------------------------------------------------------------------------------
C     SUB_EXAM_RVT -- Routine to calculate the actual offsets in Xc, Yc and Z
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) RVT:       (real array RVT(NXT,NYT)) Input RVT.
C     (>) NXT:       (integer) First dimension of RVT
C     (>) NYT:       (integer) Second dimension of RVT
C     (W) RUNS:      (real array RUNS(NYT)) Array for U.T. of ring observations.
C     (W) NORD:      (integer array NORD(NYT)) Workspace for ordering of run array
C     (>) OBSRUN:    (integer) Run number of observation cube.
C     (>) SL:        (real) Mean slope from the linear RVF fits from EXAM_RVF
C     (>) XCEN:      (real) Mean Xc of calibration rings
C     (>) YCEN:      (real) Mean Yc of calibration rings
C     (>) DISP:      (logical) Instruction to display RVT, X vs T and Y vs T.
C     (>) EST:       (integer) Method for estimating the shift.
C     (>) IGNORE:    (logical) Instruction to skip the RVT
C     (<) DX:        (real) Resulting shift in Xc
C     (<) DY:        (real) Resulting shift in Yc
C     (<) DZ:        (real) Resulting shift in Z
C     (<) DR:        (real) Resulting shift in R
C     (<) OFF_STAT:  (integer) Flag for problems in this routine
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        05-APR-1990
C     UPDATE:      16-NOV-1990 -- No longer has to 'find' CAL cube ring
C                                 as it is now always the last one.  (JRL)
C                  20-MAR-1991 -- Added IGNORE parameter (JRL)
C                  12-JUL-1991 -- Added EST = 5 method for dealing with
C                                 CAL cubes. (JRL)
C
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nxt,nyt,off_stat,est,obsrun
      integer nord(nyt)
      real rvt(nxt,nyt),runs(nyt),sl,dx,dy,dz,dr,xcen,ycen
      logical disp,ignore
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
C
C     Local variables
C
      integer status,i,ist,ifn,irem,bytes_float,totspace,
     : address,slot,ipt1
      real slope,aint,rcal,robs,r1,r2,t1,t2,x1,x2,y1,y2,xobs,yobs,
     : xmin,xmax,ymin,ymax,tmin,tmax,rmin,rmax,pad,xcal,ycal
      character softdev*16,message*80
C
C     Dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Set status in case of trouble...
C
      off_stat = 0
C
C     Set up run numbers array
C
      do i = 1,nyt
         runs(i) = rvt(1,i)
      end do
C
C     Get some workspace for the sorting routine
C
      bytes_float = dsa_typesize('float',status)
      totspace = nyt*bytes_float
      call dsa_get_workspace(totspace,address,slot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace for sort routine')
         call dsa_wrflush
         off_stat = 1
         go to 500
      end if
      ipt1 = dyn_element(address)
C
C     First sort the RVT results by run number and ditch the workspace
C
      call sort_runs(nyt,runs,dynamic_mem(ipt1),nord)
      call dsa_free_workspace(slot,status)
C
C     The calibration cube ring is always the last one...
C
      rcal = rvt(4,nyt)
      xcal = rvt(2,nyt)
      ycal = rvt(3,nyt)
      irem = nyt
C
C     Find where the observation cube fits in to all this...
C     If the observation cube run number is outside the range considered by
C     the RVT, then extrapolate at the endpoints.  Otherwise, find
C     the RVT rings on either side of the observation run number.
C
      if (obsrun .lt. nint(runs(nord(1)))) then
         if (est .le. 3) then
            ist = nord(1)
            ifn = nord(1)
         else
            ist = nord(1)
            ifn = nord(2)
         end if
         call dsa_wruser
     :   (' ***Warning -- The observation run number is outside'
     :   //' the range of the RVT')
      else if (obsrun .gt. nint(runs(nord(nyt)))) then
         if (est .le. 3) then
            ist = nord(nyt)
            ifn = nord(nyt)
         else
            ist = nord(nyt-1)
            ifn = nord(nyt)
         end if
         call dsa_wruser
     :   (' ***Warning -- The observation run number is outside'
     :   //' the range of the RVT')
      else
         i = 1
         do while ((i .lt. nyt) .and. (nint(runs(nord(i))) .lt. obsrun))
            i = i + 1
         end do
         if (i .eq. 1) i = 2
         ist = nord(i-1)
         ifn = nord(i)
      end if
      r1 = rvt(4,ist)
      r2 = rvt(4,ifn)
      t1 = runs(ist)
      t2 = runs(ifn)
C
C     Now find the appropriate radius given the chosen estimation technique
C
      if (est .eq. 1) then
         robs = r1
      else if (est .eq. 2) then
         robs = r2
      else if (est .eq. 3) then
         robs = 0.5*(r1 + r2)
      else if (est .eq. 4) then
         slope = (r2 - r1)/(t2 - t1)
         aint = r2 - slope*t2
         robs = slope*float(obsrun) + aint
      else if (est .eq. 5) then
         robs = rcal
      end if
C
C     Calculate dr and write it out (if you want)
C
      if (.not. ignore) then
         dr = rcal - robs
         write(message,100) dr
  100    format('dr =',f8.3,' pixels')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        Now calculate dz
C
         dz = (robs*robs - rcal*rcal)/sl
         call dsa_wruser('Therefore from the  R**2 = A + B*Z fits')
         call dsa_wrflush
         write(message,101) dz
  101    format('dz =',f8.3,' MPUs')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
      else
         dr = 0.0
         dz = 0.0
      end if
C
C     Find Xc for given estimation technique
C
      x1 = rvt(2,ist)
      x2 = rvt(2,ifn)
      if (est .eq. 1) then
         xobs = x1
      else if (est .eq. 2) then
         xobs = x2
      else if (est .eq. 3) then
         xobs = 0.5*(x1 + x2)
      else if (est .eq. 4) then
         slope = (x2 - x1)/(t2 - t1)
         aint = x2 - slope*t2
         xobs = slope*float(obsrun) + aint
      else if (est .eq. 5) then
         xobs = xcal
      end if
      dx = xcen - xobs
C
C     Find dy...
C
      y1 = rvt(3,ist)
      y2 = rvt(3,ifn)
      if (est .eq. 1) then
         yobs = y1
      else if (est .eq. 2) then
         yobs = y2
      else if (est .eq. 3) then
         yobs = 0.5*(y1 + y2)
      else if (est .eq. 4) then
         slope = (y2 - y1)/(t2 - t1)
         aint = y2 - slope*t2
         yobs = slope*float(obsrun) + aint
      else if (est .eq. 5) then
         yobs = ycal
      end if
      dy = ycen - yobs
C
C     Write out dx and dy results
C
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser('From the run of Xc and Yc in the RVT')
      call dsa_wrflush
      write(message,102) dx,dy
  102 format('dx =',f6.3,' pixels','  dy =',f6.3,' pixels')
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
C
C     If you want to see some of the results, then plot them
C
      if (disp) then
         call var_getchr('softdev',0,0,softdev,status)
         if (status.ne.0)then
            call dsa_wruser
     :      ('No screen device selected - use SOFT command.')
            call dsa_wrflush
            off_stat = 1
            go to 500
         end if
         call pgbegin(0,softdev,1,1)
C
C        First plot radius versus run number (if you want)
C
         if (.not. ignore) then
            rmin = max_float
            rmax = min_float
            tmin = max_float
            tmax = min_float
            do i = 1,nyt
               rmin = min(rmin,rvt(4,i))
               rmax = max(rmax,rvt(4,i))
               tmin = min(tmin,runs(i))
               tmax = max(tmax,runs(i))
            end do
            pad = (rmax - rmin)/10.0
            rmin = rmin - pad
            rmax = rmax + pad
            tmin = tmin - 1.0
            tmax = tmax + 1.0
            call pgenv(tmin,tmax,rmin,rmax,0,0)
            do i = 1,nyt
               if (i .ne. irem) then
                  call pgpoint(1,runs(i),rvt(4,i),17)
               else
                  call pgpoint(1,runs(i),rvt(4,i),21)
               end if
            end do
            call pgpoint(1,float(obsrun),robs,3)
            call pglabel('Run number','Radius (pixels)',
     :      'Filled Circles=RVT rings, Open Circle=RVF ring, '
     :      //'Star=Observation Cube ')
         end if
C
C        Now do Xc versus run number
C
         xmin = max_float
         xmax = min_float
         do i = 1,nyt
            xmin = min(xmin,rvt(2,i))
            xmax = max(xmax,rvt(2,i))
         end do
         pad = (xmax - xmin)/10.0
         xmin = xmin - pad
         xmax = xmax + pad
         call pgenv(tmin,tmax,xmin,xmax,0,0)
         do i = 1,nyt
            if (i .ne. irem) then
               call pgpoint(1,runs(i),rvt(2,i),17)
            else
               call pgpoint(1,runs(i),rvt(2,i),21)
            end if
         end do
         call pgpoint(1,float(obsrun),xobs,3)
         call pglabel('Run number','X\dc\u',
     :   'Filled Circles=RVT rings, Open Circle=RVF ring, '
     :   //'Star=Observation Cube ')
C
C        Now do Yc versus run number
C
         ymin = max_float
         ymax = min_float
         do i = 1,nyt
            ymin = min(ymin,rvt(3,i))
            ymax = max(ymax,rvt(3,i))
         end do
         pad = (ymax - ymin)/10.0
         ymin = ymin - pad
         ymax = ymax + pad
         call pgenv(tmin,tmax,ymin,ymax,0,0)
         do i = 1,nyt
            if (i .ne. irem) then
               call pgpoint(1,runs(i),rvt(3,i),17)
            else
               call pgpoint(1,runs(i),rvt(3,i),21)
            end if
         end do
         call pgpoint(1,float(obsrun),yobs,3)
         call pglabel('Run number','Y\dc\u',
     :   'Filled Circles=RVT rings, Open Circle=RVF ring, '
     :   //'Star=Observation Cube ')
         call pgend
      end if
  500 end

      subroutine sort_runs(nyt,runs,tempruns,nord)
C+--------------------------------------------------------------------------------------
C     SORT_RUNS   --- Does a quick and nasty sort of the run number array.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NYT:       (integer) Number of runs in array
C     (>) RUNS:      (real array RUNS(NYT)) Array of runs
C     (W) TEMPRUNS:  (real array TEMPRUNS(NYT)) Workspace
C     (<) NORD:      (real array NORD(NYT)) Array which gives the order number for
C                    each run entry.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        05-APR-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nyt
      integer nord(nyt)
      real runs(nyt),tempruns(nyt)
C
C     Local variables
C
      integer i,j,jrem
      real tmin
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Do the sort
C
      do i = 1,nyt
         tmin = max_int
         do j = 1,nyt
            if (runs(j) .lt. tmin) then
               tmin = runs(j)
               jrem = j
            end if
         end do
         runs(jrem) = max_float
         tempruns(jrem) = tmin
         nord(i) = jrem
      end do
C
C     Copy it all back over now...
C
      do i = 1,nyt
         runs(i) = tempruns(i)
      end do
      end
