      subroutine findpks
C+------------------------------------------------------------------------------
C
C   -------------
C   F I N D P K S
C   -------------
C
C   Description
C   -----------
C   Finds the peaks in a mask frame and writes out their coordinates.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - Data array converted to FLOAT.
C   - Magic values not supported.
C   - Quality and error arrays not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE   Name of TAUCAL file (character)(prompted for)
C
C   MASKRUN   Run number associated with matrix mask frame. (integer)
C             (prompted for).
C
C   NRAD      Radius in pixels out to which to do centroid fit. (integer)
C             (prompted for)
C
C   THRESH    Threshold in numbers of sigmas above the background above
C             which to look for peaks. (real)(prompted for).
C
C
C   Keywords (read from VARS.DST)
C   -----------------------------
C   USE_TAU   Flag for use of TAUCAL file. (prompted for)
C
C   DISP      Instruction to plot greyscale plot of the mask array.
C             (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (MASK)       NAME_MAT   Name of image containing matrix mask.
C                           (character)(read)
C
C   .MASK        NMASK      Number of masks available in TAUCAL (integer)
C                           (read/written)
C
C   .MASK.MASKn  RUN        Run number associated with mask frame (MASKRUN).
C                           (integer)(written)
C
C   .MASK.MASKn  THRESH     Threshold in numbers of sigmas above the background
C                           above which peaks were found.  (real) (written)
C
C   .MASK.MASKn  NPOS       Total number of peaks found.  (integer) (written)
C
C   .MASK.MASKn  POSX[NPOS] X positions of the centre of each peak.
C                           (real)(written)
C
C   .MASK.MASKn  POSY[NPOS] Y positions of the centre of each peak.
C                           (real)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested to see if it is 2-dimensional and
C     is mapped as FLOAT.
C   - Rough statistics are done to determine the mean background.
C   - The image is displayed if the display option has been chosen.
C   - The peak finding routine finds all pixels which are THRESH*SIG
C     above the mean background.  These are tested to make sure they
C     are not noise spikes
C   - The centroid routine does a centroid calculation for each peak.
C   - If the display option is being used, the peaks which it found are
C     displayed and the user can add or subtract any number of them.
C   - Output file 'FINDPKS.OUT' is opened and the results are dumped
C     to it. These are also dumped to the TAUCAL file.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_ADD_STRUCTURE
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_WORKSPACE
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_READ_STRUCT_DEF
C     DSA_SET_STRUCT_VAR
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_FOLD
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_BATCH
C     PAR_QUEST
C     PAR_RDKEY
C     PAR_RDVAL
C
C   PGPLOT:
C     PGCURSE
C     PGEND
C     PGPOINT
C     PGSCI
C     PGSCR
C     PGVSIZE
C     PGWINDOW
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETNUM
C     TAU_DISPLAY
C     TAU_STATS
C
C
C   Internal subroutines called
C   ---------------------------
C     CENTRES
C     CENTROID
C     CURSOR
C     HANN
C     PEAKS
C     SORT
C     SUB_FIND
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   27-MAR-90   - Original program
C   26-APR-90   - Changed to allow user to vary display grey scale.  (JRL)
C   03-JUL-90   - TAUCAL system pasted in. (JRL)
C   06-JUL-90   - Minor changes, mostly cosmetic.  (JRL)
C   25-JUL-90   - Included new 2-d display routine.  (JRL)
C   03-SEP-90   - Slight changes to plotting routines.  (JRL)
C   14-NOV-90   - Now writes to .MASK substructure in TAUCAL. (JRL)
C   19-NOV-90   - New versions of TAUCAL pasted in.  (JRL)
C   08-JAN-91   - Included new centroid routine. (JRL)
C   12-JUL-91   - Fixed so that it doens't crash if it finds no peaks. (JRL)
C   05-JAN-91   - Procedure for adding MASK substructures in TAUCAL file
C                 altered slightly. (JRL)
c   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
      logical par_batch
C
C     Maximum number of points
C
      integer npmax
      parameter (npmax=4096)
C
C     Local variables
C
      integer   address       ! Dynamic memory address
      character amask*1       ! Number of masks in TAUCAL in character form
      logical   badpix        ! Flag for magic values
      integer   bytes_float   ! Number of bytes in float data type
      integer   dims(2)       ! Dimensions of MATRIX file
      logical   disp          ! Flag for use if image display
      integer   dslot         ! Dynamic memory slot
      real      dummy         ! Dummy real variable
      logical   exist         ! TRUE if a file exists
      integer   i             ! Dummy integer counter
      integer   iptr          ! Dynamic memory pointer
      integer   maskrun       ! Run number for matrix image
      character message*80    ! Character variable for user messages
      character name_mat*80   ! Name of matrix image
      integer   ndim          ! Number of dimensions in MATRIX file
      integer   nelm          ! Number of elements in MATRIX file
      integer   nmask         ! Number of masks in TAUCAL.
      integer   npos          ! Number of peaks eventually found
      integer   nrad          ! Number of pixels for centroid radius
      integer   nx            ! X dimension of MATRIX
      integer   ny            ! Y dimension of MATRIX
      integer   status        ! Status variable for DSA and TAUCAL
      logical   tauexist      ! Flag for the existence of TAUCAL file.
      character testfile*14   ! Dummy name for output file
      integer   totspace      ! Total allocation for workspace
      real      thresh        ! Threshold for peak search (sigma above background)
      character type*8        ! Data type of input file
      integer   unit          ! Logical unit number for output file
      integer   wptr1         ! Workspace pointer
      integer   wptr2         ! Workspace pointer
      integer   wslot         ! Workspace slot
      real      xc(npmax)     ! X position of peaks
      real      yc(npmax)     ! Y position of peaks
C
C     Dynamic memory support - defines DYNAMIC_MEM
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Initialisation of DSA_ routines
C
      status=0
      call dsa_open(status)
      if (status.ne.0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get run number of MATRIX image and check to make sure this
C     is a MASK run
C
      if (tauexist) then
         call par_rdval('maskrun',1.0,float(max_int),1.0,' ',dummy)
         maskrun = nint(dummy)
      end if
      call tau_cal_check_run('cal',maskrun,'mask',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get the image name and open it
C
      call tau_cal_rdchr('cal',maskrun,'image','name_mat',tauexist,
     : .true.,80,name_mat,status)
      call dsa_named_input ('image',name_mat,status)
C
C     Get dimensions of data
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size ('image',2,ndim,dims,nelm,status)
      if (ndim.ne.2) then
         call dsa_wruser('Data is not 2-dimensional')
         call dsa_wrflush
         go to 500
      end if
      nx = dims(1)
      ny = dims(2)
C
C     Map the data
C
      call dsa_map_data ('image','read','float',address,dslot,status)
      iptr = dyn_element(address)
      if (status.ne.0) goto 500
C
C     Get workspace for dummy axis arrays.  Don't use the real axis
C     arrays as you want the results to be in pixels...
C
      bytes_float = dsa_typesize('float',status)
      totspace = (nx + ny)*bytes_float
      call dsa_get_workspace(totspace,address,wslot,status)
      if (status .ne. 0) go to 500
      wptr1 = dyn_element(address)
      wptr2 = wptr1 + nx*bytes_float
C
C     Read in parameters
C
      if (par_batch()) then
         disp = .false.
      else
         call par_rdkey('disp',.true.,disp)
      end if
      call par_rdval('nrad',0.0,float(max_int-1),4.0,'pixels',dummy)
      nrad = nint(dummy)
      call par_rdval('thresh',0.0,max_float,20.0,'sigmas',thresh)
C
C     Call subroutine to do most of the work
C
      call sub_find(dynamic_mem(iptr),nx,ny,nrad,npmax,thresh,disp,
     : dynamic_mem(wptr1),dynamic_mem(wptr2),npos,xc,yc,status)
      if (status .ne. 0) go to 500
C
C     Find an output file name...
C
      i = 99
      exist = .false.
      do while ((i .gt. 0) .and. (.not. exist))
         write(testfile,100) i
  100    format('findpks.out:',i2.2)
         inquire(file=testfile,exist=exist)
         if (.not. exist) i = i - 1
      end do
      i = i + 1
      write(testfile,100) i
C
C     Output the results
C
      call dsa_get_lu(unit,status)
      open(unit=unit,file=testfile,status='new')
      do i = 1,npos
         write(unit,*) xc(i),yc(i)
      end do
      close(unit)
      write(message,101) npos,testfile
  101 format(i8,' peaks written to file ',a14)
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
C
C     Now store things away in TAUCAL file. First increment the number of
C     mask sub-structures
C
      call tau_cal_rdnum('cal',0,'mask.nmask','nmask',
     : tauexist,.false.,0.0,float(max_int),' ',1,dummy,status)
      if (status .ne. 0) go to 500
      nmask = nint(dummy) + 1
      call tau_cal_setnum('cal',0,'mask.nmask',tauexist,1,1,1,
     : float(nmask),status)
      write(amask,102) nmask
  102 format(i1)
C
C     Now create the new mask sub-structure if using TAUCAL
C
      if (tauexist) then
         call dsa_read_struct_def('TAU_MASK_SUB.DEF',status)
         call dsa_set_struct_var('nnnn','.mask.mask'//amask,status)
         call dsa_add_structure('cal','tau_mask_sub','mask_sub_struct',
     :   status)
         if (status .ne. 0) go to 500
      end if
C
C     Now write to the new structure
C
      call tau_cal_setnum('cal',0,'mask.mask'//amask//'.run',
     : tauexist,1,1,1,float(maskrun),status)
      call tau_cal_setnum('cal',0,'mask.mask'//amask//'.thresh',
     : tauexist,1,1,1,thresh,status)
      call tau_cal_setnum('cal',0,'mask.mask'//amask//'.npos',
     : tauexist,1,1,1,float(npos),status)
      call tau_cal_setnum('cal',0,'mask.mask'//amask//'.posx',
     : tauexist,1,npos,npos,xc,status)
      call tau_cal_setnum('cal',0,'mask.mask'//amask//'.posy',
     : tauexist,1,npos,npos,yc,status)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_find(data,nx,ny,nrad,npmax,thresh,disp,xaxis,
     : yaxis,npos,xc,yc,status)
C+--------------------------------------------------------------------------------------
C     SUB_FIND   --- Does the actual work in finding the peaks.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:      (real array DATA(NX,NY)) Input data
C     (>) NX:        (integer) X dimension of data
C     (>) NY:        (integer) Y dimension of data
C     (>) NRAD:      (integer) Radius for centroiding
C     (>) NPMAX:     (integer) Maxiumum number of peaks allowed
C     (>) THRESH:    (real) Number of sigmas above background to find peaks
C     (>) DISP:      (logical) Flag for use of image display
C     (>) XAXIS:     (real array XAXIS(NX)) X axis data values
C     (>) YAXIS:     (real array YAXIS(NY)) Y axis data values
C     (<) NPOS:      (integer) Number of peaks found
C     (<) XC:        (real array XC(NPMAX)) X centroid of peak
C     (<) YC:        (real array YC(NPMAX)) Y centroid of peak
C     (!) STATUS:    (integer) Status variable
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:      26-APR-1990 -- Peaks are now marked in red.  (JRL)
C                  25-JUL-1990 -- New 2-d display routine put in.  A few
C                                 cosmetics as well.  (JRL)
C                  12-JUL-1991 -- Added STATUS parameter
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nrad,npmax,npos,status
      real data(nx,ny),thresh,xc(npmax),yc(npmax),xaxis(nx),yaxis(ny)
      logical disp
C
C     Functions
C
      integer dsa_typesize,dyn_element
C
C     Local variables
C
      integer nabove,bytes_float,bytes_int,totspace,address,
     : slot,ipt1,ipt2,ipt3,npks,slot2,ipt4,ipt5,ipt6,i,spx2d(2),
     : epx2d(2)
      real clipmean,sig,amin,amax,aminplot,sta2d(2),end2d(2),ximv(2),
     : yimv(2),low,high
      character table*4
C
C     Dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Create dummy axes
C
      do i = 1,nx
         xaxis(i) = float(i)
      end do
      do i = 1,ny
         yaxis(i) = float(i)
      end do
C
C     Do stats on the data frame to find mean background
C     and sigma.
C
      call tau_stats(data,nx,ny,thresh,clipmean,sig,nabove,amin,amax,
     : status)
      if (status .ne. 0) go to 500
C
C     Display the matrix image if you want to
C
      aminplot = clipmean + 0.25*thresh*sig
      low = aminplot
      high = amin
      table = 'grey'
      if (disp) call tau_display(data,nx,ny,low,high,xaxis,yaxis,
     : table,.false.,spx2d,epx2d,sta2d,end2d,ximv,yimv,status)
      if (status .ne. 0) go to 500
C
C     Get workspace for peak finding routine
C
      bytes_float = dsa_typesize('float',status)
      bytes_int = dsa_typesize('int',status)
      totspace = nabove*(2*bytes_int + bytes_float)
      call dsa_get_workspace(totspace,address,slot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace for PEAK.')
         call dsa_wrflush
         go to 500
      end if
      ipt1 = dyn_element(address)
      ipt2 = ipt1 + nabove*bytes_int
      ipt3 = ipt2 + nabove*bytes_int
C
C     Find the peaks above background...
C
      call peaks(data,nx,ny,clipmean,sig,thresh,nabove,npks,
     : dynamic_mem(ipt1),dynamic_mem(ipt2),dynamic_mem(ipt3))
C
C     If no peaks are found then quit...
C
      if (npks .eq. 0) then
         call dsa_wruser('No peaks found above background...')
         call dsa_wrflush
         status = 1
         go to 500
      end if
C
C     Get workspace for sorting routine
C
      totspace = npks*(2*bytes_int + bytes_float)
      call dsa_get_workspace(totspace,address,slot2,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace for SORT.')
         call dsa_wrflush
         go to 500
      end if
      ipt4 = dyn_element(address)
      ipt5 = ipt4 + npks*bytes_int
      ipt6 = ipt5 + npks*bytes_int
C
C     Sort the peaks into descending order of brightness
C
      call sort(npks,dynamic_mem(ipt1),dynamic_mem(ipt2),
     : dynamic_mem(ipt3),dynamic_mem(ipt4),dynamic_mem(ipt5),
     : dynamic_mem(ipt6))
C
C     Free the workspace from sort and get a little more
C     for the centroiding routine...
C
      call dsa_free_workspace(slot2,status)
      totspace = 2*(2*nrad + 1)*bytes_float + npks*bytes_int
      call dsa_get_workspace(totspace,address,slot2,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace for CENTROID.')
         call dsa_wrflush
         go to 500
      end if
      ipt4 = dyn_element(address)
      ipt5 = ipt4 + (2*nrad + 1)*bytes_float
      ipt6 = ipt5 + (2*nrad + 1)*bytes_float
C
C     Now determine the centroid of each peak...
C
      call centroid(data,nx,ny,npks,dynamic_mem(ipt1),dynamic_mem(ipt2),
     : nrad,npmax,dynamic_mem(ipt4),dynamic_mem(ipt5),dynamic_mem(ipt6),
     : npos,xc,yc)
C
C     Ditch the remaining workspace
C
      call dsa_free_workspace(slot,status)
      call dsa_free_workspace(slot2,status)
C
C     If you want to, then display the peaks and give the user the
C     chance to take out any of the peaks he/she/it objects to...
C     Each peak is marked with an open circle.  If it is deleted
C     it is marked with a cross. Markers drawn in red.
C
      if (disp) then
         call pgvsize(ximv(1),ximv(2),yimv(1),yimv(2))
         call pgwindow(sta2d(1),end2d(1),sta2d(2),end2d(2))
         call pgscr(2,1.0,0.0,0.0)
         call pgsci(2)
         call pgpoint(npos,xc,yc,21)
         call cursor(data,nx,ny,low,high,xaxis,yaxis,table,npos,
     :   npmax,xc,yc)
         call pgend
      end if
  500 end



      subroutine peaks(data,nx,ny,mean,sig,thresh,nabove,npks,ipx,ipy,
     : hgt)
C+--------------------------------------------------------------------------------------
C     PEAKS       --- Finds which pixels are above the threshold. Avoids
C                     noise spikes.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:      (real array DATA(NX,NY)) Input data
C     (>) NX:        (integer) X dimension of data
C     (>) NY:        (integer) Y dimension of data
C     (>) MEAN:      (real) Clipped mean background value
C     (>) SIG:       (real) Background sigma
C     (>) THRESH:    (real) Number of sigmas above background to find peaks
C     (>) NABOVE:    (integer) Number of pixels above background including those
C                              which may be noise.
C     (<) NPKS:      (integer) Number of pixels above background.
C     (<) IPX:       (integer array IPX(NABOVE)) X position of peak pixels
C     (<) IPY:       (integer array IPY(NABOVE)) Y position of peak pixels
C     (<) HGT:       (real array HGT(NABOVE)) Intensity of peak pixels
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,npks,nabove
      integer ipx(nabove),ipy(nabove)
      real data(nx,ny),mean,sig,thresh,hgt(nabove)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer jx,jy,isx,isy,jsx,jsy
      real val,cut,smooth(3,3)
      character message*80
C
C     Determine cutoff for peak finding
C
      cut = thresh*sig
      npks = 0
C
C     Loop for all pixels except for the ends.
C
      do jy = 2,ny-1
         do jx = 2,nx-1
C
C           First check to see if the value is above the threshold.
C           If it is then store the 3x3 matrix around it.
C
            val = data(jx,jy)
            if (val-mean .gt. cut) then
               do isy = 1,3
                  jsy = isy + jy - 2
                  do isx = 1,3
                     jsx = isx + jx - 2
                     smooth(isx,isy) = data(jsx,jsy)
                  end do
               end do
C
C              Smooth the matrix and check to see if the peak is still
C              above the threshold.  This should get rid of noise spikes.
C
               call hann(smooth)
               if (smooth(2,2)-mean .gt. cut) then
                  npks = npks + 1
                  ipx(npks) = jx
                  ipy(npks) = jy
                  hgt(npks) = val
               end if
            end if
         end do
      end do
C
C     A soothing little message to the user...
C
      write(message,100) npks
  100 format(i8,' pixels chosen over the background')
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
      call dsa_wruser(' ')
      call dsa_wrflush
      end


      subroutine sort(npks,ipx,ipy,hgt,tx,ty,th)
C+--------------------------------------------------------------------------------------
C     SORT        --- Does a brute force sort of the pixels into order
C                     of descending brightness
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NPKS:      (integer) Number of peaks to be sorted
C     (!) IPX:       (integer array IPX(NPKS)) X positions of peak pixels
C     (!) IPY:       (integer array IPY(NPKS)) Y positions of peak pixels
C     (!) HGT:       (real array HGT(NPKS)) Intensity of peak pixels
C     (W) TX:        (integer array TX(NPKS)) Workspace
C     (W) TY:        (integer array TY(NPKS)) Workspace
C     (W) TH:        (real array TH(NPKS)) Workspace
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer npks
      integer ipx(npks),ipy(npks),tx(npks),ty(npks)
      real hgt(npks),th(npks)
C
C     Local Variables
C
      integer i,j,jrem
      real hmax
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Sort into descending order in HGT
C
      do i = 1,npks
         hmax = min_float + 1.0
         do j = 1,npks
            if (hgt(j) .gt. hmax) then
               jrem = j
               hmax = hgt(j)
            end if
         end do
         th(i) = hgt(jrem)
         tx(i) = ipx(jrem)
         ty(i) = ipy(jrem)
         hgt(jrem) = min_float
      end do
C
C     Copy them all back over now
C
      do i = 1,npks
         hgt(i) = th(i)
         ipx(i) = tx(i)
         ipy(i) = ty(i)
      end do
      end


      subroutine centroid(data,nx,ny,npks,ipx,ipy,nrad,npmax,work1,
     : work2,ipunt,npos,xc,yc)
C+--------------------------------------------------------------------------------------
C     CENTROID    --- Does a centroid fit to the brightness distribution for each
C                     peak.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:      (real array DATA(NX,NY)) Input data
C     (>) NX:        (integer) X dimension of data
C     (>) NY:        (integer) Y dimension of data
C     (>) NPKS:      (integer) Number of peaks to be sorted
C     (>) IPX:       (integer array IPX(NPKS)) X positions of peak pixels
C     (>) IPY:       (integer array IPY(NPKS)) Y positions of peak pixels
C     (>) NRAD:      (integer) Centroid radius
C     (>) NPMAX:     (integer) Maximum number of peaks allowed
C     (W) WORK1:     (real array WORK1(2*NRAD+1)) Workspace
C     (W) WORK2:     (real array WORK2(2*NRAD+1)) Workspace
C     (W) IPUNT      (integer array IPUNT(NPKS)) Workspace
C     (<) XC:        (real array XC(NPMAX)) X centroid of peaks
C     (<) YC:        (real array YC(NPMAX)) Y centroid of peaks
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,npos,nrad,npks,npmax
      integer ipx(npks),ipy(npks),ipunt(npks)
      real data(nx,ny),work1(2*nrad+1),work2(2*nrad+1),xc(npmax),
     : yc(npmax)
C
C     Local variables
C
      integer i,status,j
      real xcn,ycn,dx,dy,r
C
C     Set up flag for pixels which have already been used as
C     part of a previous centroid.
C
      do i = 1,npks
         ipunt(i) = 1
      end do
C
C     Initialise counters
C
      i = 0
      npos = 0
C
C     Loop until you have either run out of local maxima or you
C     have as many centroids as the program can handle.
C
      do while ((i .lt. npks) .and. (npos .lt. npmax))
         i = i + 1
         if (ipunt(i) .ne. 0) then
            call centres(data,nx,ny,ipx(i),ipy(i),nrad,work1,
     :      work2,xcn,ycn,status)
C
C           If the centroiding routine failed, then don't bother including it
C
            if (status .eq. 0) then
               npos = npos + 1
               xc(npos) = xcn
               yc(npos) = ycn
C
C              Flag the pixels which were used with this peak
C
               do j = 1,npks
                  dx = xcn - float(ipx(j))
                  dy = ycn - float(ipy(j))
                  r = sqrt(dx*dx + dy*dy)
                  if (r .le. float(nrad)) ipunt(j) = 0
               end do
            end if
         end if
      end do
      end


      subroutine cursor(array,nx,ny,low,high,xaxis,yaxis,table,
     : npos,npmax,xc,yc)
C+--------------------------------------------------------------------------------------
C     CURSOR      --- Routine to allow user to add or delete any peaks
C                     he/she/it wants
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) ARRAY:     (real array ARRAY(NX,NY)) Input data frame
C     (>) NX:        (integer) X size of array
C     (>) NY:        (integer) Y size of array
C     (>) LOW:       (real) "Low" grey scale value
C     (>) HIGH:      (real) "High" grey scale value
C     (>) XAXIS:     (real array XAXIS(NX)) X axis data values
C     (>) YAXIS:     (real array YAXIS(NY)) Y axis data values
C     (>) TABLE:     (character) Colour table
C     (!) NPOS:      (integer) Total number of peaks
C     (>) NPMAX:     (integer) Maximum number of peaks allowed
C     (!) XC:        (real array XC(NPMAX)) X centroid of peaks
C     (!) YC:        (real array YC(NPMAX)) Y centroid of peaks
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:      25-JUL-1990 -- Cosmetic changes (JRL)
C                  03-SEP-1990 -- Changed to allow the user to redraw the 2-d plot.
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer npos,npmax,nx,ny
      real xc(npmax),yc(npmax),array(nx,ny),low,high,xaxis(nx),
     : yaxis(ny)
      character table*(*)
C
C     Functions
C
      logical par_quest
      integer ich_fold
      integer ich_len
C
C     Local variables
C
      integer i,irem,dumint,spx2d(2),epx2d(2),status
      real xx,yy,rmin,r,sta2d(2),end2d(2),ximv(2),yimv(2)
      character ch*1
      logical quit
      character message*80
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Loop for as many points as the user wishes to do...
C
      quit = .false.
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser(' Add or delete as many points as you wish')
      call dsa_wrflush
      call dsa_wruser(' Type ''Q'' to quit or ''?'' for help')
      call dsa_wrflush
      do while (.not. quit)
C
C        Get command and make it upper case
C
         call pgcurse(xx,yy,ch)
         dumint = ich_fold(ch)
C
C        Minimal help facility
C
         if (ch .eq. '?') then
            call dsa_wruser(' You have the choice of these commands:')
            call dsa_wrflush
            call dsa_wruser(' A: Add a point')
            call dsa_wrflush
            call dsa_wruser(' Q: Quit using the cursor')
            call dsa_wrflush
            call dsa_wruser(' R: Redraw 2-d plot')
            call dsa_wrflush
            call dsa_wruser(' Z: Delete a point')
            call dsa_wrflush
            call dsa_wruser(' ?: What you''re getting now...')
            call dsa_wrflush
C
C        Add a point with the cursor
C
         else if (ch .eq. 'A') then
            if (npos .lt. npmax) then
               npos = npos + 1
               xc(npos) = xx
               yc(npos) = yy
               call pgpoint(1,xx,yy,20)
            else
               write(message,100) npmax
  100          format(' Sorry, the maximum number allowed is',i8)
               call dsa_wruser(message(:ich_len(message)))
               call dsa_wrflush
            end if
C
C        Quit using the cursor
C
         else if (ch .eq. 'Q') then
            quit = par_quest(' Do you REALLY want to quit?',.true.)
C
C        Redraw the 2-d plot
C
         else if (ch .eq. 'R') then
            call pgend
            call tau_display(array,nx,ny,low,high,xaxis,yaxis,table,
     :      .false.,spx2d,epx2d,sta2d,end2d,ximv,yimv,status)
            if (status .ne. 0) go to 500
            call pgvsize(ximv(1),ximv(2),yimv(1),yimv(2))
            call pgwindow(sta2d(1),end2d(1),sta2d(2),end2d(2))
            call pgscr(2,1.0,0.0,0.0)
            call pgsci(2)
            call pgpoint(npos,xc,yc,21)
C
C        Zap a point the the cursor
C
         else if (ch .eq. 'Z') then
            rmin = max_float
            do i = 1,npos
               r = sqrt((xx - xc(i))**2 + (yy - yc(i))**2)
               if (r .lt. rmin) then
                  rmin = r
                  irem = i
               end if
            end do
            call pgpoint(1,xc(irem),yc(irem),17)
            do i = irem,npos-1
               xc(i) = xc(i+1)
               yc(i) = yc(i+1)
            end do
            npos = npos - 1
         end if
      end do
  500 end


      subroutine hann(array)
C+--------------------------------------------------------------------------------------
C     HANN        --- Routine to do Hanning smooth of 3,3 array.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (!) ARRAY:     (real array ARRAY(NPMAX)) Array to be smoothed
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        27-MAR-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real array(3,3)
C
C     Local variables
C
      real temp
C
C     Do the Hanning smooth
C
      temp = (array(1,1) + array(3,1) + array(1,3) + array(3,3)) +
     : 2.0*(array(2,1) + array(2,3) + array(1,2) + array(3,2)) +
     : 4.0*array(2,2)
      array(2,2) = temp/16.0
      end


      subroutine centres(image,nx,ny,ix,iy,nrad,sumx,sumy,
     : x,y,status)
C+--------------------------------------------------------------------------------------
C     CENTRES      --- Do simple unweighted centroid fit.  Based on KS's
C                      routine FIG_CPCENT.  Altered to fit into the
C                      TAURUS scheme.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IMAGE    (real array IMAGE(NX,NY)) The image data.
C     (>) NX       (integer) The number of x-pixels in IMAGE.
C     (>) NY       (integer) The number of y-pixels in IMAGE.
C     (>) IX       (integer) Estimated X coordinate.
C     (>) IY       (integer) Estimated Y coordinate.
C     (>) NRAD     (Integer) The aperture size to be used for
C                  centroiding - a number of pixels.
C     (W) SUMX     (real array SUMX(2*NRAD+1)) Workspace.
C     (W) SUMY     (real array SUMY(2*NRAD+1)) Workspace.
C     (<) X        (real) X coordinate of centroid
C     (<) Y        (real) Y coordinate of centroid
C     (<) STATUS   (integer) Returns a status code.
C                   0 => OK.
C                  -1 => Centroid sums were not positive.
C                  -2 => Centroid failed to converge.
C                  -3 => Point is too close to the edge of the data.
C                  -4 => Solution isn't in aperture defined by NRAD.
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        08-JAN-1991
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nrad,ix,iy,status
      real image(nx,ny),sumx(2*nrad+1),sumy(2*nrad+1),x,y
C
C     Local variables
C
      integer i,j,k,l,ox,oxst,oy,oyst
      real oldx,oldy,value,zx,zy,xorig,yorig
C
C     Loop through iterations
C
      x = float(ix)
      y = float(iy)
      xorig = x
      yorig = y
      l = nrad*2 + 1
      do k = 1,5
         oldx = x
         oldy = y
         ox = int(oldx + 0.5)
         oy = int(oldy + 0.5)
         oxst = ox - nrad - 1
         oyst = oy - nrad - 1
C
C        Test for edge
C
         if (((ox+nrad) .gt. nx) .or. ((ox-nrad) .lt. 1) .or.
     :      ((oy+nrad) .gt. ny) .or. ((oy-nrad) .lt. 1)) then
            status = -3
            go to 500
         end if
C
C        Marginals
C
         do i = 1,l
            sumx(i) = 0.0
            sumy(i) = 0.0
         end do
         do i = 1,l
            do j = 1,l
               value = image(oxst+i,oyst+j)
               sumx(i) = sumx(i) + value
               sumy(j) = sumy(j) + value
            end do
         end do
C
C        Initialise some counters
C
         zx = 0.0
         zy = 0.0
         x = 0.0
         y = 0.0
C
C        Centroids
C
         do i = 1,l
            x = x + (sumx(i))*float(oxst+i)
            zx = zx + sumx(i)
            y = y + (sumy(i))*float(oyst+i)
            zy = zy + sumy(i)
         end do
C
C        Negative total intensity
C
         if ((zx .le. 0.0) .or. (zy .le. 0.0)) then
            status = -1
            go to 500
         end if
C
C        Centroid solution
C
         x = x/zx
         y = y/zy
C
C        Runaway solution
C
         if (sqrt((x-xorig)**2 + (y-yorig)**2) .gt. float(nrad)) then
            status = -4
            go to 500
         end if
C
C        It has converged!
C
         if ((oldx-x)**2 +(oldy-y)**2 .lt. 1.0) then
            status = 0
            go to 500
         end if
      end do
C
C     Didn't converge in 5 iterations.
C
      status = -2
C
C     Exit
C
  500 continue
      end
