      subroutine grid
C+------------------------------------------------------------------------------
C
C   -------
C   G R I D
C   -------
C
C   Description
C   -----------
C   Takes fits found in MASK and generates an x,y grid of new pixel numbers
C   which correspond to the old pixel numbers and take into account the
C   flexure in the detector.
C
C
C   Scope of program
C   ----------------
C   - 2-d images only supported.
C   - Input data type irrelevant
C   - Quality and error arrays not supported
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE    Name of TAUCAL file (character)(prompted for)
C
C   NAME_GRID  Name of output structure with new pixel positions.  Structure has
C              dimensions of (2,nx,ny) where nx,ny are not necessarily the same
C              dimension of the input image. (character)(prompted for)
C
C   WHICH_MASK User's choice of masks from TAUCAL. (integer)(prompted for)
C
C   NX         X dimension of the output file. (integer)(prompted for)
C
C   NY         Y dimension of the output file. (integer)(prompted for)
C
C   XC         This is the position in the image file coordinates for the X
C              centre of the output grid. (real) (prompted for)
C
C   YC         This is the position in the image file coordinates for the Y
C              centre of the output grid. (real) (prompted for)
C
C
C   Keywords (read from VARS.DST)
C   -----------------------------
C   USE_TAU    Flag for use of TAUCAL file. (prompted for)
C
C   DIFFSIZE   True if the output structure has different X,Y dimensions from
C              the matrix file which was used to define the transformation.
C              (prompted for)
C
C   CENTRED    True if the output grid is centred in the same place as the
C              centre of the original matrix file. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (MASK)             NAME_MAT    Name of image containing matrix mask.
C                                  (character)(read)
C
C   .MASK              NMASK       Number of masks on offer in TAUCAL.
C                                  (integer)(read)
C
C   .MASK.MASKn        MASKRUN     Run number associated with matrix mask frame.
C                                  (integer)(prompted for).
C
C   .MASK.MASKn        NORD        Order of the fit from MASK (integer)(read)
C
C   .MASK.MASKn        COEFX[NUNK] Array of X coefficients from mask.  nunk =
C                                 ((nord+1)*(nord+2))/2.  (real)(read)
C
C   .MASK.MASKn        COEFY[NUNK] Array of Y coefficients from mask.  nunk =
C                                  ((nord+1)*(nord+2))/2.  (real)(read)
C
C   .MASK.MASKn        NGRID       Number of grids for this mask. (integer)
C                                  (read/written)
C
C   .MASK.MASKn.GRIDn  NAME_GRID   Name of output grid file. (character)
C                                  (written)
C
C   .MASK.MASKn.GRIDn  XC          Location on matrix frame of centre of this
C                                  grid. (real)(written)
C
C   .MASK.MASKn.GRIDn  YC          Location on matrix frame of centre of this
C                                  grid. (real)(written)
C
C   .MASK.MASKn.GRIDn  NX          X size of this grid (real)(written)
C
C   .MASK.MASKn.GRIDn  NY          Y size of this grid (real)(written)
C
C   .MASK.MASKn        PIXSIZE_X   Central pixel X size in microns (real)
C                                 (written)
C
C   .MASK.MASKn        PIXSIZE_Y   Central pixel Y size in microns (real)
C                                 (written)
C
C   .MASK.MASKn        ELLIPT      Central pixel ellipticity defined as
C                                  pixsize_y/pixsize_x. (real)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested for dimensions and sizes of axes.
C   - The MASK output file is read for the fit coordinates.
C   - A new output structure of size (2,nx,ny) is created.
C   - The output grid is created using the derivatives of the fits.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_ADD_STRUCTURE
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_OUTPUT
C     DSA_READ_STRUCT_DEF
C     DSA_SET_STRUCT_VAR
C     DSA_SIMPLE_OUTPUT
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDCHAR
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETCHR
C     TAU_CAL_SETNUM
C     TAU_DERIV
C
C   Internal subroutines called
C   ---------------------------
C   READ_COEFS
C   SUB_GRID
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL  or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   12-FEB-90   - Original program
C   01-APR-90   - Slight modifications, mostly cosmetic.  (JRL)
C   26-APR-90   - Modified to allow an output file which doesn't have the
C                 same dimensions or centre as the original matrix file.  (JRL)
C   03-JUL-90   - TAUCAL pasted in.  A few modifications made in order to adhere
C                 to the TAUCAL scheme, but these are very minor.  (JRL)
C   26-JUL-90   - Modified so that if there is no TAUCAL the coefs can be found
C                 from elsewhere.  (JRL)
C   15-NOV-90   - Modified to read and write TAUCAL info into mask
C                 sub-structure. (JRL)
C   19-NOV-90   - New TAUCAL routines pasted in. (JRL)
C   05-JAN-92   - Procedure for adding .MASKn.GRIDn substructures in TAUCAL 
C                 file altered slightly. (JRL)
C
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer   dsa_typesize
      integer   dyn_element
      integer   ich_len
C
C     Local Variables
C
      integer      address         ! Dynamic memory address
      real         alpha           ! Offset applied to X
      character    agrid*1         ! Number of grids in TAUCAL as character
      logical      badpix          ! Flag for magic values
      real         beta            ! Offset applied to Y
      integer      bytes_float     ! Number of bytes in float type
      logical      centred         ! True if output and input frames have diff centres
      logical      diffsize        ! True if output frame has different dims to input
      integer      dims(2)         ! Dimensions of IMAGE
      real         dummy           ! Dummy real variable
      logical      exist           ! TRUE if a file exists
      character    grid_label*80   ! Label for grid substructure
      integer      i               ! Dummy counter
      integer      len_gl          ! Length of grid sub-structure label
      integer      len_ml          ! Length of mask sub-structure label
      character    mask_label*80   ! Label for mask substructure
      integer      maskrun         ! Run number of matrix image
      character    name_grid*80    ! Name of output grid image
      character    name_mat*80     ! Name of matrix image
      integer      ndim            ! Number of dimension in IMAGE
      integer      nelm            ! Number of elements in IMAGE
      integer      ngrid           ! Number of grids in TAUCAL
      integer      nord            ! Order of fit from MASK
      integer      nunk            ! number of unknowns in fit
      integer      odims(3)        ! OUTPUT file dimensions
      integer      optr            ! OUTPUT file pointer
      real         pix_x_size      ! X size of pixel at centre of grid
      real         pix_y_size      ! Y size of pixel at centre of grid
      integer      slot            ! OUTPUT file slot
      integer      status          ! DSA and TAUCAL status variable
      logical      tauexist        ! Flag for existence of TAUCAL file
      character    testfile*11     ! Name of TMASK output file
      integer      totspace        ! Number of bytes for allocation of workspace
      character    type*8          ! Data type of input image
      integer      unit            ! Logical unit for mask.out if it's needed
      integer      wpt1            ! Dynamic memory pointer
      integer      wpt2            ! Dynamic memory pointer
      integer      wslot           ! Dynamic memory slot
      real         xc              ! X position of centre of output frame
      real         yc              ! Y position of centre of output frame
C
C     Common block for dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Get started
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file.
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find out how many masks are on offer and which one you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,mask_label,
     : len_ml,status)
      if (status .ne. 0) go to 500
C
C     Get mask run number and check to be sure it's a MASK run
C
      call tau_cal_rdnum('cal',0,mask_label(:len_ml)//'.run',
     : 'maskrun',tauexist,.false.,1.0,float(max_int),' ',1,dummy,status)
      maskrun = nint(dummy)
      call tau_cal_check_run('cal',maskrun,'mask',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Open IMAGE file
C
      call tau_cal_rdchr('cal',maskrun,'image','name_mat',tauexist,
     : .true.,80,name_mat,status)
      call dsa_named_input('image',name_mat,status)
      if (status .ne. 0) go to 500
C
C     Check image dimensions
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',2,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
C
C     Get the order of fit from TAUCAL.
C
      call tau_cal_rdnum('cal',0,mask_label(:len_ml)//'.nord','norder',
     : tauexist,.true.,float(min_int),float(max_int),' ',1,dummy,status)
      if (status .ne. 0) go to 500
      nord = nint(dummy)
C
C     Get workspace for coefs.
C
      nunk = ((nord + 1)*(nord + 2))/2
      bytes_float = dsa_typesize('float',status)
      totspace = 2*nunk*bytes_float
      call dsa_get_workspace(totspace,address,wslot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace for coefs...')
         call dsa_wrflush
         go to 500
      end if
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + nunk*bytes_float
C
C     Read the coefs into the workspace from TAUCAL if it's
C     being used
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,mask_label(:len_ml)//'.coefx',
     :   'coefx',tauexist,.false.,min_float,max_float,' ',nunk,
     :   dynamic_mem(wpt1),status)
         call tau_cal_rdnum('cal',0,mask_label(:len_ml)//'.coefy',
     :   'coefy',tauexist,.false.,min_float,max_float,' ',nunk,
     :   dynamic_mem(wpt2),status)
C
C     If no TAUCAL is being used, then try to find the coefs from the most
C     recent version of MASK.OUT.
C
      else
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('You don''t have a TAUCAL file.  I will')
         call dsa_wrflush
         call dsa_wruser('attempt to find the coefs in the most')
         call dsa_wrflush
         call dsa_wruser('recent mask.out file.')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_get_lu(unit,status)
         i = 99
         exist = .false.
         do while ((i .gt. 0) .and. (.not. exist))
            write(testfile,100) i
  100       format('mask.out:',i2.2)
            inquire(file=testfile,exist=exist)
            if (.not. exist) i = i - 1
         end do
         if (i .ne. 0) then
            write(testfile,100) i
            open(unit,file=testfile,status='old',iostat=status)
         else
            status = 1
         end if
         if (status .ne. 0) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser('Failed to open mask.out file -- ')
            call dsa_wruser('Cannot procede any further')
            call dsa_wrflush
            go to 500
         end if
         call read_coefs(unit,nunk,dynamic_mem(wpt1),dynamic_mem(wpt2),
     :   status)
         call dsa_free_lu(unit,status)
      end if
      if (status .ne. 0) go to 500
C
C     Open output file
C
      call par_rdchar('name_grid',' ',name_grid)
      call dsa_named_output('output',name_grid,' ',1,1,status)
      if (status .ne. 0) go to 500
C
C     Get size and centre information for output grid
C
      odims(1) = 2
      call par_rdkey('diffsize',.false.,diffsize)
      if (diffsize) then
         call par_rdval('nx',float(min_int+1),float(max_int-1),dims(1),
     :   'pixels',dummy)
         odims(2) = nint(dummy)
         call par_rdval('ny',float(min_int+1),float(max_int-1),dims(2),
     :   'pixels',dummy)
         odims(3) = nint(dummy)
      else
         odims(2) = dims(1)
         odims(3) = dims(2)
      end if
      call par_rdkey('centred',.true.,centred)
      if (centred) then
         xc = float(dims(1))/2.0 + 0.5
         yc = float(dims(2))/2.0 + 0.5
      else
         call par_rdval('xc',min_float,max_float,1.0,'pixels',xc)
         call par_rdval('yc',min_float,max_float,1.0,'pixels',yc)
      end if
      alpha = xc - 0.5 - float(odims(2))/2.0
      beta = yc - 0.5 - float(odims(3))/2.0
C
C     Create output structure
C
      call dsa_simple_output('output','data','float',3,odims,status)
      if (status .ne. 0) go to 500
C
C     Map output array
C
      call dsa_map_data('output','write','float',address,slot,status)
      if (status .ne. 0) go to 500
      optr = dyn_element(address)
C
C     Now create the grid...
C
      call sub_grid(nord,nunk,dynamic_mem(wpt1),dynamic_mem(wpt2),
     : odims(2),odims(3),alpha,beta,dynamic_mem(optr),pix_x_size,
     : pix_y_size)
C
C     Process TAUCAL. First find out how many grids there were and
C     increment this by one
C
      call tau_cal_rdnum('cal',0,mask_label(:len_ml)//'.ngrid',
     : 'ngrid',tauexist,.false.,min_float,max_float,' ',1,dummy,status)
      ngrid = nint(dummy) + 1
      call tau_cal_setnum('cal',0,mask_label(:len_ml)//'.ngrid',
     : tauexist,1,1,1,float(ngrid),status)
      write(agrid,101) ngrid
  101 format(i1)
C
C     Read grid .DEF file and create GRID sub-structure
C
      if (tauexist) then
         call dsa_read_struct_def('TAU_GRID_SUB.DEF',status)
         grid_label = mask_label(:len_ml)//'.grid'//agrid
         len_gl = ich_len(grid_label)
         call dsa_set_struct_var('nnnn','.'//grid_label(:len_gl),status)
         call dsa_add_structure('cal','tau_grid_sub',
     :   'grid_sub_struct',status)
         if (status .ne. 0) go to 500
      end if
C
C     Now write the grid name and some info about it...
C
      call tau_cal_setchr('cal',0,grid_label(:len_gl)//'.name_grid',
     : tauexist,80,name_grid,status)
      call tau_cal_setnum('cal',0,grid_label(:len_gl)//'.xc',
     : tauexist,1,1,1,xc,status)
      call tau_cal_setnum('cal',0,grid_label(:len_gl)//'.yc',
     : tauexist,1,1,1,yc,status)
      call tau_cal_setnum('cal',0,grid_label(:len_gl)//'.nx',
     : tauexist,1,1,1,float(odims(2)),status)
      call tau_cal_setnum('cal',0,grid_label(:len_gl)//'.ny',
     : tauexist,1,1,1,float(odims(3)),status)
C
C     Now write out the rest of the info
C
      call tau_cal_setnum('cal',0,mask_label(:len_ml)//'.pixsize_x',
     : tauexist,1,1,1,pix_x_size,status)
      call tau_cal_setnum('cal',0,mask_label(:len_ml)//'.pixsize_y',
     : tauexist,1,1,1,pix_y_size,status)
      call tau_cal_setnum('cal',0,mask_label(:len_ml)//'.ellipt',
     : tauexist,1,1,1,pix_y_size/pix_x_size,status)
C
C     Let's get outta here...
C
  500 call dsa_close(status)
      end



      subroutine sub_grid(nord,nunk,coefx,coefy,nx,ny,alpha,beta,out,
     : pix_x_size,pix_y_size)
C+-------------------------------------------------------------------------
C     SUB_GRID    -- Calculate the grid
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) NORD:      (Integer) Order of the fit.
C     (>) NUNK:      (Integer) Number of unknowns.
C     (>) COEFX:     (Real array COEFX(nunk)) X fit coefficients.
C     (>) COEFY:     (Real array COEFY(nunk)) Y fit coefficients.
C     (>) NX:        (Integer) Size of X axis.
C     (>) NY:        (Integer) Size of Y axis.
C     (>) ALPHA:     (Real) Offset in X for different file size and orientation.
C     (>) BETA:      (Real) Offset in Y for different file size and orientation.
C     (<) OUT:       (Real array OUT(2,NX,NY)) Output array with new pixel grid.
C     (<) PIX_X_SIZE:(Real) Pixel X size at centre of grid
C     (<) PIX_Y_SIZE:(Real) Pixel Y size at centre of grid
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-JAN-1990
C     UPDATE:      01-APR-1990 -- Cosmetic changes.  (JRL)
C                  26-APR-1990 -- Parameters alpha and beta introduced to allow
C                                 for an output file which isn't the same size
C                                 and doesn't have the same centre as the
C                                 file used to define the transformation. (JRL)
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nord,nunk,nx,ny
      real coefx(nunk),coefy(nunk),out(2,nx,ny),alpha,beta
C
C     Local variables
C
      real dxx,dxy,dyx,dyy,ddx,ddy,pix_x_size,pix_y_size,sumx,yt,xt
      integer ix,iy
C
C     Find central pixel size
C
      xt = float(nx/2) + alpha
      yt = float(ny/2) + beta
      call tau_deriv(nord,nunk,coefx,xt,yt,dxx,dxy)
      call tau_deriv(nord,nunk,coefy,xt,yt,dyx,dyy)
      pix_x_size = sqrt(dxx*dxx + dyx*dyx)
      pix_y_size = sqrt(dxy*dxy + dyy*dyy)
C
C     Generate the grid
C
      do iy = 1,ny
         sumx = 0.0
         yt = float(iy) + beta
         do ix = 1,nx
            xt = float(ix) + alpha
            call tau_deriv(nord,nunk,coefx,xt,yt,dxx,dxy)
            call tau_deriv(nord,nunk,coefy,xt,yt,dyx,dyy)
            ddx = sqrt(dxx*dxx + dyx*dyx)/pix_x_size
            ddy = sqrt(dxy*dxy + dyy*dyy)/pix_x_size
            sumx = sumx + ddx
            out(1,ix,iy) = sumx
            if (iy .eq. 1) then
               out(2,ix,iy) = ddy
            else
               out(2,ix,iy) = out(2,ix,iy-1) + ddy
            end if
         end do
      end do
      end


      subroutine read_coefs(lu,nunk,coefx,coefy,status)
C+-------------------------------------------------------------------------
C     READ_COEFS  --  Read the coefs if TAUCAL doesn't exist
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) LU:       (integer) Logical unit of input file
C     (>) NUNK:     (integer) Number of unknowns
C     (>) COEFX:    (real array COEFX(NUNK)) Fit X coefs
C     (>) COEFY:    (real array COEFY(NUNK)) Fit Y coefs
C     (!) STATUS:   (integer) Status flag
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        26-JUL-1990
C     UPDATE:
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer lu,nunk,status
      real coefx(nunk),coefy(nunk)
C
C     Local variables
C
      integer i
C
C     Inherited status...
C
      if (status .ne. 0) go to 500
C
C     Skip a few lines first
C
      do i = 1,3
         read(lu,*)
      end do
C
C     Now try to read X coefs
C
      read(lu,*,iostat=status) (coefx(i),i=1,nunk)
      if (status .ne. 0) then
         call dsa_wruser('Error reading X coefs')
         call dsa_wrflush
         go to 500
      end if
C
C     Now try to read Y coefs
C
      read(lu,*,iostat=status) (coefy(i),i=1,nunk)
      if (status .ne. 0) then
         call dsa_wruser('Error reading Y coefs')
         call dsa_wrflush
      end if
C
C     Exit
C
  500 end
