      subroutine lamda
C+------------------------------------------------------------------------------
C
C   ---------
C   L A M D A
C   ---------
C
C   Description
C   -----------
C   Does the wavelength calibration for TAURUS cubes
C
C
C   Scope of program
C   ----------------
C   - Dimensionality irrelevant
C   - Data type irrelevant
C   - Magic values irrelevant
C   - Quality and error arrays irrelevant
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   NOBS        Number of OBS cubes to be calibrated. (integer)
C               (prompted for).
C
C   OBSRUNS     Run numbers associated with OBS cubes. (integer[10])
C               (prompted for).
C
C   Keywords
C   --------
C   USE_TAU     Flag for use of TAUCAL file. (prompted for)
C
C   REPORT      Flag to allow user to see the calibration for
C               all the cubes rather than just the first one
C               (not prompted for).
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .RVF       NRVF       Number of RVFs on offer in TAUCAL (integer)(read)
C
C   .RVF.RVFn  WAVE       Wavelength of the calibration source. (real)(read)
C
C   .RVF.RVFn  FSR        Free spectral range in MPUs. (real)(read)
C
C   .RVF.RVFn  GAP        Etalon gap in microns. (real)(read)
C
C   .RVF.RVFn  CENPHASE   Central phase of CAL cube. (real)(read)
C
C   (CAL)      MPU0       Z axis zero point in MPUs.  (real)(read)
C
C   (CAL)      DMPU       Z axis increment in MPUs/frame (real)(read)
C
C   .RVF.RVFn  MODE       Mode of operation flag. =-1 if slopes are < 0.
C                         (integer)(read)
C
C   (OBS)      NZ         Number of pixels in spectral direction. (integer)
C                         (read)
C
C   (OBS)      WAVE       Wavelength to be centred eventually in the
C                         OBS cube.  (real)(read)
C
C   (OBS)      MPU0       Z axis zero point in MPUs.  (real)(read)
C
C   (OBS)      DMPU       Z axis increment in MPUs/frame (real)(read)
C
C   (OBS)      ALAM0      Wavelength zeropoint in microns. (real)(written)
C
C   (OBS)      DLAM       Dispersion in microns/pixel in OBS cube.
C                         (real)(written)
C
C   (OBS)      CENPHASE   Central phase of OBS cube.  (real)(written)
C
C   (OBS)      FSR        Free spectral range in OBS cube. (real)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The dispersion in the CAL cube is defined from the FSR in
C     both microns and pixels.
C   - The CAL cube wavelength zero point is defined from the
C     dispersion and the central phase.
C   - The order for the CAL cube is defined as an integer and a
C     new estimate of the gap is defined.
C   - The order for the OBS cube is defined from this new gap estimate
C     and the wavelength.
C   - The wavelength zero point in the OBS cube is defined from
C     n*lamda = const
C   - The dispersion in the OBS cube is defined from the FSR in
C     both microns and pixels.
C   - The OBS cube central phase is defined from the dispersion and
C     the wavelength zero point.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_OPEN
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library ICH:
C     ICH_CI
C     ICH_LEN
C
C   Library PAR:
C     PAR_RDARY
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDNUM
C     TAU_CAL_SETNUM
C
C
C   Internal subroutines called
C   ---------------------------
C   SUB_LAMDA
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   03-SEP-90   - Original program
C   16-NOV-90   - Modified to read info from RVF sub-structure in
C                 TAUCAL (JRL)
C   20-NOV-90   - New TAUCAL routines pasted in. (JRL)
C   30-NOV-90   - Now also writes out ALAM0. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer        ich_len
C
C     Parameter for maximum number of obs cubes to be calibrated
C
      integer nobs_max
      parameter (nobs_max=10)
C
C     Local variables
C
      real           alam0                ! Wavelegnth zeropoint in OBS cube
      real           cenphase_cal         ! Central phase of cal cube
      real           cenphase_obs         ! Central phase of obs cube
      real           dlam_obs             ! Wavelength increment in OBS cube
      real           dmpu_cal             ! Z axis increment in MPUs in cal cube
      real           dmpu_obs             ! Z axis increment in MPUs in obs cube
      real           dum(nobs_max)        ! Dummy real array
      real           dummy                ! Dummy real variable
      real           fsr_cal              ! FSR of calibration cube
      real           fsr_obs              ! FSR of OBS cube
      real           gap                  ! Etalon spacing
      integer        i                    ! Integer counter
      integer        len_rvflab           ! Length of RVF label
      character      message*80           ! Character variable for messages
      integer        mode                 ! Mode of operation variable
      real           mpu0_cal             ! Z axis zeropoint in MPUs inc cal cube
      real           mpu0_obs             ! Z axis zeropoint in MPUs inc obs cube
      integer        nobs                 ! Number of obs cubes to calibrate
      integer        nz                   ! Number of pixels in Z direction in OBS cube
      integer        obsrun               ! Run number of single observation cube
      integer        obsruns(nobs_max)    ! Run numbers of all observation cubes
      logical        report               ! Keyword to report procedings for each cube.
      integer        ringrun              ! Run number of cal cube
      character      rvf_lab*80           ! Label of RVF sub-structure
      logical        send                 ! Keyword for which cubes to report info
      integer        status               ! DSA and TAUCAL status variable
      logical        tauexist             ! Flag for existence of TAUCAL file
      real           wave_cal             ! Wavelength of calibration source
      real           wave_want            ! Wavelength to be centred in obs cube
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     If using TAUCAL, then find out how many RVFs are on offer...
C
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      if (status .ne. 0) go to 500
C
C     Get number of obs cubes to calibrate
C
      call par_rdval('nobs',1.0,float(nobs_max),1.0,' ',dummy)
      nobs = nint(dummy)
C
C     Get run numbers of obs cubes
C
      call par_rdary('obsruns',1.0,float(max_int),'n',' ',nobs,
     : nobs_max,dum)
      do i = 1,nobs
         obsruns(i) = nint(dum(i))
      end do
C
C     Ask if you want a report on all the cubes or just the first one
C
      call par_rdkey('report',.false.,report)
C
C     Get some info from TAUCAL file
C
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.wave',
     : 'wave_cal',tauexist,.true.,min_float,max_float,'microns',1,
     : wave_cal,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.fsr','fsr_cen',
     : tauexist,.true.,min_float,max_float,'MPUs',1,fsr_cal,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.gap','gap',
     : tauexist,.true.,min_float,max_float,'microns',1,gap,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.cenphase',
     : 'cenphase',tauexist,.true.,min_float,max_float,'MPUs',1,
     : cenphase_cal,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.run',
     : 'ringrun',tauexist,.false.,float(min_int),float(max_int),' ',1,
     : dummy,status)
      ringrun = nint(dummy)
      call tau_cal_rdnum('cal',ringrun,'mpu0','mpu0_cal',
     : tauexist,.true.,min_float,max_float,'MPUs',1,mpu0_cal,status)
      call tau_cal_rdnum('cal',ringrun,'dmpu','dmpu_cal',
     : tauexist,.true.,min_float,max_float,'MPUs/frame',1,dmpu_cal,
     : status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.mode','mode',
     : tauexist,.true.,-1.0,1.0,' ',1,dummy,status)
      mode = nint(dummy)
      if (status .ne. 0) go to 500
C
C     Loop now for each obs cube
C
      do i = 1,nobs
         obsrun = obsruns(i)
C
C        Get wavelength info for each obs cube
C
         call tau_cal_rdnum('cal',obsrun,'wave','wave_want',tauexist,
     :   .true.,min_float,max_float,'microns',1,wave_want,status)
         call tau_cal_rdnum('cal',obsrun,'mpu0','mpu0_obs',tauexist,
     :   .true.,min_float,max_float,'MPUs',1,mpu0_obs,status)
         call tau_cal_rdnum('cal',obsrun,'dmpu','dmpu_obs',tauexist,
     :   .true.,min_float,max_float,'MPUs/frame',1,dmpu_obs,status)
         call tau_cal_rdnum('cal',obsrun,'nz','nz',tauexist,.true.,
     :   float(min_int),float(max_int),' ',1,dummy,status)
         nz = nint(dummy)
C
C        Do you get a report this time?
C
         if (report) then
            send = .true.
         else if ((.not. report) .and. (i .eq. 1)) then
            send = .true.
         else
            send = .false.
         end if
C
C        If you get a report this time, then write out the
C        run number...
C
         if (send) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            write(message,100) obsrun
  100       format('Run number: ',i4,' ')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
         end if
C
C        Now do the calculation
C
         call sub_lamda(wave_cal,fsr_cal,gap,cenphase_cal,mpu0_cal,
     :   dmpu_cal,nz,mode,wave_want,mpu0_obs,dmpu_obs,send,
     :   cenphase_obs,alam0,dlam_obs,fsr_obs)
C
C        Process TAUCAL
C
         call tau_cal_setnum('cal',obsrun,'alam0',tauexist,1,1,1,
     :   alam0,status)
         call tau_cal_setnum('cal',obsrun,'dlam',tauexist,1,1,1,
     :   dlam_obs,status)
         call tau_cal_setnum('cal',obsrun,'cenphase',tauexist,1,1,1,
     :   cenphase_obs,status)
         call tau_cal_setnum('cal',obsrun,'fsr',tauexist,1,1,1,
     :   fsr_obs,status)
      end do
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_lamda(wave_cal,fsr_cal,gap,cenphase_cal,
     : mpu0_cal,dmpu_cal,nz,mode,wave_want,mpu0_obs,dmpu_obs,send,
     : cenphase_obs,alam0,dlam_obs,fsr_obs)
C+-----------------------------------------------------------------------------------
C     SUB_LAMDA  --- Subroutine to do the work in finding the wavelength
C                    calibration
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) WAVE_CAL:        (real) Wavelength of calibration source
C     (>) FSR_CAL:         (real) FSR of cal file.
C     (>) GAP:             (real) Etalon spacing.
C     (>) CENPHASE_CAL:    (real) Central phase of CAL cube
C     (>) MPU0_CAL:        (real) Z axis zeropoint in MPUs in CAL cube
C     (>) DMPU_CAL:        (real) Z axis increment in MPUs in CAL cube
C     (>) NZ:              (integer) Number of pixels in spectral direction
C     (>) MODE:            (integer) Mode of operation variable.
C     (>) WAVE_WANT:       (real) Wavelength to be centred in OBS cube.
C     (>) MPU0_OBS:        (real) Z axis zeropoint in MPUs in OBS cube.
C     (>) DMPU:            (real) Z axis increment in MPUs in OBS cube.
C     (>) SEND:            (logical) Instruction to print info to terminal.
C     (<) CENPHASE_OBS:    (real) Central phase of OBS cube
C     (<) ALAM0:           (real) Wavelength zeropoint in corrected OBS cube
C     (<) DLAM_OBS:        (real) Wavelength dispersion in OBS cube
C     (<) FSR_OBS:         (real) Central phase of OBS data
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           03-SEP-1990
C     UPDATE:         30-NOV-1990 -- Modified to pass ALAM0 back. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer mode,nz
      real wave_cal,fsr_cal,gap,cenphase_cal,mpu0_cal,dmpu_cal,
     : wave_want,mpu0_obs,dmpu_obs,cenphase_obs,dlam_obs,fsr_obs,
     : alam0
      logical send
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer n_cal,n_obs,nz2
      real pix_mpu_cal,pix_mpu_obs,cenphase_cal_pix,fsr_cal_pix,
     : fsr_cal_microns,dlam_cal,wave0_cal,arg,gap_new,
     : fsr_obs_pix,fsr_obs_microns,cenphase_obs_pix,wave0_obs
      character message*132
C
C     Convert mpus to Z pixels in cal and obs cubes
C
      pix_mpu_cal = 1.0/dmpu_cal
      pix_mpu_obs = 1.0/dmpu_obs
C
C     Convert central phase and FSR of CAL cube to pixels
C
      cenphase_cal_pix = (cenphase_cal - mpu0_cal)*pix_mpu_cal + 1.0
      fsr_cal_pix = fsr_cal*pix_mpu_cal
C
C     Define dispersion and wavelength zeropoint in CAL cube
C
      fsr_cal_microns = wave_cal**2/(2.0*gap)
      dlam_cal = float(mode)*fsr_cal_microns/fsr_cal_pix
      wave0_cal = wave_cal - dlam_cal*cenphase_cal_pix
C
C     Now define a new gap estimate based on the order number
C     of the obs cube
C
      arg = 2.0*gap/wave_cal
      n_cal = nint(arg)
      gap_new = float(n_cal)*wave_cal*0.5
C
C     Now find the order of the OBS cube and calculate it wavelength
C     zeropoint based on the idea that n*lamda = const.
C
      n_obs = int(2.0*gap_new/wave_want)
      wave0_obs = float(n_cal)*wave0_cal/float(n_obs)
C
C     Find dispersion in obs cube from estimates of FSR
C
      fsr_obs_pix = wave_want*fsr_cal_pix/wave_cal
      fsr_obs = fsr_obs_pix/pix_mpu_obs
      fsr_obs_microns = wave_want**2/(2.0*gap_new)
      dlam_obs = float(mode)*fsr_obs_microns/fsr_obs_pix
C
C     Now find out what the wavelength zeropoint will be in
C     the corrected cube.
C
      nz2 = nz/2
      alam0 = wave_want - float(nz2)*dlam_obs
C
C     Now find central phase of OBS cube and convert to MPUs
C
      cenphase_obs_pix = (wave_want - wave0_obs)/dlam_obs
      cenphase_obs = (cenphase_obs_pix - 1.0)/pix_mpu_obs + mpu0_obs
C
C     Now write out useful info to the user if he/she/it wants it...
C
      if (send) then
         call dsa_wruser(' ')
         call dsa_wrflush
         write(message,100) wave_cal
  100    format('The wavelength of the CAL cube is:',f9.6,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,101) fsr_cal_pix,fsr_cal_microns
  101    format('The FSR of the CAL cube is:',f8.3,' pixels and',
     :   1pg12.4,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,102) cenphase_cal_pix
  102    format('The central phase of the CAL cube is:',f8.3,
     :   ' pixels')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,103) mode
  103    format('The "mode variable" is:',i4)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser('Therefore in the CAL cube')
         call dsa_wrflush
         write(message,1040) wave0_cal
 1040    format('  wave0 =',f9.6,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,1041) dlam_cal
 1041    format('   dlam =',1pg12.4,' microns/pixel')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,105) n_cal
  105    format('The CAL cube is in order',i4)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,1060) gap
 1060    format('The original gap estimate is',f11.6,' microns ')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,1061) gap_new    
 1061    format('The new gap estimate is',f11.6,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         write(message,107) wave_want
  107    format('The wavelength of the obs cube is:',f9.6,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,108) n_obs
  108    format('The OBS cube is in order',i4)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,109) fsr_obs_pix,fsr_obs_microns
  109    format('The FSR of the OBS cube is:',f8.3,' pixels and',
     :   1pg12.4,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser
     :   ('Therefore from n*lamda = const, in the OBS cube:')
         call dsa_wrflush
         write(message,110) wave0_obs
  110    format('  wave0 =',f9.6,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser('Thus in the corrected cube...')
         call dsa_wrflush
         write(message,111) alam0
  111    format('   alam0 =',1pg12.4,' microns')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,112) dlam_obs
  112    format('   dlam =',1pg12.4,' microns/pixel')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         write(message,113) cenphase_obs_pix
  113    format('The central phase of the OBS cube is thus:',f8.3,
     :   ' pixels')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
      end if
C
C     Exit
C
      end
  
