      subroutine peakdisp
C+------------------------------------------------------------------------------
C
C   ---------------
C   P E A K D I S P
C   ---------------
C
C   Description
C   -----------
C   Displays the peaks in a mask frame as found by FINDPKS
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - Data array converted to FLOAT.
C   - Quality and error arrays not supported
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE      Name of TAUCAL file (character)(prompted for)
C
C   WHICH_MASK   If there are more than one mask on offer in the TAUCAL,
C                this parameter tells which one you want. (integer)
C                (prompted for)
C
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .MASK        NMASK        Number of masks available in TAUCAL (integer)
C                            (read)
C
C   .MASK.MASKn  MASKRUN      Run number associated with matrix mask frame.
C                             (integer)(read).
C
C   (MASK)       NAME_MAT     Name of image containing matrix mask.
C                             (character)(read)
C
C   .MASK.MASKn  THRESH       Threshold in numbers of sigmas above the
C                             background above which peaks were found.
C                             (real)(read)
C
C   .MASK.MASKn  NPOS         Total number of peaks found. (integer)(read)
C
C   .MASK.MASKn  POSX[NPOS]   X positions of the centre of each peak.
C                             (real)(read)
C
C   .MASK.MASKn  POSY[NPOS]   Y positions of the centre of each peak.
C                             (real)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested to see if it is 2-dimensional and
C     is mapped as FLOAT.
C   - The list of peaks found in FINDPKS is read
C   - Rough statistics are done to determine the mean background for use
C     in determining optimal grey scale
C   - The image is displayed and the peaks overlaid.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_LU
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDVAL
C
C   PGPLOT:
C     PGEND
C     PGPOINT
C     PGSCI
C     PGSCR
C     PGVSIZE
C     PGWINDOW
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_DISPLAY
C     TAU_STATS
C
C
C   Internal subroutines called
C   ---------------------------
C     SUB_PEAKDISP
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   01-APR-90   - Original program
C   26-APR-90   - Modified to allow user to change the grey scale.  (JRL)
C   03-JUL-90   - TAUCAL addded.  (JRL)
C   25-JUL-90   - New 2-d display routine added. (JRL)
C   26-JUL-90   - Fixed so that it will still run without a TAUCAL file (JRL)
C   03-SEP-90   - Minor modifications to SUB_PEAKDISP (JRL).
C   15-NOV-90   - Reads info from MASK sub-structure. (JRL)
C   19-NOV-90   - New versions of some TAUCAL routines pasted in.  (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
C
C     Maximum number of points
C
      integer npmax
      parameter (npmax=4096)
C
C     Local variables
C
      integer        address       ! Dynamic memory address
      logical        badpix        ! Flag for magic values
      integer        bytes_float   ! Number of bytes in float data type
      integer        dims(2)       ! Dimensions of MATRIX file
      integer        dslot         ! Dynamic memory slot
      real           dummy         ! Dummy real variable
      logical        exist         ! TRUE if a file exist
      integer        i             ! Dummy counter
      integer        iptr          ! Dynamic memory pointer
      character      label*80      ! MASK sub-structure label
      integer        length        ! Length of MASK sub-structure label
      integer        maskrun       ! Run number of mask image
      character      message*80    ! Character variable for user messages
      character      name_mat*80   ! Name of matrix image
      integer        ndim          ! Number of dimension in matrix file
      integer        nelm          ! Number of elements in MATRIX file
      integer        npos          ! Number of peaks eventually found
      integer        nx            ! X dimension of MATRIX
      integer        ny            ! Y dimension of MATRIX
      integer        status        ! DSA and TAUCAL status variable
      logical        tauexist      ! Flag for the existence of TAUCAL file
      character      testfile*14   ! Name of FINDPKS output file
      real           thresh        ! Threshold for peak search (sigma above background)
      integer        totspace      ! Total space for workspace allocation
      character      type*8        ! Data type of input file
      integer        unit          ! Logical unit for FINDPKS.OUT if needed
      integer        wptr1         ! Workspace dynamic memory pointer
      integer        wptr2         ! Workspace dynamic memory pointer
      integer        wslot         ! Workspace dynamic memory slot
      real           x(npmax)      ! X position of peaks
      real           y(npmax)      ! Y position of peaks
C
C     Dynamic memory support - defines DYNAMIC_MEM
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Initialisation of DSA_ routines
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get number of masks from TAUCAL and find which one you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,label,
     : length,status)
      if (status .ne. 0) go to 500
C
C     Get run number from TAUCAL and the input image
C
      call tau_cal_rdnum('cal',0,label(:length)//'.run',
     : 'maskrun',tauexist,.false.,float(min_int),float(max_int),' ',1,
     : dummy,status)
      maskrun = nint(dummy)
      call tau_cal_check_run('cal',maskrun,'mask',tauexist,status)
      if (status .ne. 0) go to 500
      call tau_cal_rdchr('cal',maskrun,'image','name_mat',tauexist,
     : .true.,80,name_mat,status)
      if (status .ne. 0) go to 500
C
C     Get the info on the peaks from TAUCAL
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,label(:length)//'.npos','npos',
     :   tauexist,.false.,float(min_int),float(max_int),' ',1,dummy,
     :   status)
         npos = nint(dummy)
         call tau_cal_rdnum('cal',0,label(:length)//'.posx','posx',
     :   tauexist,.false.,min_float,max_float,' ',npos,x,status)
         call tau_cal_rdnum('cal',0,label(:length)//'.posy','posy',
     :   tauexist,.false.,min_float,max_float,' ',npos,y,status)
         if (status .ne. 0) then
            call dsa_wruser('Can''t find the position info in TAUCAL')
            call dsa_wrflush
            go to 500
         end if
C
C     If TAUCAL file isn't being used, then attempt to find the
C     peaks in the file FINDPKS.OUT
C
      else
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('You are not using a TAUCAL file, so I shall ')
         call dsa_wruser('attempt to find the peak positions in the ')
         call dsa_wruser('file most recent findpks.out file')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_get_lu(unit,status)
         i = 99
         exist = .false.
         do while ((i .gt. 0) .and. (.not. exist))
            write(testfile,100) i
  100       format('findpks.out:',i2.2)
            inquire(file=testfile,exist=exist)
            if (.not. exist) i = i - 1
         end do
         if (i .ne. 0) then
            write(testfile,100) i
            open(unit,file=testfile,status='old',iostat=status)
         else
            status = 1
         end if
         if (status .ne. 0) then
            call dsa_wruser('Can''t open findpks output file.')
            call dsa_wruser('I can procede no further')
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            go to 500
         end if
C
C        If the file has been found, then try to read it...
C
         npos = 0
         do while ((status .eq. 0) .and. (npos .lt. npmax))
            npos = npos + 1
            read(unit,*,iostat=status) x(npos),y(npos)
         end do
C
C        If iostat < 0 then we have reached the end of the file
C        with no error...if iostat > 0 then we have an error in
C        the reading of the file
C
         if (status .lt. 0) then
            status = 0
         else if (status .gt. 0) then
            call dsa_wruser('An error occured in reading ')
            call dsa_wruser(testfile)
            call dsa_wrflush
            go to 500
         end if
C
C        Otherwise everything's cool, and procede as normal.
C
         npos = npos - 1
         write(message,101) npos
  101    format(i8,' peaks read from file ')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call dsa_free_lu(unit,status)
      end if
C
C     Read the threshold used in finding the peaks...
C
      call tau_cal_rdnum('cal',0,label(:length)//'.thresh','thresh',
     : tauexist,.true.,min_float,max_float,' ',1,thresh,status)
      if (status .ne. 0) go to 500
C
C     Get the image name and open it
C
      call dsa_named_input ('image',name_mat,status)
C
C     Get dimensions of data
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size ('image',2,ndim,dims,nelm,status)
      if (ndim.ne.2) then
         call dsa_wruser('Data is not 2-dimensional')
         call dsa_wrflush
         go to 500
      end if
      nx=dims(1)
      ny=dims(2)
C
C     Map the data
C
      call dsa_map_data ('image','read','float',address,dslot,status)
      iptr = dyn_element(address)
      if (status.ne.0) go to 500
C
C     Get workspace for dummy axis arrays.  You don't want the real
C     axis arrays (if they even exist) as all these results must be
C     in pixel space
C
      bytes_float = dsa_typesize('float',status)
      totspace = (nx + ny)*bytes_float
      call dsa_get_workspace(totspace,address,wslot,status)
      wptr1 = dyn_element(address)
      wptr2 = wptr1 + nx*bytes_float
      if (status .ne. 0) go to 500
C
C     Call subroutine to do most of the work
C
      call sub_peakdisp(dynamic_mem(iptr),nx,ny,npos,thresh,x,y,
     : dynamic_mem(wptr1),dynamic_mem(wptr2))
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_peakdisp(data,nx,ny,npos,thresh,x,y,xaxis,
     : yaxis)
C+--------------------------------------------------------------------------------------
C     SUB_PEAKDISP --- Does the actual work in displaying the peaks.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:      (real array DATA(NX,NY)) Input data
C     (>) NX:        (integer) X dimension of data
C     (>) NY:        (integer) Y dimension of data
C     (>) NPOS:      (integer) Number of peaks found
C     (>) THRESH:    (real) Number of sigmas above background to find peaks
C     (>) X:         (real array X(NPOS)) X centroid of peak
C     (>) Y:         (real array Y(NPOS)) Y centroid of peak
C     (>) XAXIS:     (real array XAXIS(NX)) X axis array
C     (>) YAXIS:     (real array YAXIS(NY)) Y axis array
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        01-APR-1990
C     UPDATE:      26-APR-1990 -- Peaks now marked with red circles.  (JRL)
C                  25-JUL-1990 -- New 2-d display routine used.  (JRL)
C                  03-SEP-1990 -- Slight changes to aid plotting routine. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,npos
      real data(nx,ny),thresh,x(npos),y(npos),xaxis(nx),yaxis(ny)
C
C     Local variables
C
      integer nabove,status,spx2d(2),epx2d(2),i
      real clipmean,sig,amin,amax,aminplot,sta2d(2),end2d(2),ximv(2),
     : yimv(2),low,high
      character table*4
C
C     Create dummy axis arrays
C
      do i = 1,nx
         xaxis(i) = float(i)
      end do
      do i = 1,ny
         yaxis(i) = float(i)
      end do
C
C     Repeat the stats on the image
C
      call tau_stats(data,nx,ny,thresh,clipmean,sig,nabove,amin,amax,
     : status)
      if (status .ne. 0) go to 500
C
C     Now display the peaks and the image
C
      aminplot = clipmean + 0.25*thresh*sig
      table = 'grey'
      low = aminplot
      high = amin
      call tau_display(data,nx,ny,low,high,xaxis,yaxis,
     : table,.false.,spx2d,epx2d,sta2d,end2d,ximv,yimv,status)
      if (status .ne. 0) go to 500
      call pgvsize(ximv(1),ximv(2),yimv(1),yimv(2))
      call pgwindow(sta2d(1),end2d(1),sta2d(2),end2d(2))
      call pgsci(2)
      call pgscr(2,1.0,0.0,0.0)
      call pgpoint(npos,x,y,21)
      call pgend
  500 end
