      subroutine phase_map
C+------------------------------------------------------------------------------
C
C   -----------------
C   P H A S E _ M A P
C   -----------------
C
C   Description
C   -----------
C   Finds the number of pixels to shift each spectrum in a TAURUS cube
C   and writes them out to a file.
C
C
C   Scope of program
C   ----------------
C   - Obs data dimensions irrelevant but must match
C     those of RES file and GRID file.
C   - Obs cube type irrelevant
C   - Quality and error arrays irrelevant.
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   OBSRUN      Run number associated with OBS cube. (integer)
C               (prompted for).
C
C   NAME_RES    Name of the residual phase map structure defined
C               in RESIDUAL. (character)(prompted for)
C
C   NAME_SHIFTS Name of the output file for the calculated shifts.
C               (character)(prompted for)
C
C   WHICH_RVF   Number of the RVF already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_MASK  Number of the MASK already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_GRID  Number of the GRID already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU     Flag for use of TAUCAL file. (prompted for)
C
C   USE_GRID    Instruction to either use the grid of results from
C               MASK and GRID to define pixel-space or not. (prompted for)
C
C   USE_RES     Instruction to either use the residual phase map or not.
C               (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .MASK.MASKn.GRIDn  NAME_GRID   Name of input file with grid of X,Y pixel
C                                  positions. (character)(prompted for)
C
C   .RVF               NRVF        Number of RVFs already defined in TAUCAL.
C                                  (integer)(read)
C
C   .MASK              NMASK       Number of MASKs already defined in TAUCAL.
C                                  (integer)(read)
C
C   .MASK.MASKn        NGRID       Number of GRIDs already defined in MASK
C                                  sub-structure of TAUCAL. (integer)(read)
C
C   (OBS)              IMAGE       Name of OBS cube. (character)(read)
C
C   (OBS)              WAVE        Wavelength to be eventually centred in Z in
C                                  the obs cube. (real)(read)
C
C   (OBS)              DX          The correction to XC from RVT.  (real)(read)
C
C   (OBS)              DY          The correction to YC from RVT.  (real)(read)
C
C   (OBS)              DZ          The correction to central phase from RVT
C                                  (real)(read)
C
C   (OBS)              XC          The X position of the optical axis.
C                                  (real)(read)
C
C   (OBS)              YC          The Y position of the optical axis.
C                                  (real)(read)
C
C   (OBS)              CENPHASE    Central phase of OBS cube. (real)(read)
C
C   (OBS)              MPU0        Z axis zeropoint in MPUs (real)(read)
C
C   (OBS)              DMPU        Z axis increment in MPUs/frame (real)(read)
C
C   (OBS)              FSR         The free spectral range in the obs cube.
C                                  (real)(read)
C
C   .RVF.RVFn          GAP         Etalon gap in microns. (real)(read)
C
C   .RVF.RVFn          MODE        Mode of operation flag. =-1 if slopes < 0.
C                                  (integer)(read)
C
C   .MASK.MASKn        PIXSIZE_X   Pixel size as defined by MASK. (real)(read)
C
C   .CAMERA            FOCLEN      Focal length of the camera used. (real)(read)
C
C   (OBS)              NAME_SHIFTS Name of the output file containing phase_map
C                                  (character)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The shift to be subtracted is calculated for each spectrum
C     from the definition of the phase map.
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_GET_WORKSPACE
C     DSA_GET_WORK_ARRAY
C     DSA_INPUT
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_NAMED_OUTPUT
C     DSA_OPEN
C     DSA_SIMPLE_OUTPUT
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library GEN:
C     GEN_FILL
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDCHAR
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETCHR
C     TAU_DUMMY_GRID
C
C   Internal subroutines called
C   ---------------------------
C     SUB_PHASE_MAP
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   19-MAR-90   - Original program
C   05-APR-90   - Modified to remove the RVT analysis which is now done in
C                 EXAM_RVT.  (JRL)
C   25-MAY-90   - A few cosmetic alterations.  (JRL)
C   03-JUL-90   - TAUCAL added along with minor alterations.  (JRL)
C   26-JUL-90   - Mode of operation variable added (JRL)
C   03-SEP-90   - Took out wavelength calibration section which is now
C                 done in LAMDA.  Shifts now stored in MPUs. (JRL)
C   20-NOV-90   - Now read info from RVF and MASK sub-structures in TAUCAL.
C                 Also pasted in new versions of TAUCAL routines. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer        dsa_typesize
      integer        dyn_element
C
C     Local variables
C
      integer        address        ! Dynamic memory address
      logical        badpix         ! Flag for magic values in obs file
      logical        badpixg        ! Flag for magic values in grid file
      logical        badpixr        ! Flag for magic values in residuals file
      integer        bytes_float    ! Number of bytes in float type.
      real           cen_phase_obs  ! Central phase of obs cube
      integer        dims(3)        ! Dimensions of obs cube
      real           dmpu           ! Z axis increment in MPUs
      real           dx             ! Shift in Xc from analysis of RVT
      real           dy             ! Shift in Yc from analysis of RVT
      real           dz             ! Shift in Z from analysis of RVT
      real           dummy          ! Dummy real variable
      real           focus          ! Camera focal length
      real           fscale         ! Pixel size/Focal length
      real           fsr_obs        ! FSR of obs cube (pixels)
      real           gap            ! Etalon spacing
      integer        gdims(3)       ! Grid file dimensions
      integer        gptr           ! Dynamic memory pointer for grid
      character      grid_lab*80    ! TAUCAL label for GRID
      integer        gslot          ! Dynamic memory slot for grid
      integer        irespt         ! Dynamic memory pointer for residual file
      integer        len_gridlab    ! Length of grid_lab
      integer        len_masklab    ! Length of mask_lab
      integer        len_rvflab     ! Length of rvf_lab
      character      mask_lab*80    ! TAUCAL label for MASK
      integer        mode           ! Mode of operation variable
      real           mpu0           ! Z axis zeropoint in MPUs
      character      name_grid*80   ! Name of grid structure
      character      name_obs*80    ! Name of obs file
      character      name_shifts*80 ! Name of shifts file
      integer        ndim           ! Number of dimensions in obs data
      integer        ndimr          ! Number of dimensions in residual map
      integer        nelm           ! Number of elements in a mapped array
      integer        ngdim          ! Number of grid file dimensions
      integer        obsrun         ! Run number of observation cube
      integer        odims(2)       ! Dimensions of output file
      integer        optr           ! Dynamic memory pointer for output file
      integer        oslot          ! Dynamic memory slot for output file
      real           pixsize        ! Pixel size defined in MASK
      integer        rdims(2)       ! Dimensions of residual map
      integer        reslot         ! Dynamic memory slot for residual file
      character      rvf_lab*80     ! TAUCAL label for RVF
      integer        status         ! DSA and TAUCAL status flag
      logical        tauexist       ! Flag for existence of TAUCAL file
      integer        totspace       ! Total workspace allocated in bytes.
      character      type*8         ! Data type of obs file
      character      typeg*8        ! Data type of grid file
      character      typer*8        ! Data type of residuals file
      logical        use_grid       ! Flag for existence of grid defined by MASK
      logical        use_res        ! Flag for existence of residual phase map.
      real           wave_want      ! Wavelength to be centred in obs cube
      real           xcen           ! X coord of ring centre
      real           ycen           ! Y coord of ring centre
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find which MASK and CAL images you want...
C
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      call tau_cal_find_struct('cal','mask','mask',tauexist,mask_lab,
     : len_masklab,status)
      if (status .ne. 0) go to 500
C
C     Get run number of OBS image
C
      if (tauexist) then
         call par_rdval('obsrun',1.0,float(max_int),1.0,' ',dummy)
         obsrun = nint(dummy)
      end if
C
C     Get necessary info from TAUCAL obs structure
C
      call tau_cal_rdchr('cal',obsrun,'image','name_obs',tauexist,
     : .true.,80,name_obs,status)
      call tau_cal_rdnum('cal',obsrun,'wave','wave_want',tauexist,
     : .true.,min_float,max_float,'microns',1,wave_want,status)
      call tau_cal_rdnum('cal',obsrun,'dx','dx',tauexist,.true.,
     : min_float,max_float,'pixels',1,dx,status)
      call tau_cal_rdnum('cal',obsrun,'dy','dy',tauexist,.true.,
     : min_float,max_float,'pixels',1,dy,status)
      call tau_cal_rdnum('cal',obsrun,'dz','dz',tauexist,.true.,
     : min_float,max_float,'pixels',1,dz,status)
      call tau_cal_rdnum('cal',obsrun,'xc','xcen',tauexist,.true.,
     : min_float,max_float,'pixels',1,xcen,status)
      call tau_cal_rdnum('cal',obsrun,'yc','ycen',tauexist,.true.,
     : min_float,max_float,'pixels',1,ycen,status)
      call tau_cal_rdnum('cal',obsrun,'cenphase','cenphase_obs',
     : tauexist,.true.,min_float,max_float,'MPUs',1,cen_phase_obs,
     : status)
      call tau_cal_rdnum('cal',obsrun,'fsr','fsr_obs',min_float,
     : tauexist,.true.,max_float,'MPUs',1,fsr_obs, status)
      call tau_cal_rdnum('cal',obsrun,'mpu0','mpu0',tauexist,.true.,
     : min_float,max_float,'MPUs',1,mpu0,status)
      call tau_cal_rdnum('cal',obsrun,'dmpu','dmpu',tauexist,.true.,
     : min_float,max_float,'MPUs/frame',1,dmpu,status)
      if (status .ne. 0) go to 500
C
C     Get necessary info from TAUCAL RVF structure
C
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.gap','gap',
     : tauexist,.true.,min_float,max_float,'microns',1,gap,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.mode','mode',
     : tauexist,.true.,-1.0,1.0,' ',1,dummy,status)
      mode = nint(dummy)
      if (status .ne. 0) go to 500
C
C     Get necessary info from TAUCAL mask structure
C
      call tau_cal_rdnum('cal',0,mask_lab(:len_masklab)//'.pixsize_x',
     : 'pixsize_x',tauexist,.true.,min_float,max_float,'microns',1,
     : pixsize,status)
      if (status .ne. 0) go to 500
C
C     Get camera focus and convert it to microns
C
      call tau_cal_rdnum('cal',0,'camera.foclen','camera_focus',
     : tauexist,.true.,min_float,max_float,'mm',1,focus,status)
      if (status .ne. 0) go to 500
      focus = focus*1000.0
      fscale = pixsize/focus
C
C     Open obs cube and get size info...
C
      call dsa_named_input('image',name_obs,status)
      call ndp_get_image_info('image',.true.,.false.,type,badpix,
     : status)
      call dsa_data_size('image',3,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
C
C     Get residual phase map, if it exists...Check that residual
C     map and input data array have the same sizes.
C
      call par_rdkey('use_res',.true.,use_res)
      if (use_res) then
         call dsa_input('res','name_res',status)
         call ndp_get_image_info('res',.true.,.false.,typer,badpixr,
     :   status)
         call dsa_data_size('res',2,ndimr,rdims,nelm,status)
         if (status .ne. 0) go to 500
         if ((rdims(1) .ne. dims(1)) .or. (rdims(2) .ne. dims(2))) then
            call dsa_wruser
     :      ('Input and Residual data arrays are incompatible')
            call dsa_wrflush
            go to 500
         end if
C
C        Map residual phase map.
C
         call dsa_map_data('res','read','float',address,reslot,status)
         if (status .ne. 0) go to 500
         irespt = dyn_element(address)
C
C     Otherwise create a dummy map
C
      else
         bytes_float = dsa_typesize('float',status)
         totspace = dims(1)*dims(2)*bytes_float
         call dsa_get_workspace(totspace,address,reslot,status)
         if (status .ne. 0) go to 500
         irespt = dyn_element(address)
         call gen_fill(totspace,0,dynamic_mem(irespt))
      end if
C
C     If using grid open it. If not the create a dummy one...
C
      call par_rdkey('use_grid',.true.,use_grid)
      if (use_grid) then
C
C        Find which grid you want...
C
         call tau_cal_find_struct('cal',mask_lab(:len_masklab),'grid',
     :   tauexist,grid_lab,len_gridlab,status)
         if (status .ne. 0) go to 500
C
C        Get the grid file name and open it and check its size
C
         call tau_cal_rdchr('cal',0,
     :   grid_lab(:len_gridlab)//'.name_grid',
     :   'name_grid',tauexist,.true.,80,name_grid,status)
         call dsa_named_input('grid',name_grid,status)
         call ndp_get_image_info('grid',.true.,.false.,typeg,badpixg,
     :   status)
         call dsa_data_size('grid',3,ngdim,gdims,nelm,status)
         if (status .ne. 0) go to 500
         if ((gdims(2) .ne. dims(1)) .or. (gdims(3) .ne. dims(2))) then
            call dsa_wruser
     :      ('Input and Grid data array sizes are incompatible')
            call dsa_wrflush
            go to 500
         end if
C
C        Map grid
C
         call dsa_map_data('grid','read','float',address,gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
C
C     If not using grid, then create a dummy grid
C
      else
         call dsa_get_work_array(2*dims(1)*dims(2),'float',address,
     :   gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
         call tau_dummy_grid(dims(1),dims(2),dynamic_mem(gptr))
      end if
C
C     Open output file
C
      call par_rdchar('name_shifts',' ',name_shifts)
      call dsa_named_output('output',name_shifts,' ',1,1,status)
      if (status .ne. 0) go to 500
C
C     Create output structure
C
      odims(1) = dims(1)
      odims(2) = dims(2)
      call dsa_simple_output('output','data','float',2,odims,status)
      if (status .ne. 0) go to 500
C
C     Map output array
C
      call dsa_map_data('output','write','float',address,oslot,status)
      if (status .ne. 0) go to 500
      optr = dyn_element(address)
C
C     Now do the work...
C
      call sub_phase_map(dims(1),dims(2),dims(3),gap,fsr_obs,mode,
     : mpu0,dmpu,cen_phase_obs,fscale,dynamic_mem(gptr),xcen,ycen,
     : dx,dy,dz,wave_want,dynamic_mem(irespt),dynamic_mem(optr))
C
C     Process TAUCAL
C
      call tau_cal_setchr('cal',obsrun,'name_shifts',tauexist,
     : 80,name_shifts,status)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_phase_map(nx,ny,nz,gap,fsr_obs,mode,mpu0,dmpu,
     : cen_phase_obs,fscale,coords,xcen,ycen,dx,dy,dz,wave_want,
     : resid,map)
C+-----------------------------------------------------------------------------------
C     SUB_PHASE_MAP  --- Subroutine to do the actual work in finding
C                        the total shifts
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NX:              (integer) X dimension of obs cube.
C     (>) NY:              (integer) Y dimension of obs cube.
C     (>) NZ:              (integer) Z dimension of obs cube.
C     (>) GAP:             (real) Etalon spacing.
C     (>) FSR_OBS:         (real) FSR of OBS cube.
C     (>) MODE:            (integer) Mode of operation variable
C     (>) MPU0:            (real) Z axis zeropoint in MPUs
C     (>) DMPU:            (real) Z axis increment in MPUs/frame
C     (>) CEN_PHASE_OBS:   (real) Central phase of OBS data
C     (>) FSCALE:          (real) Pixel size/Camera focal length
C     (>) COORDS:          (real array (COORDS(2,NX,NY)) Grid of coords defined by GRID
C     (>) XCEN:            (real) X centre of rings.
C     (>) YCEN:            (real) Y centre of rings.
C     (>) DX:              (real) Shift in XCEN
C     (>) DY:              (real) Shift in YCEN
C     (>) DZ:              (real) Shift in Z from analysis of RVT
C     (>) WAVE_WANT:       (real) Wavelength to be centred in obs cube.
C     (>) RESID:           (real array (RESID(NX,NY)) Residual phase map.
C     (<) MAP:             (real array (MAP(NX,NY)) Shift results.
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           19-MAR-1990
C     UPDATE:         05-APR-1990 -- Changed so that RVT shifts are read in rather
C                                    than being calculated here.  (JRL)
C                     25-MAY-1990 -- Cosmetic alterations.  (JRL)
C                     29-JUN-1990 -- Minor alterations to call.  (JRL)
C                     26-JUL-1990 -- Mode of operation variable added (JRL)
C                     03-SEP-1990 -- All wavelength calibration taken out.
C                                    Results now stored in MPUs (JRL).
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,mode
      real gap,fsr_obs,cen_phase_obs,fscale,coords(2,nx,ny),
     : xcen,ycen,dx,dy,dz,resid(nx,ny),map(nx,ny),mpu0,dmpu,wave_want
C
C     Local variables
C
      integer i,index,j
      real beta,free_shift,fscale2,fsr,r2,result,sec_theta,
     : xnew,ynew,cenphase_shift
C
C     Now calculate full shift for each spectrum
C
      free_shift = 0.5*float(nz)*dmpu
      beta = float(mode)*2.0*fsr_obs*gap/wave_want
      fscale2 = fscale*fscale
      cenphase_shift = cen_phase_obs - mpu0 + dmpu
      do j = 1,ny
         do i = 1,nx
            xnew = coords(1,i,j) - (xcen - dx)
            ynew = coords(2,i,j) - (ycen - dy)
            r2 = xnew*xnew + ynew*ynew
            sec_theta = sqrt(r2*fscale2 + 1.0)
            fsr = fsr_obs*sec_theta
            result = cenphase_shift + beta*(sec_theta - 1.0)
            result = result + resid(i,j) + dz - free_shift
            if (abs(result) .ge. abs(fsr)) then
               index = int(result/fsr)
               result = result - fsr*float(index)
            end if
            map(i,j) = result
         end do
      end do
      end
