      subroutine prering
C+------------------------------------------------------------------------------
C
C   -------------
C   P R E R I N G
C   -------------
C
C   Description
C   -----------
C   Does an initial guess at ring parameters.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-d and 3-d images only.
C   - SHORT or FLOAT data only supported.  All others
C     are converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values supported.
C   - Batch execution not supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE      Name of TAUCAL file (character)(prompted for)
C
C   RINGRUN      Run number associated with ring data cube. (integer)
C                (prompted for).
C
C   NSER         Number of series to be processed.  There can be a maximum
C                of NSERMAX.  Only necessary for cubes. (integer)(prompted for)
C
C   SER_NOS      Two frame numbers which delimit each series.
C                (integer[2])(prompted for).
C
C   FRAMES_PROC  Two frame numbers for each series which are between the
C                delimiting frame numbers (inclusive).  These are the two
C                frames to be processed in each series.  Only necessary
C                for RVFs (integer[2])(prompted for).
C
C   NBINS        The number of bins for histogramming.  Used for defining
C                the background estimate (integer)(not prompted for).
C
C   NFRAME       The frame number to be used if you are doing an RVT
C                with rings in a cube. (integer)(prompted for).
C
C   NAME_RVF     Name of the output RVF file.  (character)(prompted for).
C
C   NAME_RVT     Name of the output RVT file.  (character)(prompted for).
C
C   TABLE        Name of the LUT for 2-d display. (character)(prompted for).
C
C   USE_ALL      Flag to use all the available rings for RVT. (logical)
C                (prompted for)
C
C   CALRUN       Run number of CAL cube which contains the reference ring for
C                RVT (integer)(prompted for)
C
C   CALFRAME     Frame within CALRUN CAL cube which has the reference ring for
C                RVT (integer)(prompted for)
C
C   SIGS        Number of sigma from the radius vector past which the
C               data is thrown away.  (real)(prompted for)
C
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C   RVF          True if the output structure is an RVF. False if it's an RVT.
C                (prompted for)
C
C   RESTRICT     True if the user wants to restrict the area analysed to
C                just the area between -sigs*sigma --> sigs*sigma from
C                 the radius. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (CAL/RING)   NAME_CAL      Name of the structure containing the rings.
C                              This can either be a 2-d image with a single ring
C                              or a 3-d cube with a whole calibration cube.
C                              (character)(read)
C
C   (CAL)        MPU0          Z axis zeropoint in microprocessor units.
C                              (real)(read)
C
C   (CAL)        DMPU          Z axis increment in microprocessor units/frame
C                              (real)(read)
C
C   (RING)       WAVE          Wavelength of calibration source. (real)(read)
C
C   .RVF         NRVF          Number of RVFs which have been done. (integer)
C                              (read/written)
C
C   .RVF.RVFn    NAME_RVF      Name of the output RVF file. (character)(written)
C
C   .RVF.RVFn    WAVE          Wavelength of calibration source (real)(written)
C
C   .RVF.RVFn    RUN           Run number of CAL cube used for this RVF
C                              (integer)(written)
C
C   .RVF.RVFn    NSER          Number of series nominated in the cube.
C                              (integer)(written)
C
C   .RVF.RVFn    SER_NOS       The delimiting frame numbers for each series
C                              (integer[2,nser])(written)
C
C   .RVF.RVFn    RING_PARS     Estimated parameters for the rings.
C                              (real[10,2*nser])(written)
C
C   .RVT         NRVT          Number of RVTs which have been done. (integer)
C                              (read/written)
C
C   .RVT         NRUNS         Total number of rings available for RVT (integer)
C                              (read)
C
C   .RVT         RUNNOS        All run numbers available for RVT
C                              (integer[nruns])(read)
C
C   .RVT         NRVT          Revised number of RVTs (integer)(written)
C
C   .RVT.RVTn    NAME_RVT      Name of the output RVT file. (character)(written)
C
C   .RVT.RVTn    WAVE          Wavelength of calibration source (real)(written)
C
C   .RVT.RVTn    NRUNS         Number of runs included in this RVT. (integer)
C                              (written)
C
C   .RVT.RVTn    RUNS          Run numbers included in this RVT.
C                              (integer[NRUNS])(written)
C
C   .RVT.RVTn    FRAMES        Frames to be used with each run number included
C                              in this RVT. (integer[NRUNS])(written)
C
C   .RVT.RVTn    RUN_USED      Run number used to get estimate for this RVT
C                              (integer)(written)
C
C   .RVT.RVTn    RING_PARS     Estimated parameters for the rings. (real[10])
C                              (written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested for dimensionality
C   - The user tells the program which frames are the first and
C     last in each series and which frames in that series are to
C     be processed in detail in the program. (For RVFs only).
C   - Each frame nominated to be processed is first histogrammed
C     to define a mean background.
C   - A colour plot is displayed.  The user chooses points on the
C     ring with the cursor.
C   - A least-squares circle is fit to the points chosen.
C   - A cut through the image is taken from the centre of the circle
C     through each of the nominated points.
C   - A moments analysis defines the width of the ring and its peak.
C     A plot of the data and the resulting gaussian is displayed.
C   - Once two frames have been done for a series, an pseudo-RVF is
C     calculated and estimates for R are obtained for each ring in
C     the series.  Estimates for the peak and sigma are obtained from
C     conservation of flux considerations.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_ADD_STRUCT
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FMTCON
C     DSA_FREE_WORKSPACE
C     DSA_GET_FLAG_VALUE
C     DSA_GET_WORKSPACE
C     DSA_MAP_AXIS_DATA
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_NAMED_OUTPUT
C     DSA_OPEN
C     DSA_READ_STRUCT_DEF
C     DSA_RESHAPE_DATA
C     DSA_SET_STRUCT_VAR
C     DSA_SIMPLE_OUTPUT
C     DSA_TYPESIZE
C     DSA_UNMAP
C     DSA_USE_FLAGGED_VALUES
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_CI
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C     NDP_IMAGE_CURSOR
C     NDP_PGBIN
C
C   NAG routines:
C     E04GEF
C     F04JGF
C     X02AJF
C
C   Library PAR:
C     PAR_CNPAR
C     PAR_QUEST
C     PAR_RDARY
C     PAR_RDCHAR
C     PAR_RDKEY
C     PAR_RDVAL
C
C   PGPLOT:
C     PGADVANCE
C     PGBEGIN
C     PGBOX
C     PGEND
C     PGLABEL
C     PGLINE
C     PGSCI
C     PGSCR
C     PGSLW
C     PGWINDOW
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETCHR
C     TAU_CAL_SETNUM
C     TAU_DISPLAY
C     TAU_HISTOGRAM
C
C   Library VAR:
C     VAR_GETCHR
C
C
C   Internal subroutines called
C   ---------------------------
C   COPY_STUFF
C   DRAW_FIT
C   EXTRACT
C   IMAGE_PARS
C   LSFUN2
C   PTRANSFER_I
C   PTRANSFER_R
C   RVT_EXTRA
C   SUB_PRERING
C   WIDTHS
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C   INTEGER*2 / DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   03-MAY-90   - Original program
C   08-MAY-90   - Slight modification so RVT input is from a cube.  (JRL)
C   22-MAY-90   - Modified to cope with both SHORT and FLOAT input arrays. (JRL)
C   03-JUL-90   - TAUCAL pasted in.  Some modifications made in order to adhere
C                 to the TAUCAL scheme were made.  Mostly very minor.  (JRL)
C   25-JUL-90   - User now specifies LUT.  Minor modifications to 2-d plotting
C                 routine.  Better fitting procedure adopted. (JRL)
C   03-SEP-90   - Minor modifications to IMAGE_PARS (JRL)
C   06-NOV-90   - Modified to use RVT substructure in TAUCAL. (JRL)
C   15-NOV-90   - Modified to use RVF substructure in TAUCAL. (JRL)
C   19-NOV-90   - New TAUCAL routines pasted in.  (JRL)
C   11-DEC-90   - Modified so that estimates are done for all RVT rings. (JRL)
C   07-FEB-91   - Cleaned up a few bugs. (JRL)
C   15-JUL-91   - Modified to use new versions of PAR routines in FIGARO
C                 version 3. (JRL)
C   02-SEP-91   - RESTRICT and SIGS parameters added. Fixed bug in subroutine
C                 WIDTHS which caused the fitting routine to fall apart when
C                 there was a weak ring and a bright ring on the same 
C                 frame. (JRL)
C   05-JAN-92   - Procedure for adding .RVF.RVFn and .RVT.RVTn substructures
C                 in TAUCAL file altered slightly. (JRL)
C   29-JAN-92   - Minor modifications to include new plotting routines. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions.
C
      integer          dsa_typesize
      integer          dyn_element
      character        ich_ci*1
      integer          ich_len
      logical          par_quest
C
C     Local parameters
C
      integer nsermax
      parameter (nsermax=10)
C
C     Local variables
C
      integer        address                ! Dynamic memory address
      character      arvf*1                 ! Number of RVFs in character form
      character      arvt*1                 ! Number of RVTs in character form
      integer        axptr1                 ! Dynamic memory axis pointer
      integer        axptr2                 ! Dynamic memory axis pointer
      integer        axslot1                ! Dynamic memory axis slot
      integer        axslot2                ! Dynamic memory axis slot
      logical        badpix                 ! Flag for magic values in image file
      integer        bytes_float            ! Number of bytes in FLOAT data type
      integer        bytes_int              ! Number of bytes in INT data type
      integer        calframe               ! Frame number within cal cube for RVT
      integer        calrun                 ! Run number of cal cube
      real           dumdata(2,nsermax)     ! Dummy data array
      integer        dims(3)                ! Dimensions of image file
      integer        dimdummy(2)            ! Dimensions of dumdata
      real           dmpu                   ! Z axis increment in MPUs
      real           dum                    ! Real dummy variable
      real           dummy(2)               ! Real dummy array
      real           d1                     ! Real dummy variable
      real           d2                     ! Real dummy variable
      real           frame_nos(4096)        ! Frame numbers for TAUCAL
      integer        i                      ! integer dummy counter
      integer        idelim(2,nsermax)      ! Delimiters of series
      logical        include                ! Flag to include a given run number
      integer        iproc_frame(2,nsermax) ! Frames to be processed
      integer        iptr                   ! Dynamic memory pointer
      integer        islot                  ! Dynamic memory slot
      character      message*80             ! Character variable for user messages
      real           mpu0                   ! Z axis zero point in MPUs
      character      name_cal*80            ! Name of calibration ring cube
      character      name_rvf*80            ! Name of output RVF
      character      name_rvt*80            ! Name of output RVT
      integer        nbins                  ! Number of bins for histogramming
      integer        ndim                   ! Number of dimension in image file
      integer        nelm                   ! Number of elements in image file
      integer        nframe                 ! If doing RVT, frame number to be processed
      integer        nout                   ! Number of output lines in RVT
      integer        nruns                  ! Number of runs to process for RVT
      integer        nrvf                   ! Number of RVFs already done
      integer        nrvt                   ! Number of RVTs already done
      integer        nser                   ! Number of series
      real           numbers(4096)          ! Run numbers for single rings
      integer        odims(2)               ! Dimensions of output file
      integer        optr                   ! Dynamic memory pointer
      integer        oslot                  ! Dynamic memory slot
      logical        restrict               ! Flag to restrict region of final analysis
      integer        ringrun                ! Run number of ring cube
      logical        rvf                    ! Is this an RVF or an RVT?
      real           sigs                   ! Number of sigs for restricted analysis
      integer        status                 ! DSA status flag
      character      table*30               ! LUT for 2-d plot
      logical        tauexist               ! Flag for existence of TAUCAL file
      integer        totspace               ! Total amount of workspace needed
      character      type*8                 ! Data type of image file
      logical        use_all                ! Flag to use all runs for RVT
      real           wave                   ! Wavelength of calibration source
      real           work_num(4096)         ! Run numbers for single rings
      integer        wptr                   ! Dynamic memory workspace pointer
      integer        wpt1                   ! Dynamic memory workspace pointer
      integer        wpt2                   ! Dynamic memory workspace pointer
      integer        wslot                  ! Dynamic memory workspace slot
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find out whether the output file is an RVF or an RVT
C
      call par_rdkey('rvf',.true.,rvf)
C
C     If this is an RVF then get the run number for the cal cube
C
      if (rvf) then
         if (tauexist) then
            call par_rdval('ringrun',1.0,float(max_int),1.0,' ',dum)
            ringrun = nint(dum)
         end if
C
C     For RVT find out how many rings there are total and ask if you want to
C     use them all. If not using TAUCAL then you naturally want to use
C     all of the runs you've just typed in, otherwise, why have you
C     just typed them in?
C
      else
         call tau_cal_rdnum('cal',0,'rvt.nruns','nruns',tauexist,.true.,
     :   0.0,float(max_int),' ',1,dum,status)
         nruns = nint(dum)
         call tau_cal_rdnum('cal',0,'rvt.runnos','runnos',tauexist,
     :   .true.,0.0,float(max_int),' ',nruns,numbers,status)
         if (status .ne. 0) go to 500
         if (tauexist) then
            write(message,100) nruns
  100       format('There are',i3,' runs available')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call par_rdkey('use_all',.true.,use_all)
         else
            use_all = .true.
         end if
C
C        If you aren't using them all, then find out which ones you want...
C
         nout = 0
         do i = 1,nruns
            if (.not. use_all) then
               write(message,105) nint(numbers(i))
  105          format('Include run',i6,'?')
               include = par_quest(message,.true.)
            else
               include = .true.
            end if
            if (include) then
               nout = nout + 1
               work_num(nout) = numbers(i)
            end if
         end do
         do i = 1,nout
            numbers(i) = work_num(i)
         end do
C
C       Find out which frame number of the cal cube to use
C
         call par_rdval('calrun',float(min_int),float(max_int),1.0,
     :   ' ',dum)
         calrun = nint(dum)
         call par_rdval('calframe',float(min_int),float(max_int),1.0,
     :   ' ',dum)
         calframe = nint(dum)
C
C        Put cal cube things into your arrays
C
         nout = nout + 1
         numbers(nout) = float(calrun)
         frame_nos(nout) = float(calframe)
C
C        Process the first ring in the series
C
         ringrun = nint(numbers(1))
      end if
C
C     Get LUT file name
C
      call par_rdchar('table',' ',table)
C
C     Ask if you want to restrict the region for the fit.  This may be necessary in
C     the event of a weak ring competing against a strong one on the same frame.
C
      call par_rdkey('restrict',.false.,restrict)
      if (restrict) call par_rdval('sigs',0.0,max_float,3.0,'sigmas',
     : sigs)
C
C     Check and make sure the input file is the right type
C
      if (rvf) then
         call tau_cal_check_run('cal',ringrun,'cal',tauexist,status)
      else
         call tau_cal_check_run('cal',ringrun,'ring',tauexist,status)
      end if
      if (status .ne. 0) go to 500
C
C     Open IMAGE file for ring cube.
C
      call tau_cal_rdchr('cal',ringrun,'image','name_cal',tauexist,
     : .true.,80,name_cal,status)
      call dsa_named_input('image',name_cal,status)
      if (status .ne. 0) go to 500
C
C     Check image dimensions
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',3,ndim,dims,nelm,status)
      if (badpix) call dsa_use_flagged_values('image',status)
      if (status .ne. 0) go to 500
      if (ndim .lt. 2) then
         call dsa_wruser(' Image has unacceptable dimensions.')
         call dsa_wrflush
         go to 500
      end if
      if (ndim .eq. 2) dims(3) = 1
C
C     If doing RVT and the input data is 3-d, then specify the frame number now
C
      if ((.not. rvf) .and. (ndim .ne. 2)) then
         call par_rdval('nframe',1.0,float(dims(3)),1.0,' ',dum)
         nframe = nint(dum)
      else
         nframe = 1
      end if
C
C     Map IMAGE data array.
C
      if (type .eq. 'SHORT') then
         call dsa_map_data('image','read','short',address,islot,status)
      else
         call dsa_map_data('image','read','float',address,islot,status)
      end if
      if (status .ne. 0) go to 500
      iptr = dyn_element(address)
C
C     Map the image axes
C
      call dsa_map_axis_data('image',1,'read','float',address,axslot1,
     : status)
      axptr1 = dyn_element(address)
      call dsa_map_axis_data('image',2,'read','float',address,axslot2,
     : status)
      axptr2 = dyn_element(address)
C
C     Get the frame numbers which delimit each series if this is
C     in fact a calibration cube.
C
      if (rvf) then
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         write(message,110) nsermax
  110    format('A maximum of',i5,' series are allowed.')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         call par_rdval('nser',1.0,float(nsermax),2.0,' ',dum)
         nser = nint(dum)
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('Type the frame numbers which delimit each')
         call dsa_wruser(' series.  ')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         do i = 1,nser
            write(message,115) i
  115       format('First and last frames in series no.',i4)
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call par_cnpar('ser_nos')
            call par_rdary('ser_nos',0.0,float(dims(3)),'n',
     :      'frame numbers',2,2,dummy)
            idelim(1,i) = nint(dummy(1))
            idelim(2,i) = nint(dummy(2))
         end do
C
C        Now get the frame numbers for the rings to process in
C        each series...
C
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser
     :   ('Now for each series nominate two ring frames')
         call dsa_wrflush
         call dsa_wruser('to be processed')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         do i = 1,nser
            d1 = float(idelim(1,i))
            d2 = float(idelim(2,i))
            write(message,120) i,idelim(1,i),idelim(2,i)
  120       format('Series no.',i3,' frame numbers between',i4,' and',
     :      i4)
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call par_cnpar('frames_proc')
            call par_rdary('frames_proc',d1,d2,'n','frame numbers',
     :      2,2,dummy)
            iproc_frame(1,i) = nint(dummy(1))
            iproc_frame(2,i) = nint(dummy(2))
         end do
C
C        Get Z axis zero point and increment in MPUs
C
         call tau_cal_rdnum('cal',ringrun,'mpu0','mpu0',tauexist,
     :   .true.,min_float,max_float,'MPUs',1,mpu0,status)
         call tau_cal_rdnum('cal',ringrun,'dmpu','dmpu',tauexist,
     :   .true.,min_float,max_float,'MPUs/frame',1,dmpu,status)
         if (status .ne. 0) go to 500
      end if
C
C     Get number of bins for background histogramming
C
      call par_rdval('nbins',float(min_int+1),float(max_int-1),
     : 100.0,' ',dum)
      nbins = nint(dum)
C
C     Open OUTPUT file for radius-vs-frame or radius-vs-time data.
C
      if (rvf) then
         call par_rdchar('name_rvf',' ',name_rvf)
         call dsa_named_output('output',name_rvf,' ',1,1,status)
      else
         call par_rdchar('name_rvt',' ',name_rvt)
         call dsa_named_output('output',name_rvt,' ',1,1,status)
      end if
C
C     Create output structure with a data array.
C
      odims(1) = 10
      if (rvf) then
         odims(2) = dims(3)
      else
         odims(2) = nout
      end if
      call dsa_simple_output('output','data','float',2,odims,status)
      if (status .ne. 0) go to 500
C
C     Map output array
C
      call dsa_map_data('output','write','float',address,oslot,
     : status)
      if (status .ne. 0) go to 500
      optr = dyn_element(address)
C
C     Now do the work...
C
      call sub_prering(iptr,type,dims(1),dims(2),dims(3),restrict,sigs,
     : nsermax,nser,idelim,iproc_frame,nframe,rvf,nbins,ringrun,
     : dynamic_mem(axptr1),dynamic_mem(axptr2),mpu0,dmpu,odims(1),
     : odims(2),table,badpix,dynamic_mem(optr))
C
C     Unmap axis arrays
C
      call dsa_unmap(axslot1,status)
      call dsa_unmap(axslot2,status)
      if (status .ne. 0) go to 500
C
C     If this is an RVT then get an initial estimate for the rest
C     of the rings...First close the old file
C
      if (.not. rvf) then
         frame_nos(1) = float(nframe)
         do i = 2,nout
            call dsa_unmap(islot,status)
            if (badpix) call dsa_post_process_flagged_values('image',
     :      status)
            call dsa_close_structure('image',status)
            if (status .ne. 0) go to 500
C
C           Open IMAGE file for ring
C
            if (.not. tauexist) then
               write(message,125) nint(numbers(i))
  125          format('Run number',i6)
               call dsa_wruser(message(:ich_len(message)))
               call dsa_wrflush
            end if
            ringrun = nint(numbers(i))
            call tau_cal_rdchr('cal',ringrun,'image','name_cal',
     :      tauexist,.true.,80,name_cal,status)
            call dsa_named_input('image',name_cal,status)
            if (status .ne. 0) go to 500
C
C           Check image dimensions
C
            call ndp_get_image_info('image',.true.,.false.,type,badpix,
     :      status)
            call dsa_data_size('image',3,ndim,dims,nelm,status)
            if (badpix) call dsa_use_flagged_values('image',status)
            if (status .ne. 0) go to 500
            if (ndim .lt. 2) then
               call dsa_wruser(' Image has unacceptable dimensions.')
               call dsa_wrflush
               go to 500
            end if
            if (ndim .eq. 2) dims(3) = 1
C
C           If the input data is 3-d, then specify the frame number now
C
            if (i .ne. nout) then
               if (ndim .ne. 2) then
                  call par_cnpar('nframe')
                  call par_rdval('nframe',1.0,float(dims(3)),1.0,' ',
     :            dum)
                  nframe = nint(dum)
               else
                  nframe = 1
               end if
               frame_nos(i) = float(nframe)
            else
               nframe = frame_nos(nout)
            end if
C
C           Map the input data
C
            if (type .eq. 'SHORT') then
               call dsa_map_data('image','read','short',address,islot,
     :         status)
            else
               call dsa_map_data('image','read','float',address,islot,
     :         status)
            end if
            if (status .ne. 0) go to 500
            iptr = dyn_element(address)
C
C           Next get a bit of workspace
C
            bytes_float = dsa_typesize('float',status)
            bytes_int = dsa_typesize('int',status)
            totspace = dims(1)*dims(2)*bytes_float + nbins*bytes_int
            call dsa_get_workspace(totspace,address,wslot,status)
            if (status .ne. 0) go to 500
            wpt1 = dyn_element(address)
            wpt2 = wpt1 + dims(1)*dims(2)*bytes_float
C
C           Now do the work
C
            call rvt_extra(iptr,type,dims(1),dims(2),dims(3),nbins,
     :      nframe,restrict,sigs,dynamic_mem(wpt1),dynamic_mem(wpt2),
     :      dynamic_mem(axptr1),dynamic_mem(axptr2),table,badpix,
     :      numbers(i),i,odims(1),odims(2),dynamic_mem(optr))
C
C           Release the workspace
C
            call dsa_free_workspace(wslot,status)
            if (status .ne. 0) go to 500
         end do
      end if
C
C     Process TAUCAL file for RVF
C
      if (rvf) then
         if (tauexist) then
C
C           First find out how many RVFs have already been done and increment
C           by one...
C
            call tau_cal_rdnum('cal',0,'rvf.nrvf','nrvf',
     :      tauexist,.false.,0.0,float(max_int),' ',1,dum,status)
            nrvf = nint(dum)
            if (status .ne. 0) go to 500
            nrvf = nrvf + 1
            dum = float(nrvf)
            call tau_cal_setnum('cal',0,'rvf.nrvf',tauexist,1,1,1,dum,
     :      status)
            arvf = ich_ci(nrvf)
C
C           Now create a sub-structure for this RVF.
C
            call dsa_read_struct_def('TAU_RVF_SUB.DEF',status)
            call dsa_set_struct_var('nnnn','.rvf.rvf'//arvf,status)
            call dsa_add_structure('cal','tau_rvf_sub','rvf_sub_struct',
     :      status)
            if (status .ne. 0) go to 500
            call tau_cal_rdnum('cal',ringrun,'wave','wave',tauexist,
     :      .false.,min_float,max_float,'microns',1,wave,status)
         end if
C
C        Write name of RVF file, its wavelength and the run number of the
C        cal cube used
C
         call tau_cal_setchr('cal',0,'rvf.rvf'//arvf//'.name_rvf',
     :   tauexist,80,name_rvf,status)
         call tau_cal_setnum('cal',0,'rvf.rvf'//arvf//'.wave',
     :   tauexist,1,1,1,wave,status)
         call tau_cal_setnum('cal',0,'rvf.rvf'//arvf//'.run',
     :   tauexist,1,1,1,float(ringrun),status)
C
C        Write out info on the series
C
         call tau_cal_setnum('cal',0,'rvf.rvf'//arvf//'.nser',
     :   tauexist,1,1,1,float(nser),status)
         do i = 1,nser
            dumdata(1,i) = float(idelim(1,i))
            dumdata(2,i) = float(idelim(2,i))
         end do
         dimdummy(1) = 2
         dimdummy(2) = nser
         call tau_cal_setnum('cal',0,'rvf.rvf'//arvf//'.ser_nos',
     :   tauexist,2,dimdummy,2*nser,dumdata,status)
C
C        Get a bit of workspace
C
         bytes_float = dsa_typesize('float',status)
         totspace = 2*nser*10.0*bytes_float
         call dsa_get_workspace(totspace,address,wslot,status)
         wptr = dyn_element(address)
C
C        Copy the ring estimates over to TAUCAL.
C
         do i = 1,nser
            call copy_stuff(dynamic_mem(optr),odims(1),odims(2),
     :      iproc_frame(1,i),float(iproc_frame(1,i)),10,2*nser,2*i-1,
     :      dynamic_mem(wptr))
            call copy_stuff(dynamic_mem(optr),odims(1),odims(2),
     :      iproc_frame(2,i),float(iproc_frame(2,i)),10,2*nser,2*i,
     :      dynamic_mem(wptr))
         end do
         dimdummy(1) = 10
         dimdummy(2) = 2*nser
         call tau_cal_setnum('cal',0,'rvf.rvf'//arvf//'.ring_pars',
     :   tauexist,2,dimdummy,20*nser,dynamic_mem(wptr),status)
         call dsa_free_workspace(wslot,status)
         if (status .ne. 0) go to 500
C
C     For RVT fits...
C
      else
         if (tauexist) then
C
C           First find out how many RVTs have already been done and increment
C           by one...
C
            call tau_cal_rdnum('cal',0,'rvt.nrvt','nrvt',
     :      tauexist,.false.,0.0,float(max_int),' ',1,dum,status)
            nrvt = nint(dum)
            if (status .ne. 0) go to 500
            nrvt = nrvt + 1
            dum = float(nrvt)
            call tau_cal_setnum('cal',0,'rvt.nrvt',tauexist,1,1,1,dum,
     :      status)
            arvt = ich_ci(nrvt)
C
C           Now create a sub-structure for this RVT and include the name
C           of the output file and its wavelength.
C
            call dsa_read_struct_def('TAU_RVT_SUB.DEF',status)
            call dsa_set_struct_var('nnnn','.rvt.rvt'//arvt,status)
            call dsa_add_structure('cal','tau_rvt_sub','rvt_sub_struct',
     :      status)
            if (status .ne. 0) go to 500
            call tau_cal_setchr('cal',0,'rvt.rvt'//arvt//'.name_rvt',
     :      tauexist,80,name_rvt,status)
            call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.run_used',
     :      tauexist,1,1,1,float(ringrun),status)
            call tau_cal_rdnum('cal',ringrun,'wave','wave',tauexist,
     :      .false.,min_float,max_float,'microns',1,wave,status)
            call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.wave',
     :      tauexist,1,1,1,wave,status)
         end if
C
C        Copy the run numbers to TAUCAL.
C
         call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.nruns',
     :   tauexist,1,1,1,float(nout),status)
         call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.runs',
     :   tauexist,1,nout,nout,numbers,status)
         call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.frames',
     :   tauexist,1,nout,nout,frame_nos,status)
         call tau_cal_setnum('cal',0,'rvt.rvt'//arvt//'.ring_pars',
     :   tauexist,2,odims,odims(1)*odims(2),dynamic_mem(optr),status)
         if (status .ne. 0) go to 500
      end if
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end



      subroutine sub_prering(iptr,type,nx,ny,nz,restrict,sigs,nsermax,
     : nser,idelim,iproc_frame,nframe,rvf,nbins,ringrun,xaxis,yaxis,
     : mpu0,dmpu,nox,noy,table,badpix,output)
C+--------------------------------------------------------------------------------------
C     SUB_PRERING --- Routine to do the actual work in finding the rough
C                     parameters of the rings.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IPTR:       (integer) Dynamic memory pointer for input data.
C     (>) TYPE:       (character) Input data type
C     (>) NX:         (integer) X dimension of data.
C     (>) NY:         (integer) Y dimension of data.
C     (>) NZ:         (integer) Z dimension of data.
C     (>) RESTRICT:   (logical) TRUE for restricted analysis
C     (>) SIGS:       (real) Number of sigmas for restricted analysis
C     (>) NSERMAX:    (integer) Maximum number of series allowed
C     (>) NSER:       (integer) Number of series defined by the user
C     (>) IDELIM:     (integer array IDELIM(2,NSERMAX)) Defines the endpoints of each series
C     (>) IPROC_FRAME:(integer array IPROC_FRAME(2,NSERMAX)) Frames from each series
C                      to be processed.
C     (>) NFRAME      (integer) Frame to process if doing RVT
C     (>) RVF:        (logical) True if output file is RVF. False if output is RVT.
C     (>) NBINS:      (integer) Number of bins for histogramming.
C     (>) RINGRUN:    (integer) Run number of RVT ring being fit
C     (>) XAXIS:      (real array XAXIS(NX)) Array of X axis data values.
C     (>) YAXIS:      (real array YAXIS(NY)) Array of Y axis data values.
C     (>) MPU0:       (real) Z axis zeropoint in MPUs
C     (>) DMPU:       (real) Z axis increment in MPUs/pixel
C     (>) NOX:        (integer) First dimension of output array
C     (>) NOY:        (integer) Second dimension of output array
C     (>) TABLE:      (character) LUT for 2-d plots
C     (>) BADPIX:     (logical) Flag for magic values
C     (<) OUTPUT:     (real array OUTPUT(NOX,NOY)) Array with output parameters.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        03-MAY-1990
C     UPDATE:      22-MAY-1990 -- Modified to cope with either SHORT or FLOAT
C                                 input structures.  (JRL)
C                  25-JUL-1990 -- User specified LUT included.  (JRL)
C                  06-NOV-1990 -- Changed to cope with new RVT algorithm. (JRL)
C                  02-SEP-1991 -- Added RESTRICT parameter (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,nox,noy,nbins,nsermax,nser,iptr,nframe,ringrun
      integer idelim(2,nsermax),iproc_frame(2,nsermax)
      real output(nox,noy),xaxis(nx),yaxis(ny),mpu0,dmpu,sigs
      logical rvf,restrict,badpix
      character type*(*),table*(*)
C
C     Functions.
C
      integer          dyn_element
      integer          dsa_typesize
C
C     Local variables
C
      integer i,bytes_float,bytes_int,totspace,address,
     : slot,wpt1,wpt2,j,status
      real slope,aint,xc1,yc1,radius(2),sigma(2),peaks(2),flag,
     : back1,r_sig_const,peak_const,xc,yc,rad,sig,peak,back,r2
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Initialise output array
C
      do i = 1,noy
         if (rvf) then
            output(1,i) = mpu0 + float(i-1)*dmpu
         else
            output(1,i) = float(ringrun)
         end if
         do j = 2,9
            output(j,i) = 0.0
         end do
         output(10,i) = 2.0
      end do
C
C     Get workspace for the fitting and histogramming routines...
C
      bytes_float = dsa_typesize('float',status)
      bytes_int = dsa_typesize('int',status)
      totspace = nx*ny*bytes_float + nbins*bytes_int
      call dsa_get_workspace(totspace,address,slot,status)
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + nx*ny*bytes_float
C
C     Now get some rough fit parameters for the rings.  If it's an
C     RVT you're after, then just fit the first one. (All the others
C     should be roughly the same...)
C
      if (.not. rvf) then
         call image_pars(iptr,type,nx,ny,nz,nbins,nframe,restrict,sigs,
     :   dynamic_mem(wpt1),dynamic_mem(wpt2),xaxis,yaxis,table,.false.,
     :   badpix,xc,yc,rad,sig,peak,back)
         output(2,1) = xc
         output(3,1) = yc
         output(4,1) = rad
         output(5,1) = sig
         output(6,1) = peak
         output(7,1) = back
         output(8,1) = 1.0
         output(9,1) = 0.0
         output(10,1) = -1.0
C
C     Get image parameters for RVF output file...
C     First get parameters for two rings in each series.
C
      else
         do i = 1,nser
            xc = 0.0
            yc = 0.0
            back = 0.0
            do j = 1,2
               call image_pars(iptr,type,nx,ny,nz,nbins,
     :         iproc_frame(j,i),restrict,sigs,dynamic_mem(wpt1),
     :         dynamic_mem(wpt2),xaxis,yaxis,table,.false.,badpix,xc1,
     :         yc1,radius(j),sigma(j),peaks(j),back1)
               xc = xc + xc1
               yc = yc + yc1
               back = back + back1
            end do
            xc = 0.5*xc
            yc = 0.5*yc
            back = 0.5*back
C
C           Calculate pseudo RVF
C
            slope = (radius(2)**2 - radius(1)**2)/
     :      (float(iproc_frame(2,i)-iproc_frame(1,i)))
            aint = radius(1)**2 - slope*float(iproc_frame(1,i))
C
C           Calculate constants on the basis that:
C               radius*width = const
C
            r_sig_const = 0.5*(radius(1)*sigma(1) + radius(2)*sigma(2))
            peak_const = 0.5*(peaks(1) + peaks(2))
C
C           Write estimates to output file
C
            do j = idelim(1,i),idelim(2,i)
               output(2,j) = xc
               output(3,j) = yc
               r2 = slope*float(j) + aint
               flag = -1.0
               if (r2 .gt. 0.0) then
                  rad = sqrt(r2)
               else
                  rad = 0.0
                  flag = 2.0
               end if
               output(4,j) = rad
               if (rad .ne. 0.0) then
                  sig = r_sig_const/rad
               else
                  sig = 0.0
                  flag = 2.0
               end if
               output(5,j) = sig
               peak = peak_const
               output(6,j) = peak
               output(7,j) = back
               output(8,j) = 1.0
               output(9,j) = 0.0
               output(10,j) = flag
            end do
         end do
      end if
      call dsa_free_workspace(slot,status)
      end


      subroutine image_pars(iptr,type,nx,ny,nz,nbins,iframe,restrict,
     : sigs,temp,freq,xaxis,yaxis,table,repeat,badpix,xc,yc,rad,sig,
     : peak,back)
C+--------------------------------------------------------------------------------------
C     IMAGE_PARS  --- Subroutine to calculate the parameters of the rings.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IPTR:      (integer) Dynamic memory pointer for input data
C     (>) TYPE:      (character) Input data type
C     (>) NX:        (integer) X dimension of data.
C     (>) NY:        (integer) Y dimension of data.
C     (>) NZ:        (integer) Z dimension of data.
C     (>) NBINS:     (integer) Number of bins for histogramming.
C     (>) IFRAME:    (integer) Frame number to be analysed
C     (>) RESTRICT:  (logical) TRUE for restricted analysis
C     (>) SIGS:      (real) Number of sigma for restricted analysis
C     (W) TEMP:      (real array TEMP(NX,NY)) Workspace.
C     (W) FREQ:      (integer array FREQ(NBINS)) Workspace
C     (>) XAXIS:     (real array XAXIS(NX)) Array of X axis data values.
C     (>) YAXIS:     (real array YAXIS(NY)) Array of Y axis data values.
C     (>) TABLE:     (character) Name of LUT for 2-d display routine.
C     (>) REPEAT:    (logical) TRUE if only the only the final fit is being repeated
C     (>) BADPIX:    (logical) Flag for magic values in data
C     (<) XC:        (real) Calculated value of X centre
C     (<) YC:        (real) Calculated value of Y centre
C     (<) RAD:       (real) Calculated value of radius
C     (<) SIG:       (real) Calculated value of ring width.
C     (<) PEAK:      (real) Calculated value of ring peak
C     (<) BACK:      (real) Calculated value of background.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        03-MAY-1990
C     UPDATE:      25-JUL-1990 -- Added in LUT for 2-d plotting routine (JRL)
C                  03-SEP-1990 -- Minor modifications to aid plotting routine (JRL)
C                  11-DEC-1990 -- Added REPEAT option. (JRL)
C                  02-SEP-1991 -- Added RESTRICT parameter. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iframe,nbins,iptr
      integer freq(nbins)
      real xc,yc,rad,sig,peak,back,temp(nx,ny),xaxis(nx),
     : yaxis(ny),sigs
      logical repeat,restrict,badpix
      character type*(*),table*(*)
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
      double precision x02ajf
C
C     Local parameters
C
      integer npts
      parameter (npts=20)
C
C     Local variables
C
      integer i,nin,nused,hist_stat,irank,ifail,spx2d(2),epx2d(2),
     : bytes_float,totspace,address,slot,wpt1,wpt2,size,ipts,
     : status,wpt3,wpt4,bytes_double
      real amin,amax,hiclip,loclip,low,up,mean,sigma,xpos(npts),
     : ypos(npts),r2,sta2d(2),end2d(2),ximv(2),yimv(2),x,y,xlast,ylast,
     : lowplt,highplt,d1,d2,d3,d4,magic_float
      double precision condition(npts,3),rhs(npts),tol,error,
     : nag_work(12)
      character message*80
      logical svd,quit
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Common block for data pointers
C
      common /pointers/ wpt1,wpt2
C
C     First transfer data to a real array...Get min and max values.
C
      if (type(1:5) .eq. 'SHORT') then
         call ptransfer_i(dynamic_mem(iptr),nx,ny,nz,iframe,temp,amin,
     :   amax)
      else
         call ptransfer_r(dynamic_mem(iptr),nx,ny,nz,iframe,temp,amin,
     :   amax)
      end if
C
C     Get magic value
C
      call dsa_get_flag_value('FLOAT',magic_float,status)
      if (status .ne. 0) go to 500
C
C     If doing the full 2-d analysis...
C
      if (.not. repeat) then
C
C        Now histogram the frame to find the background value
C
         hiclip = amax + 0.1
         loclip = amin - 0.1
         call tau_histogram(temp,nx,ny,loclip,hiclip,magic_float,nbins,
     :   nin,nused,low,up,freq,mean,sigma,hist_stat)
C
C        Do an non-symmetric clip to get rid of the ring pixels
C
         hiclip = mean + 2.0*sigma
         call tau_histogram(temp,nx,ny,loclip,hiclip,magic_float,nbins,
     :   nin,nused,low,up,freq,mean,sigma,hist_stat)
         back = mean
C
C        Do a display.
C
         lowplt = amin
         highplt = amax
         call tau_display(temp,nx,ny,lowplt,highplt,xaxis,yaxis,
     :   table,badpix,spx2d,epx2d,sta2d,end2d,ximv,yimv,status)
         if (status .ne. 0) then
            call dsa_wruser('Plotting routine failure...')
            call dsa_wrflush
            go to 500
         end if
C
C        Now get the cursor positions for up to npts points on the ring.
C        A minimum of 3 points are needed for the fit...
C
         ipts = 0
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser
     :   ('Use the cursor to identify at least 3 points')
         call dsa_wrflush
         write(message,100) npts
  100    format('The maximum number of points allowed is:',i4)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
         do while (ipts .lt. 3)
            call ndp_image_cursor(xaxis,yaxis,nx,ny,sta2d,end2d,spx2d,
     :      epx2d,npts,'cp',ximv,yimv,xpos,ypos,ipts,xlast,ylast,quit)
            if (ipts .lt. 3) then
               call dsa_wruser
     :         ('At least 3 points are needed.  Start again please')
               call dsa_wrflush
            end if
         end do
C
C        Now do a fit to find a least-squares circle.  First set up
C        the condition equations...
C
         do i = 1,ipts
            x = xpos(i)
            y = ypos(i)
            condition(i,1) = dble(x*x + y*y)
            condition(i,2) = dble(x)
            condition(i,3) = dble(y)
            rhs(i) = 1.0d0
         end do
         tol = x02ajf()
         ifail = 1
         call f04jgf(ipts,3,condition,npts,rhs,tol,svd,error,irank,
     :    nag_work,12,ifail)
         if (ifail .ne. 0) then
            call dsa_wruser('Fit failed...')
            call dsa_wrflush
            go to 500
         end if
C
C        Now find the circle parameters and write them out
C
         xc = -0.5*sngl(rhs(2)/rhs(1))
         yc = -0.5*sngl(rhs(3)/rhs(1))
         r2 = sngl(1.0d0/rhs(1)) + xc*xc + yc*yc
         if (r2 .le. 0.0) then
            call dsa_wruser('Fitted radius < 0...')
            call dsa_wrflush
            go to 500
         end if
         rad = sqrt(r2)
         call dsa_wruser('Fitted circle has the following parameters')
         call dsa_wrflush
         write(message,110) xc,yc,rad
  110    format('Centre at:',2f9.3,'  Radius =',f9.3)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        Draw the fit...
C
         call draw_fit(xc,yc,rad)
C
C        Find the maximum distance from the centre to a corner of the frame
C
         d1 = sqrt((xc-1.0)**2 + (yc-1.0)**2)
         d2 = sqrt((xc-float(nx))**2 + (yc-1.0)**2)
         d3 = sqrt((xc-float(nx))**2 + (yc-float(ny))**2)
         d4 = sqrt((xc-1.0)**2 + (yc-float(ny))**2)
         d1 = max(d1,d2,d3,d4)
      end if
C
C     Find the width of the ring pattern now...
C
      bytes_float = dsa_typesize('float',status)
      bytes_double = dsa_typesize('double',status)
      size = nint(d1)
      totspace = 2*size*bytes_float + size*4*bytes_double
      call dsa_get_workspace(totspace,address,slot,status)
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + size*bytes_float
      wpt3 = wpt2 + size*bytes_float
      wpt4 = wpt3 + size*bytes_double
      call widths(temp,nx,ny,xc,yc,xpos,ypos,ipts,size,restrict,sigs,
     : dynamic_mem(wpt1),dynamic_mem(wpt2),dynamic_mem(wpt3),
     : dynamic_mem(wpt4),peak,sig,back,rad,status)
      if (status .ne. 0) then
         status = 0
         call dsa_wruser('Gauss fit failed....')
         call dsa_wrflush
      end if
      call dsa_free_workspace(slot,status)
  500 end


      subroutine widths(array,nx,ny,xc,yc,xpos,ypos,ipts,size,restrict,
     : sigs,work1,work2,rhs,condition,peak,width,back,rad,status)
C+--------------------------------------------------------------------------------------
C     WIDTHS      --- Routine to calculate the width and peak of a ring.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) ARRAY:     (real array ARRAY(NX,NY)) Input data frame.
C     (>) NX:        (integer) X dimension of data.
C     (>) NY:        (integer) Y dimension of data.
C     (>) XC:        (real) X centre
C     (>) YC:        (real) Y centre
C     (>) XPOS:      (real array XPOS(IPTS)) X position of points defining ring
C     (>) YPOS:      (real array YPOS(IPTS)) Y position of points defining ring
C     (>) IPTS:      (integer) Number of points defining the ring
C     (>) SIZE:      (integer) Size of workspace
C     (>) RESTRICT:  (logical) TRUE for restricted analysis
C     (>) SIGS:      (real) Number of sigma for restricted analysis
C     (W) WORK1:     (real array WORK1(SIZE)) Workspace
C     (W) WORK2:     (real array WORK2(SIZE)) Workspace
C     (W) RHS:       (double array RHS(SIZE)) Workspace
C     (W) CONDITION: (double array CONDITION(SIZE,3)) workspace
C     (<) PEAK:      (real) Calculated value of ring peak
C     (<) WIDTH:     (real) Calculated value of ring width.
C     (!) BACK:      (real) Revised estimate of background
C     (!) RAD:       (real) Revised estimate of radius
C     (!) STATUS:    (integer) Status flag
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        03-MAY-1990
C     UPDATE:      25-JUL-1990 -- Added in better fitting routine (JRL)
C                  02-SEP-1991 -- Added RESTRICT parameter. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,size,ipts,status
      real array(nx,ny),xc,yc,work1(size),work2(size),
     : width,xpos(ipts),ypos(ipts),peak,back,rad,sigs
      double precision condition(size,3),rhs(size)
      logical restrict
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
      double precision x02ajf
C
C     Local variables
C
      integer i,j,ix,iy,n,irem,ist,ifn,nfit,bytes_double,
     : totspace,address,slot,wptr,iw,lw,bytes_float,ifail,irank,ninc,
     : i1,i2
      real dx,dy,theta,x,y,low,w,ymax,x0,high,sum,aveback,rmin,diff,cut,
     : cut1,cut2,cut3,cut4,arg
      double precision tol,error,work(12),coefs(4),fsumsq
      logical svd,init
      character message*80,softdev*30
      data init /.true./
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Get current PGPLOT screen device on initial entry
C
      if (init) then
         init = .false.
         call var_getchr('softdev',0,0,softdev,status)
         if (status .ne. 0) then
            call dsa_wruser
     :      ('No screen device selected -- use SOFT command')
            call dsa_wrflush
            go to 500
         end if
      end if
C
C     First zero the work arrays
C
      do i = 1,size
         work1(i) = 0.0
         work2(i) = 0.0
      end do
      status = 0
C
C     Collect all the values along the radius to each point
C
      do i = 1,ipts
         dy = ypos(i) - yc
         dx = xpos(i) - xc
         theta = atan2(dy,dx)
         do j = 1,size
            x = xc + float(j-1)*cos(theta)
            y = yc + float(j-1)*sin(theta)
            ix = nint(x)
            iy = nint(y)
            if ((ix .ge. 1) .and. (ix .le. nx) .and. (iy .ge. 1)
     :      .and. (iy .le. ny)) then
               work1(j) = work1(j) + array(ix,iy)
               work2(j) = work2(j) + 1.0
            end if
         end do
      end do
C
C     Average the values now.
C
      n = 0
      low = max_float
      do i = 1,size
         if ((work2(i) .gt. 0.0) .and. (work1(i) .gt. 0.0)) then
            n = n + 1
            work2(n) = work1(i)/work2(i)
            work1(n) = float(i-1)
            low = min(low,work2(n))
         end if
      end do
C
C     Find the location of the radius
C
      rmin = max_float
      do i = 1,n
         diff = abs(rad-work1(i))
         if (diff .lt. rmin) then
            rmin = diff
            irem = i
         end if
      end do
      high = work2(irem)
C
C     Subtract off the background...
C
      do i = 1,n
         work2(i) = work2(i) - low
      end do
      high = high - low
C
C     Find region over which to do fit
C
      cut = 0.14*high
      i = irem
      do while ((work2(i) .ge. cut) .and. (i .ge. 1))
         i = i - 1
      end do
      ist = i + 1
      i = irem
      do while ((work2(i) .gt. cut) .and. (i .le. n))
         i = i + 1
      end do
      ifn = i - 1
C
C     Get some rough guesses in case all these fits fail...
C
      width = max(0.5*(work2(ifn)-work2(irem)),
     : 0.5*(work2(irem)-work2(ist)))
      peak = high
      back = low
C
C     Set up condition equations
C
      nfit = 0
      do i = ist,ifn
         if (work2(i) .gt. 0.0) then
            nfit = nfit + 1
            x = work1(i)
            condition(nfit,1) = 1.0d0
            condition(nfit,2) = dble(x)
            condition(nfit,3) = dble(x*x)
            rhs(nfit) = dble(log(work2(i)))
         end if
      end do
C
C     Now fit a gaussian in log space with zero background to get
C     rough parameter estimates...
C
      tol = x02ajf()
      ifail = 1
      call f04jgf(nfit,3,condition,size,rhs,tol,svd,error,irank,
     : work,12,ifail)
      if ((ifail .ne. 0) .or. (rhs(3) .gt. 0.0)) then
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('***Rough fit failed***')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         status = 1
         go to 500
      end if
C
C     Get the gaussian parameters back
C
      width = sqrt(-0.5/sngl(rhs(3)))
      x0 = width*width*sngl(rhs(2))
      peak = exp(sngl(rhs(1))+0.5*x0*x0/(width*width))
      rad = x0
C
C     Find the location of the new radius
C
      rmin = max_float
      do i = 1,n
         diff = abs(x0-work1(i))
         if (diff .lt. rmin) then
            rmin = diff
            irem = i
         end if
      end do
C
C     Do a rough guess at the background intensity now...
C
      ninc = 0
      sum = 0.0
      cut1 = x0 - 3.0*width
      cut2 = x0 - 5.0*width
      cut3 = x0 + 3.0*width
      cut4 = x0 + 5.0*width
      do i = 1,n
         if (((work1(i) .le. cut1) .and. (work1(i) .ge. cut2)) .or.
     :   ((work1(i) .ge. cut3) .and. (work1(i) .le. cut4))) then
            sum = sum + work2(i)
            ninc = ninc + 1
         end if
      end do
      if (ninc .ne. 0) then
         aveback = sum/float(ninc)
      else
         aveback = 0.0
      end if
      back = low + aveback
C
C     Give user an idea of the results...
C
      call dsa_wruser
     : ('Initial gaussian fit with the following parameters:')
      call dsa_wrflush
      write(message,100) peak,x0,width,back
  100 format('Peak =',f8.2,'  Centre =',f8.2,'  Sigma =',f8.2,
     : '  Back =',f8.2)
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
C
C     Now set up data to do a better fit. If you want to restrict the analysis
C     then do that here.
C
      if (restrict) then
         cut = x0 - sigs*width
         i1 = 0
         i = irem
         do while ((i .ge. 1) .and. (work1(i) .gt. cut))
            i = i - 1
         end do
         i1 = i + 1
         cut = x0 + sigs*width
         i2 = n + 1
         i = irem
         do while ((i .le. n) .and. (work1(i) .lt. cut))
            i = i + 1
         end do
         i2 = i - 1
         j = 0
         n = i2 - i1 + 1
         do i = i1,i2
            j = j + 1
            work1(j) = work1(i)
            work2(j) = work2(i) + low
         end do
      else
         do i = 1,n
            work2(i) = work2(i) + low
         end do
      end if
C
C     Set up initial guesses
C
      coefs(1) = dble(peak)
      coefs(2) = dble(x0)
      coefs(3) = dble(width)
      coefs(4) = dble(back)
C
C     Get some workspace
C
      lw = 64 + 11*n
      bytes_double = dsa_typesize('double',status)
      totspace = lw*bytes_double
      call dsa_get_workspace(totspace,address,slot,status)
      if (status .ne. 0) go to 500
      wptr = dyn_element(address)
C
C     Do a proper fit now...
C
      ifail = 1
      call e04gef(n,4,coefs,fsumsq,iw,1,dynamic_mem(wptr),lw,
     : ifail)
C
C     If proper fit fails, tell the user and try to procede with
C     the rough estimates
C
      if ((ifail .gt. 5)  .or. (ifail .eq. 1)) then
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('***Proper fit failed -- attempting to')
         call dsa_wruser(' continue with rough parameters')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
C
C     If the fit worked, then set up new estimates
C
      else
         peak = sngl(coefs(1))
         x0 = sngl(coefs(2))
         width = sngl(coefs(3))
         back = sngl(coefs(4))
         rad = x0
C
C        Give user an idea of the final results...
C
         call dsa_wruser
     :   ('Final gaussian fit with the following parameters:')
         call dsa_wrflush
         write(message,100) peak,x0,width,back
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
      end if
C
C     Delete old workspace and get new...
C
      call dsa_free_workspace(slot,status)
      bytes_float = dsa_typesize('float',status)
      totspace = 2*n*bytes_float
      call dsa_get_workspace(totspace,address,slot,status)
      if (status .ne. 0) go to 500
      wptr = dyn_element(address)
C
C     Plot the resulting data and its gaussian fit
C
      ymax = max(peak+back+1.0,high+low+1.0)
      call pgend
      call pgbegin(0,softdev,1,1)
      call pgadvance
      call pgwindow(work1(1),work1(n),low-1.0,ymax)
      call pgbox('BCNT',0.0,0,'BCNTS',0.0,0)
      call ndp_pgbin(work1,work2,dynamic_mem(wptr),
     : dynamic_mem(wptr+bytes_float*n),n,.false.,.true.)
      call dsa_free_workspace(slot,status)
      do i = 1,n
         w = work1(i)
         dx = w - x0
         arg = -0.5*dx*dx/(width*width)
         if (abs(arg) .lt. 40.0) then
            work2(i) = peak*exp(-0.5*dx*dx/(width*width)) + back
         else
            work2(i) = back
         end if
      end do
      call pgslw(3)
      call pgline(n,work1,work2)
      call pgslw(1)
      call pglabel('Distance from centre','Mean intensity',
     : 'Averaged profile')
      call pgend
  500 end


      subroutine draw_fit(xc,yc,r)
C+--------------------------------------------------------------------------------------
C     DRAW_FIT    --- Subroutine to draw the fitted ring onto the colour plot
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) XC:        (real) X centre.
C     (>) YC:        (real) Y centre.
C     (>) R:         (real) Radius of the ring.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        03-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      real xc,yc,r
C
C     Local variables
C
      integer i
      real xfit(101),yfit(101),dx,y2,y1
C
C     Define range in x
C
      dx = r/25.0
C
C     Now evaluate the fit
C
      do i = 1,50
         xfit(i) = -r + float(i-1)*dx + xc
         xfit(101-i+1) = xfit(i)
         y1 = r*r - (xfit(i) - xc)**2
         if (y1 .ge. 0.0) then
            y2 = sqrt(r*r - (xfit(i) - xc)**2)
         else
            y2 = 0.0
         end if
         yfit(i) = y2 + yc
         yfit(101-i+1) = -y2 + yc
      end do
      xfit(51) = r + xc
      yfit(51) = yc
      call pgscr(2,1.0,1.0,1.0)
      call pgsci(2)
      call pgline(101,xfit,yfit)
      end


      subroutine ptransfer_i(input,nx,ny,nz,iframe,temp,amin,amax)
C+--------------------------------------------------------------------------------------
C     PTRANSFER_I  --- Routine to transfer data from one array to another,
C                     performing a conversion if necessary.  Also find min and max.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (short array INPUT(NX,NY,NZ)) Input data cube.
C     (>) NX:        (integer) X dimension of input data
C     (>) NY:        (integer) Y dimension of input data
C     (>) NZ:        (integer) Z dimension of input data
C     (>) IFRAME:    (integer) Frame to be taken from the cube
C     (<) TEMP:      (real array TEMP(NX,NY)) Output frame
C     (<) AMIN:      (real) Minimum value in TEMP
C     (<) AMAX:      (real) Maximum value in TEMP
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        22-MAY-1990
C     UPDATE:      31-JAN-1994 -- Replace call to CNV_FMTCON with DSA_FMTCON.
C                                 This should probably be rewritten to take
C                                 advantage of the latter's array capability
C                                 (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iframe
      integer*2 input(nx,ny,nz)
      real temp(nx,ny),amin,amax
C
C     Local variables
C
      integer i,j,nbad,status
      integer*2 ival
      real val,magic_float
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Transfer data and find min and max (not including magic values...)
C
      amin = max_float
      amax = min_float
      call dsa_get_flag_value('FLOAT',magic_float,status)
      do j = 1,ny
         do i = 1,nx
            ival = input(i,j,iframe)
            call dsa_fmtcon(.true.,2,4,ival,val,1,nbad)
            if (val .ne. magic_float) then
               amin = min(amin,val)
               amax = max(amax,val)
            end if
            temp(i,j) = val
         end do
      end do
      end


      subroutine ptransfer_r(input,nx,ny,nz,iframe,temp,amin,amax)
C+--------------------------------------------------------------------------------------
C     PTRANSFER_R  --- Routine to transfer data from one array to another,
C                     performing a conversion if necessary.  Also find min and max.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (real array INPUT(NX,NY,NZ)) Input data cube.
C     (>) NX:        (integer) X dimension of input data
C     (>) NY:        (integer) Y dimension of input data
C     (>) NZ:        (integer) Z dimension of input data
C     (>) IFRAME:    (integer) Frame to be taken from the cube
C     (<) TEMP:      (real array TEMP(NX,NY)) Output frame
C     (<) AMIN:      (real) Minimum value in TEMP
C     (<) AMAX:      (real) Maximum value in TEMP
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        22-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iframe
      real temp(nx,ny),amin,amax,input(nx,ny,nz)
C
C     Local variables
C
      integer i,j,status
      real val,magic_float
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Transfer data and find min and max (not including magic values...)
C
      amin = max_float
      amax = min_float
      call dsa_get_flag_value('FLOAT',magic_float,status)
      do j = 1,ny
         do i = 1,nx
            val = input(i,j,iframe)
            if (val .ne. magic_float) then
               amin = min(amin,val)
               amax = max(amax,val)
            end if
            temp(i,j) = val
         end do
      end do
      end


      subroutine lsfun2(nres,nvar,coefs,fvecc,fjacc,ljc)
C+--------------------------------------------------------------------------------------
C     LSFUN2      --- Routine with fitting function and Jacobean for NAG
C                     routine E04GEF.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NRES:      (integer) Number of data points
C     (>) NVAR:      (integer) Number of variables
C     (>) COEFS:     (double precision array COEFS(NVAR)) Fit coefficients
C     (<) FVECC:     (double precision array FVECC(NRES)) Value of fitting function
C     (<) FJACC:     (double precision array FJACC(LJC,NVAR)) Value of Jacobean
C     (>) LJC:       (integer) First dimension of Jacobean matrix
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        25-JUL-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nvar,nres,ljc
      double precision coefs(nvar),fvecc(nres),fjacc(ljc,nvar)
C
C     Local variables
C
      integer i,wpt1,wpt2
      double precision x,arg,exparg,y
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Common block with data for fitting routine
C
      common /pointers/ wpt1,wpt2
C
C     Form residuals and Jacobean
C
      do i = 1,nres
         call extract(i,dynamic_mem(wpt1),dynamic_mem(wpt2),nres,x,y)
         arg = (0.5*(x - coefs(2))**2)/(coefs(3)*coefs(3))
         if (arg .lt. 40.0) then
            exparg = exp(-arg)
         else 
            exparg = 0.0d0
         end if
         fvecc(i) = coefs(1)*exparg + coefs(4) - y
         fjacc(i,1) = exparg
         fjacc(i,2) = coefs(1)*exparg*((x - coefs(2))/(coefs(3)**2))
         fjacc(i,3) = coefs(1)*exparg*(2.0*arg/coefs(3))
         fjacc(i,4) = 1.0d0
      end do
      end


      subroutine extract(ifind,work1,work2,n,x,y)
C+--------------------------------------------------------------------------------------
C     EXTRACT     --- Routine to extract data from work arrays for fitting function.
C                     Necessary due to NAG routine's ridiculous parameter list.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IFIND:     (integer) Element wanted
C     (>) WORK1:     (real array WORK1(N)) First data array
C     (>) WORK2:     (real array WORK2(N)) Second data array
C     (>) N          (integer) Dimension of data arrays
C     (<) X:         (double precision) Extracted data from first array
C     (<) Y:         (double precision) Extracted data from second array
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        25-JUL-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer ifind,n
      real work1(n),work2(n)
      double precision x,y
C
C     Extract the data
C
      x = dble(work1(ifind))
      y = dble(work2(ifind))
      end


      subroutine copy_stuff(input,nix,niy,mframe,extra,nox,noy,nframe,
     : output)
C+--------------------------------------------------------------------------------------
C     COPY_STUFF  --- Routine to copy RVT information from one array to another
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:      (real array INPUT(NIX,NIY)) Input information
C     (>) NIX:        (integer) X dimension of input data
C     (>) NIY:        (integer) Y dimension of input data
C     (>) MFRAME:     (integer) Y location for input data
C     (>) EXTRA:      (real) Information to be included in the first position of output
C     (>) NOX:        (integer) X dimension of output data
C     (>) NOY:        (integer) Y dimension of output data
C     (>) NFRAME:     (integer) Y location for output data
C     (<) OUTPUT:     (real array OUTPUT(NOX,NOY)) Output information
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        06-NOV-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nix,niy,nox,noy,nframe,mframe
      real input(nix,niy),output(nox,noy),extra
C
C     Local variables
C
      integer i
C
C     Do the copying
C
      do i = 2,nox
         output(i,nframe) = input(i,mframe)
      end do
C
C     Add in the extra item
C
      output(1,nframe) = extra
      end


      subroutine rvt_extra(iptr,type,nx,ny,nz,nbins,nframe,restrict,
     : sigs,temp,freq,xaxis,yaxis,table,badpix,ringnum,nring,nox,noy,
     : output)
C+--------------------------------------------------------------------------------------
C     RVT_EXTRA  --- Subroutine do a bit of extra work on RVT rings
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IPTR:      (integer) Dynamic memory pointer for input data
C     (>) TYPE:      (character) Input data type
C     (>) NX:        (integer) X dimension of data.
C     (>) NY:        (integer) Y dimension of data.
C     (>) NZ:        (integer) Z dimension of data.
C     (>) NBINS:     (integer) Number of bins for histogramming.
C     (>) NFRAME:    (integer) Frame number to be analysed
C     (>) RESTRICT:  (logical) TRUE for restricted analysis
C     (>) SIGS:      (real) Number of sigmas for restricted analysis
C     (W) TEMP:      (real array TEMP(NX,NY)) Workspace.
C     (W) FREQ:      (integer array FREQ(NBINS)) Workspace
C     (>) XAXIS:     (real array XAXIS(NX)) Array of X axis data values.
C     (>) YAXIS:     (real array YAXIS(NY)) Array of Y axis data values.
C     (>) TABLE:     (character) Name of LUT for 2-d display routine.
C     (>) BADPIX:    (logical) Flag for magic values in data.
C     (>) RINGNUM:   (real) Run number of ring
C     (>) NRING:     (integer) Location in output file for results
C     (>) NOX:       (integer) X dimension of output file
C     (>) NOY:       (integer) Y dimension of output file
C     (<) OUTPUT:    (real array OUTPUT(NOX,NOY)) Output array
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        11-DEC-1990
C     UPDATE:      02-SEP-1991 -- Added RESTRICT and SIGS parameters (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer iptr,nx,ny,nz,nbins,nframe,nring,nox,noy
      integer freq(nbins)
      real temp(nx,ny),xaxis(nx),yaxis(ny),output(nox,noy),ringnum,sigs
      character type*(*),table*(*)
      logical restrict,badpix
C
C     Local variables
C
      real xc,yc,rad,sig,peak,back
C
C     First, copy the results from the first ring into some local variables
C
      xc = output(2,1)
      yc = output(3,1)
      rad = output(4,1)
      sig = output(5,1)
      peak = output(6,1)
      back = output(7,1)
C
C     Now run IMAGE_PARS without the 2-d plotting option...
C
      call image_pars(iptr,type,nx,ny,nz,nbins,nframe,restrict,sigs,
     : temp,freq,xaxis,yaxis,table,.true.,badpix,xc,yc,rad,sig,peak,
     : back)
C
C     Now copy the results from IMAGE_PARS into the correct spot in the
C     output file...
C
      output(1,nring) = ringnum
      output(2,nring) = xc
      output(3,nring) = yc
      output(4,nring) = rad
      output(5,nring) = sig
      output(6,nring) = peak
      output(7,nring) = back
      output(8,nring) = 1.0
      output(9,nring) = 0.0
      output(10,nring) = -1.0
      end
