      subroutine print_rv
C+------------------------------------------------------------------------------
C
C   ---------------
C   P R I N T _ R V
C   ---------------
C
C   Description
C   -----------
C   Utility program to print RVF or RVT file out to ASCII file
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - Data mapped as FLOAT.
C   - Error arrays supported
C   - Quality arrays not supported
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read or written)
C   ----------------------------
C   IMAGE     Name of the structure containing the RVF or RVT
C             (character)(prompted for).
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested for dimensionality and
C     is mapped as FLOAT.
C   - An ASCII file with the same name as the structure and
C     the extension '.out' and a version number is opened.
C   - The contents of the RV file is dumped to the ASCII file
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_LU
C     DSA_GET_LU
C     DSA_NAMED_INPUT
C     DSA_MAP_DATA
C     DSA_MAP_ERRORS
C     DSA_OPEN
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C  Library ICH:
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDCHAR
C
C   Internal subroutines called
C   ---------------------------
C   SUB_PRINT_RV
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   14-FEB-90   - Original program
C   22-MAY-90   - Slight cosmetic modifications.  (JRL)
C   21-FEB-91   - Changed so that a header line and the error
C                 array is also written (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer dyn_element
C
C     Local variables
C
      integer    address   ! Dyamic memory address
      logical    badpix    ! Flag for magic values
      integer    dims(2)   ! Dimensions of input file
      integer    eptr      ! Dynamic memory pointer (error array)
      integer    eslot     ! Dynamic memory slot (error array)
      integer    iptr      ! Dynamic memory pointer (data array)
      character  name*80   ! Name of input structure
      integer    ndim      ! Number of dimensions input file
      integer    nelm      ! Number of elements in input file
      integer    slot      ! Dynamic memory slot (data array)
      integer    status    ! DSA status flag
      character  type*8    ! Input file type
C
C     Common block for dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Get things started...
C
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Get the name of the file, open it and get its size
C
      call par_rdchar('name_rv',' ',name)
      call dsa_named_input('image',name,status)
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',2,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
      if (ndim .ne. 2) then
         call dsa_wruser('This is not a 2-d image...')
         call dsa_wrflush
         go to 500
      end if
C
C     Map the image file
C
      call dsa_map_data('image','read','float',address,slot,status)
      if (status .ne. 0) go to 500
      iptr = dyn_element(address)
C
C     Map the error array
C
      call dsa_map_errors('image','read','float',address,eslot,status)
      if (status .ne. 0) go to 500
      eptr = dyn_element(address)
C
C     Do the work...
C
      call sub_print_rv(dynamic_mem(iptr),dynamic_mem(eptr),dims(1),
     : dims(2),name)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end

      subroutine sub_print_rv(array,errors,nx,ny,name)
C+--------------------------------------------------------------------------------------
C     SUB_PRINT_RV --- Routine to do the actual printing of the RV file
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) ARRAY:     (real array ARRAY(NX,NY)) Input RV file data
C     (>) ERRORS:    (real array ERRORS(NX,NY)) Error array
C     (>) NX:        (integer) X dimension of input file
C     (>) NY:        (integer) Y dimension of input file
C     (>) NAME:      (character) Name of input file
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        14-FEB-1990
C     UPDATE:      22-MAY-1990 -- Output file is now named after the input (JRL)
C                  21-FEB-1991 -- Now writes out error array. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny
      real array(nx,ny),errors(nx,ny)
      character name*(*)
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer unit,status,i,j
      logical exist
      character testfile*80
C
C     Find an output file name...
C
      i = 99
      exist = .false.
      do while ((i .gt. 0) .and. (.not. exist))
         write(testfile,100) name(:ich_len(name)),i
  100    format(a,'.out:',i2.2)
         inquire(file=testfile(:ich_len(testfile)),exist=exist)
         if (.not. exist) i = i - 1
      end do
      i = i + 1
      write(testfile,100) name(:ich_len(name)),i
C
C     Open the output file...
C
      call dsa_get_lu(unit,status)
      open(unit=unit,file=testfile(:ich_len(testfile)),
     : status='new',iostat=status)
      if (status .ne. 0) then
         call dsa_wruser('Couldn''t open output file...')
         call dsa_wrflush
         go to 500
      end if
C
C     Write a line of info to tell you want you're gettting...
C
      write(unit,101)
  101 format('   Z axis     Xc       Yc        R      WIDTH     PEAK  ',
     : '   BACK    ELLIPT   ERROR     FLAG  ')
C
C     Now write out the info...
C
      do i = 1,ny
         write(unit,110) (array(j,i),j=1,nx)
         write(unit,110) (errors(j,i),j=1,nx)
  110    format(' ',<nx>f9.3)
         write(unit,120)
  120    format(' ')
      end do
      close(unit)
C
C     Exit
C
  500 call dsa_free_lu(unit,status)
      end
