      subroutine residual
C+------------------------------------------------------------------------------
C
C   ---------------
C   R E S I D U A L
C   ---------------
C
C   Description
C   -----------
C   Calculates and outputs the residual phase map.
C
C
C   Scope of program
C   ----------------
C   - Handles 3-d images only.
C   - SHORT and FLOAT data only supported, all others
C     are converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE      Name of TAUCAL file (character)(prompted for)
C
C   WINDOW       Window in Z pixels over which to search for the
C                peak of the emission line.  The whole search
C                window will be 2*WINDOW + 1 pixels. (integer)
C                (prompted for)
C
C   SMOOTH_SIZE  Size of box (in pixels) over which to average the
C                residuals. (integer)(prompted for)
C
C   NAME_RES     Name of structure for output residual map. (character)
C                (prompted for)
C
C   WHICH_RVF   Number of the RVF already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_MASK  Number of the MASK already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_GRID  Number of the GRID already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C   USE_GRID     Instruction to either use the grid of results from
C                MASK and GRID to define pixel-space or not. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .RVF.RVFn         RINGRUN      Run number associated with CAL cube.
C                                  (integer)(prompted for).
C
C   .MASK.MASKn.GRIDn NAME_GRID    Name of input file with grid of X,Y pixel
C                                  positions. (character)(prompted for)
C
C   .RVF              NRVF         Number of RVFs already defined in TAUCAL.
C                                  (integer)(read)
C
C   .MASK             NMASK        Number of MASKs already defined in TAUCAL.
C                                  (integer)(read)
C
C   .MASK.MASKn       NGRID        Number of GRIDs already defined in MASK
C                                  sub-structure of TAUCAL. (integer)(read)
C
C   (CAL)             IMAGE        Name of calibration cube. (character)(read)
C
C   .MASK.MASKn       PIXSIZE_X    Pixel size as defined by MASK. (real)(read)
C
C   .CAMERA           FOCLEN       Focal length of the camera used. (real)(read)
C
C   .RVF.RVFn         WAVE         Wavelength of the calibration source. (real)
C                                  (read)
C
C   .RVF.RVFn         FSR          Free spectral range for CAL cube in MPUs
C                                  (real)(read)
C
C   .RVF.RVFn         GAP          Etalon gap in microns. (real)(read)
C
C   .RVF.RVFn         XC           X position of optical axis (real)(read)
C
C   .RVF.RVFn         YC           Y position of optical axis (real)(read)
C
C   .RVF.RVFn         CENPHASE     Central phase of cal cube in MPUs
C                                  (real)(read)
C
C   (CAL)             MPU0         Z axis zero point in MPUs.  (real)(read)
C
C   (CAL)             DMPU         Z axis increment in MPUs/frame (real)(read)
C
C   (CAL)             DX           Mean RVT shift in X for cal cube (real)(read)
C
C   (CAL)             DY           Mean RVT shift in Y for cal cube (real)(read)
C
C   .RVF.RVFn         MODE         Mode of operation flag. =-1 if slopes  < 0.
C                                  (integer)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - Calibration cube is opened, its dimensions tested (3-d only).
C   - Calibration cube Z axis data is mapped as REAL.
C   - If use of grid is desired, then the GRID file is opened.  If
C     not then an empty workspace is created.
C   - OUTPUT is set up with the same dimensions as IMAGE.
C   - Find the zero phase point from the linear fits from EXAM_RVF.
C   - For each pixel in the box defined by NSIZE use the theoretical
C     phase function to define a position in Z axis coordinates for
C     the predicted maximum of the emission line.
C   - Sum up the weighted difference between the Z position and the
C     theoretical position over the window size.  This is summed over
C     all the pixels in the box.
C   - Find the mean for the box and write it to the residuals map.
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_GET_WORK_ARRAY
C     DSA_MAP_DATA
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_OUTPUT
C     DSA_SIMPLE_OUTPUT
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_DUMMY_GRID
C
C   Internal subroutines called
C   ---------------------------
C     RTRANSFER_I
C     RTRANSFER_R
C     SUB_RESIDUAL
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   INTEGER*2
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   09-MAR-90   - Original program
C   22-MAR-90   - Modifications to centroiding algorithm  (JRL)
C   26-MAR-90   - Addition of smoothing box parameter  (JRL)
C   24-MAY-90   - Modified to allow for more than two RVF series  (JRL)
C   03-JUL-90   - TAUCAL system pasted in with a few minor modifications.  (JRL)
C   26-JUL-90   - Added MODE parameter. Also modified so that input data of
C                 any type is now allowed. (JRL)
C   03-SEP-90   - Residuals now stored in MPUs. (JRL)
C   20-NOV-90   - Info now read from RVF and MASK sub-structures.  New
C                 TAUCAL routines also pasted in. (JRL)
C   06-DEC-90   - Improved algorithm for calculating residual added. (JRL)
C   11-JAN-91   - DX, DY now taken into account.  (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer           dyn_element
C
C     Local Variables
C
      integer           address      ! Dynamic memory address
      logical           badpix       ! Flag for magic values in image array
      logical           badpixg      ! Flag for magic values in grid array
      real              cenphase     ! Central phase of cube
      integer           dims(3)      ! Dimensions of IMAGE
      real              dmpu         ! Z axis increment in MPUs
      real              dummy        ! Dummy real variable
      real              dx           ! RVT X shift
      real              dy           ! RVT Y shift
      real              focus        ! Camera focal length
      real              fscale       ! Pixelsize/focal length
      real              fsr_cen      ! Central free spectral range
      real              gap          ! Etalon spacing
      integer           gdims(3)     ! Dimensions of GRID
      integer           gptr         ! Dynamic memory pointer for GRID
      character         grid_lab*80  ! Label for GRID sub-structure
      integer           gslot        ! Dynamic memory slot for GRID
      integer           iptr         ! Dynamic memory pointer for IMAGE
      integer           islot        ! Dynamic memory slot for IMAGE
      integer           len_gridlab  ! Length of GRID label
      integer           len_masklab  ! Length of MASK label
      integer           len_rvflab   ! Length of RVF label
      character         mask_lab*80  ! Label for MASK sub-structure
      integer           mode         ! Mode of operation variable
      real              mpu0         ! Z axis zeropoint in MPUs
      character         name_cal*80  ! Name of calibration cube
      character         name_grid*80 ! Name of GRID file
      integer           ndim         ! Number of dimensions in IMAGE
      integer           nelm         ! Number of elements in IMAGE
      integer           ngdim        ! Number of dimensions in GRID
      integer           nsize        ! Size of smoothing box
      integer           odims(2)     ! Dimensions of OUTPUT
      integer           optr         ! Dynamic memory pointer for OUTPUT
      integer           oslot        ! Dynamic memory slot for OUTPUT
      real              pixsize      ! Central X pixel size
      integer           ringrun      ! Run number of ring file
      character         rvf_lab*80   ! Label for RVF sub-structure
      integer           status       ! DSA and TAUCAL status flag
      logical           tauexist     ! Flag for existence of TAUCAL file
      character         type*8       ! Data type of image array
      character         typeg*8      ! Data type of grid array
      logical           use_grid     ! Flag for whether using grid or not
      real              wave         ! Calibration source wavelength
      integer           window       ! Size of window for searching for z peaks
      integer           wptr         ! Dynamic memory pointer for workspace
      integer           wslot        ! Dynamic memory slot for workspace
      real              xcen         ! Mean x-centre of the rings
      real              ycen         ! Mean y-centre of the rings
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find out which RVF you want
C
      call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     : len_rvflab,status)
      if (status .ne. 0) go to 500
C
C     Find out which mask you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,mask_lab,
     : len_masklab,status)
      if (status .ne. 0) go to 500
C
C     Get the run numbers of the CAL image
C
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.run',
     : 'ringrun',tauexist,.false.,float(min_int),float(max_int),
     : ' ',1,dummy,status)
      ringrun = nint(dummy)
C
C     Make sure this run is the right type.
C
      call tau_cal_check_run('cal',ringrun,'cal',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Open IMAGE file for calibration cube.
C
      call tau_cal_rdchr('cal',ringrun,'image','name_cal',tauexist,
     : .true.,80,name_cal,status)
      call dsa_named_input('image',name_cal,status)
      if (status .ne. 0) go to 500
C
C     Check image dimensions
C
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',3,ndim,dims,nelm,status)
      if (status .ne. 0) go to 500
      if (ndim .ne. 3) then
         call dsa_wruser(' Image has unacceptable dimensions.')
         call dsa_wrflush
         go to 500
      end if
C
C     Map IMAGE data array.  If it's SHORT, then map it as such.  Otherwise
C     map it as FLOAT.
C
      if (type .eq. 'SHORT') then
         call dsa_map_data('image','read','short',address,islot,status)
      else
         call dsa_map_data('image','read','float',address,islot,status)
      end if
      if (status .ne. 0) go to 500
      iptr = dyn_element(address)
C
C     Get some parameters
C
      call par_rdval('window',0.0,float(max_int),5,'z pixels',dummy)
      window = nint(dummy)
      call par_rdval('smooth_size',0.0,float(max_int),5,'pixels',dummy)
      nsize = nint(dummy)
C
C     Get some parameters from TAUCAL
C
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.wave',
     : 'wave_cal',tauexist,.true.,min_float,max_float,'microns',1,wave,
     : status)
      call tau_cal_rdnum('cal',0,mask_lab(:len_masklab)//'.pixsize_x',
     : 'pixsize_x',tauexist,.true.,min_float,max_float,'microns',1,
     : pixsize,status)
      call tau_cal_rdnum('cal',0,'camera.foclen','camera_focus',
     : tauexist,.true.,min_float,max_float,'mm',1,focus,status)
      if (status .ne. 0) go to 500
      focus = focus*1000.0
      fscale = pixsize/focus
C
C     If using grid...
C
      call par_rdkey('use_grid',.true.,use_grid)
      if (use_grid) then
C
C        Find which grid you want...
C
         call tau_cal_find_struct('cal',mask_lab(:len_masklab),'grid',
     :   tauexist,grid_lab,len_gridlab,status)
         if (status .ne. 0) go to 500
C
C        Get the grid file name and open it
C
         call tau_cal_rdchr('cal',0,
     :   grid_lab(:len_gridlab)//'.name_grid',
     :   'name_grid',tauexist,.true.,80,name_grid,status)
         call dsa_named_input('grid',name_grid,status)
         call ndp_get_image_info('grid',.true.,.false.,typeg,badpixg,
     :   status)
         call dsa_data_size('grid',3,ngdim,gdims,nelm,status)
         if (status .ne. 0) go to 500
         if ((gdims(2) .ne. dims(1)) .or. (gdims(3) .ne. dims(2))) then
            call dsa_wruser
     :      ('Dimensions of grid and image files disagree')
            call dsa_wrflush
            go to 500
         end if
         call dsa_map_data('grid','read','float',address,gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
      else
         call dsa_get_work_array(2*dims(1)*dims(2),'float',address,
     :   gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
         call tau_dummy_grid(dims(1),dims(2),dynamic_mem(gptr))
      end if
C
C     Get info from EXAM_RVF run
C
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.fsr','fsr_cen',
     : tauexist,.true.,min_float,max_float,'MPUs',1,fsr_cen,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.gap','gap',
     : tauexist,.true.,min_float,max_float,'microns',1,gap,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.xc','xcen',
     : tauexist,.true.,min_float,max_float,'pixels',1,xcen,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.yc','ycen',
     : tauexist,.true.,min_float,max_float,'pixels',1,ycen,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.cenphase',
     : 'cenphase',tauexist,.true.,min_float,max_float,'MPUs',1,cenphase,
     : status)
      call tau_cal_rdnum('cal',ringrun,'mpu0','mpu0',tauexist,.true.,
     : min_float,max_float,'MPUs',1,mpu0,status)
      call tau_cal_rdnum('cal',ringrun,'dmpu','dmpu',tauexist,.true.,
     : min_float,max_float,'MPUs/frame',1,dmpu,status)
      call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.mode','mode',
     : tauexist,.true.,-1.0,1.0,' ',1,dummy,status)
      mode = nint(dummy)
C
C     Get info from EXAM_RVT run
C
      call tau_cal_rdnum('cal',ringrun,'dx','dx',tauexist,.true.,
     : min_float,max_float,'pixels',1,dx,status)
      call tau_cal_rdnum('cal',ringrun,'dy','dy',tauexist,.true.,
     : min_float,max_float,'pixels',1,dy,status)
C
C     Get a work array
C
      call dsa_get_work_array(dims(3),'float',address,wslot,status)
      if (status .ne. 0) go to 500
      wptr = dyn_element(address)
C
C     Open output file
C
      call dsa_output('output','name_res',' ',1,1,status)
      if (status .ne. 0) go to 500
C
C     Create output structure
C
      odims(1) = dims(1)
      odims(2) = dims(2)
      call dsa_simple_output('output','data','float',2,odims,status)
      if (status .ne. 0) go to 500
C
C     Map output array
C
      call dsa_map_data('output','write','float',address,oslot,status)
      if (status .ne. 0) go to 500
      optr = dyn_element(address)
C
C     Now do the work
C
      call sub_residual(iptr,type,dims(1),dims(2),dims(3),gap,
     : fsr_cen,wave,fscale,xcen,ycen,dx,dy,nsize,dynamic_mem(gptr),
     : cenphase,mpu0,dmpu,mode,window,dynamic_mem(wptr),
     : dynamic_mem(optr))
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end



      subroutine sub_residual(iptr,type,nx,ny,nz,gap,fsr_cen,wave,
     : fscale,xcen,ycen,dx,dy,nsize,coords,cenphase,mpu0,dmpu,mode,
     : window,temp,res)
C+-----------------------------------------------------------------------------------
C     SUB_RESIDUAL   --- Subroutine to do the actual work in finding
C                        the residual phase map.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) IPTR:       (integer) Input data array dynamic memory pointer.
C     (>) TYPE:       (character) Input data array type.
C     (>) NX:         (integer) X dimension of DATA
C     (>) NY:         (integer) Y dimension of DATA
C     (>) NZ:         (integer) Z dimension of DATA
C     (>) GAP:        (real) Etalon gap in microprocessor units
C     (>) FSR_CEN:    (real) Central free spectral range in microprocessor units
C     (>) WAVE:       (real) Calibration source wavelength
C     (>) FSCALE:     (real) Pixel size/camera focal length
C     (>) XCEN:       (real) X centre of rings
C     (>) YCEN:       (real) Y centre of rings
C     (>) DX:         (real) RVT shift in X
C     (>) DY:         (real) RVT shift in Y
C     (>) NSIZE:      (integer) Size of box for smoothing of results
C     (>) COORDS:     (real array COORDS(2,NX,NY)) If (use_grid) then this contains
C                     the coordinate grid defined by MASK and GRID.  If not then this
C                     is a dummy array.
C     (>) CENPHASE:   (real) Central phase of the cube
C     (>) MPU0:       (real) Z axis zeropoint in MPUs
C     (>) DMPU:       (real) Z axis increment in MPUs
C     (>) MODE:       (integer) Mode of operation variable.
C     (>) WINDOW:     (integer) The window in Z over which the program searches for the
C                     peak emission.  The real window is 2*WINDOW + 1.
C     (W) TEMP:       (real array TEMP(NZ)) Workspace
C     (<) RES:        (real array RES(NX,NY)) The output results array.
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           09-MAR-1990
C     UPDATE:         22-MAR-1990 -- Now searches for peak before doing
C                                    centroid.  (JRL)
C                     26-MAR-1990 -- Now adds the intensity in the smoothing box together
C                                    before finding the peak.  (JRL)
C                     24-MAY-1990 -- Slight modifications to cope with more
C                                    than two series in the RVF. (JRL)
C                     26-JUL-1990 -- Added in MODE variable.  Also now caters for
C                                    all data types. (JRL)
C                     03-SEP-1990 -- Residuals now stored in MPUs.  (JRL)
C                     06-DEC-1990 -- Modified algorithm which should cut down on rounding
C                                    errors. (JRL)
C                     11-JAN-1991 -- Now takes DX and DY into account. (JRL)
C
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,nsize,window,mode,iptr
      real gap,fsr_cen,wave,fscale,coords(2,nx,ny),res(nx,ny),
     : xcen,ycen,temp(nz),cenphase,mpu0,dmpu,dx,dy
      character type*(*)
C
C     Local Variables
C
      integer ifsr,index,izst,jx,jy,jxend,jyend,k,z_look_pix,jyst,
     : jxst,iy,ix,iz,krem,half,i,idrem,nfsr
      real beta,dz,fscale2,fsr,r2,result,xnew,ynew,z0,zmax,zmid,dfsr,
     : z_look,z_pre,sumres,sum,sec_theta,tmax,zval1,diff,mindiff,val
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     A few useful variables
C
      zval1 = mpu0
      zmax = mpu0 + float(nz-1)*dmpu
      zmid = mpu0 + (float(nz/2) - 1.0)*dmpu
      z0 = cenphase
      dz = dmpu
      beta = float(mode)*2.0*gap*fsr_cen/wave
      fscale2 = fscale*fscale
      half = nsize/2 + 1
C
C     Loop through the map in steps of NSIZE
C
      do jy = half,ny-half+1,nsize
         jyst = jy
         jyend = min(ny,jyst+nsize-1)
         do jx = half,nx-half+1,nsize
            jxst = jx
            jxend = min(nx,jxst+nsize-1)
C
C           Sum together the intensity in the smoothing box and
C           convert it to real if necessary.
C
            if (type .eq. 'SHORT') then
               call rtransfer_i(dynamic_mem(iptr),nx,ny,nz,jxst,jxend,
     :         jyst,jyend,temp)
            else
               call rtransfer_r(dynamic_mem(iptr),nx,ny,nz,jxst,jxend,
     :         jyst,jyend,temp)
            end if
C
C           Initialize a few variables
C
            sumres = 0.0
            sum = 0.0
C
C           Loop now for each pixel in the smoothing box
C
            do iy = jyst,jyend
               do ix = jxst,jxend
C
C                 Get the grid coordinates minus centre
C
                  xnew = coords(1,ix,iy) - (xcen - dx)
                  ynew = coords(2,ix,iy) - (ycen - dy)
C
C                 Find the predicted phase for XNEW,YNEW.
C
                  r2 = xnew*xnew + ynew*ynew
                  sec_theta = sqrt(r2*fscale2 + 1.0)
                  fsr = fsr_cen*sec_theta
                  ifsr = abs(int(fsr/dz))
                  nfsr = abs(nint(fsr/dz))
                  dfsr = abs(fsr/dz) - float(ifsr)
                  z_pre = z0 + beta*(sec_theta - 1.0)
C
C                 Find out whether the predicted position overruns the spectrum
C
                  if (z_pre .gt. zmax) then
                     index = int((z_pre - zmax)/fsr) + 1
                  else if (z_pre .lt. mpu0) then
                     index = int((z_pre - mpu0)/fsr) - 1
                  else
                     index = 0
                  end if
                  z_look = z_pre - fsr*float(index)
C
C                 Now find out if you are better off looking one FSR away...
C
                  mindiff = max_float
                  do i = -1,1
                     diff = abs(z_look - float(i)*fsr - zmid)
                     if (diff .lt. mindiff) then
                       idrem = i
                       mindiff = diff
                     end if
                  end do
                  z_look = z_look - float(idrem)*fsr
C
C                 Define where in pixel space to find the peak (theoretically)
C
                  z_look_pix = nint((z_look - zval1)/dz + 1.0)
C
C                 Search for maximum within the defined window.
C
                  tmax = min_float
                  izst = z_look_pix - window - 1
                  do k = 1,2*window+1
                     izst = izst + 1
                     if (izst .lt. 1) then
                        iz = izst + nfsr
                     else if (izst .gt. nz) then
                        iz = izst - nfsr
                     else
                        iz = izst
                     end if
                     if (temp(iz) .gt. tmax) then
                        tmax = temp(iz)
                        krem = izst
                     end if
                  end do
C
C                 Now sum the weighted residuals where the weight is the
C                 intensity at Z and the residual is the Z axis value at Z
C                 minus the predicted phase.
C
                  izst = krem - window - 1
                  do k = 1,2*window+1
                     izst = izst + 1
                     if (izst .lt. 1) then
                        iz = izst + ifsr
                        val = temp(iz)*(1.0 - dfsr) + temp(iz+1)*dfsr
                     else if (izst .gt. nz) then
                        iz = izst - ifsr
                        val = temp(iz)*(1.0 - dfsr) + temp(iz-1)*dfsr
                     else
                        iz = izst
                        val = temp(iz)
                     end if
                     sumres = sumres +
     :               val*(zval1 + float(izst-1)*dz - z_look)
                     sum = sum + val
                  end do
               end do
            end do
C
C           Find the mean residual and write it to the output file
C
            if (sum .ne. 0.0) then
               result = sumres/sum
            else
               result = 0.0
            end if
            do iy = jyst,jyend
               do ix = jxst,jxend
                  res(ix,iy) = result
               end do
            end do
         end do
      end do
C
C     Take care of edges...
C
      do iy = half,ny-half+1
         do ix = 1,half-1
            res(ix,iy) = res(half,iy)
         end do
         do ix = nx-half+2,nx
            res(ix,iy) = res(nx-half+1,iy)
         end do
      end do
      do iy = 1,half-1
         do ix = half,nx-half+1
            res(ix,iy) = res(ix,half)
         end do
      end do
      do iy = ny-half+2,ny
         do ix = half,nx-half+1
            res(ix,iy) = res(ix,ny-half+1)
         end do
      end do
C
C     Now the corners...
C
      do iy = 1,half-1
         do ix = 1,half-1
            res(ix,iy) = res(half,half)
         end do
      end do
      do iy = 1,half-1
         do ix = nx-half+2,nx
            res(ix,iy) = res(nx-half+1,half)
         end do
      end do
      do iy = ny-half+2,ny
         do ix = 1,half-1
            res(ix,iy) = res(half,ny-half+1)
         end do
      end do
      do iy = ny-half+2,ny
         do ix = nx-half+2,nx
            res(ix,iy) = res(nx-half+1,ny-half+1)
         end do
      end do
      end


      subroutine rtransfer_i(data,nx,ny,nz,jxst,jxend,jyst,jyend,temp)
C+-------------------------------------------------------------------------------
C     RTRANSFER_I     --- Subroutine to sum short data in a box from the input
C                        cube into an output vector, converting it to real.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:       (short array DATA(NX,NY,NZ)) Input data cube
C     (>) NX:         (integer) X dimension of data cube.
C     (>) NY:         (integer) Y dimension of data cube.
C     (>) NZ:         (integer) Z dimension of data cube.
C     (>) JXST:       (integer) Starting point in X for smoothing box
C     (>) JXEND:      (integer) Ending point in X for smoothing box
C     (>) JYST:       (integer) Starting point in Y for smoothing box
C     (>) JYEND:      (integer) Ending point in Y for smoothing box
C     (<) TEMP:       (real array TEMP(NZ)) Output summed array.
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,jxst,jxend,jyst,jyend
      integer*2 data(nx,ny,nz)
      real temp(nz)
C
C     Local variables
C
      integer k,ix,iy
C
C     Initialise output array
C
      do k = 1,nz
         temp(k) = 0.0
      end do
C
C     Copy the data over and convert as you go along...
C
      do iy = jyst,jyend
         do ix = jxst,jxend
            do k = 1,nz
               temp(k) = temp(k) + float(data(ix,iy,k))
            end do
         end do
      end do
      end



      subroutine rtransfer_r(data,nx,ny,nz,jxst,jxend,jyst,jyend,temp)
C+-------------------------------------------------------------------------------
C     RTRANSFER_R     --- Subroutine to sum real data in a box from the input
C                        cube into an output vector.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DATA:       (real array DATA(NX,NY,NZ)) Input data cube
C     (>) NX:         (integer) X dimension of data cube.
C     (>) NY:         (integer) Y dimension of data cube.
C     (>) NZ:         (integer) Z dimension of data cube.
C     (>) JXST:       (integer) Starting point in X for smoothing box
C     (>) JXEND:      (integer) Ending point in X for smoothing box
C     (>) JYST:       (integer) Starting point in Y for smoothing box
C     (>) JYEND:      (integer) Ending point in Y for smoothing box
C     (<) TEMP:       (real array TEMP(NZ)) Output summed array.
C
C     AUTHOR:         Jim Lewis -- RGO
C     DATE:           26-JUL-1990
C     UPDATE:
C+------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,jxst,jxend,jyst,jyend
      real data(nx,ny,nz),temp(nz)
C
C     Local variables
C
      integer k,ix,iy
C
C     Initialise output array
C
      do k = 1,nz
         temp(k) = 0.0
      end do
C
C     Copy the data over and convert as you go along...
C
      do iy = jyst,jyend
         do ix = jxst,jxend
            do k = 1,nz
               temp(k) = temp(k) + data(ix,iy,k)
            end do
         end do
      end do
      end
