      subroutine rings
C+------------------------------------------------------------------------------
C
C   ---------
C   R I N G S
C   ---------
C
C   Description
C   -----------
C   Do fits to ring frames for calibration cubes for RVF or RVT
C
C
C
C   Scope of program
C   ----------------
C   - 2-d and 3-d only supported.
C   - FLOAT or SHORT data only supported.  All others
C     are converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   COMP_FAC    A factor for compressing the input arrays.  If, for
C               example, comp_fac = 2 then only every second line
C               and every second column will be looked at. (integer)
C              (prompted for)
C
C   WEIGHT      Flag for the weighting scheme to be employed.
C               (1) Sqrt(n), (2) Equal Weights.  (integer)(prompted for)
C
C   TRIVIAL     Definition of a trivial change in chi-squared for use in
C               deciding if a fit has converged. (real)(no prompt)
C
C   NTRIV       The number of times the change in chi-squared has to be
C               below TRIVIAL before the fit is defined has converged.
C               (integer)(no prompt)
C
C   TOLCENT     Tolerance for the error in the determination of the centre
C               of the ring. (real)(prompted for)
C
C   TOLRAD      Tolerance for the error in the determination of the radius
C               of the ring. (real)(prompted for)
C
C   SIGS        Number of sigma from the radius vector past which the
C               data is thrown away.  (real)(prompted for)
C
C   WHICH_RVT   Number of the RVT already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_RVF   Number of the RVF already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_MASK  Number of the MASK already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_GRID  Number of the GRID already defined in TAUCAL to use.
C               (integer)(prompted for)
C
C   WHICH_FRAME Number of the frame to use if doing an RVT with a cube.
C               (integer)(prompted for)
C
C
C   Keywords (read from VARS.DST)
C   -----------------------------
C   USE_TAU       Flag for use of TAUCAL file. (prompted for)
C
C   USE_GRID      Instruction to either use the grid of results from
C                 MASK and GRID to define pixel-space or not. (prompted for)
C
C   RVF           True if you are doing an RVF this time around.  (prompted for)
C
C   RESTRICT      True if the user wants to restrict the area analysed to
C                 just the area between -sigs*sigma --> sigs*sigma from
C                 the radius. (prompted for)
C
C   CIRCLE        True if the user want to constrain the axial ratio of the
C                 ellipse to 1. (no prompt)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .RVF.RVFn          RINGRUN    Run number associated with ring data cube.
C                                 (integer)(read).
C
C   .RVF.RVFn          NAME_RVF   Name of output RVF file (character)(read)
C
C   .RVF               NRVF       Number of RVFs already defined in TAUCAL.
C                                 (integer)(read)
C
C   .MASK              NMASK      Number of MASKs already defined in TAUCAL.
C                                 (integer)(read)
C
C   .MASK.MASKn        NGRID      Number of GRIDs already defined in MASK
C                                 sub-structure of TAUCAL. (integer)(read)
C
C   .MASK.MASKn.GRIDn  NAME_GRID  Name of input file with grid of X,Y pixel
C                                 positions. (character)(read)
C
C   .RVT.RVTn          NAME_RVT   Name of output RVT file (character)(read)
C
C   .RVT               NRVT       Number of RVTs already defined in TAUCAL.
C                                 (integer)(read)
C
C   .RVT.RVTn          NRUNS      Number of rings used in a particular RVT.
C                                 (integer)(read)
C
C   .RVT.RVTn          RUN_USED   Run number of ring used in PRERING for this
C                                 RVT. (integer)(read)
C
C   (RING)             NAME_USED  Image used in PRERING for this RVT.
C                                 (integer)(read)
C
C   .RVT.RVTn          RUNS       Run numbers of all the rings in an RVT.
C                                 (integer[nruns])(read)
C
C   .RVT.RVTn          FRAMES     Run numbers of all the rings in an RVT.
C                                 (integer[nruns])(read)
C
C   (CAL/RING)         NAME_CAL   Name of image containing the rings
C                                 (character)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The IMAGE structure is tested for dimensionality.
C   - Image data is mapped according to its input type.
C   - The output file is mapped as FLOAT.
C   - The program reads rough parameters from the output file
C     which were calculated by PRERING.
C   - A function which is a ring of variable axial ratio with a
C     gaussian distribution about the radius is fit to the intensity
C     distribution of the chosen pixels.  A constant background
C     is assumed.
C   - Output is to RVF/RVT file.  Both have the same number of data
C     points for each entry.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_CLOSE_STRUCTURE
C     DSA_DATA_SIZE
C     DSA_DATA_TYPE
C     DSA_FMTCON
C     DSA_FREE_WORKSPACE
C     DSA_GET_FLAG_VALUE
C     DSA_GET_WORKSPACE
C     DSA_GET_WORK_ARRAY
C     DSA_MAP_DATA
C     DSA_MAP_ERRORS
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_POST_PROCESS_FLAGGED_VALUES
C     DSA_TYPESIZE
C     DSA_UNMAP
C     DSA_USE_FLAGGED_VALUES
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library GEN:
C     GEN_FILL
C     GEN_MOVE
C
C   Library ICH:
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_DUMMY_GRID
C     TAU_CONVERGE
C     MRQMIN (TAU_PRESS)
C
C   Internal subroutines called
C   ---------------------------
C     FIT
C     RFILL
C     RING_FUNC
C     SUB_RINGS
C     TABLE
C     VALS_I
C     VALS_R
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   INTEGER*2 / DO WHILE
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL  or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   27-FEB-90   - Original program
C   15-MAR-90   - Added RVT file  (JRL)
C   02-APR-90   - Added axial ratio as a free parameter  (JRL)
C   02-MAY-90   - Major changes so that program interfaces with
C                 PRERING.  (JRL)
C   08-MAY-90   - Added RVF keyword  (JRL)
C   10-MAY-90   - Added facility to restrict region aroung ring.  (JRL)
C   21-MAY-90   - Another major overhaul.  Now caters for both SHORT
C                 and FLOAT data.  A new fitting routine derived from
C                 Press et.al. introduced which is MUCH faster than NAG
C                 and works in single precision.  A few cosmetics as well. (JRL)
C   03-JUL-90   - TAUCAL pasted in.  Also slight changes in convergence
C                 procedure.  (JRL)
C   16-OCT-90   - Ability to compress the input array added in.  Also added
C                 ability to change the weighting scheme.  (JRL)
C   06-NOV-90   - Changed to cope with new RVT strategy.  (JRL)
C   08-NOV-90   - Output files now have error arrays. (JRL)
C   14-NOV-90   - Added ability to constrain axial ratio to 1. (JRL)
C   16-NOV-90   - RVF and GRID info now read from TAUCAL RVF and
C                 MASK sub-structures.  (JRL)
C   19-NOV-90   - New TAUCAL routines pasted in (JRL)
C   20-NOV-90   - Slight changes in convergence procedure. (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions.
C
      integer          dsa_typesize
      integer          dyn_element
C
C     Default values for NTRIV and TRIVIAL.
C
      integer ntriv_def
      parameter (ntriv_def=3)
      real trivial_def
      parameter (trivial_def=0.01)
C
C     Local variables.
C
      integer          address         ! Dynamic memory address
      logical          badpix          ! Flag for magic values image file
      logical          badpixg         ! Flag for magic values in grid file
      logical          badpixo         ! Flag for magic values in output file
      logical          circle          ! True if you want to constrain ellipticity to 1.0
      integer          comp_fac        ! Compression factor
      character        data_type*8     ! Data type of the structure used in PRERING for RVT
      integer          dims(3)         ! Array of image dimensions
      real             dummy           ! Dummy real variable
      integer          eptr            ! Dynamic memory pointer for error array
      integer          eslot           ! Dynamic memory slot for error array
      real             frames(20)      ! Frames to be used in RVT
      integer          gdims(3)        ! Dimensions of GRID array
      integer          gptr            ! Dynamic memory GRID pointer
      character        grid_lab*80     ! TAUCAL label for GRID
      integer          gslot           ! Dynamic memory GRID slot
      integer          i               ! Integer counter
      integer          iptr            ! Dynamic memory IMAGE pointer
      integer          islot           ! Dynamic memory IMAGE slot
      integer          len_gridlab     ! Length of grid_lab
      integer          len_masklab     ! Length of mask_lab
      integer          len_rvflab      ! Length of rvf_lab
      integer          len_rvtlab      ! Length of rvt_lab
      character        mask_lab*80     ! TAUCAL label for MASK
      character        name_cal*80     ! Name of calibration structure
      character        name_grid*80    ! Name of grid structure
      character        name_used*80    ! Name of structure used in PRERING for RVT
      integer          ndim            ! Number of dimensions in IMAGE
      integer          nelm            ! Total number of elements
      integer          nf              ! Frame number to be used
      integer          ngdim           ! Number of dimension in GRID
      integer          nodim           ! Number of dimensions in output grid
      integer          nruns           ! Number of rings in a given RVT
      integer          ntriv           ! Number of times d(chi-squared) < trivial for convergence
      real             numbers(20)     ! Run numbers of rings included in RVT
      integer          nux             ! X dimension of structure used in PRERING for RVT
      integer          nuy             ! Y dimension of structure used in PRERING for RVT
      integer          odims(2)        ! Array with RVF output file dimensions
      integer          optr            ! Dynamic memory OUTPUT pointer
      integer          oslot           ! Dynamic memory OUTPUT slot
      logical          restrict        ! Flag to restrict area of analysis
      integer          ringrun         ! Run number for calibration ring structure
      integer          run_used        ! Run number of ring used for PRERING estimate for RVT
      logical          rvf             ! Flag for type of output file
      character        rvf_lab*80      ! TAUCAL label for RVF
      character        rvt_lab*80      ! TAUCAL label for RVT
      real             sigs            ! Number of sigma from radius for restricted analysis
      integer          status          ! DSA status flag
      logical          struct          ! Flag for whether data is a structured type
      character        structure*80    ! Name of RVT structure
      logical          tauexist        ! Flag for existence of TAUCAL file
      real             tolcent         ! Tolerance in error of ring centre
      real             tolrad          ! Tolerance in error of ring radius
      integer          totspace        ! Total allocation for workspace
      real             trivial         ! Definition of a trivial change in chi-squared
      character        type*8          ! Type of data in image file
      character        typeg*8         ! Type of data in grid file
      character        typeo*8         ! Type of data in output file
      integer          type_size       ! Number of bytes in data type
      logical          use_grid        ! Flag for whether the results from GRID are being used
      integer          weight          ! Flag for the weighting scheme to be used
      integer          wptr            ! Dynamic memory workspace pointer
      integer          wslot           ! Dynamic memory workspace slot
C
C     Numeric ranges.
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Common block for pointers
C
      common /pointers/ iptr
C
C     Open DSA system.
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get RVF flag
C
      call par_rdkey('rvf',.true.,rvf)
C
C     If this is not an RVF then find out which RVT you're using and
C     how many runs are to be included
C
      if (.not. rvf) then
C
C        If using TAUCAL, then find out how many RVTs are on offer...
C
         call tau_cal_find_struct('cal','rvt','rvt',tauexist,rvt_lab,
     :   len_rvtlab,status)
         if (status .ne. 0) go to 500
C
C        Get the number of rings for that RVT
C
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.nruns',
     :   'nruns',tauexist,.true.,0.0,float(max_int),' ',1,dummy,status)
         if (status .ne. 0) go to 500
         nruns = nint(dummy)
         if (nruns .eq. 0) then
            call dsa_wruser
     :      ('You can''t have an RVT with no rings!!')
            call dsa_wrflush
            go to 500
         end if
C
C        Find out which ring was used to get the initial estimates and
C        get its map size
C
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.run_used',
     :   'run_used',tauexist,.false.,float(min_int),float(max_int),' ',
     :   1,dummy,status)
         run_used = nint(dummy)
         call tau_cal_rdchr('cal',run_used,'image','name_used',tauexist,
     :   .true.,80,name_used,status)
         if (status .ne. 0) go to 500
         call dsa_named_input('image',name_used,status)
         call dsa_data_size('image',3,ndim,dims,nelm,status)
         call dsa_data_type('image',data_type,struct,status)
         nux = dims(1)
         nuy = dims(2)
         call dsa_close_structure('image',status)
         if (status .ne. 0) go to 500
C
C        Get some workspace
C
         type_size = dsa_typesize(data_type,status)
         totspace = nux*nuy*nruns*type_size
         call dsa_get_workspace(totspace,address,wslot,status)
         wptr = dyn_element(address)
C
C        Now open the individual ring files and copy the data into the workspace
C
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.runs',
     :   'runs',tauexist,.false.,float(min_int),float(max_int),' ',
     :   nruns,numbers,status)
         call tau_cal_rdnum('cal',0,rvt_lab(:len_rvtlab)//'.frames',
     :   'frames',tauexist,.false.,float(min_int),float(max_int),
     :   ' ',nruns,frames,status)
         do i = 1,nruns
            call tau_cal_rdchr('cal',nint(numbers(i)),'image',
     :      'name_cal',tauexist,.true.,80,name_cal,status)
            call dsa_named_input('image',name_cal,status)
            if (status .ne. 0) go to 500
C
C           Check image dimensions
C
            call ndp_get_image_info('image',.true.,.false.,type,badpix,
     :      status)
            call dsa_data_size('image',3,ndim,dims,nelm,status)
            if (badpix) call dsa_use_flagged_values('image',status)
            if (status .ne. 0) go to 500
            if (ndim .lt. 2) then
               call dsa_wruser(' Image has unacceptable dimensions.')
               call dsa_wrflush
               go to 500
            end if
            if ((dims(1) .ne. nux) .or. (dims(2) .ne. nuy) .or.
     :      (type .ne. data_type)) then
               call dsa_wruser('Image conflicts with main ring image')
               call dsa_wrflush
               go to 500
            end if
C
C           If this is a 3-d image, then find out which frame to use
C
            if (tauexist) then
               nf = nint(frames(i))
            else
               if (ndim .eq. 3) then
                  call par_rdval('which_frame',1.0,float(dims(3)),1.0,
     :            ' ',dummy)
                  nf = nint(dummy)
               else
                  nf = 1
               end if
            end if
C
C           Map IMAGE data array.
C
            if (type(1:5) .eq. 'SHORT') then
               call dsa_map_data('image','read','short',address,islot,
     :         status)
            else
               call dsa_map_data('image','read','float',address,islot,
     :         status)
            end if
            iptr = dyn_element(address)
C
C           Copy it all over now...
C
            call gen_move(nux*nuy*type_size,
     :      dynamic_mem(iptr+(nf-1)*nux*nuy*type_size),
     :      dynamic_mem(wptr+(i-1)*nux*nuy*type_size))
C
C           Now close the file
C
            call dsa_unmap(islot,status)
            if (badpix)
     :      call dsa_post_process_flagged_values('image',status)
            call dsa_close_structure('image',status)
            if (status .ne. 0) go to 500
         end do
         iptr = wptr
         dims(1) = nux
         dims(2) = nuy
         dims(3) = nruns
C
C     If this is for an RVF....
C
      else
C
C        If using TAUCAL, then find out how many RVFs are on offer...
C
         call tau_cal_find_struct('cal','rvf','rvf',tauexist,rvf_lab,
     :   len_rvflab,status)
         if (status .ne. 0) go to 500
C
C        Get the run number for that RVF and its associated image
C
         call tau_cal_rdnum('cal',0,rvf_lab(:len_rvflab)//'.run',
     :   'ringrun',tauexist,.false.,0.0,float(max_int),' ',1,dummy,
     :   status)
         ringrun = nint(dummy)
         call tau_cal_rdchr('cal',ringrun,'image','name_cal',tauexist,
     :   .true.,80,name_cal,status)
         call dsa_named_input('image',name_cal,status)
         if (status .ne. 0) go to 500
C
C        Check image dimensions
C
         call ndp_get_image_info('image',.true.,.false.,type,badpix,
     :   status)
         call dsa_data_size('image',3,ndim,dims,nelm,status)
         if (badpix) call dsa_use_flagged_values('image',status)
         if (status .ne. 0) go to 500
         if (ndim .lt. 2) then
            call dsa_wruser(' Image has unacceptable dimensions.')
            call dsa_wrflush
            go to 500
         end if
         if (ndim .eq. 2) dims(3) = 1
C
C        Map IMAGE data array.
C
         if (type(1:5) .eq. 'SHORT') then
            call dsa_map_data('image','read','short',address,islot,
     :      status)
         else
            call dsa_map_data('image','read','float',address,islot,
     :      status)
         end if
         if (status .ne. 0) go to 500
         iptr = dyn_element(address)
      end if
C
C     Get parameters.
C
      call par_rdval('comp_fac',1.0,float(max_int-1),1.0,' ',dummy)
      comp_fac = nint(dummy)
      call par_rdval('tolcent',0.0,max_float,1.0,' ',tolcent)
      call par_rdval('tolrad',0.0,max_float,1.0,' ',tolrad)
      call par_rdval('weight',1.0,2.0,1.0,' ',dummy)
      weight = nint(dummy)
      call par_rdval('trivial',0.0,max_float,trivial_def,' ',trivial)
      call par_rdval('ntriv',0.0,float(max_int-1),ntriv_def,' ',dummy)
      ntriv = nint(dummy)
      call par_rdkey('circle',.false.,circle)
C
C     Decide whether or not to use a grid
C
      call par_rdkey('use_grid',.true.,use_grid)
      if (use_grid) then
C
C        First find out which mask you want
C
         call tau_cal_find_struct('cal','mask','mask',tauexist,mask_lab,
     :   len_masklab,status)
         if (status .ne. 0) go to 500
C
C        Now find out which grid you want
C
         call tau_cal_find_struct('cal',mask_lab(:len_masklab),'grid',
     :   tauexist,grid_lab,len_gridlab,status)
         if (status .ne. 0) go to 500
C
C        Open the grid file and map it
C
         call tau_cal_rdchr('cal',0,
     :   grid_lab(:len_gridlab)//'.name_grid','name_grid',
     :   tauexist,.true.,80,name_grid,status)
         call dsa_named_input('grid',name_grid,status)
         call ndp_get_image_info('grid',.true.,.false.,typeg,badpixg,
     :   status)
         if (status .ne. 0) go to 500
         call dsa_data_size('grid',3,ngdim,gdims,nelm,status)
         if ((gdims(2) .ne. dims(1)) .or. (gdims(3) .ne. dims(2))) then
            call dsa_wruser('Grid file has wrong dimensions...')
            call dsa_wrflush
            go to 500
         end if
         call dsa_map_data('grid','read','float',address,gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
C
C     If not using a pre-defined grid, then create a dummy grid.
C
      else
         call dsa_get_work_array(dims(1)*dims(2)*2,'float',address,
     :   gslot,status)
         if (status .ne. 0) go to 500
         gptr = dyn_element(address)
         call tau_dummy_grid(dims(1),dims(2),dynamic_mem(gptr))
      end if
C
C     If you want to restrict the area of analysis, then get the number
C     of sigmas (initial estimate from PRERING) away from the radius.
C     This will define an annulus for the analysis.
C
      call par_rdkey('restrict',.false.,restrict)
      if (restrict) then
         call par_rdval('sigs',0.0,max_float,2.0,'sigmas',sigs)
      end if
C
C     Get output file name
C
      if (rvf) then
         call tau_cal_rdchr('cal',0,rvf_lab(:len_rvflab)//'.name_rvf',
     :   'name_rvf',tauexist,.true.,80,structure,status)
         if (status .ne. 0) go to 500
      else
         call tau_cal_rdchr('cal',0,rvt_lab(:len_rvtlab)//'.name_rvt',
     :   'name_rvt',tauexist,.true.,80,structure,status)
         if (status .ne. 0) go to 500
      end if
C
C     Open OUTPUT file
C
      call dsa_named_input('output',structure,status)
      if (status .ne. 0) then
         call dsa_wruser('***Output structure doens''t exist...')
         call dsa_wrflush
         call dsa_wruser('Try running PRERING to create it***')
         call dsa_wrflush
         go to 500
      end if
      call ndp_get_image_info('output',.true.,.false.,typeo,badpixo,
     : status)
      call dsa_data_size('output',2,nodim,odims,nelm,status)
C
C     Check to make sure that the number of frames in the output file is
C     the same as for the input file...
C
      if (odims(2) .ne. dims(3)) then
         call dsa_wruser('***Number of frames in input structure ')
         call dsa_wrflush
         call dsa_wruser
     :   ('   doesn''t match the number in the output file***')
         call dsa_wrflush
         go to 500
      end if
C
C     Map output array and its error array
C
      call dsa_map_data('output','update','float',address,oslot,
     : status)
      optr = dyn_element(address)
      call dsa_map_errors('output','write','float',address,eslot,
     : status)
      eptr = dyn_element(address)
      if (status .ne. 0) go to 500
C
C     Fill the error array with zeros...
C
      totspace = 10*odims(2)*dsa_typesize('float',status)
      call gen_fill(totspace,0,dynamic_mem(eptr))
C
C     Now fit rings to each frame.
C
      call sub_rings(dims(1),dims(2),dims(3),dynamic_mem(gptr),type,
     : restrict,sigs,trivial,ntriv,tolcent,tolrad,comp_fac,weight,
     : circle,odims(2),dynamic_mem(optr),dynamic_mem(eptr))
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end



      subroutine sub_rings(nx,ny,nz,coords,type,restrict,sigs,
     : trivial,ntriv,tolcent,tolrad,comp_fac,weight,circle,nrvf,rvfar,
     : errs)
C+-------------------------------------------------------------------------------------
C     SUB_RINGS    --- Does the actual fitting of the ring profiles to the
C                      intensity distribution for each frame.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NX:        (integer) X dimension of arrays
C     (>) NY:        (integer) Y dimension of arrays
C     (>) NZ:        (integer) Z dimension of arrays
C     (>) COORDS:    (real array COORDS(2,NX,NY)) Either the coordinate grid
C                     defined in GRID or a workspace.
C     (>) TYPE:      (character) Type of input data.
C     (>) RESTRICT:  (logical) True if the user wants to restrict the analysis region
C     (>) SIGS:      (real) Number of sigma to define annulus of analysis
C     (>) TRIVIAL:   (real) Definition of a trivial change in chi-squared
C     (>) NTRIV      (integer) Number of times d(chi_squared) < trivial to signal convergence
C     (>) TOLCENT:   (real) Tolerance of the error of coords of ring centre
C     (>) TOLRAD:    (real) Tolerance of the error of ring radius
C     (>) COMP_FAC:  (integer) Compression factor
C     (>) WEIGHT:    (integer) Flag for weighting scheme.
C     (>) CIRCLE:    (logical) True if the user wants axial ratio constrained to 1.0
C     (>) NRVF:      (integer) The number of elements in the RVF/RVT array
C     (!) RVFAR:     (real array RVFAR(10,NRVF)) On entry: array of rough
C                     parameter from PRERING.  On exit: array with RVF/RVT results.
C                    (1) Z, (2) Xc, (3) Yc, (4) Radius, (5) Width, (6) Peak,
C                    (7) Background, (8) Axial ratio, (9) Mean error of fit,
C                    (10) Flag for quality of fit used by FSR_GAP.
C     (<) ERRS:      (real array ERRS(10,NRVF)) Output file error array.
C
C     AUTHOR:        Jim Lewis -- RGO
C     DATE:          27-FEB-1990
C     UPDATE:        02-APR-1990 -- Added in axial ratio as a parameter. (JRL)
C                    02-MAY-1990 -- Guessing routine taken out.  Rough estimates
C                                   of ring parameters are now read in from output
C                                   structure.  (JRL)
C                    14-MAY-1990 -- Both SHORT and FLOAT input data useable now. (JRL)
C                    21-MAY-1990 -- New fitting routines.  (JRL)
C                    16-OCT-1990 -- Data compression and weighting schemes added. (JRL)
C                    08-NOV-1990 -- Output error array added (JRL)
C                    14-NOV-1990 -- Ability to constrain axial ratio to 1 added. (JRL)
C                    20-NOV-1990 -- Modified to take runaway solutions into account. (JRL)
C
C+-------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,ntriv,nrvf,comp_fac,weight
      real coords(2,nx,ny),rvfar(10,nrvf),trivial,tolrad,
     : tolcent,sigs,errs(10,nrvf)
      character type*(*)
      logical restrict,circle
C
C     Functions
C
      integer ich_len
      integer dsa_typesize
      integer dyn_element
C
C     Number of variables in the fit
C
      integer nvar
      parameter (nvar=7)
C
C     Minimum number of points for fit
C
      integer minpoints
      parameter (minpoints=nvar+1)
C
C     Local Variables
C
      integer i,ncoord,iframe,errsum,index,address,iwslt,fit_status,
     : bytes_float,status,iwpt1,ixerr,iyerr,irerr,iptr,iwpt2,
     : totspace,bytes_int,iwslt2,iwpt3,iwpt4,iwpt5,iwpt6,size
      real par(nvar),errcoef(nvar),meanerr,dr,r1,r2,xc,yc
      character message*80
      logical alloc,init,runaway,alloc2
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Common block with pointers
C
      common /pointers/ iptr
C
C     Common block to flag for initialisation of look up table in
C     RING_FUNC.
C
      common /flag/ init
C
C     Initialise the look up table in first entry to RING_FUNC
C
      init = .true.
C
C     Label for parameters to help the user...
C
      write(message,101)
  101 format(' Frame   Max      Xc      Yc       R      Sigma     ',
     : 'Back     Axial')
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
C
C     Allocate a bit of workspace for the fitting routines...
C
      alloc2 = .false.
      bytes_float = dsa_typesize('float',status)
      bytes_int = dsa_typesize('int',status)
      totspace = 2*nvar*nvar*bytes_float + nvar*bytes_int
      call dsa_get_workspace(totspace,address,iwslt2,status)
      if (status .ne. 0) go to 600
      iwpt4 = dyn_element(address)
      iwpt5 = iwpt4 + nvar*nvar*bytes_float
      iwpt6 = iwpt5 + nvar*nvar*bytes_float
      alloc2 = .true.
      size = (nx/comp_fac + 1)*(ny/comp_fac + 1)
C
C     Loop for each frame.
C
      do iframe = 1,nz
         index = iframe
         if (nint(rvfar(10,index)) .eq. 2) then
            write(message,150) iframe
  150       format(I4,5X,
     :      'not used: Frame was flagged for skipping')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            go to 500
         end if
C
C        Get some workspace for the data arrays
C
         alloc = .false.
         totspace = 4*size*bytes_float
         call dsa_get_workspace(totspace,address,iwslt,status)
         if (status .ne. 0) then
            call dsa_wruser('Failed to allocate workspace')
            call dsa_wrflush
            go to 500
         end if
         iwpt1 = dyn_element(address)
         iwpt2 = iwpt1 + 2*size*bytes_float
         iwpt3 = iwpt2 + size*bytes_float
         alloc = .true.
C
C        Calculate the radii for restricted areas
C
         dr = sigs*rvfar(5,iframe)
         r1 = rvfar(4,iframe) - dr
         r2 = rvfar(4,iframe) + dr
         xc = coords(1,nint(rvfar(2,iframe)),nint(rvfar(3,iframe)))
         yc = coords(2,nint(rvfar(2,iframe)),nint(rvfar(3,iframe)))
C
C        Put the data into some work arrays
C
         call rfill(size,comp_fac,nx,ny,nz,iframe,coords,restrict,r1,r2,
     :   xc,yc,type,iptr,dynamic_mem(iwpt1),dynamic_mem(iwpt2),ncoord)
C
C        Guesses for parameters
C
         par(1) = rvfar(6,index)        !maximum intensity
         par(2) = rvfar(2,index)        !centre x coord
         par(3) = rvfar(3,index)        !centre y coord
         par(4) = rvfar(4,index)        !radius
         par(5) = rvfar(5,index)        !width
         par(6) = rvfar(7,index)        !background intensity
         par(7) = rvfar(8,index)        !axial ratio
C
C        Do 2-d fit
C
         call fit(ncoord,dynamic_mem(iwpt1),dynamic_mem(iwpt2),
     :   dynamic_mem(iwpt3),nvar,dynamic_mem(iwpt4),dynamic_mem(iwpt5),
     :   dynamic_mem(iwpt6),weight,trivial,ntriv,circle,par,errcoef,
     :   meanerr,runaway,fit_status)
C
C        Reject if the fitting routine failed.
C
         if (fit_status .ne. 0) then
            write(message,170) iframe
  170       format(i4,5x,'not used: 2-d fit failed...')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            rvfar(10,index) = 2.0
            go to 500
         end if
C
C        Write out results of the fit
C
         write(message,180) iframe,(par(i),i=1,nvar)
  180    format(i4,<nvar>f9.3,'')
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        Reject if the maximum intensity is negative
C
         if (par(1) .lt. 0.0) then
            write(message,191) iframe
  191       format(i4,5x,'not used: negative intensity for peak')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            rvfar(10,index) = 2.0
            go to 500
         end if
C
C        Reject if radius is negative
C
         if (par(4) .lt. 0.0) then
            write(message,192) iframe
  192       format(i4,5x,'not used: negative radius found')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            rvfar(10,index) = 2.0
            go to 500
         end if
C
C        Write out the errors
C
         write(message,180) iframe,(errcoef(i),i=1,nvar)
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C        Define how much the errors are over the tolerances
C
         ixerr = int(errcoef(2)/tolcent)
         iyerr = int(errcoef(3)/tolcent)
         irerr = int(errcoef(4)/tolrad)
C
C        If the errors are more than twice the tolerance errors,
C        the reject the whole thing.
C
         if ((ixerr .ge. 2) .or. (iyerr .ge. 2) .or.
     :      (irerr .ge. 2)) then
            write(message,190) iframe
  190       format(i4,5x,'not used: errors in 2-d fit are too big')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            rvfar(10,index) = 2.0
            go to 500
         end if
C
C        Store the results
C
         rvfar(2,index) = par(2)
         errs(2,index) = errcoef(2)
         rvfar(3,index) = par(3)
         errs(3,index) = errcoef(3)
         rvfar(4,index) = par(4)
         errs(4,index) = errcoef(4)
         rvfar(5,index) = par(5)
         errs(5,index) = errcoef(5)
         rvfar(6,index) = par(1)
         errs(6,index) = errcoef(1)
         rvfar(7,index) = par(6)
         errs(7,index) = errcoef(6)
         rvfar(8,index) = par(7)
         errs(8,index) = errcoef(7)
         rvfar(9,index) = meanerr
         rvfar(10,index) = 0.0
C
C        If 1 out of the 3 is higher than the tolerance then flag the point.
C
         errsum = ixerr + iyerr + irerr
         if (errsum .gt. 1) then
            rvfar(10,index) = 1.0
         end if
C
C        If this was a runaway solution, then flag it as half weight...
C
         if (runaway) then
            write(message,193) iframe
  193       format(i4,5x,'Runaway solution. May still be ok')
            call dsa_wruser(message(:ich_len(message)))
            call dsa_wrflush
            rvfar(10,index) = 1.0
         end if
         call dsa_wruser(' ')
         call dsa_wrflush
C
C        if workspace was allocated, then free it...
C
  500    if (alloc) then
            call dsa_free_workspace(iwslt,status)
            alloc = .false.
         end if
      end do
  600 if (alloc2) then
         call dsa_free_workspace(iwslt2,status)
         alloc2 = .false.
      end if
      end



      subroutine rfill(size,comp_fac,nx,ny,nz,iframe,coords,restrict,
     : r1,r2,xc,yc,type,iptr,xy,aintens,ncoord)
C+--------------------------------------------------------------------------------------
C     RFILL        --- Routine to fill in the work arrays with values of x,y,i
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) SIZE:      (integer) Total number pixels to be used
C     (>) COMP_FAC:  (integer) Compression factor
C     (>) NX:        (integer) X dimension of the frame
C     (>) NY:        (integer) Y dimension of the frame
C     (>) NZ:        (integer) Z dimension of the frame
C     (>) IFRAME:    (integer) The frame in question
C     (>) COORDS:    (real array COORDS(2,NX,NY)) Grid of X,Y
C     (>) RESTRICT:  (logical) Keyword to restrict area to cocentric annuli
C     (>) R1:        (real) Inner radius of restricted region
C     (>) R2:        (real) Outer radius of restricted region
C     (>) XC:        (real) Nominal X coord of centre
C     (>) YC:        (real) Nominal Y coord of centre
C     (>) TYPE:      (character) Input data type
C     (>) IPTR:      (integer) Input data pointer
C     (<) XY:        (real array XY(2,SIZE)) The array of XY positions for fit
C     (<) AINTENS:   (real array AINTENS(SIZE)) The array of intensities for fit
C     (<) NCOORD:    (integer) The actual number of pixels which fit the criteria
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        21-MAY-1990
C     UPDATE:      16-OCT-1990 -- Major changes.  What was originally two transfer
C                                 routines have been run together into this one. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer size,comp_fac,nx,ny,nz,iframe,iptr,ncoord
      real coords(2,nx,ny),r1,r2,xc,yc,xy(2,size),aintens(size)
      logical restrict
      character type*(*)
C
C     Local variables
C
      integer jx,jy,nbad,status
      integer*2 val_i
      real xx,yy,x,y,r,val_r,magic_float
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Count the number of included pixels
C
      ncoord = 0
C
C     Get magic value
C
      call dsa_get_flag_value('FLOAT',magic_float,status)
C
C     Loop through the frame
C
      do jy = 1,ny,comp_fac
         do jx = 1,nx,comp_fac
C
C           Find radius
C
            xx = coords(1,jx,jy)
            yy = coords(2,jx,jy)
            x = xx - xc
            y = yy - yc
            r = sqrt(x*x + y*y)
C
C           Decide whether to include this pixel...
C
            if ((.not. restrict) .or. ((restrict) .and.
     :      ((r .ge. r1) .and. (r .le. r2)))) then
C
C              Do the type conversion if necessary
C
               if (type .eq. 'SHORT') then
                  call vals_i(dynamic_mem(iptr),nx,ny,nz,jx,jy,iframe,
     :            val_i)
                  call dsa_fmtcon(.true.,2,4,val_i,val_r,1,nbad)
               else
                  call vals_r(dynamic_mem(iptr),nx,ny,nz,jx,jy,iframe,
     :            val_r)
               end if
C
C              If this is a magic value, then skip it
C
               if (val_r .gt. magic_float) then
                  ncoord = ncoord + 1
                  xy(1,ncoord) = xx
                  xy(2,ncoord) = yy
                  aintens(ncoord) = val_r
               end if
            end if
         end do
      end do
      end


      subroutine vals_i(input,nx,ny,nz,jx,jy,iframe,val_i)
C+--------------------------------------------------------------------------------------
C     VALS_I      --- Routine to grab a value from a SHORT input array
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (short array INPUT(NX,NY,NZ)) Input data array
C     (>) NX:        (integer) X dimension of the frame
C     (>) NY:        (integer) Y dimension of the frame
C     (>) NZ:        (integer) Z dimension of the frame
C     (>) JX:        (integer) X pixel wanted
C     (>) JY:        (integer) Y pixel wanted
C     (>) IFRAME:    (integer) The frame in question
C     (<) VAL_I:     (short) Output value.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        16-OCT-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iframe,jx,jy
      integer*2 input(nx,ny,nz),val_i
C
C     Get the value you want
C
      val_i = input(jx,jy,iframe)
      end


      subroutine vals_r(input,nx,ny,nz,jx,jy,iframe,val_r)
C+--------------------------------------------------------------------------------------
C     VALS_R      --- Routine to grab a value from a REAL input array
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (real array INPUT(NX,NY,NZ)) Input data array
C     (>) NX:        (integer) X dimension of the frame
C     (>) NY:        (integer) Y dimension of the frame
C     (>) NZ:        (integer) Z dimension of the frame
C     (>) JX:        (integer) X pixel wanted
C     (>) JY:        (integer) Y pixel wanted
C     (>) IFRAME:    (integer) The frame in question
C     (<) VAL_R:     (real) Output value.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        16-OCT-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nz,iframe,jx,jy
      real input(nx,ny,nz),val_r
C
C     Get the value you want
C
      val_r = input(jx,jy,iframe)
      end


      subroutine table(darg,ndiv,expon)
C+--------------------------------------------------------------------------------------
C     TABLE       --- Routine to set up a look-up table for values of the exponential
C                     function
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) DARG:      (real) Size of the bins
C     (>) NDIV:      (integer) Number of bins
C     (<) EXPON:     (real array EXPON(NDIV+1)) Ouput array with the table
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        14-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer ndiv
      real darg,expon(ndiv+1)
C
C     Local variables
C
      integer i
      real arg
C
C     Create the table.
C
      do i = 1,ndiv
         arg = darg*float(i-1)
         expon(i) = exp(-arg)
      end do
      expon(ndiv+1) = 0.0
      end


      subroutine fit(ncoord,xy,aintens,sig,nvar,covar,alpha,lista,
     : weight,trivial,ntriv,circle,par,erpar,rms,runaway,status)
C+--------------------------------------------------------------------------------------
C     FIT         --- Routine to do the actual iterated fit
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) NCOORD:    (integer) Total number of pixels to be fit
C     (>) XY:        (real array XY(2,NCOORD))  X,Y positions for fit
C     (>) AINTENS:   (real array AINTENS(NCOORD)) Intensities for fit
C     (W) SIG:       (real array SIG(NCOORD)) Array of intensity errors.
C     (>) NVAR:      (integer) Number of variables in the fit
C     (W) COVAR:     (real array COVAR(NVAR,NVAR)) Covariance matrix workspace.
C     (W) ALPHA:     (real array ALPHA(NVAR,NVAR)) Curvature matrix workspace.
C     (W) LISTA:     (integer array LISTA(NVAR)) Flags for holding particular parameters
C                    constant in the fit.  Always set so that all parameters vary.
C     (>) WEIGHT:    (integer) Flag for weighting scheme.
C     (>) TRIVIAL:   (real) Definition of a trivial change in chi-squared
C     (>) NTRIV:     (integer) The number of times d(chi-squared) < trivial for the
C                    fit to be considered to have converged.
C     (>) CIRCLE:    (logical) True if the user wants to constrain the axial ratio to 1.0
C     (<) PAR:       (real array PAR(NVAR)) Array of fitted parameters
C     (<) ERPAR:     (real array ERPAR(NVAR)) Array of parameter errors
C     (<) RMS:       (real) RMS error in the fit
C     (<) RUNAWAY:   (logical) Flag for runaway fit.
C     (<) STATUS:    (integer) Status flag for bad fit.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        21-MAY-1990
C     UPDATE:      03-JUL-1990 -- Slight changes to convergence procedure.  (JRL)
C                  16-OCT-1990 -- New weighting schemes added. (JRL)
C                  14-NOV-1990 -- Ability to constrain axial ratio added. (JRL)
C                  20-NOV-1990 -- Again, slight changes to convergence procedure. (JRL)
C                  10-DEC-1990 -- Fixed bug in RMS estimation. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nvar,ncoord,status,ntriv,weight
      integer lista(nvar)
      real xy(2,ncoord),aintens(ncoord),par(nvar),covar(nvar,nvar),
     : alpha(nvar,nvar),sig(ncoord),rms,erpar(nvar),trivial
      logical circle,runaway
C
C     Functions
C
      integer dyn_element
      integer dsa_typesize
C
C     Local variables
C
      integer i,bytes_float,bytes_int,totspace,address,iwslt,
     : iwpt1,iwpt2,iwpt3,iwpt4,iwpt5,itriv,ilam,mfit,iter,dsa_status
      real alamda,chisq,ochisq,olamda,sumwts
      logical continue,alloc
      external ring_func
C
C     Common block for dynamic memory.
C
      include 'DYNAMIC_MEMORY'
C
C     Initialise status variables
C
      status = 0
      dsa_status = 0
      alloc = .false.
C
C     Initialise the sigma for the data points.  Also set up a flag
C     array for the fitting routine. NB:  The sqrt(n) weights have
C     been made sqrt(n+1) because of pixels which have an intensity
C     of zero.  These are real data and shouldn't have zero weight.
C     I'm trying to think of a more satisfying solution...
C
      sumwts = 0.0
      do i = 1,ncoord
         if (weight .eq. 1) then
            sig(i) = 1.0/sqrt(aintens(i)+1.0)
         else if (weight .eq. 2) then
            sig(i) = 1.0
         end if
         sumwts = sumwts + 1.0/(sig(i)**2)
      end do
      do i = 1,nvar
         lista(i) = i
      end do
      mfit = nvar
C
C     If you want to constrain axial ratio, then do so...
C
      if (circle) then
         lista(7) = 0
         par(7) = 1.0
         erpar(7) = 0.0
         mfit = mfit - 1
      end if
C
C     Get a bit more workspace for the fitting routines...
C
      bytes_float = dsa_typesize('float',dsa_status)
      bytes_int = dsa_typesize('int',dsa_status)
      totspace = 4*nvar*bytes_float + 3*nvar*bytes_int
      call dsa_get_workspace(totspace,address,iwslt,dsa_status)
      if (dsa_status .ne. 0) go to 500
      alloc = .true.
      iwpt1 = dyn_element(address)
      iwpt2 = iwpt1 + nvar*bytes_float
      iwpt3 = iwpt2 + nvar*bytes_float
      iwpt4 = iwpt3 + nvar*bytes_float
      iwpt5 = iwpt4 + nvar*bytes_float
C
C     Do initial fit
C
      alamda = -1.0
      call mrqmin(xy,2,aintens,sig,ncoord,par,nvar,lista,mfit,covar,
     : alpha,nvar,chisq,ring_func,alamda,dynamic_mem(iwpt1),
     : dynamic_mem(iwpt2),dynamic_mem(iwpt3),dynamic_mem(iwpt4),
     : dynamic_mem(iwpt5),status)
C
C     Now do iterative fit.  Iterate until the change in chi-squared
C     is a trivial amount for the second time...
C
      ochisq = chisq
      olamda = alamda
      itriv = 0
      iter = 0
      continue = .true.
      do while (continue)
         call mrqmin(xy,2,aintens,sig,ncoord,par,nvar,lista,mfit,covar,
     :   alpha,nvar,chisq,ring_func,alamda,dynamic_mem(iwpt1),
     :   dynamic_mem(iwpt2),dynamic_mem(iwpt3),dynamic_mem(iwpt4),
     :   dynamic_mem(iwpt5),status)
         if (status .ne. 0) go to 500
         iter = iter + 1
         ilam = nint(olamda/alamda)
         call tau_converge(iter,ilam,alamda,chisq,itriv,ntriv,trivial,
     :   runaway,continue)
         olamda = alamda
      end do
C
C     Now call the fitting routine one last time to get the errors in the
C     fit parameters...
C
      alamda = 0.0
      call mrqmin(xy,2,aintens,sig,ncoord,par,nvar,lista,mfit,covar,
     : alpha,nvar,chisq,ring_func,alamda,dynamic_mem(iwpt1),
     : dynamic_mem(iwpt2),dynamic_mem(iwpt3),dynamic_mem(iwpt4),
     : dynamic_mem(iwpt5),status)
      rms = sqrt(chisq/sumwts)
      do i = 1,mfit
         if (covar(i,i) .ge. 0.0) then
            erpar(i) = rms*sqrt(covar(i,i))
         else
            erpar(i) = -rms*sqrt(abs(covar(i,i)))
         end if
      end do
C
C     Tidy and exit
C
  500 if (alloc) then
         call dsa_free_workspace(iwslt,dsa_status)
         alloc = .false.
      end if
      end


      subroutine ring_func(xx,nx,par,intens,dydpar,nvar)
C+--------------------------------------------------------------------------------------
C     RING_FUNC   --- Routine to used by the fitting routines to fit the ring
C                     pattern.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) XX:        (real array XX(NX)) X,Y positions
C     (>) NX:        (integer) Dimension of XX
C     (>) Y:         (real) Y position
C     (>) PAR:       (real array PAR(NVAR)) Input ring parameters
C     (<) INTENS:    (real) Model intensity
C     (<) DYDPAR:    (real array DYDPAR(NVAR)) First derivatives wrt each parameter
C     (>) NVAR:      (integer) Number of parameters in the fit
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        21-MAY-1990
C     UPDATE:
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nvar,nx
      real par(nvar),dydpar(nvar),intens,xx(nx)
C
C     Number of divisions in look-up table.  (Based on the idea that
C     exp(-arg) where arg > 25.0 = 0
C
      integer ndiv
      parameter (ndiv=2501)
C
C     Local variables
C
      integer nplus,iarg1,iarg2
      real expon(ndiv+1),lamda,darg,sig,dx,dy,r,rho,arg,
     : exponential,slope,aint,e1,e2,x,y
      logical init
C
C     Common flag for initialisation of look up table
C
      common /flag/ init
C
C     On initial entry, set up the look up table for the Gaussian
C
      if (init) then
         init = .false.
         darg = 25.0/float(ndiv-1)
         call table(darg,ndiv,expon)
         nplus = ndiv + 1
      end if
C
C     Otherwise fit a gaussian with a constant background to the
C     distribution in R.
C
      x = xx(1)
      y = xx(2)
      sig = 1.0/par(5)**2
      dx = x - par(2)
      dy = (y - par(3))*par(7)
      r = sqrt(dx*dx + dy*dy)
      rho = (r - par(4))
      arg = rho*rho*sig*0.5
C
C     Find where on the gaussian look-up table to find your argument
C
      iarg1 = int(arg/darg) + 1
      iarg2 = iarg1 + 1
      iarg1 = min(iarg1,nplus)
      iarg2 = min(iarg2,nplus)
C
C     If the argument > 25.0 then the gaussian is automatically set to 0.0
C
      if ((iarg1 .eq. nplus) .and. (iarg2 .eq. nplus)) then
         exponential = 0.0
         slope = 0.0
C
C     Otherwise interpolate between adjacent values
C
      else
         e1 = expon(iarg1)
         e2 = expon(iarg2)
         slope = (e2 - e1)/darg
         aint = e2 - slope*float(iarg2-1)*darg
         exponential = slope*arg + aint
      end if
C
C     Now calculate the model and its Jacobean.
C
      intens = par(1)*exponential + par(6)
      dydpar(1) = exponential
      lamda = -par(1)*slope*rho*sig
      dydpar(2) = lamda*dx/r
      dydpar(3) = lamda*dy*par(7)/r
      dydpar(4) = lamda
      dydpar(5) = lamda*rho/par(5)
      dydpar(6) = 1.0
      dydpar(7) = -lamda*dy*dy/(r*par(7))
      end
