      subroutine sizes
C+------------------------------------------------------------------------------
C
C   ---------
C   S I Z E S
C   ---------
C
C   Description
C   -----------
C   Utility program to create a map of the pixel sizes in either
C   coordinate as a function of x,y.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - Data type irrelevant
C   - Quality and Error arrays not supported
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   AXIS        If = 1, then calculate the X pixel size.  If = 2
C               then do the Y. (integer)(prompted for).
C
C   NX          Size in the X direction of the output frame.
C               (integer)(prompted for).
C
C   NY          Size in the Y direction of the output frame.
C               (integer)(prompted for).
C
C   PIX_MAP     Name of the output structure. (character)(prompted for).
C
C   WHICH_MASK  Users choice of mask (integer)(prompted for)
C
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   .MASK.MASKn NORD        Order of the fit from MASK (integer)(read)
C
C   .MASK       NMASK       Number of masks on offer in TAUCAL. (integer)(read)
C
C   .MASK.MASKn COEFX[NUNK] Array of X coefficients from mask.  nunk =
C                           ((nord+1)*(nord+2))/2.  (real)(read)
C
C   .MASK.MASKn COEFY[NUNK] Array of Y coefficients from mask.  nunk =
C                           ((nord+1)*(nord+2))/2.  (real)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The fit coefficients are from MASK are read in
C   - The total derivative is performed to find the
C     pixel size.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_FREE_LU
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_MAP_DATA
C     DSA_OPEN
C     DSA_OUTPUT
C     DSA_SIMPLE_OUTPUT
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library PAR:
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDNUM
C     TAU_DERIV
C
C   Internal subroutines called
C   ---------------------------
C   SREAD_COEFS
C   SUB_SIZES
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   29-MAR-90   - Original program
C   03-JUL-90   - TAUCAL system pasted in. (JRL)
C   26-JUL-90   - Modified so that it may still run if there is no TAUCAL
C                 file being used.  (JRL)
C   15-NOV-90   - Now reads info from MASK sub-structure in TAUCAL. (JRL)
C   19-NOV-90   - New TAUCAL routines pasted in. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
C
C     Local variables
C
      integer     address    ! Dynamic memory address
      integer     axis       ! Flag for axis for which pixel sizes is determined
      integer     dims(2)    ! Dimensions of output file
      real        dummy      ! Dummy real variable
      logical     exist      ! TRUE if a file exists
      integer     i          ! Dummy counter
      character   label*80   ! MASK sub-structure label
      integer     length     ! Length of MASK sub-structure label
      integer     nord       ! Order of fit from MASK
      integer     nunk       ! Number of unknowns in fit from MASK
      integer     optr       ! Dynamic memory pointer
      integer     slot       ! Dynamic memory slot
      integer     status     ! DSA and TAUCAL status variable
      logical     tauexist   ! Flag for existence of TAUCAL file
      character   testfile*11! Name of TMASK output file
      integer     totspace   ! Space in bytes allocated for output file
      integer     unit       ! Logical unit for MASK.OUT if it's needed.
      integer     wptr       ! Dynamic memory pointer
      integer     wptr2      ! Dynamic memory pointer
      integer     wslot      ! Dynamic memory slot
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Common block for dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Get started
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file.
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find out how many masks are on offer and which one you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,label,
     : length,status)
      if (status .ne. 0) go to 500
C
C     Get the order of fit from TAUCAL.
C
      call tau_cal_rdnum('cal',0,label(:length)//'.nord','norder',
     : tauexist,.true.,1.0,float(max_int),' ',1,dummy,status)
      if (status .ne. 0) go to 500
      nord = nint(dummy)
C
C     Which axis to create grid? Also how big a map?
C
      call par_rdval('axis',1.0,2.0,1.0,' ',dummy)
      axis = nint(dummy)
      call par_rdval('nx',0.0,float(max_int),256.0,' ',dummy)
      dims(1) = nint(dummy)
      call par_rdval('ny',0.0,float(max_int),256.0,' ',dummy)
      dims(2) = nint(dummy)
C
C     Get some workspace
C
      nunk = ((nord + 2)*(nord + 1))/2
      totspace = 2*nunk*dsa_typesize('float',status)
      call dsa_get_workspace(totspace,address,wslot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace...')
         call dsa_wrflush
         go to 500
      end if
      wptr = dyn_element(address)
      wptr2 = wptr + nunk*dsa_typesize('float',status)
C
C     Read the coefs into the workspace from TAUCAL if it exists
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,label(:length)//'.coefx',
     :   'coefx',tauexist,.false.,min_float,max_float,' ',nunk,
     :   dynamic_mem(wptr),status)
         call tau_cal_rdnum('cal',0,label(:length)//'.coefy',
     :   'coefy',tauexist,.false.,min_float,max_float,' ',nunk,
     :   dynamic_mem(wptr2),status)
C
C     If no TAUCAL is being used, then try to find the coefs from the most
C     recent version of MASK.OUT.
C
      else
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('You don''t have a TAUCAL file.  I will')
         call dsa_wrflush
         call dsa_wruser('attempt to find the coefs in the most')
         call dsa_wrflush
         call dsa_wruser('recent mask.out file.')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_get_lu(unit,status)
         i = 99
         exist = .false.
         do while ((i .gt. 0) .and. (.not. exist))
            write(testfile,100) i
  100       format('mask.out:',i2.2)
            inquire(file=testfile,exist=exist)
            if (.not. exist) i = i - 1
         end do
         if (i .ne. 0) then
            write(testfile,100) i
            open(unit,file=testfile,status='old',iostat=status)
         else
            status = 1
         end if
         if (status .ne. 0) then
            call dsa_wruser(' ')
            call dsa_wrflush
            call dsa_wruser('Failed to open mask.out file -- ')
            call dsa_wruser('Cannot procede any further')
            call dsa_wrflush
            go to 500
         end if
         call sread_coefs(unit,nunk,dynamic_mem(wptr),
     :   dynamic_mem(wptr2),status)
         call dsa_free_lu(unit,status)
      end if
      if (status .ne. 0) go to 500
C
C     Create output file
C
      call dsa_output('output','pix_map',' ',1,1,status)
      call dsa_simple_output('output','data','float',2,dims,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t create output file')
         call dsa_wrflush
         go to 500
      end if
C
C     Map the output array.
C
      call dsa_map_data('output','write','float',address,slot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t map output data')
         call dsa_wrflush
         go to 500
      end if
      optr = dyn_element(address)
C
C     Now create the map
C
      call sub_sizes(nord,axis,dims(1),dims(2),dynamic_mem(wptr),
     : dynamic_mem(optr))
C
C     Tidy and exit...
C
  500 call dsa_close(status)
      end



      subroutine sub_sizes(nord,axis,nx,ny,coefs,data)
C+-------------------------------------------------------------------------
C     SUB_SIZES   --  Do the actual calculation of pixel sizes.
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) NORD:      (Integer) Order of the fit
C     (>) AXIS:      (Integer) = 1 for X pixel size, = 2 for Y
C     (>) NX:        (Integer) X dimension of output map
C     (>) NY:        (Integer) Y dimension of output map
C     (>) COEFS:     (Real array COEFS(((NORD+1)*(NORD+2))/2,2) Fit coefs
C     (<) DATA:      (Real array DATA(NX,NY)) Output map
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-MAR-1990
C     UPDATE:
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nord,axis,nx,ny
      real data(nx,ny),coefs(((nord+1)*(nord+2))/2,2)
C
C     Local variables
C
      integer nunk,i,j
      real y,x,dxx,dxy,dyx,dyy,pixsize
C
C     Calculate number of unknowns
C
      nunk = ((nord+1)*(nord+2))/2
C
C     Calculate the sizes now...
C
      do j = 1,ny
         y = float(j)
         do i = 1,nx
            x = float(i)
            call tau_deriv(nord,nunk,coefs(1,1),x,y,dxx,dxy)
            call tau_deriv(nord,nunk,coefs(1,2),x,y,dyx,dyy)
            if (axis .eq. 1) then
               pixsize = sqrt(dxx*dxx + dyx*dyx)
            else
               pixsize = sqrt(dyy*dyy + dxy*dxy)
            end if
            data(i,j) = pixsize
         end do
      end do
      end


      subroutine sread_coefs(lu,nunk,coefx,coefy,status)
C+-------------------------------------------------------------------------
C     SREAD_COEFS  --  Read the coefs if TAUCAL doesn't exist
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) LU:       (integer) Logical unit of input file
C     (>) NUNK:     (integer) Number of unknowns
C     (>) COEFX:    (real array COEFX(NUNK)) Fit X coefs
C     (>) COEFY:    (real array COEFY(NUNK)) Fit Y coefs
C     (!) STATUS:   (integer) Status flag
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        26-JUL-1990
C     UPDATE:
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer lu,nunk,status
      real coefx(nunk),coefy(nunk)
C
C     Local variables
C
      integer i
C
C     Inherited status...
C
      if (status .ne. 0) go to 500
C
C     Skip a few lines first
C
      do i = 1,3
         read(lu,*)
      end do
C
C     Now try to read X coefs
C
      read(lu,*,iostat=status) (coefx(i),i=1,nunk)
      if (status .ne. 0) then
         call dsa_wruser('Error reading X coefs')
         call dsa_wrflush
         go to 500
      end if
C
C     Now try to read Y coefs
C
      read(lu,*,iostat=status) (coefy(i),i=1,nunk)
      if (status .ne. 0) then
         call dsa_wruser('Error reading Y coefs')
         call dsa_wrflush
      end if
C
C     Exit
C
  500 end
