      subroutine tmask
C+------------------------------------------------------------------------------
C   ---------
C   T M A S K
C   ---------
C
C   Description
C   -----------
C   Takes the peaks found in FINDPKS, fits a bivariate polynomial to find
C   the transformation between pixel postions and position within the grid,
C   then takes the first derivative of the polynomial to find the scale.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-D images only.
C   - Data array converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE     Name of TAUCAL file (character)(prompted for)
C
C   WHICH_MASK  Users choice of mask (integer)(prompted for)
C
C   NORDER      The order of the polynomial to fit. (integer)(prompted for).
C
C
C   Keywords
C   --------
C   USE_TAU     Flag for use of TAUCAL file. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (MASK)      NAME_MAT    Name of image containing matrix mask.
C                           (character)(read)
C
C   .MASK.MASKn MASKRUN     Run number associated with matrix mask frame.
C                           (integer)(read)
C
C   .MASK.MASKn NPOS        Total number of peaks found in mask frame. (integer)
C                           (read)
C
C   .MASK.MASKn POSX[NPOS]  X positions of centroids of the peaks. (real)(read)
C
C   .MASK.MASKn POSY[NPOS]  Y positions of centroids of the peaks. (real)(read)
C
C   .MASK       MASK_DIST   Physical separation in microns between the mask
C                           pinholes (real)(read)
C
C   .CAMERA     CAMFOC      Focal length of the camera in millimetres.
C                           (real)(read)
C
C   .COLLIMATOR COLFOC      Focal length of the collimator in millimetres.
C                           (real)(read)
C
C   .MASK.MASKn NORD        Order of the the polynomial fit.  (integer)(written)
C
C   .MASK.MASKn COEFX[NUNK] Polynomial fit coefficients for X -| NUNK =
C                           solution (real)(written)           | (NORD+1)*
C                                                              | (NORD+2)/2
C   .MASK.MASKn COEFY[NUNK] Polynomial fit coefficients for Y  |
C                           solution (real)(written)          -|
C
C   .MASK.MASKn PHI         Angle of rotation of grid. (real)(written)
C
C   .MASK.MASKn SIGX        Mean error in X. (real)(written)
C
C   .MASK.MASKn SIGY        Mean error in Y. (real)(written)
C
C
C   Propagation of data structure
C   -----------------------------
C   Not relevant.
C
C
C   Method
C   ------
C   - The mask image is opened to get its dimensions.
C   - Peaks are read from TAUCAL file.
C   - A rough rotation for the mask with respect to the X,Y pixel
C     coordinate system is found.  The peaks are assigned coordinates
C     in microns in that rotated grid.
C   - Polynomials are fit for a transformation between pixel coordinates
C     and the grid coordinates in microns.
C   - The first derivative of the transformation is then performed which
C     gives the scale (e.g. the number of microns per pixel).  This is
C     calculated at a grid of 100 points covering the mask frame.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_DATA_SIZE
C     DSA_FREE_LU
C     DSA_FREE_WORKSPACE
C     DSA_GET_LU
C     DSA_GET_WORKSPACE
C     DSA_NAMED_INPUT
C     DSA_OPEN
C     DSA_TYPESIZE
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library ICH:
C     ICH_LEN
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_CHECK_RUN
C     TAU_CAL_FIND_STRUCT
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C     TAU_CAL_RDNUM
C     TAU_CAL_SETNUM
C     TAU_DERIV
C     TAU_LSQ
C
C   Internal subroutines called
C   ---------------------------
C     ASSIGN_DIST
C     SET_EQ
C     SUB_MASK
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYNAMIC_MEMORY'
C   INCLUDE 'NDP_NUMERIC_RANGES'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters
C
C
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   23-JAN-90   - Original program
C   29-MAR-90   - Major overhaul, including creation of subroutine SUB_MASK
C                 and a number of cosmetic changes.  (JRL)
C   03-JUL-90   - TAUCAL pasted in.  (JRL)
C   06-JUL-90   - Cosmetic changes made as a result of introduction of TAUCAL.
C                 (JRL)
C   26-JUL-90   - Altered so that if TAUCAL isn't being used, it is still
C                 possible to read in the peak positions from a file. (JRL)
C   01-AUG-90   - Changed so that user is now informed of the maximum order
C                 of the fit which is allowed by the number of peaks. (JRL)
C   08-NOV-90   - Changed so that a few more things get written out to TAUCAL.
C                 (JRL)
C   15-NOV-90   - Changed to write results to .MASK structure (JRL)
C   19-NOV-90   - New versions of TAUCAL routines pasted in.  (JRL)
C   07-MAY-91   - Changed name from MASK to TMASK to avoid confusion
C                 with FIGARO program of the same name. (JRL)
C   31-JAN-94   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Maximum number of peaks allowed
C
      integer npmax
      parameter (npmax=4096)
C
C     Functions
C
      integer dsa_typesize
      integer dyn_element
      integer ich_len
C
C     Local variables
C
      integer   address                  ! Dynamic memory address
      real      arg                      ! Dummy real variable
      logical   badpix                   ! Flag for magic values
      integer   bytes_float              ! Number of bytes in float type
      real      camfoc                   ! Camera focal length
      real      colfoc                   ! Collimator focal length
      integer   dims(2)                  ! Array with dimensions of image
      real      dist                     ! Projected distance between adjacent peaks (microns)
      real      dummy                    ! Dummy variable
      logical   exist                    ! TRUE if a file exists
      integer   i                        ! Dummy counter
      character label*80                 ! Label for MASK sub-tructure
      integer   length                   ! Length of label for MASK sub-structure
      integer   maskrun                  ! Run number of MATRIX image.
      character message*80               ! Character variable for user messages
      character name_mat*80              ! Name of MATRIX image
      integer   ndim                     ! Number of dimensions in image
      real      meanerrx                 ! Mean error in X
      real      meanerry                 ! Mean error in Y
      integer   nelm                     ! Total number of elements in image
      integer   nord                     ! Order of the fit
      integer   nordmax                  ! Maximum order allowed by number of peaks
      integer   npos                     ! Number of peaks
      integer   nunk                     ! Number of unknowns given order
      integer   nx                       ! X dimension of image
      integer   ny                       ! Y dimension of image
      real      ord                      ! Nord in real
      real      phi_deg                  ! Angle of rotation of grid
      character testfile*14              ! Name of FINDPKS output file
      integer   slot                     ! Dynamic memory slot
      integer   status                   ! DSA status flag
      logical   tauexist                 ! Flag for existence of TAUCAL file
      integer   totspace                 ! Total space for allocated workspace
      character type*8                   ! Data type of matrix image
      integer   unit                     ! Logical unit for FINDPKS.OUT if needed
      integer   wpt1                     ! Dynamic memory pointer
      integer   wpt2                     ! Dynamic memory pointer
      real      x(npmax)                 ! X position (pixels) of peaks
      real      y(npmax)                 ! Y position (pixels) of peaks
C
C     Common block for dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Get DSA_ started
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Find out how many masks are on offer and which one you want
C
      call tau_cal_find_struct('cal','mask','mask',tauexist,label,
     : length,status)
      if (status .ne. 0) go to 500
C
C     Get run number of MATRIX image and check to it's a MASK run
C
      call tau_cal_rdnum('cal',0,label(:length)//'.run',
     : 'maskrun',tauexist,.false.,1.0,float(max_int),' ',1,dummy,
     : status)
      maskrun = nint(dummy)
      call tau_cal_check_run('cal',maskrun,'mask',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get the image name
C
      call tau_cal_rdchr('cal',maskrun,'image','name_mat',tauexist,
     : .true.,80,name_mat,status)
      if (status .ne. 0) go to 500
C
C     Get mask image and its size
C
      call dsa_named_input('image',name_mat,status)
      call ndp_get_image_info('image',.true.,.false.,type,badpix,status)
      call dsa_data_size('image',2,ndim,dims,nelm,status)
      if (ndim .ne. 2) then
         call dsa_wruser('Data is not 2-dimensional')
         call dsa_wrflush
         go to 500
      end if
      nx = dims(1)
      ny = dims(2)
C
C     Read peaks from TAUCAL file if it exists.
C
      if (tauexist) then
         call tau_cal_rdnum('cal',0,label(:length)//'.npos','npos',
     :   tauexist,.false.,0.0,float(npmax),' ',1,dummy,status)
         npos = nint(dummy)
         call tau_cal_rdnum('cal',0,label(:length)//'.posx','posx',
     :   tauexist,.false.,min_float,max_float,'pixels',npos,x,status)
         call tau_cal_rdnum('cal',0,label(:length)//'.posy','posy',
     :   tauexist,.false.,min_float,max_float,'pixels',npos,y,status)
         if (status .ne. 0) go to 500
         write(message,101) npos
  101    format(i8,' peaks read from file ')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
C
C     Otherwise, try to find them in FINDPKS.OUT
C
      else
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser('You are not using a TAUCAL file, so I shall ')
         call dsa_wruser('attempt to find the peak positions in the ')
         call dsa_wruser('most recent findpks.out file')
         call dsa_wrflush
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_get_lu(unit,status)
         i = 99
         exist = .false.
         do while ((i .gt. 0) .and. (.not. exist))
            write(testfile,100) i
  100       format('findpks.out:',i2.2)
            inquire(file=testfile,exist=exist)
            if (.not. exist) i = i - 1
         end do
         if (i .ne. 0) then
            write(testfile,100) i
            open(unit,file=testfile,status='old',iostat=status)
         else
            status = 1
         end if
         if (status .ne. 0) then
            call dsa_wruser('Can''t open findpks output file.')
            call dsa_wruser('I can procede no further')
            call dsa_wrflush
            call dsa_wruser(' ')
            call dsa_wrflush
            go to 500
         end if
C
C        If the file has been found, then try to read it...
C
         npos = 0
         do while ((status .eq. 0) .and. (npos .lt. npmax))
            npos = npos + 1
            read(unit,*,iostat=status) x(npos),y(npos)
         end do
C
C        If iostat < 0 then we have reached the end of the file
C        with no error...if iostat > 0 then we have an error in
C        the reading of the file
C
         if (status .lt. 0) then
            status = 0
         else if (status .gt. 0) then
            call dsa_wruser('An error occured in reading ')
            call dsa_wruser(testfile)
            call dsa_wrflush
            go to 500
         end if
C
C        Otherwise everything's cool, and procede as normal.
C
         npos = npos - 1
         write(message,101) npos
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_free_lu(unit,status)
      end if
C
C     Get mask pinhole distance, camera focal length and colliamtor
C     focal length from TAUCAL.  Project the pinhole distance
C
      call tau_cal_rdnum('cal',0,'mask.separation','mask_dist',
     : tauexist,.true.,min_float,max_float,'microns',1,dist,status)
      call tau_cal_rdnum('cal',0,'camera.foclen','camfoc',
     : tauexist,.true.,min_float,max_float,'mm',1,camfoc,status)
      call tau_cal_rdnum('cal',0,'collimator.foclen','colfoc',
     : tauexist,.true.,min_float,max_float,'mm',1,colfoc,status)
      if (status .ne. 0) go to 500
      dist = dist*camfoc/colfoc
C
C     Calculate from the number of peaks the maximum order for the fit
C
      arg = -1.5 + sqrt(2.0*float(npos) + 0.25)
      nordmax = int(arg+0.00001)
      write(message,102) nordmax
  102 format('Maximum order of fit allowed is:',i4,'')
      call dsa_wruser(' ')
      call dsa_wrflush
      call dsa_wruser(message(:ich_len(message)))
      call dsa_wrflush
C
C     Get order of fit
C
      call par_rdval('norder',0.0,float(nordmax),1.0,' ',ord)
      nord = nint(ord)
C
C     Get workspace for coefficients
C
      nunk = ((nord+1)*(nord+2))/2
      bytes_float = dsa_typesize('float',status)
      totspace = 4*nunk*bytes_float
      call dsa_get_workspace(totspace,address,slot,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace')
         call dsa_wrflush
         go to 500
      end if
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + 2*nunk*bytes_float
C
C     Now do the work
C
      call sub_mask(npos,nord,nx,ny,dist,x,y,dynamic_mem(wpt1),
     : dynamic_mem(wpt2),phi_deg,meanerrx,meanerry)
C
C     Update TAUCAL file
C
      call tau_cal_setnum('cal',0,label(:length)//'.nord',
     : tauexist,1,1,1,float(nord),status)
      call tau_cal_setnum('cal',0,label(:length)//'.coefx',
     : tauexist,1,nunk,nunk,dynamic_mem(wpt1),status)
      call tau_cal_setnum('cal',0,label(:length)//'.coefy',
     : tauexist,1,nunk,nunk,dynamic_mem(wpt1+nunk*bytes_float),status)
      call tau_cal_setnum('cal',0,label(:length)//'.phi',
     : tauexist,1,1,1,phi_deg,status)
      call tau_cal_setnum('cal',0,label(:length)//'.sigx',
     : tauexist,1,1,1,meanerrx,status)
      call tau_cal_setnum('cal',0,label(:length)//'.sigy',
     : tauexist,1,1,1,meanerry,status)
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end


      subroutine sub_mask(npos,nord,nx,ny,dist,x,y,coefs,ercoef,
     : phi_deg,meanerrx,meanerry)
C+-------------------------------------------------------------------------
C     SUB_MASK    -- Do the actual work in finding the pixel sizes
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) NPOS:     (integer) Number of data points
C     (>) NORD:     (integer) Order of the fit
C     (>) NX:       (integer) X size of MATRIX image
C     (>) NY:       (integer) Y size of MATRIX image
C     (>) DIST:     (real) Physical distance in microns between the peaks.
C     (>) X:        (real) X pixel position of the peaks
C     (>) Y:        (real) Y pixel position of the peaks
C     (<) COEFS:    (real array COEFS(((NORD+2)*(NORD+1))/2,2) Fit coefs.
C     (<) ERCOEF:   (real array ERCOEF(((NORD+2)*(NORD+1))/2,2) Error in coefs.
C     (<) PHI_DEG:  (real) Angle of rotation of grid in degrees
C     (<) MEANERRX: (real) Mean error in X
C     (<) MEANERRY: (real) Mean error in Y
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        29-MAR-1990
C     UPDATE:      08-NOV-1990 -- Modified to pass a few more things back
C                                 to calling routine (JRL)
C
C+-------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer npos,nord,nx,ny
      real dist,x(npos),y(npos),coefs(((nord+1)*(nord+2))/2,2),
     : ercoef(((nord+1)*(nord+2))/2,2),phi_deg,meanerrx,meanerry
C
C     Functions
C
      integer dsa_typesize,dyn_element,ich_len
C
C     Local variables
C
      integer nunk,nplus,bytes_float,status,totspace,address,slot1,
     : wpt1,wpt2,slot2,wpt3,wpt4,wpt5,wpt6,bytes_double,slot3,
     : wpt7,wpt8,wpt9,wpt10,lsq_status,ng,i,j,ist,ifn,unit
      real xgrid(100),ygrid(100),pix_x_size(100),
     : pix_y_size(100),dxx,dxy,dyx,dyy,dx,dy,yy,xx,ellip(100)
      character message*80,testfile*11
      logical exist
C
C     Common block for dynamic memory
C
      include 'DYNAMIC_MEMORY'
C
C     Set up a few variables
C
      nunk = ((nord+1)*(nord+2))/2
      nplus = nunk + 1
C
C     Get some workspace
C
      bytes_float = dsa_typesize('float',status)
      totspace = 2*npos*bytes_float
      call dsa_get_workspace(totspace,address,slot1,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace.')
         call dsa_wrflush
         go to 500
      end if
      wpt1 = dyn_element(address)
      wpt2 = wpt1 + npos*bytes_float
C
C     Assign a grid distance in microns to each peak
C
      call assign_dist(npos,x,y,dist,dynamic_mem(wpt1),
     : dynamic_mem(wpt2),phi_deg)
C
C     Get some more work space
C
      totspace = (2*npos*nplus + 2*npos)*bytes_float
      call dsa_get_workspace(totspace,address,slot2,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace.')
         call dsa_wrflush
         go to 500
      end if
      wpt3 = dyn_element(address)
      wpt4 = wpt3 + npos*nplus*bytes_float
      wpt5 = wpt4 + npos*bytes_float
      wpt6 = wpt5 + npos*nplus*bytes_float
C
C     Now set up the condition equations...
C
      call set_eq(nord,npos,nplus,x,y,dynamic_mem(wpt1),
     : dynamic_mem(wpt2),dynamic_mem(wpt3),dynamic_mem(wpt4),
     : dynamic_mem(wpt5),dynamic_mem(wpt6))
C
C     Free some of the workspace...
C
      call dsa_free_workspace(slot1,status)
C
C     Get a bit more workspace...
C
      bytes_double = dsa_typesize('double',status)
      totspace = nunk*(nplus + nunk + (nplus + 1))*bytes_double +
     : npos*bytes_float
      call dsa_get_workspace(totspace,address,slot3,status)
      if (status .ne. 0) then
         call dsa_wruser(' Couldn''t get workspace.')
         call dsa_wrflush
         go to 500
      end if
      wpt7 = dyn_element(address)
      wpt8 = wpt7 + nunk*nplus*bytes_double
      wpt9 = wpt8 + nunk*nunk*bytes_double
      wpt10 = wpt9 + nunk*(nplus+1)*bytes_double
C
C     Now do the fit for X
C
      call tau_lsq(dynamic_mem(wpt3),dynamic_mem(wpt4),npos,nplus,npos,
     : nunk,dynamic_mem(wpt7),dynamic_mem(wpt8),dynamic_mem(wpt9),
     : coefs(1,1),ercoef(1,1),dynamic_mem(wpt10),meanerrx,lsq_status)
      if (lsq_status .ne. 0) then
         call dsa_wruser('X solution failed...')
         call dsa_wrflush
         go to 500
      else
         write(message,101) meanerrx
  101    format('X solution fit with mean error:',f8.3,' microns')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
      end if
C
C     Do the same for Y
C
      call tau_lsq(dynamic_mem(wpt5),dynamic_mem(wpt6),npos,nplus,npos,
     : nunk,dynamic_mem(wpt7),dynamic_mem(wpt8),dynamic_mem(wpt9),
     : coefs(1,2),ercoef(1,2),dynamic_mem(wpt10),meanerry,lsq_status)
      if (lsq_status .ne. 0) then
         call dsa_wruser('Y solution failed...')
         call dsa_wrflush
         go to 500
      else
         write(message,102) meanerry
  102    format('Y solution fit with mean error:',f8.3,' microns')
         call dsa_wruser(' ')
         call dsa_wrflush
         call dsa_wruser(message(:ich_len(message)))
         call dsa_wrflush
      end if
C
C     Toss the remaining workspace...
C
      call dsa_free_workspace(slot3,status)
      call dsa_free_workspace(slot2,status)
C
C     Find an output file name...
C
      i = 99
      exist = .false.
      do while ((i .gt. 0) .and. (.not. exist))
         write(testfile,100) i
  100    format('mask.out:',i2.2)
         inquire(file=testfile,exist=exist)
         if (.not. exist) i = i - 1
      end do
      i = i + 1
      write(testfile,100) i
C
C     Open output file
C
      call dsa_get_lu(unit,status)
      open(unit=unit,file=testfile,status='new')
C
C     Create a grid across the image at which to evaluate the derivatives
C
      dx = float(nx-1)/9.0
      dy = float(ny-1)/9.0
      ng = 0
      do i = 1,10
         yy = float(i-1)*dy + 1
         do j = 1,10
            xx = float(j-1)*dx + 1
            ng = ng + 1
            xgrid(ng) = xx
            ygrid(ng) = yy
         end do
      end do
C
C     Calculate derivative and the corresponding pixel sizes for each fit
C     at each grid point.
C
      do i = 1,100
         call tau_deriv(nord,nunk,coefs(1,1),xgrid(i),ygrid(i),dxx,dxy)
         call tau_deriv(nord,nunk,coefs(1,2),xgrid(i),ygrid(i),dyx,dyy)
         pix_x_size(i) = sqrt(dxx*dxx + dyx*dyx)
         pix_y_size(i) = sqrt(dyy*dyy + dxy*dxy)
      end do
C
C     Calculate the pixel ellipticity
C
      do i = 1,100
         ellip(i) = pix_y_size(i)/pix_x_size(i)
      end do
C
C     Write out order and fit coefs
C
      write(unit,*) ' Order of fit'
      write(unit,*) nord
      write(unit,*) ' Fit coefs'
      write(unit,*) (coefs(i,1),i=1,nunk)
      write(unit,*) (coefs(i,2),i=1,nunk)
C
C     Write out x pixel_size results
C
      write(unit,*) '                    x pixel results'
      write(unit,105)
  105 format(30x,'x')
      write(unit,110) ((dx*float(j-1)+1),j=1,10)
  110 format(13x,10f6.1)
      write(unit,115)
  115 format('-------------------------------------------',
     : '------------------------------')
      do j = 1,10
         ist = 10*(j-1) + 1
         ifn = ist + 9
         if (j .eq. 5) then
            write(unit,120) (dy*float(j-1)+1),(pix_x_size(i),i=ist,ifn)
         else
            write(unit,125) (dy*float(j-1)+1),(pix_x_size(i),i=ist,ifn)
         end if
  120 format(2x,'y',2x,f6.1,1x,'|',10f6.1)
  125 format(5x,f6.1,1x,'|',10f6.1)
      end do
C
C     Write out y pixel_size results
C
      write(unit,*) ' '
      write(unit,*) ' '
      write(unit,*) '                    y pixel results'
      write(unit,105)
      write(unit,110) ((dx*float(j-1)+1),j=1,10)
      write(unit,115)
      do j = 1,10
         ist = 10*(j-1) + 1
         ifn = ist + 9
         if (j .eq. 5) then
            write(unit,120) (dy*float(j-1)+1),(pix_y_size(i),i=ist,ifn)
         else
            write(unit,125) (dy*float(j-1)+1),(pix_y_size(i),i=ist,ifn)
         end if
      end do
C
C     Write out ellipticity results
C
      write(unit,*) ' '
      write(unit,*) ' '
      write(unit,*) '                    ellipticity results'
      write(unit,105)
      write(unit,110) ((dx*float(j-1)+1),j=1,10)
      write(unit,115)
      do j = 1,10
         ist = 10*(j-1) + 1
         ifn = ist + 9
         if (j .eq. 5) then
            write(unit,130) (dy*float(j-1)+1),(ellip(i),i=ist,ifn)
         else
            write(unit,135) (dy*float(j-1)+1),(ellip(i),i=ist,ifn)
         end if
      end do
  130 format(2x,'y',2x,f6.1,1x,'|',10f6.3)
  135 format(5x,f6.1,1x,'|',10f6.3)
      close(unit)
      call dsa_free_lu(unit,status)
  500 end


      subroutine assign_dist(npos,x,y,dist,xg,yg,phi_deg)
C+--------------------------------------------------------------------------
C     ASSIGN_DIST -- Assign each peak a grid position in microns
C
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) NPOS      (Integer) Number of peaks
C     (>) X         (Real array X(NPOS)) X pixel positions of peaks
C     (>) Y         (Real array Y(NPOS)) Y pixel positions of peaks
C     (>) DIST      (Real) Distance in microns between adjacent mask grid
C                          points at the focal plane.
C     (<) XG        (Real array XG(NPOS)) X grid positions (microns) of peaks
C     (<) YG        (Real array YG(NPOS)) Y grid positions (microns) of peaks
C     (<) PHI_DEG   (real) Angle of rotation of the grid in degrees
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-JAN-1990
C     UPDATE:      08-NOV-1990 -- Modified to pass phi back to calling routine (JRL).
C
C
C+--------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer npos
      real x(npos),y(npos),xg(npos),yg(npos),dist,phi_deg
C
C     Functions
C
      integer ich_len
C
C     Local variables
C
      integer irem,i,jrem,ifound
      real sumx,sumy,avex,avey,dx,dy,r,rmin,phi,radeg,xx,yy,
     : theta,xc,yc,xdis,ydis,sumphi
      character message*80
C
C     Find peak closest to the middle of the frame.
C
      sumx = 0.0
      sumy = 0.0
      do i = 1,npos
         sumx = sumx + x(i)
         sumy = sumy + y(i)
      end do
      avex = sumx/float(npos)
      avey = sumy/float(npos)
      rmin = 1.0e10
      do i = 1,npos
         dx = x(i) - avex
         dy = y(i) - avey
         r = sqrt(dx*dx + dy*dy)
         if (r .lt. rmin) then
            irem = i
            rmin = r
         end if
      end do
C
C     Find peak next to the right of the central one in the same grid row.
C
      rmin = 1.0e10
      do i = 1,npos
         if (i .ne. irem) then
            dx = x(i) - x(irem)
            dy = y(i) - y(irem)
            if ((dx .gt. 0.0) .and. (abs(dx) .gt. abs(dy))) then
               r = sqrt(dx*dx + dy*dy)
               if (r .lt. rmin) then
                  jrem = i
                  rmin = r
               end if
            end if
         end if
      end do
C
C     Calculate rough rotation from these two peaks
C
      dx = x(jrem) - x(irem)
      dy = y(jrem) - y(irem)
      phi = atan(dy/dx)
C
C     Now look at all four points adjacent to the middle one and
C     calculate a better value of the rotation.  Phi is always defined
C     with respect to the pixel X axis and is always -45 <= PHI <= 45.
C
      ifound = 0
      sumphi = 0.0
      do i = 1,npos
         dx = abs(x(i) - x(irem))
         dy = abs(y(i) - y(irem))
         r = sqrt(dx*dx + dy*dy)
         if ((r .le. rmin+2.0) .and. (r .ge. rmin-2.0)) then
            if (dx .gt. dy) then
               sumphi = sumphi + atan(dy/dx)
            else
               sumphi = sumphi + atan(dx/dy)
            end if
            ifound = ifound + 1
         end if
      end do
C
C     Average the measurements of PHI and let the user know...
C
      sumphi = sumphi/float(ifound)
      phi = sign(sumphi,phi)
      radeg = 45.0/atan(1.0)
      phi_deg = phi*radeg
      call dsa_wruser(' ')
      call dsa_wrflush
      write(message,100) ifound,phi_deg
  100 format('From',i3,' points, rough rotation of mask is',f8.3,
     : ' degrees')
      call dsa_wruser(message(:ich_len(message)))
C
C     calculate grid position for each of the peaks now...
C
      xc = x(irem)
      yc = y(irem)
      do i = 1,npos
         if (i .eq. irem) then
            xg(i) = 0.0
            yg(i) = 0.0
         else
            xx = x(i)
            yy = y(i)
            dx = xx - xc
            dy = yy - yc
            theta = atan2(dy,dx)
            r = sqrt(dx*dx + dy*dy)
            xdis = r*cos(theta-phi)
            ydis = r*sin(theta-phi)
            xg(i) = float(nint(xdis/rmin))*dist
            yg(i) = float(nint(ydis/rmin))*dist
         end if
      end do
      end


      subroutine set_eq(nord,npos,nplus,x,y,xg,yg,condx,
     : weightx,condy,weighty)
C+-------------------------------------------------------------------------
C     SET_EQ      -- Set up the condition equations for the fits.
C
C
C     Parameters -  (">" input, "<" output, "W" Workspace, "!" Changed)
C
C     (>) NORD      (Integer) Order of the fit
C     (>) NPOS      (Integer) Number of peaks
C     (>) NPLUS     (Integer) Number of unknowns + 1
C     (>) X         (Real array X(NPOS)) Pixel X coordinates
C     (>) Y         (Real array Y(NPOS)) Pixel Y coordinates
C     (>) XG        (Real array XG(NPOS)) Grid X coordinates
C     (>) YG        (Real array YG(NPOS)) Grid Y coordinates
C     (<) CONDX     (Real array CONDX(NPOS,NPLUS)) X condition equations
C                   including the left-hand side.
C     (<) WEIGHTX   (Real array WEIGHTX(NPOS)) Weights for X solution.
C     (<) CONDY     (Real array CONDY(NPOS,NPLUS)) Y condition equations
C                   including the left-hand side.
C     (<) WEIGHTY   (Real array WEIGHTY(NPOS)) Weights for Y solution.
C
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        23-JAN-1990
C     UPDATE:
C
C+----------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nord,npos,nplus
      real x(npos),y(npos),xg(npos),yg(npos),condx(npos,nplus),
     : condy(npos,nplus),weightx(npos),weighty(npos)
C
C     Local variables
C
      integer j,k,iterm,i,nt
      real xx,yy
C
C     set up condition equations for x and y together
C
      do i = 1,npos
         weightx(i) = 1.0
         weighty(i) = 1.0
         iterm = 0
         xx = x(i)
         yy = y(i)
         do j = 0,nord
            nt = j + 1
            do k = 1,nt
               iterm = iterm + 1
               condx(i,iterm) = (xx**(nt-k))*(yy**(k-1))
               condy(i,iterm) = condx(i,iterm)
            end do
         end do
         condx(i,iterm+1) = -xg(i)
         condy(i,iterm+1) = -yg(i)
      end do
      end
