      subroutine tau_cal_rdnum(ref_name,runnum,object,par_name,taucal,
     : ask_par,vmin,vmax,units,nitems,value,status)
C+-----------------------------------------------------------------------------
C
C   -------------------------
C   T A U _ C A L _ R D N U M
C   -------------------------
C
C   Description
C   -----------
C   Reads a real variable from TAUCAL structure.
C
C
C   Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C   (>) REF_NAME:  (character) Reference name for TAUCAL file.
C   (>) RUNNUM:    (integer) Run number of the substructure to be searched
C   (>) OBJECT:    (character) The element name of the variable to be read
C   (>) PAR_NAME:  (character) Parameter name of the variable to be read
C                              (in case there is no TAUCAL)
C   (>) TAUCAL:    (logical) TRUE if TAUCAL file exist.
C   (>) ASK_PAR:   (logical) TRUE if you want to be prompted for the value if
C                            TAUCAL file doesn't exist.
C   (>) VMIN:      (real) Minimum value for the object
C   (>) VMAX:      (real) Maximum value for the object
C   (>) UNITS:     (character) Units of the object
C   (>) NITEMS:    (integer) Number of items to be read
C   (<) VALUE:     (real array VALUE(NITEMS)) Actual value(s) of the desired variable
C   (!) STATUS:    (integer) Status flag
C
C
C   Prompts
C   -------
C   "PAR_NAME" --     If the object cannot be found in the TAUCAL
C                     structure, then the PAR system takes over.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C   Library DTA:
C     DTA_RDVARF
C
C   Library ICH:
C     ICH_CI
C     ICH_FOLD
C     ICH_LEN
C
C   Library PAR:
C     PAR_BATCH
C     PAR_CNPAR
C     PAR_GIVEN
C     PAR_RDARY
C     PAR_RDCHAR
C     PAR_RDKEY
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'TAU_COMMON'
C
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / Names > 6 characters
C
C
C   SUN-specific statements
C   -----------------------
C   None.
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   02-JUL-1990   - Original subroutine
C   19-NOV-1990   - Added parameter ASK_PAR. (JRL)
C   06-DEC-1990   - Cosmetic changes. (JRL)
C   09-JAN-1991   - Now checks for parameter on command line in batch. (JRL)
C   18-FEB-1991   - Checks ASK_PAR in event of read failure. (JRL)
C   15-JUL-1991   - Modified to use version 3 PAR routines. (JRL)
C   03-JAN-1992   - Included TAUCAL common block. (JRL)
C   28-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C+-----------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer runnum,nitems,status
      real vmin,vmax,value(nitems)
      character ref_name*(*),object*(*),par_name*(*),units*(*)
      logical taucal,ask_par
C
C     Functions
C
      character ich_ci*8
      integer ich_fold
      integer ich_len
      logical par_batch
      logical par_given
C
C     Local variables
C
      integer dumint,ldta,dta_status
      character number*8,dta_run*80,objup*80,par_name_up*80
      logical vallog,batch_job,given
C
C     TAUCAL common block
C
      include 'TAU_COMMON'
C
C     Inherited status
C
      if (status .ne. 0) go to 500
C
C     Batch job?
C
      batch_job = par_batch()
C
C     If the TAUCAL file exists then try to read the parameter first from
C     the TAUCAL file...
C
      if (taucal) then
         par_name_up = par_name
         dumint = ich_fold(par_name_up)
C
C        Construct full name.  If this is from a run the append the run number.
C        If not then create the proper name from the object name
C
         objup = object
         dumint = ich_fold(objup)
         if (runnum .ne. 0) then
            number = ich_ci(runnum)
            dta_run = dta_name(:lendta)//'.RUN'//
     :      number(:ich_len(number))//'.'//objup(:ich_len(objup))
         else
            dta_run = dta_name(:lendta)//'.'//objup(:ich_len(objup))
         end if
         ldta = ich_len(dta_run)
C
C        Now read the data
C
         call dta_rdvarf(dta_run(:ldta),nitems,value,dta_status)
C
C        If there was an error getting the information from the TAUCAL file
C        then try to get it from the parameter system.
C
         if (dta_status .ne. 0) then
            call dsa_wruser('Couldn''t find value of '//
     :      par_name_up(:ich_len(par_name_up)))
            call dsa_wrflush
C
C           If this is a batch job, then check to see that the parameter is
C           given on the command line.
C
            if (batch_job) then
               given = par_given(par_name)
               if (.not. given) then
                  call dsa_wruser(par_name_up(:ich_len(par_name_up))//
     :            ' parameter undefined')
                  call dsa_wrflush
                  status = 1
                  go to 500
               end if
            end if
C
C           If you still want to ask the parameter system in the event
C           of read failure, then do so.  Otherwise, escape with error status
C
            if (ask_par) then
               if (.not. batch_job) call par_cnpar(par_name)
               if (nitems .eq. 1) then
                  call par_rdval(par_name,vmin,vmax,1.0,units,value)
               else
                  call par_rdary(par_name,vmin,vmax,'n',units,nitems,
     :            nitems,value)
               end if
            else
               status = 1
               go to 500
            end if
         end if
C
C     If no TAUCAL file exists, then just get the values from the parameter
C     system (as long as that's what you want...)
C
      else if (ask_par) then
C
C        If this is a batch job, then check to see that the parameter is
C        given on the command line.
C
         if (batch_job) then
            given = par_given(par_name)
            if (.not. given) then
               call dsa_wruser(par_name_up(:ich_len(par_name_up))//
     :         ' parameter undefined')
               call dsa_wrflush
               status = 1
               go to 500
            end if
         end if
C
C        For single items, if the unit is 'key' then read a keyword from
C        the parameter system otherwise read a floating point value
C
         if (.not. batch_job) call par_cnpar(par_name)
         if (nitems .eq. 1) then
            if ((units .eq. 'key') .or. (units .eq. 'KEY')) then
               call par_rdkey(par_name,.true.,vallog)
               value(1) = 1.0
               if (.not. vallog) value(1) = 0.0
            else
               call par_rdval(par_name,vmin,vmax,1.0,units,value)
            end if
         else
            call par_rdary(par_name,vmin,vmax,'n',units,nitems,
     :      nitems,value)
         end if
      end if
C
C     Exit
C
  500 continue
      end
