      subroutine tau_histogram(array,nx,ny,loclip,hiclip,rnotinfld,
     : nbins,ninfield,nused,min1,max1,freqar,mean,sigma,status)
C+-----------------------------------------------------------------------------
C
C   -------------------------
C   T A U _ H I S T O G R A M
C   -------------------------
C
C   Description
C   -----------
C   Does histogram binning of 2-d images of real data
C
C
C   Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C   (>) ARRAY:      (real array ARRAY(NX,NY)) pixels to be histogrammed
C   (>) NX:         (integer) number of pixels in x-direction
C   (>) NY:         (integer) number of pixels in y-direction
C   (>) LOCLIP:     (real) low clipping level
C   (>) HICLIP:     (real) high clipping level
C   (>) RNOTINFLD:  (real) value indicating point outside field of view
C   (>) NBINS:      (integer) number of bins
C   (<) NINFIELD:   (integer) number of pixels inside the field
C   (<) NUSED:      (integer) number of pixels between clipping levels
C   (<) MIN:        (real) lowest value found between clipping levels
C   (<) MAX:        (real) highest value found between clipping levels
C   (<) FREQAR:     (integer array FREQAR(NBINSMAX)) number of pixels in the bins
C   (<) MEAN:       (real) mean value of pixels between clipping levels
C   (<) SIGMA:      (real) mean error per pixel
C   (<) STATUS:     (integer) value indicating an error
C
C
C   Prompts
C   -------
C   None
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_WRFLUSH
C     DSA_WRUSER
C
C
C   INCLUDE statements
C   ------------------
C   None
C                                                
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / END DO / Names > 6 characters
C
C
C   SUN-specific statements
C   -----------------------
C   None.
C
C
C   Author/s
C   --------
C   Ton Schoenmaker - Kapteyn Sterrenwacht - Roden
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   26-JAN-1987   - Original subroutine
C   06-FEB-1990   - Binning algorithm replaced. (JRL)
C   27-FEB-1990   - Took out maximum number of bins declaration. (JRL)
C   04-SEP-1990   - Took out a few VAX specific things and added in
C                   inherited status flag.  (JRL)
C   07-MAY-1991   - Now does mean and sigma arithmetic in double precision
C                   to avoid roundoff errors from large numbers in CCD data. (JRL)
C   28-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C   Other Remarks
C   -------------
C   This routine desparately needs to be replaced by
C   something a little more worthwhile.
C
C
C+-----------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,nbins,ninfield,nused,status,freqar(nbins)
      real array(nx,ny),loclip,hiclip,mean,sigma,rnotinfld
C
C     Local variables
C
      integer index,i,j
      real min1,max1,step,val
      double precision sum,sumsq,dval,dmean
C
C     Inherited status
C
      if (status .ne. 0) go to 500
C
C     Some inits
C
      max1 = hiclip
      min1 = loclip
      ninfield = 0
      nused = 0
      sum = 0d0
      sumsq = 0d0
      do i = 1,nbins
         freqar(i) = 0
      end do
C
C     Nonsense ranges...
C
      if (((max1-min1) .lt. 1e-3) .or. (nbins .lt. 2)) then
         freqar(1) = nused
         mean = max1
         sigma = 0d0
         status = 1
         go to 500
      end if
C
C     Fill frequency array
C
      step = (max1 - min1)/(real(nbins))
      do j = 1,ny
         do i = 1,nx
            val = array(i,j)
            if (val .gt. rnotinfld) ninfield = ninfield + 1
            if ((val .ge. loclip) .and. (val .lt. hiclip)) then
               nused = nused + 1
               dval = dble(val)
               sum = sum + dval
               sumsq = sumsq + dval*dval
               index = int((val-loclip)/step) + 1
               index = max(index,1)
               index = min(index,nbins)
               freqar(index) = freqar(index) + 1
            end if
         end do
      end do
C
C     Reasonable values ?
C
      if (nused.eq.0.) then
         call dsa_wruser('No points histogrammed')
         call dsa_wrflush
         status=1
         goto 500
      end if
C
C     Calculate mean and sigma
C
      dmean = sum/float(nused)
      mean = sngl(dmean)
      if (nused .eq. 1) then
        sigma = 0e0
      else
        sigma = sqrt(sumsq/float(nused) - dmean*dmean)
      end if
      status=0
  500 end
                         
