      subroutine tau_lsq(p,w,n,m,nlim,mlim1,q,s,r,t,er,res,me,status)
C+-----------------------------------------------------------------------------
C
C   -------------
C   T A U _ L S Q
C   -------------
C
C   Description
C   -----------
C   Performs a least-squares solution
C
C
C   Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C   (>) P:       (real array P(NLIM,MLIM)) condition equations, 
C                including right-hand side
C   (>) W:       (real array W(NLIM)) weight for each of the condition equations
C   (>) N:       (integer) number of condition equations
C   (>) M:       (integer) number of unknowns + 1
C   (>) NLIM:    (integer) maximum number of equations
C   (>) MLIM1:   (integer) maximum number of unknowns
C   (>) Q:       (real*8 array Q(MLIM1,MLIM)) workspace 
C   (>) S:       (real*8 array S(MLIM1,MLIM1)) workspace 
C   (>) R:       (real*8 array R(MLIM1,MLIM2)) workspace 
C   (<) T:       (real array T(MLIM1)) m-1 solved unknowns
C   (<) ER:      (real array ER(MLIM1)) m-1 error of the unknowns
C   (<) RES:     (real array RES(NLIM)) the n residuals (o-c)
C   (<) ME:      (real) unit weight error of the solution
C   (<) STATUS:  (integer) value returned indicating an error (if lt 0)
C
C   Prompts
C   -------
C   None
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   None
C
C
C   INCLUDE statements
C   ------------------
C   None
C                                                
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / Names > 6 characters
C
C
C   SUN-specific statements
C   -----------------------
C   None.
C
C
C   Author/s
C   --------
C   Ton Schoenmaker - Kapteyn Sterrenwacht Werkgroep - Roden
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   14-AUG-1984   - Original subroutine
C   21-FEB-1985
C   05-SEP-1986   - New limits for m and n (AAS)
C   23-JAN-1990   - q,s,r all passed in subroutine call. This takes away the 
C                   need for parameter statements defining the array 
C                   sizes.  (JRL)
C   04-SEP-1990   - Cleaned out a few VAX specific things and added an 
C                   inherited status flag.  (JRL)
C
C   Other Remarks
C   -------------
C   This routine desparately needs to be replaced
C   by something worthwhile!!!
C
C+-----------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer mlim1,nlim,status,m,n
      real p(nlim,mlim1+1),w(nlim),t(mlim1),er(mlim1),res(nlim),me
      double precision q(mlim1,mlim1+1),s(mlim1,mlim1),r(mlim1,mlim1+2)
C
C     Local variables
C
      integer i,j,k
      double precision u,l,kk
C
C     Inherited status
C
      if (status .ne. 0) go to 500
C
C     Check number of equations vs. number of unknowns
C
      if (n.lt.(m-1)) then
        status=-1
        goto 500
      end if
C
C     Check number of equations
C
      if (n.gt.nlim) then
        status=-1
        goto 500
      end if
C
C     Check number of unknowns
C
      if (m.gt.mlim1+1) then
        status=-1
        goto 500
      end if
C
C     Set normal equations to zero
C
      do 100 i=1,m-1
        do 90 j=1,m
          q(i,j)=0d0
   90   continue
  100 continue
C
C     Compose normal equations
C
      do 200 i=1,m-1
        do 190 k=1,n
          do 180 j=1,m
            q(i,j)=q(i,j)+p(k,i)*p(k,j)*w(k)
  180     continue
  190   continue
  200 continue
      do 250 j=1,m-1
        do 240 k=1,j
          q(j,k)=q(k,j)
  240   continue
  250 continue
C
C     Elimination of the normal equations
C
      do 300 j=2,m+1
        r(1,j)=q(1,j-1)
  300 continue
      if (abs(q(1,1)).lt.1d-30) then
        status=-2
        goto 500
      end if
      r(1,1)=1d0/q(1,1)
      do 400 i=2,m-1
        do 370 j=1,i-1
          if (abs(r(j,j+1)).lt.1d-30) then
            status=-3
            goto 500
          end if
          r(i,j)=-r(j,i+1)/r(j,j+1)
  370   continue
        do 390 j=i+1,m+1
          r(i,j)=q(i,j-1)
          do 380 k=1,i-1
            r(i,j)=r(i,j)+r(i,k)*r(k,j)
  380     continue
  390   continue
        if (abs(r(i,i+1)).lt.1d-30) then
          status=-4
          goto 500
        end if
        r(i,i)=1d0/(r(i,i+1))
  400 continue
C
C     Calculation of unknowns
C
      if (abs(r(m-1,m)).lt.1d-30) then
        status=-5
        goto 500
      end if
      t(m-1)=-r(m-1,m+1)/r(m-1,m)
      do 450 i=m-2,1,-1
        t(i)=r(i,m+1)
        do 440 k=m-1,i+1,-1
          t(i)=t(i)+t(k)*r(i,k+1)
  440   continue
        if (abs(r(i,i+1)).lt.1d-30) then
          status=-6
          goto 500
        end if
        t(i)=-t(i)/r(i,i+1)
  450 continue
C
C     Calculation of residuals
C
      u=0d0
      do 510 i=1,n
        kk=0d0
        do 490 j=1,m-1
          kk=kk+t(j)*p(i,j)
  490   continue
        l=kk+p(i,m)
        res(i)=-l*w(i)
        u=u+l*l*w(i)
  510 continue
      if (n.eq.m-1) then
        me=0e0
        do 550 i=1,m-1
          er(i)=0e0
  550   continue
        status=0
        goto 500
      else
        u=u/(dble(n-m+1))
        me=sqrt(u)
      end if
C
C     Calculation of mean error of unknowns
C
      do 600 i=m-1,1,-1
        do 590 j=i,1,-1
          if (j.eq.i) then
            s(i,j)=r(i,j)
          else
            s(i,j)=0d0
          end if
          do 580 k=j+1,m-1
            s(i,j)=s(i,j)+r(k,j)*s(i,k)
  580     continue
          s(j,i)=s(i,j)
  590   continue
  600 continue
      do 700 i=1,m-1
C       if (s(i,i).lt.1d-20) s(i,i)=0d0
        s(i,i)=sqrt(s(i,i)*u)
        er(i)=s(i,i)
  700 continue
C
      status=0
      return
  500 end
