      subroutine mrqmin(x,nx,y,sig,ndata,a,ma,lista,mfit,covar,alpha,
     : nca,chisq,funcs,alamda,atry,beta,da,dyda,iwork,status)
C+-----------------------------------------------------------------------------
C     MRQMIN      --- Routine taken from Press et. al. to do the non-linear
C                     fitting.  This has been slightly modified in three ways:
C                     (1) The independent variable (X) is now multi-dimensional
C                     to allow for the fact that this is a multi-d fit. (2) The
C                     work arrays have all be passed in which takes away
C                     the need to dimension them explicitly with a parameter
C                     statement in the subroutines.  (3) A status flag has
C                     been included so that if something drastic goes wrong
C                     the routine doesn't just stop dead in its tracks.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:         (real array X(NX,NDATA)) Array of independent variables
C     (>) NX:        (integer) Number of independent variables
C     (>) Y:         (real array Y(NDATA)) Array of dependent variables
C     (>) SIG:       (real array SIG(NDATA)) Array of errors in Y
C     (>) NDATA:     (integer) Number of pixels in the fit
C     (!) A:         (real array A(MA)) Fit parameters
C     (>) MA:        (integer) Dimensions of A
C     (>) LISTA:     (integer array LISTA(MA)) Flags for allowing some parameters to
C                    remain constant.
C     (>) MFIT:      (integer) Total number of parameters to vary.
C     (!) COVAR:     (real array COVAR(NCA,NCA)) Covariance matrix
C     (!) ALPHA:     (real array ALPHA(NCA,NCA)) Curvature matrix
C     (>) NCA:       (integer) Dimensions of COVAR and ALPHA. NCA >= MFIT
C     (!) CHISQ:     (real) Chi squared
C     (>) FUNCS:     (external routine) Name of external subroutine to evaluate function and
C                    its Jacobean.
C     (!) ALAMDA:    (real) Estimator of scale of change required for the next iteration
C     (W) ATRY:      (real array ATRY(MA)) Workspace
C     (W) BETA:      (real array BETA(MA)) Workspace
C     (W) DA:        (real array DA(MA)) Workspace
C     (W) DYDA:      (real array DYDA(MA)) Workspace
C     (W) IWORK:     (real array IWORK(3*MA)) Workspace
C     (!) STATUS:    (integer) Status flag 1=problems with
C                    LISTA, 2=Gauss-Jordan routine failed.
C     
C
C     AUTHOR:      Press et. al. "Numerical Recipes" pp. 521-528
C     UPDATE:      21-MAY-1990 - As outlined above (JRL)
C                  30-OCT-1990 - Changes to status flag values.  (JRL)
C                  28-JAN-1994 -- Modified for use with standalone UNIX FIGARO
C                                 v4.0 (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer ma,status,mfit,nca,ndata,nx
      integer lista(ma),iwork(3*ma)
      real x(nx,ndata),y(ndata),sig(ndata),a(ma),covar(nca,nca),
     : alpha(nca,nca),chisq,alamda,atry(ma),beta(ma),da(ma),dyda(ma)
      external funcs
C
C     Local variables
C
      integer ihit,j,k,kk
      real ochisq
C
C     Initialise the status flag
C
      status = 0
C
C     For alamda < 0.0 do some initialisations
C
      if (alamda .lt. 0.0) then
         kk = mfit + 1
C
C        Check which parameters will be held constant and make sure LISTA
C        has a proper permutation of the coefficients.
C
         do j = 1,ma
            ihit = 0
            do k = 1,mfit
               if (lista(k) .eq. j) ihit = ihit + 1
            end do
            if (ihit .eq. 0) then
               lista(kk) = j
               kk = kk + 1
            else if (ihit .gt. 1) then
               call dsa_wruser('Improper permutation in LISTA')
               call dsa_wrflush
               status = 1
               go to 500
            end if
         end do
         if (kk .ne. (ma+1)) then
            call dsa_wruser('Improper permutation in LISTA')
            call dsa_wrflush
            status = 1
            go to 500
         end if
C
C        Set up a first guess at the covariance matrix
C
         alamda = 0.001
         call mrqcof(x,nx,y,sig,ndata,a,ma,lista,mfit,alpha,beta,nca,
     :   dyda,chisq,funcs)
         ochisq = chisq
         do j = 1,ma
            atry(j) = a(j)
         end do
      end if
C
C     Alter linearised fitting matrix by augmenting diagonal elements.
C
      do j = 1,mfit
         do k = 1,mfit
            covar(j,k) = alpha(j,k)
         end do
         covar(j,j) = alpha(j,j)*(1.0 + alamda)
         da(j) = beta(j)
      end do
C
C     Matrix solution
C
      call gaussj(covar,mfit,nca,da,1,1,iwork(1),iwork(ma+1),
     : iwork(2*ma+1),status)
      if (status .ne. 0) go to 500
C
C     Once converged evaluate covariance matrix with alamda = 0.0
C
      if (alamda .eq. 0.0) then
         call covsrt(covar,nca,ma,lista,mfit)
C
C     Otherwise look at success of most recent solution.  If chisq has
C     decreased, then accept it.  If not the increase alamda and return
C     the same parameters
C
      else
         do j = 1,mfit
            atry(lista(j)) = a(lista(j)) + da(j)
         end do
         call mrqcof(x,nx,y,sig,ndata,atry,ma,lista,mfit,covar,da,nca,
     :   dyda,chisq,funcs)
         if (chisq .lt. ochisq) then
            alamda = 0.1*alamda
            ochisq = chisq
            do j = 1,mfit
               do k = 1,mfit
                  alpha(j,k) = covar(j,k)
               end do
               beta(j) = da(j)
               a(lista(j)) = atry(lista(j))
            end do
         else
            alamda = 10.0*alamda
            chisq = ochisq
         end if
      end if
  500 end


      subroutine mrqcof(x,nx,y,sig,ndata,a,ma,lista,mfit,alpha,beta,
     : nalp,dyda,chisq,funcs)
C+--------------------------------------------------------------------------------------
C     MRQCOF      --- Routine taken from Press et. al. to evaluate the linearised 
C                     fitting matrix ALPHA and vector BETA. 
C
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) X:         (real array X(NX,NDATA)) Array of independent variables
C     (>) NX:        (integer) Number of independent variables
C     (>) Y:         (real array Y(NDATA)) Array of dependent variables
C     (>) SIG:       (real array SIG(NDATA)) Array of errors in Y
C     (>) NDATA:     (integer) Number of pixels in the fit
C     (>) A:         (real array A(MA)) Fit parameters
C     (>) MA:        (integer) Dimensions of A
C     (>) LISTA:     (integer array LISTA(MA)) Flags for allowing some parameters to
C                    remain constant.
C     (>) MFIT:      (integer) Total number of parameters to vary.
C     (<) ALPHA:     (real array ALPHA(NCA,NCA)) Curvature matrix
C     (<) BETA:      (real array BETA(MA)) Workspace
C     (>) NALP:      (integer) Size of ALPHA and BETA
C     (<) DYDA:      (real array DYDA(MA)) Array of first derivatives of fitting function
C     (<) CHISQ:     (real) Chi squared
C     (>) FUNCS:     (external routine) Name of external subroutine to evaluate function and
C                    its Jacobean.
C     
C
C     AUTHOR:      Press et. al. "Numerical Recipes" pp. 521-528
C     UPDATE:      21-MAY-1990 - Slightly tarted up. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer ndata,ma,mfit,nalp,nx
      integer lista(ma)      
      real x(nx,ndata),y(ndata),sig(ndata),alpha(nalp,nalp),beta(ma),
     : a(ma),chisq,dyda(ma)
      external funcs
C
C     Local variables
C
      integer i,j,k
      real dy,sig2i,wt,ymod
C
C     Initialise (symmetric) alpha and beta
C
      do j = 1,mfit
         do k = 1,j
            alpha(j,k) = 0.0
         end do
         beta(j) = 0.0
      end do
      chisq = 0.0
C
C     Summation loop over all data.
C
      do i = 1,ndata
         call funcs(x(1,i),nx,a,ymod,dyda,ma)
         sig2i = 1.0/(sig(i)*sig(i))
         dy = y(i) - ymod
         do j = 1,mfit
            wt = dyda(lista(j))*sig2i
            do k = 1,j
               alpha(j,k) = alpha(j,k) + wt*dyda(lista(k))
            end do
            beta(j) = beta(j) + dy*wt
         end do
         chisq = chisq + dy*dy*sig2i
      end do
C
C     Fill in symmetric side
C
      do j = 2,mfit
         do k = 1,j-1
            alpha(k,j) = alpha(j,k)
         end do
      end do
      end


      subroutine gaussj(a,n,np,b,m,mp,ipiv,indxr,indxc,status)
C+--------------------------------------------------------------------------------------
C     GAUSSJ      --- Routine for Gauss-Jordan elimination with full pivoting
C                     from Press et.al.  Very slight modifications
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (!) A:         (real array A(NP,NP)) Input array
C     (>) N:         (integer) Number of rows and columns in A
C     (>) NP:        (integer) Physical dimensions of A
C     (!) B:         (real array B(NP,MP)) Input array with RHS vectors
C     (>) M:         (integer) Number of RHS vectors
C     (>) MP:        (integer) Physical dimensions of B
C     (W) IPIV:      (integer array IPIV(NP)) Workspace
C     (W) INDXR:     (integer array INDXR(NP)) Workspace
C     (W) INDXC:     (integer array INDXC(NP)) Workspace
C     (!) STATUS:    (integer) Status flag
C     
C
C     AUTHOR:      Press et. al. "Numerical Recipes" pp. 24-29.
C     UPDATE:      21-MAY-1990 -- Integer workspace passed in and status flag included
C                                 so that routine doesn't just stop if there is a problem
C                                 Also slightly tarted up. (JRL)
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer n,np,m,mp,status
      integer ipiv(np),indxr(np),indxc(np)
      real a(np,np),b(np,mp)
C
C     Local variables
C
      integer i,icol,irow,j,k,l,ll
      real big,dum,pivinv
C
C     Initialisations
C
      do j = 1,n
         ipiv(j) = 0
      end do
C
C     This is the main loop over the columns to be reduced.
C
      do i = 1,n
         big = 0.0
C
C        This is the outer loop for the search for a pivot element.
C
         do j = 1,n
            if (ipiv(j).ne.1) then
               do k = 1,n
                  if (ipiv(k) .eq. 0) then
                     if (abs(a(j,k)) .ge. big) then
                        big = abs(a(j,k))
                        irow = j
                        icol = k
                     end if
                  else if (ipiv(k) .gt. 1) then
                     status = 2
                     go to 500
                  end if
               end do
            end if
         end do
         ipiv(icol) = ipiv(icol) + 1
C
C        We now have the pivot element, so we interchange rows, if needed to put the
C        pivot element on the diagonal.  The columns are not physically interchanged,
C        only relabeled:  INDXC(I), the column of the Ith pivot element, is the Ith 
C        column that is reduced, while INDXR(I) is the row in which that pivot element
C        was originally located.  If INDXR(I) .NE. INDXC(I) there is an implied column
C        interchange.  With this form of bookkeeping, the solution B's will end up in 
C        correct order, and the inverse matrix will be scrambled by columns.
C
         if (irow.ne.icol) then
            do l = 1,n
               dum = a(irow,l)
               a(irow,l) = a(icol,l)
               a(icol,l) = dum
            end do
            do l = 1,m
               dum = b(irow,l)
               b(irow,l) = b(icol,l)
               b(icol,l) = dum
            end do
         end if
C
C        We are now ready to divide the pivot row by the pivot element located
C        at IROW and ICOL.
C
         indxr(i) = irow
         indxc(i) = icol
         if (a(icol,icol) .eq. 0.0) THEN
            status = 2
            go to 500
         end if
         pivinv = 1.0/a(icol,icol)
         a(icol,icol) = 1.0
         do l = 1,n
            a(icol,l) = a(icol,l)*pivinv
         end do
         do l = 1,m
            b(icol,l) = b(icol,l)*pivinv
         end do
C
C        Next we reduce the rows...except for the pivot one, of course...
C
         do ll = 1,n
            if (ll.ne.icol) then
               dum = a(ll,icol)
               a(ll,icol) = 0.0
               do l = 1,n
                  a(ll,l) = a(ll,l) - a(icol,l)*dum
               end do
               do l = 1,m
                  b(ll,l) = b(ll,l) - b(icol,l)*dum
               end do
            end if
         end do
      end do
C
C     It only remains to unscramble the solution in view of the column
C     interchanges.  We do this by interchanging pairs of columns in the
C     reverse order that the permutation was built up
C
      do l = n,1,-1
         if (indxr(l).ne.indxc(l)) then
            do k = 1,n
               dum = a(k,indxr(l))
               a(k,indxr(l)) = a(k,indxc(l))
               a(k,indxc(l)) = dum
            end do
         end if
      end do
  500 end


      subroutine covsrt(covar,ncvm,ma,lista,mfit)
C+--------------------------------------------------------------------------------------
C     COVSRT      --- Routine from Press et.al. to repack the covariance matrix to
C                     the true order of parameters.
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (!) COVAR:     (real array COVAR(NCVM,NCVM)) Input covariance matrix
C     (>) NCVM:      (integer) Physical dimensions of COVAR
C     (>) MA:        (integer) Total number of parameters in fit
C     (>) LISTA:     (integer array LISTA(MFIT)) List of parameters to be varied
C     (>) MFIT:      (integer) Number of parameters to be varied.
C     
C
C     AUTHOR:      Press et. al.  "Numerical Recipes" p. 515.
C     UPDATE:      
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer ncvm,ma,mfit
      integer lista(mfit)
      real covar(ncvm,ncvm)
C
C     Local variables
C
      integer i,j
      real swap
C
C     Zero all elements below diagonal
C
      do j = 1,ma-1
         do i = j + 1,ma
            covar(i,j) = 0.0
         end do
      end do
C
C     Repack off-diagonal elements of fit into correct locations below diagonal
C
      do i = 1,mfit-1
         do j = i+1,mfit
            if (lista(j) .gt. lista(i)) then
               covar(lista(j),lista(i)) = covar(i,j)
            else
               covar(lista(i),lista(j)) = covar(i,j)
            end if
         end do
      end do
C
C     Temporarily store original diagonal elements in top row, and zero
C     the diagonal
C
      swap = covar(1,1)
      do j = 1,ma
         covar(1,j) = covar(j,j)
         covar(j,j) = 0.0
      end do
      covar(lista(1),lista(1)) = swap
C
C     Now sort elements into proper order on diagonal
C
      do j = 2,mfit
         covar(lista(j),lista(j)) = covar(1,j)
      end do
C
C     Finally fill in above diagonal by symmetry.
C
      do j = 2,ma
         do i = 1,j-1
            covar(i,j) = covar(j,i)
         end do
      end do
      end
