      subroutine tau_stats(data,nx,ny,thresh,clipmean,sig,nabove,amin,
     : amax,status)
C+-----------------------------------------------------------------------------
C
C   -----------------
C   T A U _ S T A T S
C   -----------------
C
C   Description
C   -----------
C   Finds which TAUCAL sub-structure the user wants to use.  It
C   first checks how many sub-structures there are.  If there are
C   none, then an error occurs.  If there is more than one then
C   the user is prompted for the number of the sub-structure to
C   be used.  Output is a label for the TAUCAL sub-structure for
C   other TAUCAL routines to use.
C
C
C   Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C   (>) DATA:      (real array DATA(NX,NY)) Input data
C   (>) NX:        (integer) X dimension of data
C   (>) NY:        (integer) Y dimension of data
C   (>) THRESH:    (real) Number of sigmas above background to find peaks
C   (<) CLIPMEAN:  (real) Clipped mean background value
C   (<) SIG:       (real) Background sigma
C   (<) NABOVE     (integer) Total number of pixels above THRESH*SIG (may
C                  include noise spikes.
C   (<) AMIN:      (real) Minimum value of array
C   (<) AMAX:      (real) Maximum value of array
C   (!) STATUS:    (integer) Status flag
C
C
C   Prompts
C   -------
C   None
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   None
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'NDP_NUMERIC_RANGES'
C                                                
C
C   Extensions to FORTRAN77
C   -----------------------
C   IMPLICIT NONE / INCLUDE / END DO / Names > 6 characters
C
C
C   SUN-specific statements
C   -----------------------
C   None.
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (jrl@uk.ac.cam.ast.mail)
C
C
C   History
C   -------
C   27-MAR-1990   - Original subroutine
C   25-JUL-1990   - Changed to find min and max of array as well (JRL)
C   19-MAR-1991   - Included STATUS parameter (JRL)
C   12-JUL-1991   - Mean and sigma now calculated in double precision. (JRL)
C   28-JAN-1994   - Modified for use with standalone UNIX FIGARO v4.0 (JRL)
C
C
C+-----------------------------------------------------------------------------
      implicit none
C     
C     Parameters
C
      integer nx,ny,nabove,status
      real data(nx,ny),thresh,clipmean,sig,amin,amax
C
C     Local variables
C
      integer i,npts,iy,ix
      real amean(2),delta,cut,val
      double precision sum,sumsq,dval,ave,sd
C
C     Numeric ranges
C
      include 'NDP_NUMERIC_RANGES'
C
C     Initialise mean and sigma so that on the first
C     go 'round nothing gets clipped.
C
      status = 0
      amean(1) = 0.0
      sig = max_float/(thresh + 1.0)
      amin = max_float
      amax = min_float
C
C     Calculate mean using asymmetric clip.
C
      do i = 1,2
         sum = 0.0d0
         sumsq = 0.0d0
         npts = 0
         do iy = 1,ny
            do ix = 1,nx
               val = data(ix,iy)
               dval = dble(val)
               amin = min(amin,val)
               amax = max(amax,val)
               delta = val - amean(1)
               if (delta .le. thresh*sig) then
                  npts = npts + 1
                  sum = sum + dval
                  sumsq = sumsq + dval*dval
               end if
            end do
         end do
         ave = sum/dble(float(npts))
         amean(i) = sngl(ave)
         sd = sumsq/dble(float(npts)) - ave**2
         if (sd .lt. 0.0) then
            call dsa_wruser('Non-real sigma found for background --')
            call dsa_wruser(' Can''t continue.  Try setting the')
            call dsa_wruser(' threshold a bit higher')
            call dsa_wrflush
            status = 1
            go to 500
         end if
         sig = sngl(sqrt(sd))
      end do
C
C     Now find the number of pixels above the threshold
C
      cut = thresh*sig
      nabove = 0
      do iy = 1,ny
         do ix = 1,nx
            delta = data(ix,iy) - amean(2)
            if (delta .gt. cut) nabove = nabove + 1
         end do
      end do
      clipmean = amean(2)
C
C     If there are no pixels above the threshold, then you can't continue!
C
      if (nabove .eq. 0) then
         call dsa_wruser('No pixels above the threshold!  Can''t')
         call dsa_wruser(' continue.  Try setting the threshold a bit')
         call dsa_wruser(' lower')
         call dsa_wrflush
         status = 1
      end if
  500 end                     
