      subroutine build
C+------------------------------------------------------------------------------
C
C   ---------
C   B U I L D
C   ---------
C
C   Description
C   -----------
C   Takes individual frames from either 2 or 3-d arrays and builds
C   a cube from them for use as RVT ring file.
C
C
C   Scope of program
C   ----------------
C   - Handles 2-d and 3-d images only.
C   - SHORT and FLOAT only supported.  All others
C     are converted to FLOAT.
C   - Quality and error arrays not supported.
C   - Magic values not supported.
C   - Batch execution supported.
C
C
C   Environment
C   -----------
C   FIGARO
C
C
C   Parameters (read from VARS.DST)
C   -------------------------------
C   CALFILE      Name of TAUCAL file (character)(prompted for)
C
C   RUNNUM       Run number associated with matrix mask frame. (integer)
C                (prompted for).
C
C   OUTPUT       Name of the structure for output.  This can either be new or
C                old.  Must be either 2-d or 3-d. (character)(prompted for).
C
C   WHICH_FRAME  If the input file is 3-d then this tells which frame is to
C                be extracted.  (integer)(prompted for)
C
C   Keywords
C   --------
C   USE_TAU      Flag for use of TAUCAL file. (prompted for)
C
C   NEW_FILE     Flag for the output file being new. (prompted for)
C
C
C   Parameters (read from or written to TAUCAL file)
C   ------------------------------------------------
C
C   Parameters which reside in .RUN structures have the run type given in
C   parentheses.  Those without a run type reside in the top level of the
C   calibration structure.
C
C   (RVT/RING/CAL)  IMAGE        Name of input file(s).  Files must exist.  
C                                Can be either 2-d or 3-d (character)(read)
C
C
C   Propagation of data structure
C   -----------------------------
C   If a new file is created, then it is created with the same data array
C   size and type as the first input file.  Each subsequent input file
C   must then match this size and type.
C
C
C   Method
C   ------
C   - Output file name and flag for new output file read
C   - If not a new output file, then the output file is opened
C     and tested for dimensionality and type.
C   - First input file is opened and tested for type and dimensionality
C   - If the output file is new, then it is created with the same type
C     and x,y dimensions as the first input file.
C   - Otherwise the input is checked to make sure it has the same type
C     and dimensions as the output file.
C   - Frame is copied.
C   - As many input files as the user wants can be done all in the same run.
C
C
C   External functions & subroutines called
C   ---------------------------------------
C   Library DSA:
C     DSA_CLOSE
C     DSA_CLOSE_STRUCTURE
C     DSA_DATA_SIZE
C     DSA_NAMED_INPUT
C     DSA_NAMED_OUTPUT
C     DSA_MAP_AXIS_DATA
C     DSA_MAP_DATA
C     DSA_OPEN
C     DSA_RESHAPE_AXIS
C     DSA_RESHAPE_DATA
C     DSA_SIMPLE_OUTPUT
C     DSA_UNMAP
C     DSA_WRUSER
C
C   Library DYN:
C     DYN_ELEMENT
C
C   Library NDP:
C     NDP_GET_IMAGE_INFO
C
C   Library PAR:
C     PAR_RDCHAR
C     PAR_RDKEY
C     PAR_RDVAL
C
C   Library TAU:
C     TAU_CAL_OPEN
C     TAU_CAL_RDCHR
C
C   Internal subroutines called
C   ---------------------------
C   SUB_BUILDI
C   SUB_BUILDR
C
C
C   INCLUDE statements
C   ------------------
C   INCLUDE 'DYN_SOURCE:DYNAMIC_MEMORY.INC'
C   INCLUDE 'NDP_SOURCE:NDP_NUMERIC_RANGES.INC'
C                                                
C
C   Extensions to FORTRAN77
C   -----------------------                           
C   END DO / IMPLICIT NONE / INCLUDE / Names > 6 characters /
C   DO WHILE / INTEGER*2
C
C                           
C   Possible future upgrades
C   ------------------------
C
C
C   Author/s
C   --------
C   Jim Lewis    RGO  (CAVAD::JRL  or JRL@UK.AC.CAM.AST-STAR)
C
C
C   History
C   -------
C   08-MAY-90   - Original program
C   03-JUL-90   - TAUCAL pasted in.  (JRL)
C   09-NOV-90   - Run number is now put into Z axis array (JRL)
C   20-NOV-90   - New versions of TAUCAL routines pasted in (JRL)
C
C
C   Other Comments
C   --------------
C
C+------------------------------------------------------------------------------
      implicit none
C
C     Functions
C
      integer       dyn_element
C
C     Local variables
C
      integer       address    ! Dynamic memory address
      integer       axptr      ! Dynamic memory axis pointer
      logical       badpix     ! Flag for magic values
      logical       badpixi    ! Flag for magic values
      integer       dims(3)    ! Dimensions of output file
      integer       dimsi(3)   ! Dimensions of input file
      real          dummy      ! Dummy real variable
      character     filein*80  ! Input file name
      integer       frame      ! Frame number to processed in input cube
      integer       i          ! Dummy counter
      integer       iptr       ! Dynamic memory pointer
      integer       ndim       ! Number of dimensions in output file
      integer       ndimi      ! Number of dimensions in input file
      integer       nelm       ! Number of elements in a file
      logical       new_file   ! Flag for output file's being new
      integer       optr       ! Dynamic memory pointer
      character     output*80  ! Output file name
      integer       runnum     ! Run number of input file
      integer       slot       ! Dynamic memory slot
      integer       slot2      ! Dynamic memory slot
      integer       slot3      ! Dynamic memory slot
      integer       status     ! Status flag
      logical       tauexist   ! Flag for existence of TAUCAL file
      character     type*8     ! Output file type
      character     typei*8    ! Input file type
C
C     Numeric ranges.
C
      include 'ndp_source:ndp_numeric_ranges.inc'
C
C     Common block for dynamic memory.
C
      include 'dyn_source:dynamic_memory.inc'
C
C     Get DSA started
C
      status = 0
      call dsa_open(status)
      if (status .ne. 0) go to 500
C
C     Open TAUCAL file
C
      call tau_cal_open('cal','calfile',tauexist,status)
      if (status .ne. 0) go to 500
C
C     Get output file name and keyword for new file
C
      call par_rdchar('output',' ',output)
      call par_rdkey('new_file',.true.,new_file)
C
C     If the output file already exists, then open it and check its dimensions.
C
      if (.not. new_file) then
         call dsa_named_input('output',output,status)
         call ndp_get_image_info('output',.true.,.false.,type,badpix,
     :   status)
         call dsa_data_size('output',3,ndim,dims,nelm,status)
         if (status .ne. 0) go to 500
         if (ndim .lt. 2) then
            call dsa_wruser('Output image must be at least 2-d\n')
            go to 500
         else if (ndim .eq. 2) then
            dims(3) = 1
         end if
         if (badpix) then
            call dsa_wruser('I can''t cope with magic values (yet)\n')
            go to 500
         end if
      end if
C
C     Read the run number of each input file in turn.  A run number of zero
C     ends it all...
C
      i = 0
      call par_rdval('runnum',float(min_int),float(max_int),1.0,
     : ' ',dummy)
      runnum = nint(dummy)
      do while (runnum .ne. 0)
C
C        Get name of ring file.  
C
         call tau_cal_rdchr('cal',runnum,'image','filein',tauexist,
     :   .true.,80,filein,status)
         if (status .ne. 0) go to 500
C
C        Open the input file and check its dimensions 
C
         call dsa_named_input('input',filein,status)
         call ndp_get_image_info('input',.true.,.false.,typei,badpixi,
     :   status)
         call dsa_data_size('input',3,ndimi,dimsi,nelm,status)
         if (status .ne. 0) go to 500
         if (badpix) then
            call dsa_wruser('I can''t cope with magic values (yet)\n')
            go to 500
         end if
         if (ndimi .lt. 2) then
            call dsa_wruser('Input file needs to be at least 2-d\n')
            go to 500
         end if
         if (ndimi .eq. 2) dimsi(3) = 1
C
C        If the output file is new and this is the first time through
C        the loop, then create the new output file with the same 
C        dimemsions as the first input file
C
         if ((new_file) .and. (i .eq. 0)) then
            i = i + 1
            dims(1) = dimsi(1)
            dims(2) = dimsi(2)
            dims(3) = 1   
            type = typei
            call dsa_named_output('output',output,'input',1,1,status)
            if (type .eq. 'SHORT') then
               call dsa_simple_output('output',
     :         'd,a3(float)','short',3,dims,status)
            else
               call dsa_simple_output('output','d,a3','float',3,
     :         dims,status)
            end if
C
C        Otherwise check to make sure the input file has the same dimensions
C        and type as the output file
C
         else
            if (typei .ne. type) then
               call dsa_wruser
     :         ('Input and output file types don''t match\n')
               go to 500
            end if
            if ((dimsi(1) .ne. dims(1)) .or. 
     :      (dimsi(2) .ne. dims(2))) then
               call dsa_wruser
     :         ('Input data array dimensions don''t match output\n')
               go to 500
            end if  
C
C           Now reshape the axis and data arrays in order to add on another frame
C
            dims(3) = dims(3) + 1
            call dsa_reshape_data('output','output',3,dims,status)
            call dsa_reshape_axis('output',3,'output',3,1,dims(3),
     :      status)
         end if
C
C        If the input frame is in fact a cube, then nominate the 
C        frame to be copied over...
C
         if (ndimi .eq. 3) then
            call par_rdval('which_frame',float(min_int),float(max_int),
     :      1.0,' ',dummy)
            frame = nint(dummy)
         else
            frame = 1
         end if
C
C        Now map the output file and its z axis as well as the input file
C
         call dsa_map_data('output','update',type,address,slot,status)         
         optr = dyn_element(address)
         call dsa_map_axis_data('output',3,'update','float',address,
     :   slot2,status)
         axptr = dyn_element(address)
         call dsa_map_data('input','read',typei,address,slot3,status)
         iptr = dyn_element(address)
         if (status .ne. 0) go to 500
C
C        Now call the subroutine appropriate to the type of file...
C
         if (type .eq. 'SHORT') then
            call sub_buildi(dynamic_mem(iptr),dims(1),dims(2),dimsi(3),
     :      dynamic_mem(axptr),dims(3),runnum,frame,dynamic_mem(optr))
         else
            call sub_buildr(dynamic_mem(iptr),dims(1),dims(2),dimsi(3),
     :      dynamic_mem(axptr),dims(3),runnum,frame,dynamic_mem(optr))
         end if
C
C        Unmap all of the files in preparation for the next one...
C
         call dsa_unmap(slot,status)
         call dsa_unmap(slot2,status)
         call dsa_unmap(slot3,status)
         call dsa_close_structure('input',status)
         if (status .ne. 0) go to 500
C
C        Get next runnumber
C
         call par_rdval('runnum',float(min_int),float(max_int),1.0,
     :   ' ',dummy)
         runnum = nint(dummy)
      end do
C
C     Tidy and exit
C
  500 call dsa_close(status)
      end

      subroutine sub_buildi(input,nx,ny,niz,axis,noz,run,frame,output)
C+--------------------------------------------------------------------------------------
C     SUB_BUILDI  --- Routine to copy a frame input another structure. 
C                     SHORT version
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (short array INPUT(NX,NY,NIZ)) Input data cube
C     (>) NX:        (integer) X dimension of data array
C     (>) NY:        (integer) Y dimension of data array
C     (>) NIZ:       (integer) Z dimension of input data array
C     (!) AXIS:      (real array AXIS(NOZ)) Output Z axis data array
C     (>) NOZ:       (integer) Z dimension of output data array
C     (>) RUN:       (integer) Run number of input file
C     (>) FRAME:     (integer) Input frame number
C     (<) OUTPUT:    (short array OUTPUT(NX,NY,NOZ)) Output data cube
C     
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        08-MAY-1990
C     UPDATE:      09-NOV-1990 -- Modified to add run number to z axis
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,niz,noz,frame,run
      integer*2 input(nx,ny,niz),output(nx,ny,noz)
      real axis(noz)
C
C     Local variables
C
      integer i,j                  
C
C     Give the Z axis the value of the UT of the observation.
C
      axis(noz) = float(run)
C      
C     Now transfer the data over...
C
      do j = 1,ny
         do i = 1,nx
            output(i,j,noz) = input(i,j,frame)
         end do
      end do
      end


      subroutine sub_buildr(input,nx,ny,niz,axis,noz,run,frame,output)
C+--------------------------------------------------------------------------------------
C     SUB_BUILDR  --- Routine to copy a frame input another structure. 
C                     FLOAT version
C
C     Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
C
C     (>) INPUT:     (real array INPUT(NX,NY,NIZ)) Input data cube
C     (>) NX:        (integer) X dimension of data array
C     (>) NY:        (integer) Y dimension of data array
C     (>) NIZ:       (integer) Z dimension of input data array
C     (!) AXIS:      (real array AXIS(NOZ)) Output Z axis data array
C     (>) NOZ:       (integer) Z dimension of output data array
C     (>) RUN:       (integer) Run number of input data
C     (>) FRAME:     (integer) Input frame number
C     (<) OUTPUT:    (real array OUTPUT(NX,NY,NOZ)) Output data cube
C     
C
C     AUTHOR:      Jim Lewis -- RGO
C     DATE:        08-MAY-1990
C     UPDATE:      09-NOV-1990 -- Modified to add run number to z axis
C
C+--------------------------------------------------------------------------------------
      implicit none
C
C     Parameters
C
      integer nx,ny,niz,noz,frame,run
      real input(nx,ny,niz),output(nx,ny,noz)
      real axis(noz)
C
C     Local variables
C
      integer i,j                  
C
C     Give the Z axis the value of the UT of the observation.
C
      axis(noz) = float(run)
C      
C     Now transfer the data over...
C
      do j = 1,ny
         do i = 1,nx
            output(i,j,noz) = input(i,j,frame)
         end do
      end do
      end
